
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, February 1999.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "absbal.h"
#include "statq.h"
#include "statstaq.h"
#include "dynstaq.h"
#include "setfam.h"


abstractBalancedFNW::abstractBalancedFNW(TNode nn) throw()
{
    n1 = nn;

    Q = NULL;
    prop = NULL;
    petal = NULL;
    base = NULL;

    LogEntry(LOG_MEM,"...Abstract balanced flow network allocated");
}


abstractBalancedFNW::~abstractBalancedFNW() throw()
{
    ReleaseCycles();
    ReleaseProps();
    ReleasePetals();
    ReleaseBlossoms();

    LogEntry(LOG_MEM,"...Abstract balanced flow network disallocated");
}


unsigned long abstractBalancedFNW::Allocated() throw()
{
    unsigned long tmpSize = 0;

    if (Q!=NULL)            tmpSize += n*sizeof(TArc);
    if (prop!=NULL)         tmpSize += n*sizeof(TArc);
    if (petal!=NULL)        tmpSize += n*sizeof(TArc);
    if (base!=NULL)         tmpSize += n1*sizeof(TArc);
    
    return tmpSize;
}                  


void abstractBalancedFNW::WriteBalancedFNW(const char* wrFile,TOption options)
    throw(ERFile)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    if (CT.logIO)
    {
        sprintf(CT.logBuffer,
            "Writing balanced flow network to \"%s\"...",wrFile);
        LogEntry(LOG_IO,CT.logBuffer);
    }
    
    goblinExport F(wrFile,CT);
    
    F.StartTuple("balanced_fnw",0);
    
    F.StartTuple("definition",0);
    
    F.StartTuple("nodes",1);
    F.MakeIntItem(n,0);
    F.MakeIntItem(n,0);
    F.MakeIntItem(ni,0);
    F.EndTuple();
    
    F.StartTuple("arcs",1);
    F.MakeIntItem(m,0);
    F.EndTuple();
    
    WriteIncidences(&F);

    WriteUCap(&F);
    WriteLCap(&F);
    WriteDemand(&F);
    WriteOrientation(&F);
    
    F.EndTuple();   // definition
    
    WriteObjectives(&F);
    WriteGeometry(&F);
    WriteLayout(&F);
    WriteSolutions(&F);
    
    CT.sourceNode = Source();
    CT.targetNode = Target();
    CT.rootNode = Root();
    F.WriteConfiguration(this);

    F.EndTuple();   // balanced

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
}


bool abstractBalancedFNW::Outer(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Outer",v);

    #endif

    return !(v^1);
}                 


bool abstractBalancedFNW::Inner(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Inner",v);

    #endif

    return v^1;
}                    


TNode abstractBalancedFNW::ComplNode(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("ComplNode",v);

    #endif

    return v^1;
}


TArc abstractBalancedFNW::ComplArc(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("ComplArc",a);

    #endif

    return a^2;
}


void abstractBalancedFNW::InitCycles() throw()
{
    if (Q==NULL)
    {
        Q = new TArc[n];
        LogEntry(LOG_MEM,"...Odd length cycles allocated");
    }
    else
    {
        #if defined(_LOGGING_)
    
        Error(MSG_WARN,"InitCycles","Odd length cycles are already present");

        #endif
    }
    
    for (TNode i=0;i<n;i++) Q[i] = NoArc;
}


void abstractBalancedFNW::ReleaseCycles() throw()
{
    if(Q!=NULL)
    {
        delete[] Q;
        Q = NULL;
        LogEntry(LOG_MEM,"...Odd length cycles disallocated");
    }
}


void abstractBalancedFNW::InitBlossoms() throw()
{
    if (base==NULL)
    {
        base =  new TNode[n1];
        partition = new goblinSetFamily<TNode>(n1,CT);
        LogEntry(LOG_MEM,"...Blossoms allocated");
    }
    else
    {
        #if defined(_LOGGING_)

        partition -> Init();
        Error(MSG_WARN,"InitBlossoms","Blossoms are already present");

        #endif
    }
}


void abstractBalancedFNW::ReleaseBlossoms() throw()
{
    if (base!=NULL)
    {
        delete[] base;
        base = NULL;
        delete partition;
        partition = NULL;
        LogEntry(LOG_MEM,"...Blossoms disallocated");
    }
}


TNode abstractBalancedFNW::Base(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Base",v);

    #endif
    
    TNode u = partition->Find(v/2);

    if (u==NoNode) return NoNode;
    else return base[u];
}


TNode abstractBalancedFNW::Pred(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Pred",v);

    #endif
    
    return Base(StartNode(prop[v]));
}


void abstractBalancedFNW::Bud(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Bud",v);

    #endif
    
    partition -> Bud(v/2);
    base[v/2] = v;
}


void abstractBalancedFNW::Shrink(TNode u,TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("Shrink",u);

    if (v>=n) NoSuchNode("Shrink",v);

    #endif
    
    TNode x = partition->Find(u/2);
    TNode y = partition->Find(v/2);

    partition -> Merge(x,y);
    base[partition->Find(x)] = base[x];
}


void abstractBalancedFNW::InitProps() throw()
{
    if (prop==NULL)
    {
        prop = new TArc[n];
        LogEntry(LOG_MEM,"...Props allocated");
    }
    else
    {
        #if defined(_LOGGING_)

        Error(MSG_WARN,"InitProps","Odd length cycles are already present");

        #endif
    }
    
    for (TNode i=0;i<n;i++) prop[i] = NoArc;
}


void abstractBalancedFNW::ReleaseProps() throw()
{
    if (prop!=NULL)
    {
        delete[] prop;
        prop = NULL;
        LogEntry(LOG_MEM,"...Props disallocated");
    }
}


void abstractBalancedFNW::InitPetals() throw()
{
    if (petal==NULL)
    {
        petal = new TArc[n];
        LogEntry(LOG_MEM,"...Petals allocated");
    }
    else
    {
        #if defined(_LOGGING_)

        Error(MSG_WARN,"InitPetals","Odd length cycles are already present");

        #endif
    }
    
    for (TNode i=0;i<n;i++) petal[i] = NoArc;
}


void abstractBalancedFNW::ReleasePetals() throw()
{
    if (petal!=NULL)
    {
        delete[] petal;
        petal = NULL;
        LogEntry(LOG_MEM,"...Petals disallocated");
    }
}


TFloat abstractBalancedFNW::BalCap(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("BalCap",a);

    #endif
    
    if(a&1) return BalFlow(a)-LCap(a);
    else return UCap(a)-BalFlow(a);
}


void abstractBalancedFNW::BalAugment(TNode u,TNode v,TFloat Lambda)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("BalAugment",u);

    if (v>=n) NoSuchNode("BalAugment",v);

    if (Lambda<=0) Error(ERR_REJECTED,"BalAugment","Amount should be positive");

    if (P==NULL) Error(ERR_REJECTED,"BalAugment","No path specified");

    TNode PathLength = 0;

    #endif

    TNode w = v;

    #if defined(_LOGGING_)

    THandle LH = NoHandle;

    if (CT.logMeth>1)
    {
        sprintf(CT.logBuffer,"Augmenting by %g units of flow...",2*Lambda);
        LogEntry(LOG_METH2,CT.logBuffer);
        LogEntry(LOG_METH2,"Path in reverse order:");
        OpenFold();
        sprintf(CT.logBuffer,"(%ld",w);
        LH = LogStart(LOG_METH2,CT.logBuffer);
    }

    #endif
    
    while (w!=u)
    {
        TArc a = P[w];
	
        #if defined(_FAILSAVE_)

        if (PathLength>=n || a==NoArc)
            InternalError("BalAugment","Missing start node");

        PathLength++;
	
        #endif

        BalPush(a,Lambda);
        P[w] = NoArc;
        w = StartNode(a);

        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"[%ld]%ld",a,w);
            LogAppend(LH,CT.logBuffer);
        }

        #endif
    }
    
    #if defined(_LOGGING_)

    if (CT.logMeth>1)
    {
        LogEnd(LH,")");
        CloseFold();
    }

    #endif
}
 
 
TFloat abstractBalancedFNW::FindBalCap(TNode u,TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("FindBalCap",u);

    if (v>=n) NoSuchNode("FindBalCap",v);

    if (P==NULL) Error(ERR_REJECTED,"FindBalCap","No path specified");

    TNode PathLength = 0;

    #endif

    TNode w = v;
    TFloat Lambda = InfFloat;

    while (w!=u)
    {
        TArc a = P[w];

        #if defined(_FAILSAVE_)

        if (PathLength>=n || a==NoArc)
        {
            InternalError("FindBalCap","Missing start node");
        }

        PathLength++;
	
        #endif

        TFloat LambdaNeu = BalCap(a);
        w = StartNode(a);

        if (P[w^1]==(a^2)) LambdaNeu = floor(LambdaNeu/2);
        if (LambdaNeu<Lambda) Lambda = LambdaNeu;

        #if defined(_FAILSAVE_)

        if (Lambda==0)      
        {
            sprintf(CT.logBuffer,"Arc %ld has capacity %g",a,BalCap(a));
            Error(ERR_CHECK,"FindBalCap",CT.logBuffer);
        }

        #endif
    }
    
    return Lambda;
}

 
void abstractBalancedFNW::CancelEven() throw()
{
    LogEntry(LOG_METH,"Cancelling even length cycles...");
    OpenFold(ModCycleCancel);

    InitPredecessors();
    InitCycles();
    Symmetrize();
    
    #if defined(_LOGGING_)

    THandle LH = NoHandle;

    if (CT.logMeth>1)
    {
        LogEntry(LOG_METH2,"Traversed nodes (arcs):");
        LH = LogStart(LOG_METH2,"");
    }
    
    TNode k = 0;

    #endif

    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    for (TNode i=0;i<n1;i++)
    {
        TNode root = 2*i;
        TNode v = root;
        
        while (v!=root || I.Active(v))
        {
            TArc a = I.Read(v);
            TNode u = EndNode(a);
            
            if (BalCap(a)!=floor(BalCap(a))+0.5 ||
                P[v]==(a^1) || Q[v]==(a^1) || Q[u]==a) continue;
                
            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer," %ld (%ld)",v,a);
                LogAppend(LH,CT.logBuffer);
            }

            #endif

            if (P[u]!=NoArc || u==root)
            {
                BalPush(a^1,0.5);
                MakeIntegral(P,u,v);
                v = u;

                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer," %ld",u);
                    LogEnd(LH,CT.logBuffer);
                    LogEntry(LOG_METH2,"...Fractional arcs cancelled");
                    LH = LogStart(LOG_METH2,"");
                }

                #endif

                CT.Ping(Handle(),1);
                
                continue;
            }

            TNode cu = u^1;

            if (P[cu]==NoArc && cu!=root)
            {
                P[u] = a;
                v = u;
                
                continue;
            }

            P[u] = a;
            TNode w = u;

            while (w!=cu && Q[w]==NoArc) w = StartNode(P[w]);

            if (Q[w]==NoArc)
            {
                v = u;

                while (v!=cu)
                {
                    a = P[v];
                    Q[v] = a;
                    Q[v^1] = a^3;
                    P[v] = NoArc;
                    v = StartNode(a);
                }

                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer," %ld",u);
                    LogEnd(LH,CT.logBuffer);
                    LogEntry(LOG_METH2,"...New odd cycle found");
                    LH = LogStart(LOG_METH2,"");
                    k++;
                }

                #endif
            }
            else
            {
                MakeIntegral(Q,w^1,w);
                TNode x = u;

                while (x!=w)
                {
                    TArc a = P[x];
                    BalPush(a^1,0.5);
                    P[x] = NoArc;
                    x = StartNode(a);   
                }

                while (w!=cu)
                {
                    a = P[w];
                    BalPush(a,0.5);
                    P[w] = NoArc;
                    if (w!=u) P[w^1] = NoArc;
                    w = StartNode(a);   
                }

                v = cu;

                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer," %ld",u);
                    LogEnd(LH,CT.logBuffer);
                    LogEntry(LOG_METH2,"...Odd cycle cancelled");
                    LH = LogStart(LOG_METH2,"");
                    k--;
                }

                #endif

                CT.Ping(Handle(),1);
            }
        }
    }

    Close(H);
    ReleasePredecessors();

    #if defined(_LOGGING_)

    if (CT.logMeth>1)
    {
        LogEnd(LH);
        sprintf(CT.logBuffer,"...Keeping %ld odd cycles uncancelled",k);
        LogEntry(LOG_METH2,CT.logBuffer);
    }

    #endif
    
    CloseFold(ModCycleCancel);

    if (pi!=NULL)
    {
        LogEntry(LOG_METH,"Symmetrizing node potentials...");

        for (TNode i=0;i<n1;i++)
        {
            TFloat thisPi = (pi[2*i]-pi[2*i+1])/2;
            pi[2*i] = thisPi;
            pi[2*i+1] = -thisPi;
        }
    }
}

 
void abstractBalancedFNW::MakeIntegral(TArc *PP,TNode u,TNode v) throw(ERRange)
{
    TNode w = v;
    while (w!=u)
    {
        TArc a = PP[w];
        BalPush(a^1,0.5);
        PP[w] = NoArc;
        PP[w^1] = NoArc;
        w = StartNode(a);   
    }
}
