
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "absbig.h"


abstractBiGraph::abstractBiGraph(TNode nn1,TNode nn2) throw(ERRange)
{
    n1 = nn1;
    n2 = nn2;
    CheckLimits();

    LogEntry(LOG_MEM,"...Abstract bigraph object allocated");
}


abstractBiGraph::~abstractBiGraph() throw()
{
    LogEntry(LOG_MEM,"...Abstract bigraph object disallocated");
}


unsigned long abstractBiGraph::Allocated() throw()
{
    return 0;
}


void abstractBiGraph::CheckLimits() throw(ERRange)
{
    #if defined(_FAILSAVE_)
    
    abstractMixedGraph::CheckLimits();
    
    if (2*n1>=CT.MaxNode())
    {
        InternalError("RandomArcs","Number of outer nodes is out of range");
    }

    if (2*n2>=CT.MaxNode())
    {
        InternalError("RandomArcs","Number of inner nodes is out of range");
    }

    #endif
}


bool abstractBiGraph::Outer(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=N()) NoSuchNode ("Outer",v);

    #endif

    return v<N1();
}


bool abstractBiGraph::Inner(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=N()) NoSuchNode ("Inner",v);

    #endif

    return v>=N1();
}


void abstractBiGraph::RandomArcs(TArc mm) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (m+mm>=CT.MaxArc() || (!CT.randParallels && m+mm>TArc(n1*n2)))
        Error(ERR_REJECTED,"RandomArcs","Number of arcs is out of range");

    #endif

    if (CT.logMan && mm>0)
    {
        sprintf(CT.logBuffer,"Generating %ld edges...",m);
        LogEntry(LOG_MAN,CT.logBuffer);
    }
 
    TArc i=0;
    while (i<mm)
    {
        TNode u = (TNode)CT.Rand(n1);
        TNode v = n1+(TNode)CT.Rand(n2);

        TArc a = NoArc;
        if (!CT.randParallels && m>0) a = Adjacency(u,v);
        
        if (a==NoArc)
        {
            ReleaseIterators();
            InsertArc(u,v);
            i++;
        }
    }
}
    

void abstractBiGraph::WriteBiGraph(const char* wrFile,TWriteInc options) 
    throw(ERFile)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    if (CT.logIO)
    {
        sprintf(CT.logBuffer,"Writing bigraph to \"%s\"...",wrFile);
        LogEntry(LOG_IO,CT.logBuffer);
    }
    
    goblinExport F(wrFile,CT);
    
    if (options==WRITE_INCIDENCES) F.StartTuple("bigraph",0);
    else F.StartTuple("dense_bigraph",0);
    
    F.StartTuple("definition",0);
    
    F.StartTuple("nodes",1);
    F.MakeIntItem(n,0);
    F.MakeIntItem(n1,0);
    F.MakeIntItem(ni,0);
    F.EndTuple();
    
    if (options==WRITE_INCIDENCES)
    {
        F.StartTuple("arcs",1);
        F.MakeIntItem(m,0);
        F.EndTuple();
    
        WriteIncidences(&F);
    }

    WriteUCap(&F);
    WriteLCap(&F);
    WriteDemand(&F);
    WriteOrientation(&F);
    
    F.EndTuple();   // definition
    
    WriteObjectives(&F);
    WriteGeometry(&F);
    WriteLayout(&F);
    WriteSolutions(&F);
    
    CT.sourceNode = Source();
    CT.targetNode = Target();
    CT.rootNode = Root();
    F.WriteConfiguration(this);

    F.EndTuple();   // bigraph

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
}
