
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "absdig.h"
#include "statq.h"


abstractDiGraph::abstractDiGraph() throw()
{
    delta = 0;

    LogEntry(LOG_MEM,"...Abstract digraph allocated");
}


abstractDiGraph::~abstractDiGraph() throw()
{
    LogEntry(LOG_MEM,"...Abstract digraph disallocated");
}


unsigned long abstractDiGraph::Allocated() throw()
{
    return 0;
}


void abstractDiGraph::WriteDiGraph(const char* wrFile,TWriteInc options)
    throw(ERFile)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    if (CT.logIO)
    {
        sprintf(CT.logBuffer,"Writing digraph to \"%s\"...",wrFile);
        LogEntry(LOG_IO,CT.logBuffer);
    }
    
    goblinExport F(wrFile,CT);
    
    if (options==WRITE_INCIDENCES) F.StartTuple("digraph",0);
    else F.StartTuple("dense_digraph",0);
    
    F.StartTuple("definition",0);
    
    F.StartTuple("nodes",1);
    F.MakeIntItem(n,0);
    F.MakeIntItem(n,0);
    F.MakeIntItem(ni,0);
    F.EndTuple();
    
    if (options==WRITE_INCIDENCES)
    {
        F.StartTuple("arcs",1);
        F.MakeIntItem(m,0);
        F.EndTuple();
    
        WriteIncidences(&F);
    }

    WriteUCap(&F);
    WriteLCap(&F);
    WriteDemand(&F);
    WriteOrientation(&F);
    
    F.EndTuple();   // definition
    
    WriteObjectives(&F);
    WriteGeometry(&F);
    WriteLayout(&F);
    WriteSolutions(&F);
    
    CT.sourceNode = Source();
    CT.targetNode = Target();
    CT.rootNode = Root();
    F.WriteConfiguration(this);

    F.EndTuple();   // digraph

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
}


TFloat abstractDiGraph::RedLength(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("RedLength",a);

    #endif

    if (pi==NULL) return Length(a);
    else return Length(a)+pi[StartNode(a)]-pi[EndNode(a)];
}


void abstractDiGraph::Push(TArc a,TFloat Lambda) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("Push",a);

    #endif

    if (a&1) OmitArc(a,Lambda);
    else AddArc(a,Lambda);
}


TFloat abstractDiGraph::ResCap(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("ResCap",a);

    #endif
    
    if (a&1) return Flow(a)-LCap(a);
    else
    {
        if (UCap(a)==InfCap) return InfCap;
        else return UCap(a)-Flow(a);
    }
}


bool abstractDiGraph::Eligible(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("Eligible",a);

    #endif
    
    return ResCap(a)>delta;
}


void abstractDiGraph::Augment(TNode u,TNode v,TFloat Lambda)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("Augment",u);

    if (v>=n) NoSuchNode("Augment",v);

    if (Lambda<=0) Error(ERR_REJECTED,"Augment","Amount should be positive");

    if (P==NULL) Error(ERR_REJECTED,"Augment","No path specified");

    TNode PathLength = 0;
    
    #endif

    TNode w = v;

    #if defined(_LOGGING_)

    THandle LH = NoHandle;

    if (CT.logMeth>1)
    {
        sprintf(CT.logBuffer,"Augmenting by %g units of flow...",Lambda);
        LogEntry(LOG_METH2,CT.logBuffer);
        LogEntry(LOG_METH2,"Path in reverse order:");
        OpenFold();
        sprintf(CT.logBuffer,"(%ld",w);
        LH = LogStart(LOG_METH2,CT.logBuffer);
    }
    
    #endif
    
    bool searching = true;

    while (searching)
    {
        TArc a = P[w];
    
        #if defined(_FAILSAVE_)

        if (PathLength>=n || a==NoArc)
        {  
            InternalError("Augment","Missing start node");
        }

        PathLength++;

        #endif

        Push(a,Lambda);
        P[w] = NoArc;
        w = StartNode(a);
        if (w==u) searching = false;

        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"[%ld]%ld",a,w);
            LogAppend(LH,CT.logBuffer);
        }

        #endif
    }

    #if defined(_LOGGING_)

    if (CT.logMeth>1)
    {
        LogEnd(LH,")");
        CloseFold();
    }

    #endif
}
 
 
TFloat abstractDiGraph::FindCap(TNode u,TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("FindCap",u);

    if (v>=n) NoSuchNode("FindCap",v);

    if (P==NULL) Error(ERR_REJECTED,"FindCap","No path specified");

    TNode PathLength = 0;
    
    #endif

    TNode w = v;
    TFloat Lambda = InfFloat;
    bool searching = true;

    while (searching)
    {
        TArc a = P[w];
    
        #if defined(_FAILSAVE_)

        if (PathLength>=n || a==NoArc)
        {
            InternalError("FindCap","Missing start node");
        }

        PathLength++;

        #endif

        if (ResCap(a)<Lambda) Lambda = ResCap(a);
        w = StartNode(a);
        if (w==u) searching = false;
    }
    
    return Lambda;
}

 
void abstractDiGraph::AdjustDivergence(TArc a,TFloat Lambda)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("AdjustDivergence",a);

    if (Lambda<0)
        Error(ERR_REJECTED,"AdjustDivergence","Amount should be non-negative");

    #endif
          
    TNode u = StartNode(a);
    TNode v = EndNode(a);

    if (sDegIn!=NULL)
    {
        if (a&1)
        {
            sDegIn[u] -= Lambda;
            sDegOut[v] -= Lambda;
        }
        else
        {
            sDegOut[u] += Lambda;
            sDegIn[v] += Lambda;
        }
    }
}


TFloat abstractDiGraph::FlowValue(TNode s,TNode t) throw(ERRange,ERCheck)
{
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("FlowValue",s);

    if (t>=n) NoSuchNode("FlowValue",t);

    #endif
    
    LogEntry(LOG_METH,"Checking the flow labels...");

    for (TNode v=0;v<n;v++)
    {
        if (Divergence(v)!=0 && !(v==s || v==t))
            Error(ERR_CHECK,"FlowValue","Not a legal st-flow");
    }

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Flow value: %g",Divergence(t));
        LogEntry(LOG_RES,CT.logBuffer);
    }

    return Divergence(t);
}


bool abstractDiGraph::Compatible() throw()
{
    LogEntry(LOG_METH,"Checking reduced length labels...");

    bool ret = true;
    for (TArc a=0;(ret && a<2*m);a++)
    {
        if (ResCap(a)>0 && RedLength(a)<0) ret = false;
    }
        
    if (CT.logRes)
    {
        if (ret) LogEntry(LOG_RES,"...solutions are compatible");
        else LogEntry(LOG_RES,"...solutions are not compatible");
    }
    
    return ret;
}


void abstractDiGraph::InitFlow() throw()
{
    for (TArc a=0;a<m;a++) Push(2*a+1,ResCap(2*a+1));
}
