
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1999.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "exp2xfig.h"
#include "exp2tk.h"
#include "spsmix.h"
#include "statq.h"
#include "statstaq.h"
#include "setfam.h"
#include "basheap.h"
#include "fibheap.h"
#include "binheap.h"


// ------------------------------------------------------


abstractMixedGraph::abstractMixedGraph(TNode nn,TArc mm) throw()
{
    n = nn;
    ni = 0;
    m = mm;
    CheckLimits();
    
    d = NULL;
    P = NULL;
    pi = NULL;
    colour = NULL;
    itCounter = 0;
    LH = RH = NoHandle;
    pIterator = NULL;
    partition = NULL;
    nHeap = NULL;
    adj = NULL;
    
    sDeg = NULL;
    sDegIn = NULL;
    sDegOut = NULL;
    
    face = NULL;
    aExterior = NoArc;
    
    mu = 0;
    
    sourceNode = NoNode;
    targetNode = NoNode;
    rootNode = NoNode;

    LogEntry(LOG_MEM,"...Abstract mixed graph allocated");
}


abstractMixedGraph::~abstractMixedGraph() throw()
{
    ReleaseIterators();
    ReleasePredecessors();
    ReleaseLabels();
    ReleasePartition();
    ReleasePotentials();
    ReleaseColours();
    ReleaseDegrees();
    ReleaseAdjacencies();
    ReleaseEmbedding();

    delete[] pIterator;

    LogEntry(LOG_MEM,"...Abstract mixed graph disallocated");
}


void abstractMixedGraph::CheckLimits() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (n>=CT.MaxNode())
    {
        sprintf(CT.logBuffer,"Number of nodes is out of range: %ld",n);
        Error(ERR_REJECTED,"CheckLimits",CT.logBuffer);
    }

    if (2*m>CT.MaxArc()-2)
    {
        sprintf(CT.logBuffer,"Number of arcs is out of range: %ld",m);
        Error(ERR_REJECTED,"CheckLimits",CT.logBuffer);
    }

    #endif
}


unsigned long abstractMixedGraph::Allocated() throw()
{
    unsigned long tmpSize = 0;

    if (d!=NULL)            tmpSize += n*sizeof(TFloat);
    if (pi!=NULL)           tmpSize += n*sizeof(TFloat);
    if (colour!=NULL)       tmpSize += n*sizeof(TNode);
    if (P!=NULL)            tmpSize += n*sizeof(TArc);
    if (pIterator!=NULL)    tmpSize += itCounter*sizeof(goblinIterator**);
    if (sDeg!=NULL)         tmpSize += n*sizeof(TFloat);
    if (sDegIn!=NULL)       tmpSize += n*sizeof(TFloat);
    if (sDegOut!=NULL)      tmpSize += n*sizeof(TFloat);
    if (face!=NULL)         tmpSize += 2*m*sizeof(TNode);
    
    return tmpSize;
}


goblinQueue<TNode,TFloat> * abstractMixedGraph::NewNodeHeap() throw(ERRejected)
{
    if (CT.methPQ==0) return new basicHeap<TNode,TFloat>(n,CT);
    if (CT.methPQ==1) return new binaryHeap<TNode,TFloat>(n,CT);
    if (CT.methPQ==2) return new fibonacciHeap<TNode,TFloat>(n,CT);
    
    #if defined(_FAILSAVE_)

    UnknownOption("NewNodeHeap",CT.methPQ);

    #endif

    throw ERRejected();
}


goblinQueue<TArc,TFloat> * abstractMixedGraph::NewArcHeap() throw(ERRejected)
{
    if (CT.methPQ==0) return new basicHeap<TArc,TFloat>(m,CT);
    if (CT.methPQ==1) return new binaryHeap<TArc,TFloat>(m,CT);
    if (CT.methPQ==2) return new fibonacciHeap<TArc,TFloat>(m,CT);
    
    #if defined(_FAILSAVE_)

    UnknownOption("NewArcHeap",CT.methPQ);

    #endif

    throw ERRejected();
}


void abstractMixedGraph::RandomArcs(TArc mm) throw(ERRejected)
{                             
    #if defined(_FAILSAVE_)

    if (m+mm>=CT.MaxArc())
    {
        sprintf(CT.logBuffer,"Number of arcs is out of range: %ld",m);
        Error(ERR_REJECTED,"RandomArcs",CT.logBuffer);
    }

    #endif

    if (CT.logMan && mm>0)
    {
        sprintf(CT.logBuffer,"Generating %ld edges...",mm);
        LogEntry(LOG_MAN,CT.logBuffer);
    }
 
    TArc i=0;

    while (i<mm)
    {
        TNode u = (TNode)CT.Rand(n);
        TNode v = (TNode)CT.Rand(n);
        
        TArc a = NoArc;

        if (!CT.randParallels && m>0) a = Adjacency(u,v);

        if ((a==NoArc || UCap(a)==0) && (CT.randParallels || u!=v))
        {
            ReleaseIterators();
            InsertArc(u,v);
            i++;
        }
    }
}


void abstractMixedGraph::RandEulerian(TArc mm) throw(ERRejected)
{                             
    #if defined(_FAILSAVE_)

    if (m+mm>=CT.MaxArc() || mm<2)
    {
        sprintf(CT.logBuffer,"Number of arcs is out of range: %ld",m);
        Error(ERR_REJECTED,"RandEulerian",CT.logBuffer);
    }

    #endif

    if (CT.logMan)
    {
        sprintf(CT.logBuffer,"Generating eulerian cycle of length %ld...",mm);
        LogEntry(LOG_MAN,CT.logBuffer);
    }

    TNode start = (TNode)CT.Rand(n);
    TNode u = start;
    TNode v = start;
    
    TArc i;
    for (i=0;i<mm-1;i++)
    {
        for (char timeOut=100;timeOut>0;timeOut--)
        {
            while (u==v) v = TNode(CT.Rand(n));

            if (!CT.randParallels && m>0)
            {
                TArc a = Adjacency(u,v);
                if (a==NoArc || UCap(a)==0) timeOut = 0;
            }
            else timeOut = 0;
        }

        ReleaseIterators();
        InsertArc(u,v);
        u = v;
    }
    
    InsertArc(v,start);
}


void abstractMixedGraph::RandRegular(TNode k) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (n&k&1)
    {
        sprintf(CT.logBuffer,"Parity mismatch: n=%ld, k=%ld",n,k);
        Error(ERR_REJECTED,"RandRegular",CT.logBuffer);
    }

    #endif

    TNode minDeg = 0;
    TNode nMin = n;
    TNode nDef = n;
    
    TNode* dg = new TNode[n];

    for (TNode v=0;v<n;v++) dg[v]=0;
    
    while (minDeg<k)
    {
        while (nMin>0)
        {
            TNode shift = 1+CT.Rand(nMin);
            TNode v = n;

            while (shift>0)
            {
                v--;
                if (dg[v]==minDeg) shift--;
            }

            shift = 1+CT.Rand(nDef-1);
            TNode w = n;

            while (shift>0)
            {
                w--;
                if (dg[w]<k && v!=w) shift--;
            }

            nMin--;

            if (dg[w]==minDeg) nMin--;

            InsertArc(v,w);
            dg[v]++;
            dg[w]++;

            if (dg[v]==k) nDef--;
            if (dg[w]==k) nDef--;
            
        }
        
        minDeg++;

        for (TNode v=0;v<n;v++)
        {
            if (dg[v]==minDeg) nMin++;
        }
    }

    delete[] dg;
    
    if (CT.logMan)
    {
        sprintf(CT.logBuffer,"...Random %ld-regular graph generated",k);
        LogEntry(LOG_MAN,CT.logBuffer);
    }
}


void abstractMixedGraph::Merge(abstractMixedGraph& G) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (G.IsDense())
        Error(ERR_REJECTED,"Merge","Cannot merge dense graphs");

    #endif

    TFloat xShift = 1.5*CMax(0);
    TNode nOrig = n;

    for (TNode v=0;v<G.N();v++)
    {
        InsertNode();
        SetC(n-1,0,G.C(v,0)+xShift);
        SetC(n-1,1,G.C(v,1));
    }

    for (TArc a=0;a<G.M();a++)
    {
        TNode u = nOrig+G.StartNode(2*a);
        TNode v = nOrig+G.EndNode(2*a);
        
        TArc a2 = InsertArc(u,v,G.UCap(2*a),G.LCap(2*a),G.Length(2*a));

        if (!IsDirected() && !IsUndirected())
            SetOrientation(2*a2,G.Orientation(2*a));
    }
    
    if (CT.logMan)
    {
        sprintf(CT.logBuffer,"...Graph \"%s\" merged",G.Label());
        LogEntry(LOG_MAN,CT.logBuffer);
    }
}
    

void abstractMixedGraph::SetSourceNode(TNode s) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (s>=n && s!=NoNode) NoSuchNode("SetSourceNode",s);

    #endif

    sourceNode = s;
}
    

void abstractMixedGraph::SetTargetNode(TNode t) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (t>=n && t!=NoNode) NoSuchNode("SetTargetNode",t);

    #endif

    targetNode = t;
}
    

void abstractMixedGraph::SetRootNode(TNode r) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (r>=n && r!=NoNode) NoSuchNode("SetRootNode",r);

    #endif

    rootNode = r;
}
    

TNode abstractMixedGraph::EndNode(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("EndNode",a);

    #endif

    return StartNode(a^1);
}
                                  

TArc abstractMixedGraph::Reverse(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("Reverse",a);

    #endif

    return a^1;
}


bool abstractMixedGraph::Outer(TNode v) throw(ERRange,ERRejected)
{
    TNode dd = (TNode)floor(d[v]);

    #if defined(_FAILSAVE_)

    if (v>=N()) NoSuchNode("Outer",v);

    if (d==NULL || d[v]>=n || d[v]!=dd)
        Error(ERR_REJECTED,"Outer","Distance labels are currupted");

    #endif

    return dd%2==0;
}


bool abstractMixedGraph::Inner(TNode v) throw(ERRange,ERRejected)
{
    TNode dd = (TNode)floor(d[v]);

    #if defined(_FAILSAVE_)

    if (v>=N()) NoSuchNode("Inner",v);

    if (d==NULL || d[v]>=n || d[v]!=dd)
        Error(ERR_REJECTED,"Inner","Distance labels are currupted");
    
    #endif

    return dd%2==1;
}


TCap abstractMixedGraph::MaxUCap() throw()
{
    TCap max = 0;
    
    for (TArc a=0;a<m;a++)
        if (UCap(2*a)>max) max = UCap(2*a);

    return max;
}


TFloat abstractMixedGraph::MaxLength() throw()
{
    TFloat max = 0;
    
    for (TArc a=0;a<m;a++)
    {
        TFloat thisLength = Length(2*a);

        if (thisLength>max) max = thisLength;
        if (-thisLength>max) max = -thisLength;
    }

    return max;
}


TFloat abstractMixedGraph::RedLength(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("RedLength",a);

    #endif
    
    if (pi==NULL) return Length(a);
    else return Length(a)+pi[StartNode(a)]+pi[EndNode(a)];
}


TCap abstractMixedGraph::MaxLCap() throw()
{
    TCap max = 0;
    
    for (TArc a=0;a<m;a++)
        if (LCap(2*a)>max) max = LCap(2*a);

    return max;
}


bool abstractMixedGraph::HiddenArc(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("HiddenArc",a);

    #endif

    return HiddenNode(StartNode(a)) || HiddenNode(EndNode(a));
}


TArc abstractMixedGraph::Adjacency(TNode u,TNode v) throw(ERRange)
{                       
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("Adjacency",u);

    if (v>=n) NoSuchNode("Adjacency",v);

    #endif
    
    TArc a = NoArc;
    if (adj==NULL)
    {
        if (CT.methAdjacency==1)
        {
            if (m==0) a = NoArc;
            else
            {
                adj = new goblinHashTable<TArc,TArc>(n*n,2*m,NoArc,CT);

                for (TArc a1=0;a1<2*m;a1++)
                {
                    TNode u0 = StartNode(a1);
                    TNode v0 = EndNode(a1);
                    TArc a0 = u0*n+v0;

                    if (!Blocking(a1) && adj->Key(a0)==NoArc)
                        adj -> ChangeKey(a0,a1);
                }

                a = adj->Key(u*n+v);
            }
        }
        else
        {
            TArc a0 = First(u);

            if (EndNode(a0)==v && !Blocking(a0)) a = a0;

            while (Right(a0,u)!=First(u))
            {
                a0 = Right(a0,u);
                if (EndNode(a0)==v && !Blocking(a0) && a0<a) a = a0;
            }

        }
    }
    else a = adj->Key(u*n+v);
    
    if (a==NoArc)
    {
        #if defined(_LOGGING_)

        if (CT.logWarn)
        {
            sprintf(CT.logBuffer,"Nodes are non-adjacent: %ld, %ld",u,v);
            Error(MSG_WARN,"Adjacency",CT.logBuffer);
        }

        #endif
    
        return NoArc;
    }
    else
    {
        #if defined(_LOGGING_)

        if (CT.logRes>2)
        {
            sprintf(CT.logBuffer,
                "The nodes %ld and %ld are adjacent by the arc %ld",u,v,a);
            LogEntry(LOG_RES2,CT.logBuffer);
        }

        #endif
    
        return a;
    }
}

    
TCap abstractMixedGraph::Cardinality() throw()
{
    TCap sum = 0;

    for (TArc a=0;a<m;a++) sum += Sub(2*a);

    return sum;
}

    
TFloat abstractMixedGraph::Weight() throw()
{
    TFloat sum = 0;

    for (TArc a=0;a<m;a++) sum += Sub(2*a)*Length(2*a);

    return sum;
}

    
TFloat abstractMixedGraph::Length() throw()
{
    if (P==NULL) return 0;
    
    TFloat sum = 0;

    for (TNode v=0;v<n;v++)
        if (P[v]!=NoArc) sum += Length(P[v]);

    return sum;
}

    
void abstractMixedGraph::AddToSubgraph(TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (v>=n && v!=NoNode) NoSuchNode("AddToSubgraph",v);

    #endif
    
    if (v==NoNode)
    {
        for (TNode w=0;w<n;w++)
        {
            TArc a = P[w];

            if (a!=NoArc && LCap(a)==0) AddArc(a,1);
        }
    }
    else
    {
        TNode u = v;
        TArc a = NoArc;

        while ((a = P[u])!=NoArc && (u = StartNode(a))!=v) AddArc(a,1);

        if (a!=NoArc) AddArc(a,1);
    }
}

    
void abstractMixedGraph::ExtractTrees() throw(ERCheck)
{
    #if defined(_LOGGING_)

    LogEntry(LOG_METH2,"Extracting forest from subgraph...");

    #endif

    InitPredecessors();
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    for (TNode v=0;v<n;v++)
    {
        if (P[v]==NoArc)
        {
            TNode w = v;

            while (I.Active(w) || w!=v)
            {
                if (I.Active(w))
                {
                    TArc a = I.Read(w);
                    TNode u = EndNode(a);

                    if (Sub(a)>0 && a!=(P[w]^1))
                        if (P[u]==NoArc)
                        {
                            P[u] = a;
                            w = u;
                        }
                        else Error(ERR_CHECK,"ExtractTrees",
                                "Subgraph contains cycles");
                }
                else w = StartNode(P[w]);
            }
        }
    }

    Close(H);
}

    
void abstractMixedGraph::ExtractTree(TNode r,TMinTree options) throw(ERCheck)
{
    #if defined(_LOGGING_)

    LogEntry(LOG_METH2,"Extracting tree from subgraph...");

    #endif

    InitPredecessors();
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    TNode w = r;

    while (I.Active(w) || w!=r)
    {
        if (I.Active(w))
        {
            TArc a = I.Read(w);
            TNode u = EndNode(a);

            if (Sub(a)>0 && a!=(P[w]^1))
            {
                if (P[u]==NoArc)
                {
                    P[u] = a;

                    if (u!=r) w = u;
                }
                else
                {
                    Close(H);
                    Error(ERR_CHECK,"ExtractTree",
                        "Subgraph is neither a tree nor a 1-tree");
                }
            }
        }
        else w = StartNode(P[w]);
    }

    Close(H);
    
    for (TNode w=0;w<n;w++)
        if (w!=r && P[w]==NoArc)
            Error(ERR_CHECK,"ExtractTree","Subgraph is disconnected");

    if (P[r]==NoArc)
    {
        #if defined(_LOGGING_)
    
        LogEntry(LOG_RES2,"...Subgraph is a tree");

        #endif
        
        if ((options & MTREE_1TREE))
            Error(ERR_CHECK,"ExtractTree","Subgraph is not a 1-tree");
    }
    else
    {
        #if defined(_LOGGING_)
    
        LogEntry(LOG_RES2,"...Subgraph is a 1-tree");

        #endif
        
        if (!(options & MTREE_1TREE))
            Error(ERR_CHECK,"ExtractTree","Subgraph is not a tree");
    }
}

    
TNode abstractMixedGraph::ExtractPath(TNode u,TNode v) throw(ERCheck)
{
    #if defined(_LOGGING_)

    if (CT.logMeth>1)
    {
        sprintf(CT.logBuffer,"Extracting (%ld,%ld)-path from subgraph...",u,v);
        LogEntry(LOG_METH2,CT.logBuffer);
    }

    #endif

    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    TNode w = u;
    TNode count = 0;
    
    while (I.Active(w))
    {
        if (I.Active(w))
        {
            TArc a = I.Read(w);
            TNode x = EndNode(a);

            if (Sub(a)>0 && !Blocking(a) && a!=(P[w]^1))
            {
                if (P[x]==NoArc)
                {
                    P[x] = a;
                    w = x;
                    count++;

                    if (w==v) break;
                }
                else
                {
                    Close(H);
                    Error(ERR_CHECK,"ExtractPath","Unexpected branch");
                }
            }
        }
        else
        {
            Close(H);
            Error(ERR_CHECK,"ExtractPath","Missing end node");
        }
    }

    Close(H);

    #if defined(_LOGGING_)

    if (CT.logRes>1)
    {
        sprintf(CT.logBuffer,"...Path of length %ld found",count);
        LogEntry(LOG_RES2,CT.logBuffer);
    }

    #endif

    return count;
}
    

void abstractMixedGraph::Extract1Matching() throw(ERCheck)
{
    #if defined(_LOGGING_)

    LogEntry(LOG_METH2," Extracting 1-factor from subgraph...");

    #endif

    InitPredecessors();
    
    for (TArc a=0;a<m;a++)
    {
        if (Sub(2*a)>0)
        {
            TNode u = StartNode(2*a);
            TNode v = EndNode(2*a);

            if (P[u]==NoArc && P[v]==NoArc && Sub(2*a)==1) P[v] = 2*a;
            else
            {
                #if defined(_LOGGING_)

                Error(ERR_CHECK,"Extract1Matching",
                    "Subgraph contains adjacent arcs");

                #endif
            }
        }
    }

    #if defined(_LOGGING_)

    LogEntry(LOG_RES2,"...Subgraph is a 1-matching");

    #endif
}
    

TNode abstractMixedGraph::ExtractCycles() throw(ERCheck)
{
    #if defined(_LOGGING_)

    LogEntry(LOG_METH2,"Extracting 2-factor from subgraph...");

    #endif

    InitPredecessors();
    
    TNode count = 0;
    for (TNode v=0;v<n;v++)
    {
        if (Deg(v)+DegIn(v)+DegOut(v)>2 || DegIn(v)>1 || DegOut(v)>1)
            Error(ERR_CHECK,"ExtractCycles","Subgraph is not a 2-factor");
        
        if (P[v]==NoArc) {ExtractPath(v,v); count++;};
    }

    #if defined(_LOGGING_)

    if (CT.logRes>1)
    {
        sprintf(CT.logBuffer,"...Subgraph splits into %ld cycles",count);
        LogEntry(LOG_RES2,CT.logBuffer);
    }

    #endif
    
    return count;
}

    
void abstractMixedGraph::ExtractCut() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (d==NULL) Error(ERR_REJECTED,"ExtractCut","No distance labels found");

    #endif
    
    #if defined(_LOGGING_)

    LogEntry(LOG_METH2,"Extracting cut from distance labels...");

    #endif

    InitColours();
    
    for (TNode v=0;v<n;v++)
        if (d[v]==InfFloat) colour[v] = 1;
        else colour[v] = 0;
}

    
void abstractMixedGraph::ExtractBipartition() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (d==NULL) Error(ERR_REJECTED,"ExtractBipartition","No distance labels found");

    #endif
    
    #if defined(_LOGGING_)

    LogEntry(LOG_METH2,"Extracting bipartition from distance labels...");

    #endif

    InitColours();
    
    for (TNode v=0;v<n;v++)
        if ((((int)d[v])&1) || d[v]==InfFloat) colour[v] = 1;
        else colour[v] = 0;
}

    
void abstractMixedGraph::ExtractColours() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (partition==NULL) Error(ERR_REJECTED,"ExtractColours","No partition found");

    #endif
    
    #if defined(_LOGGING_)

    LogEntry(LOG_METH2,"Extracting colours from node partition...");

    #endif

    InitColours();
    
    TNode i = 0;
    for (TNode v=0;v<n;v++)
    {
        if (colour[Find(v)]==NoNode) colour[Find(v)] = i++;
        colour[v] = colour[Find(v)];
    }

    #if defined(_LOGGING_)

    if (CT.logRes>1)
    {
        sprintf(CT.logBuffer,"...%ld colour classes found",i);
        LogEntry(LOG_RES2,CT.logBuffer);
    };

    #endif
}

    
goblinIterator *abstractMixedGraph::NewIterator() throw()
{                             
    return new iGraph(*this);
}


goblinIterator &abstractMixedGraph::Iterator(THandle H) throw(ERRejected)
{                             
    #if defined(_FAILSAVE_)

    if ((H>=itCounter)||(pIterator[H]==NULL)) NoSuchHandle("Iterator",H);

    #endif
    
    return *pIterator[H];
}


THandle abstractMixedGraph::Investigate() throw(ERRejected)
{
    if (LH!=NoHandle)
    {
        THandle tmpHandle = LH;
        LH = NoHandle;
        pIterator[tmpHandle] -> Reset();
        return tmpHandle;
    }
    
    if (RH!=NoHandle)
    {
        THandle tmpHandle = RH;
        RH = NoHandle;
        pIterator[tmpHandle] -> Reset();
        return tmpHandle;
    }
    
    THandle i = 0;
    for (;((i<itCounter)&&(pIterator[i]!=NULL));i++) {};
    
    if (i==itCounter)
    {
        if (itCounter==NoHandle)
            Error(ERR_REJECTED,"Investigate","No more handles available");

        pIterator = (goblinIterator**)GoblinRealloc(pIterator,sizeof(goblinIterator*)*(itCounter+1));
        pIterator[itCounter] = NewIterator();
        return itCounter++;
    }
    else
    {
        pIterator[i] = NewIterator();
        return i;
    }
}


void abstractMixedGraph::Reset(THandle H) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if ((H>=itCounter)||(pIterator[H]==NULL)) NoSuchHandle("Reset",H);

    #endif
    
    pIterator[H]->Reset();
}


void abstractMixedGraph::Reset(THandle H, TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if ((H>=itCounter)||(pIterator[H]==NULL)) NoSuchHandle("Reset",H);

    #endif
    
    pIterator[H]->Reset(v);
}


TArc abstractMixedGraph::Read(THandle H, TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if ((H>=itCounter)||(pIterator[H]==NULL)) NoSuchHandle("Read",H);

    #endif
    
    return pIterator[H]->Read(v);
}


TArc abstractMixedGraph::Peek(THandle H, TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if ((H>=itCounter)||(pIterator[H]==NULL)) NoSuchHandle("Peek",H);

    #endif
    
    return pIterator[H]->Peek(v);
}


bool abstractMixedGraph::Active(THandle H, TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if ((H>=itCounter)||(pIterator[H]==NULL)) NoSuchHandle("Active",H);

    #endif
    
    return pIterator[H]->Active(v);
}


void abstractMixedGraph::Close(THandle H) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if ((H>=itCounter)&&(pIterator[H]==NULL)) NoSuchHandle("Close",H);

    #endif
    
    if ((LH==NoHandle)||(RH==NoHandle))
    {
        if (LH==NoHandle) LH = H;
        else RH = H;
        LogEntry(LOG_MEM,"Iterator cached");
    }
    else
    {
        delete pIterator[H];
        pIterator[H] = NULL;
        LogEntry(LOG_MEM,"Iterator disallocated");
    }
}


void abstractMixedGraph::ReleaseIterators() throw()
{                                
    if (LH!=NoHandle)
    {
        delete pIterator[LH];
        pIterator[LH] = NULL;
        LH = NoHandle;
    }
    
    if (RH!=NoHandle)
    {
        delete pIterator[RH];
        pIterator[RH] = NULL;
        RH = NoHandle;
    }

    for (THandle i=0;i<itCounter;i--)
        if (pIterator[i-1]!=NULL)
        {
            delete pIterator[i];
            pIterator[i] = NULL;
        }
    
    delete[] pIterator;
    pIterator = NULL;
    itCounter = 0;
    
    LogEntry(LOG_MEM,"Iterator cache cleared");
}


void abstractMixedGraph::StripIterators() throw()
{                                
    THandle i = itCounter;
    for (;((i>0)&&(pIterator[i-1]==NULL));i--) {};
    
    pIterator = (goblinIterator**)GoblinRealloc(pIterator,sizeof(goblinIterator*)*i);
    itCounter = i;
}


void abstractMixedGraph::Bud(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Bud",v);

    #endif
    
    partition -> Bud(v);                                
}


void abstractMixedGraph::Merge(TNode u,TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("Merge",u);

    if (v>=n) NoSuchNode("Merge",v);

    #endif
    
    partition -> Merge(u,v);                                
}


TNode abstractMixedGraph::Find(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Find",v);

    #endif
    
    if (partition==NULL) return v;
    else return partition -> Find(v);                                
}


TFloat abstractMixedGraph::Dist(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Dist",v);

    #endif
    
    if (d==NULL) return InfFloat;
    else return d[v];                                
}


TFloat abstractMixedGraph::Pi(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Pi",v);

    #endif
    
    if (pi==NULL) return 0;
    else return pi[v];                                
}


TNode abstractMixedGraph::Colour(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Colour",v);

    #endif
    
    if (colour==NULL) return NoNode;
    else return colour[v];                                
}


TArc abstractMixedGraph::EdgeColour(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("EdgeColour",a);

    #endif
    
    TFloat cc = Sub(a);

    if (cc>=m) return NoArc;
    else return TArc(cc);                                
}


TArc abstractMixedGraph::Pred(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Pred",v);

    #endif
    
    if (P==NULL) return NoArc;
    else return P[v];                                
}


TNode abstractMixedGraph::Face(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("Face",a);

    #endif
    
    if (!face) ExtractEmbedding();
    
    return face[a];
}


void abstractMixedGraph::SetSub(TArc a,TFloat multiplicity)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("SetSub",a);

    if (multiplicity>UCap(a) || multiplicity<LCap(a))
         AmountOutOfRange("SetSub",multiplicity);

    #endif
    
    if (multiplicity>Sub(a)) AddArc(a,multiplicity-Sub(a));
    else OmitArc(a,Sub(a)-multiplicity);
}


void abstractMixedGraph::SetDist(TNode v,TFloat thisLabel) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("SetDist",v);

    #endif
    
    if (d==NULL)
    {
        if (thisLabel!=InfFloat)
        {
            InitLabels();
            d[v] = thisLabel;
        }
    }
    else d[v] = thisLabel;
}


void abstractMixedGraph::SetPotential(TNode v,TFloat thisPi) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("SetPotential",v);

    #endif
    
    if (pi==NULL)
    {
        if (thisPi!=0)
        {
            InitPotentials();
            pi[v] = thisPi;
        }
    }
    else pi[v] = thisPi;
}


void abstractMixedGraph::SetColour(TNode v,TNode thisColour) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("SetColour",v);

    if (thisColour>=n && thisColour!=NoNode)
    {
        sprintf(CT.logBuffer,"Illegal assignment: %ld",thisColour);
        Error(MSG_WARN,"SetColour",CT.logBuffer);
    }

    #endif
    
    if (colour==NULL)
    {
        if (thisColour!=NoNode)
        {
            InitColours();
            colour[v] = thisColour;
        };
    }
    else colour[v] = thisColour;
}


void abstractMixedGraph::SetEdgeColour(TArc a,TArc thisColour) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("SetEdgeColour",a);

    if (thisColour>=m && thisColour!=NoArc)
    {
        sprintf(CT.logBuffer,"Illegal assignment: %ld",thisColour);
        Error(MSG_WARN,"SetColour",CT.logBuffer);
    }

    #endif
    
    if (!CUCap() || MaxUCap()!=m) SetCUCap(TCap(m));
    if (!CLCap() || MaxLCap()!=0) SetCUCap(0);
    
    if (thisColour==NoArc) thisColour = m;
    
    if (thisColour<Sub(a))
        OmitArc(a,Sub(a)-thisColour);
    else AddArc(a,thisColour-Sub(a));
}


void abstractMixedGraph::SetPred(TNode v,TArc thisPred)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("SetPred",v);

    if (thisPred>=2*m && thisPred!=NoArc) NoSuchArc("SetPred",thisPred);

    if (thisPred!=NoArc && EndNode(thisPred)!=v)
        Error(ERR_REJECTED,"SetPred","Mismatching end node");

    #endif
    
    if (P==NULL)
    {
        if (thisPred!=NoArc)
        {
            InitPredecessors();
            P[v] = thisPred;
        }
    }
    else P[v] = thisPred;
}


void abstractMixedGraph::PushPotential(TNode v,TFloat Epsilon) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("PushPotential",v);

    #endif
    
    if (pi==NULL) InitPotentials();
    pi[v] += Epsilon;                                
}


TFloat abstractMixedGraph::Deg(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Deg",v);

    #endif
    
    if (!sDeg) InitDegrees();
    
    return sDeg[v];
}


TFloat abstractMixedGraph::DegIn(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("DegIn",v);

    #endif
    
    if (!sDegIn) InitDegIO();
    
    return sDegIn[v];
}


TFloat abstractMixedGraph::DegOut(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("DegOut",v);

    #endif
    
    if (!sDegOut) InitDegIO();
    
    return sDegOut[v];
}


void abstractMixedGraph::AdjustDegrees(TArc a,TFloat Lambda) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("AdjustDegrees",a);

    #endif
          
    TNode u = StartNode(a);
    TNode v = EndNode(a);

    if (sDeg && !Blocking(a) && !Blocking(a^1))
    {
        sDeg[u] += Lambda;
        sDeg[v] += Lambda;
    }
    else
    {
        if (sDegIn)
        {
            if (!Blocking(a^1)) sDegIn[u] += Lambda;
            else sDegOut[u] += Lambda;

            if (!Blocking(a)) sDegIn[v] += Lambda;
            else sDegOut[v] += Lambda;
        }
    }
}


void abstractMixedGraph::InitDegrees() throw()
{
    if (sDeg==NULL)
    {
        LogEntry(LOG_MEM,"Generating degree labels...");
        sDeg = new TFloat[n];
        
        THandle H = Investigate();                                            
        goblinIterator &I = Iterator(H);

        for (TNode u=0;u<n;u++)
        {
            TFloat thisDeg = 0;

            while (I.Active(u))
            {
                TArc a = I.Read(u);
                if (!Blocking(a) && !Blocking(a^1)) thisDeg += Sub(a);
            }

            sDeg[u] = thisDeg;
        }

        Close(H);             
    }
}


void abstractMixedGraph::InitDegIO() throw()
{
    if (!sDegIn)
    {
        sDegIn =  new TFloat[n];
        sDegOut =  new TFloat[n];
        LogEntry(LOG_MEM,"Generating IO degree labels...");

        THandle H = Investigate();
        goblinIterator &I = Iterator(H);

        for (TNode v=0;v<n;v++)
        {
            TFloat thisDegIn = 0;
            TFloat thisDegOut = 0;

            while (I.Active(v))
            {
                TArc a = I.Read(v);
                if (Blocking(a)) thisDegIn += Sub(a);
                if (Blocking(a^1)) thisDegOut += Sub(a);
            }

            sDegIn[v] = thisDegIn;
            sDegOut[v] = thisDegOut;
        }

        Close(H);
    }
}


void abstractMixedGraph::InitLabels(TFloat def) throw()
{
    if (!d)
    {
        d =  new TFloat[n];
        LogEntry(LOG_MEM,"...Distance labels allocated");
    }
    else
    {
        #if defined(_LOGGING_)
    
        Error(MSG_WARN,"InitLabels","Labels are already present");

        #endif
    }
    
    for (TNode i=0;i<n;i++) d[i] = def;
}


void abstractMixedGraph::InitPotentials(TFloat def) throw()
{
    if (!pi)
    {
        pi = new TFloat[n];
        LogEntry(LOG_MEM,"...Node potentials allocated");
    }
    else
    {
        #if defined(_LOGGING_)

        Error(MSG_WARN,"InitPotentials","Node potentials are already present");

        #endif
    }
    
    for (TNode i=0;i<n;i++) pi[i] = def;
}


void abstractMixedGraph::InitColours(TNode def) throw()
{
    if (!colour)
    {
        colour = new TNode[n];
        LogEntry(LOG_MEM,"...Node colours allocated");
    }
    else
    {
        #if defined(_LOGGING_)

        Error(MSG_WARN,"InitColours","Node colours are already present");

        #endif
    }
    
    for (TNode i=0;i<n;i++) colour[i] = def;
}


void abstractMixedGraph::InitEdgeColours(TArc def) throw()
{
    for (TArc i=0;i<m;i++) SetEdgeColour(2*i,def);
}


void abstractMixedGraph::InitPredecessors() throw()
{
    if (!P)
    {
        P = new TArc[n];
        LogEntry(LOG_MEM,"...Predecessor labels allocated");
    }
    else
    {
        #if defined(_LOGGING_)

        Error(MSG_WARN,"InitPredecessors","Predecessor labels are already present");

        #endif
    }
    
    for (TNode i=0;i<n;i++) P[i] = NoArc;
}


void abstractMixedGraph::UpdatePotentials(TFloat alpha) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (!d)
        Error(ERR_REJECTED,"UpdatePotentials","No distance labels present");

    #endif

    if (pi)
    {
        for (TNode i=0;i<n;i++)
        {
            if (d[i]<alpha) pi[i]+=d[i];
            else pi[i]+=alpha;
        }
    }
    else
    {
        pi = new TFloat[n];

        for (TNode i=0;i<n;i++)
        {
            if (d[i]<alpha) pi[i]=d[i];
            else pi[i]=alpha;
        }
    }

    LogEntry(LOG_MEM,"...Node potentials updated");
}


void abstractMixedGraph::InitPartition() throw()
{
    if (!partition)
    {
        partition = new goblinSetFamily<TNode>(n,CT);
        LogEntry(LOG_MEM,"...Partition allocated");
    }
    else
    {
        partition -> Init();
    
        #if defined(_LOGGING_)

        Error(MSG_WARN,"InitPartition","A partition is already present");

        #endif
    }
}


void abstractMixedGraph::InitSubgraph() throw()
{
    for (TArc a=0;a<m;a++) OmitArc(2*a,Sub(2*a)-LCap(2*a));
    
    ReleaseDegrees();
}


void abstractMixedGraph::ReleaseAdjacencies() throw()
{
    if (adj)
    {
        delete adj;
        adj = NULL;
        LogEntry(LOG_MEM,"...Adjacencies disallocated");
    }
}


void abstractMixedGraph::ReleaseDegrees() throw()
{
    if (sDegIn)
    {
        delete[] sDegIn;
        sDegIn = NULL;
        LogEntry(LOG_MEM,"...Indegree labels disallocated");
    }

    if (sDegOut)
    {
        delete[] sDegOut;
        sDegOut = NULL;
        LogEntry(LOG_MEM,"...Outdegree labels disallocated");
    }

    if (sDeg)
    {
        delete[] sDeg;
        sDeg = NULL;
        LogEntry(LOG_MEM,"...Degree labels disallocated");
    }
}


void abstractMixedGraph::ReleaseLabels() throw()
{
    if (d)
    {
        delete[] d;
        d = NULL;
        LogEntry(LOG_MEM,"...Distance labels disallocated");
    }
}


void abstractMixedGraph::ReleasePotentials() throw()
{
    if (pi)
    {
        delete[] pi;
        pi = NULL;
        LogEntry(LOG_MEM,"...Node potentials disallocated");
    }
}


void abstractMixedGraph::ReleaseColours() throw()
{
    if (colour)
    {
        delete[] colour;
        colour = NULL;
        LogEntry(LOG_MEM,"...Node colours disallocated");
    }
}


void abstractMixedGraph::ReleaseEdgeColours() throw()
{
    InitSubgraph();
}


void abstractMixedGraph::ReleasePredecessors() throw()
{
    if (P)
    {
        delete[] P;
        P = NULL;
        LogEntry(LOG_MEM,"...Predecessor labels disallocated");
    }
}


void abstractMixedGraph::ReleaseEmbedding() throw()
{
    if (face)
    {
        delete[] face;
        face = NULL;
        LogEntry(LOG_MEM,"...Dual incidences disallocated");
    }
}


void abstractMixedGraph::ReleasePartition() throw()
{
    if (partition)
    {
        delete partition;
        partition = NULL;
        LogEntry(LOG_MEM,"...Partition disallocated");
    }
}


void abstractMixedGraph::ReleaseSolutions() throw()
{
    ReleaseColours();
    ReleaseLabels();
    ReleasePredecessors();
    ReleasePotentials();
    ReleasePartition();
    ReleaseEmbedding();
}


void abstractMixedGraph::WriteMixedGraph(const char* wrFile) throw(ERFile)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    sprintf(CT.logBuffer,"Writing mixed graph to \"%s\"...",wrFile);
    LogEntry(LOG_IO,CT.logBuffer);
    
    goblinExport F(wrFile,CT);
    
    F.StartTuple("mixed",0);
    F.StartTuple("definition",0);
    
    F.StartTuple("nodes",1);
    F.MakeIntItem(n,0);
    F.MakeIntItem(n,0);
    F.MakeIntItem(ni,0);
    F.EndTuple();
    
    F.StartTuple("arcs",1);
    F.MakeIntItem(m,0);
    F.EndTuple();
    
    WriteIncidences(&F);

    WriteUCap(&F);
    WriteLCap(&F);
    WriteDemand(&F);
    WriteOrientation(&F);
    
    F.EndTuple();   // definition
    
    WriteObjectives(&F);
    WriteGeometry(&F);
    WriteLayout(&F);
    WriteSolutions(&F);
    
    CT.sourceNode = Source();
    CT.targetNode = Target();
    CT.rootNode = Root();
    F.WriteConfiguration(this);

    F.EndTuple();   // mixed

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
}


void abstractMixedGraph::WriteIncidences(goblinExport* F) throw()
{
    F -> StartTuple("incidences",0);
    
    char mLength = CT.ExternalIntLength(2*m-1);

    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    for (TNode v=0;v<n;v++)
    {
        F -> StartTuple(v,10);
        
        while (I.Active(v))
        { 
            TArc a = I.Read(v);
            F -> MakeIntItem(a,mLength);
        }
                             
        F -> EndTuple();
    }

    Close(H);

    F -> EndTuple();
}


void abstractMixedGraph::WriteUCap(goblinExport* F) throw()
{
    if (CUCap())
    {
        F -> StartTuple("ucap",1);
        if (MaxUCap()==InfCap) F -> MakeNoItem(0);
        else F -> MakeFloatItem((TFloat)MaxUCap(),0);
    }
    else
    {
        char lLength = CT.ExternalFloatLength((TFloat)MaxUCap());
        F -> StartTuple("ucap",10);

        for (TArc a=0;a<m;a++)
            F -> MakeFloatItem((TFloat)UCap(2*a),lLength);
    }
    
    F -> EndTuple();
}


void abstractMixedGraph::WriteLCap(goblinExport* F) throw()
{
    if (CLCap())
    {
        F -> StartTuple("lcap",1);
        F -> MakeFloatItem((TFloat)MaxLCap(),0);
    }
    else
    {
        char lLength = CT.ExternalFloatLength((TFloat)MaxLCap());
        F -> StartTuple("lcap",10);

        for (TArc a=0;a<m;a++)
            F -> MakeFloatItem((TFloat)LCap(2*a),lLength);
    }
    
    F -> EndTuple();
}


void abstractMixedGraph::WriteDemand(goblinExport* F) throw()
{
    if (CDemand())
    {
        F -> StartTuple("demand",1);
        F -> MakeFloatItem(Demand(0),0);
    }
    else
    {
        char lLength = CT.ExternalFloatLength((TCap)MaxDemand());
        F -> StartTuple("demand",10);

        for (TNode v=0;v<n+ni;v++)
            F -> MakeFloatItem(Demand(v),lLength);
    }
    
    F -> EndTuple();
}


void abstractMixedGraph::WriteOrientation(goblinExport* F) throw()
{
    if (COrientation())
    {
        F -> StartTuple("directed",1);
        F -> MakeIntItem(Orientation(),0);
    }
    else
    {
        char lLength = 1;
        F -> StartTuple("directed",10);

        for (TArc a=0;a<m;a++)
            F -> MakeIntItem(Orientation(2*a),lLength);
    }
    
    F -> EndTuple();
}


void abstractMixedGraph::WriteObjectives(goblinExport* F) throw()
{
    F -> StartTuple("objectives",0);

    F -> StartTuple("commodities",1);
    F -> MakeIntItem(1,0);
    F -> EndTuple();

    F -> StartTuple("bound",1);
    F -> MakeNoItem(0);
    F -> EndTuple();

    F -> StartTuple("length",0);

    int swap = CT.methGeometry;
    CT.methGeometry = 0;

    if (CLength())
    {
        F -> StartTuple("comm0",1);
        F -> MakeFloatItem(MaxLength(),0);
    }
    else
    {
        char lLength = CT.ExternalFloatLength(MaxLength());
        F -> StartTuple("comm0",10);
        for (TArc a=0;a<m;a++)
            F -> MakeFloatItem(Length(2*a),lLength);
    }

    CT.methGeometry = swap;
    F -> EndTuple(); // 0
                                
    F -> EndTuple(); // length
                                
    F -> EndTuple(); // objectives
}


void abstractMixedGraph::WriteGeometry(goblinExport* F) throw()
{
    char* header = "geometry";
    F -> StartTuple(header,0);
    
    header = "metrics";
    F -> StartTuple(header,1);
    F -> MakeIntItem(CT.methGeometry,0);
    F -> EndTuple();
    
    header = "dim";
    F -> StartTuple(header,1);
    F -> MakeIntItem(Dim(),0);
    F -> EndTuple();
    
    if (Dim()==2)
    {
        header = "coordinates";
        F -> StartTuple(header,0);
        
        char xLength = 0;

        for (TNode v=0;v<n+ni;v++)
        {
            char thisLength = CT.ExternalFloatLength(C(v,0));

            if (thisLength>xLength) xLength = thisLength;

            thisLength = CT.ExternalFloatLength(C(v,1));

            if (thisLength>xLength) xLength = thisLength;
        }

        header = "axis0";
        F -> StartTuple(header,10);

        for (TNode v=0;v<n+ni;v++) F -> MakeFloatItem(C(v,0),xLength);

        F -> EndTuple();
        
        header = "axis1";
        F -> StartTuple(header,10);

        for (TNode v=0;v<n+ni;v++) F -> MakeFloatItem(C(v,1),xLength);

        F -> EndTuple();
        
        F -> EndTuple(); // coordinates
    }

    F -> EndTuple();
}


void abstractMixedGraph::WriteLayout(goblinExport* F) throw()
{
    F -> StartTuple("layout",0);

    F -> StartTuple("style",1);
    F -> MakeIntItem(CT.arcStyle,0);
    F -> EndTuple();

    char l = CT.ExternalIntLength(n+ni);

    if (CAlign())
    {
        F -> StartTuple("align",1);
        F -> MakeNoItem(0);
    }
    else
    {
        F -> StartTuple("align",10);

        for (TArc a=0;a<m;a++)
        {
            TNode x = Align(2*a);

            if (x==NoNode) F->MakeNoItem(l);
            else F -> MakeIntItem(x,l);
        }
    }

    F -> EndTuple();
    
    if (CInterpolate())
    {
        F -> StartTuple("interpolate",1);
        F -> MakeNoItem(0);
    }
    else
    {
        F -> StartTuple("interpolate",10);

        for (TNode v=0;v<n+ni;v++)
        {
            TNode x = Interpolate(v);
            if (x==NoNode) F->MakeNoItem(l);
            else F -> MakeIntItem(x,l);
        }
    }

    F -> EndTuple();
                                
    F -> EndTuple(); // layout
}


void abstractMixedGraph::WriteSolutions(goblinExport* F) throw()
{
    F -> StartTuple("solutions",0);
    
    WriteLabels(F);
    WritePredecessors(F);
    WriteSubgraph(F);
    WritePotentials(F);
    WriteColours(F);
    
    F -> EndTuple(); // solutions
}


void abstractMixedGraph::WriteLabels(goblinExport* F) throw()
{
    if (d==NULL)
    {
        F -> StartTuple("label",1);
        F -> MakeNoItem(0);
    }
    else
    {
        F -> StartTuple("label",10);
    

        char Length = 1;

        for (TNode v=0;v<n;v++)
        {
            char thisLength = CT.ExternalFloatLength(d[v]);

            if (thisLength>Length) Length = thisLength;
        }
    
        for (TNode v=0;v<n;v++)
            if (d[v]!=InfFloat)
                F->MakeFloatItem(d[v],Length);
            else F->MakeNoItem(10);
    }
    
    F -> EndTuple();
}


void abstractMixedGraph::WritePredecessors(goblinExport* F) throw()
{
    if (P==NULL)
    {
        F -> StartTuple("predecessor",1);
        F -> MakeNoItem(0);
    }
    else
    {
        char mLength = CT.ExternalIntLength(2*m-1);

        F -> StartTuple("predecessor",10);
    
        for (TNode v=0;v<n;v++)
            if (P[v]!=NoArc)
                F->MakeIntItem(P[v],mLength);
            else F->MakeNoItem(mLength);
    }
    
    F -> EndTuple();
}


void abstractMixedGraph::WriteSubgraph(goblinExport* F) throw()
{
    bool voidSubgraph = true;

    for (TArc a=0;a<m;a++)
        if (Sub(2*a)>0) voidSubgraph = 0;

    if (voidSubgraph)
    {
        F -> StartTuple("subgraph",1);
        F -> MakeFloatItem(TFloat(0),0);
    }
    else
    {
        F -> StartTuple("subgraph",10);

        char Length = 1;

        for (TArc a=0;a<m;a++)
        {
            char thisLength = CT.ExternalFloatLength(Sub(2*a));

            if (thisLength>Length) Length = thisLength;
        }

        for (TArc a=0;a<m;a++) F->MakeFloatItem(Sub(2*a),Length);
    }
    
    F -> EndTuple();
}


void abstractMixedGraph::WritePotentials(goblinExport* F) throw()
{
    if (pi==NULL)
    {
        F -> StartTuple("potential",1);
        F -> MakeNoItem(0);
    }
    else
    {
        F -> StartTuple("potential",10);
    
        char Length = 1;

        for (TNode v=0;v<n;v++)
        {
            char thisLength = CT.ExternalFloatLength(pi[v]);

            if (thisLength>Length) Length = thisLength;
        }
    
        for (TNode v=0;v<n;v++)
            if (pi[v]!=InfFloat)
                F->MakeFloatItem(pi[v],Length);
            else F->MakeNoItem(10);
    }
    
    F -> EndTuple();
}


void abstractMixedGraph::WriteColours(goblinExport* F) throw()
{
    if (colour==NULL)
    {
        F -> StartTuple("colour",1);
        F -> MakeNoItem(0);
    }
    else
    {
        F -> StartTuple("colour",10);
    
        char Length = 1;

        for (TNode v=0;v<n;v++)
        {
            char thisLength = CT.ExternalIntLength(colour[v]);

            if (thisLength>Length) Length = thisLength;
        }
    
        for (TNode v=0;v<n;v++)
            if (colour[v]!=NoNode)
                F->MakeIntItem(colour[v],Length);
            else F->MakeNoItem(10);
    };
    
    F -> EndTuple();
}


void abstractMixedGraph::ReadSolutions(goblinImport& F) throw()
{
    F.Scan("solutions");
    
    ReadLabels(&F);
    ReadPredecessors(&F);
    ReadSubgraph(&F);
    ReadPotentials(&F);
    ReadColours(&F);
    
    F.Scan(); // solutions
}


void abstractMixedGraph::ReadLabels(goblinImport* F) throw()
{
    ReleaseLabels();

    F -> Scan("label");
    d = F->GetTFloatTuple(n);

    if (F->Constant())
    {
        delete[] d;
        d = NULL;
    }
    else LogEntry(LOG_MEM,"...Distance labels allocated");
}


void abstractMixedGraph::ReadPredecessors(goblinImport* F) throw()
{
    ReleasePredecessors();

    F -> Scan("predecessor");
    P = F->GetTArcTuple(n);

    if (F->Constant())
    {
        delete[] P;
        P = NULL;
    }
    else LogEntry(LOG_MEM,"...Predecessor labels allocated");
}


void abstractMixedGraph::ReadSubgraph(goblinImport* F) throw(ERRange)
{
    F -> Scan("subgraph");
    TFloat *tmpSub = F->GetTFloatTuple(m);
    
    for (TArc a=0;a<m;a++)
    {
        TFloat thisSub = InfFloat;

        if (!F->Constant()) thisSub = tmpSub[a]; 
        else thisSub = tmpSub[0];
        
        if (thisSub<LCap(2*a) || thisSub>UCap(2*a))
        {
            sprintf(CT.logBuffer,"Arc multiplicity exeeds capacity bounds: %ld",a);
            Error(ERR_RANGE,"ReadSubgraph",CT.logBuffer);
        }

        TFloat delta = thisSub-Sub(2*a);

        if (delta<0) OmitArc(2*a,-delta);
        if (delta>0) AddArc(2*a,delta);
    }
        
    LogEntry(LOG_MEM,"...Subgraph loaded");
    
    delete[] tmpSub;
}


void abstractMixedGraph::ReadPotentials(goblinImport* F) throw()
{
    ReleasePotentials();

    F -> Scan("potential");
    pi = F->GetTFloatTuple(n);

    if (F->Constant())
    {
        delete[] pi;
        pi = NULL;
    }
    else LogEntry(LOG_MEM,"...Node potentials allocated");
}


void abstractMixedGraph::ReadColours(goblinImport* F) throw()
{
    ReleaseColours();

    F -> Scan("colour");
    colour = F->GetTNodeTuple(n);

    if (F->Constant())
    {
        delete[] colour;
        colour = NULL;
    }
    else LogEntry(LOG_MEM,"...Node colours allocated");
}


bool abstractMixedGraph::Eligible(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("Eligible",a);

    #endif
    
    return (UCap(a)>0 && !(Blocking(a)));
}


char* abstractMixedGraph::Display() throw(ERRejected,ERFile)
{
    if (CT.displayMode==0)
    {
        TextDisplay(NoNode);
        return NULL;
    }

    #if defined(_TRACING_)

    CT.IncreaseCounter();

    if (CT.displayMode==3)
    {
        char* gobName = new char[strlen(CT.Label())+15];
        sprintf(gobName,"%s.trace%d.gob",CT.Label(),CT.fileCounter);
        Write(gobName);
        if (CT.traceEventHandler) CT.traceEventHandler(gobName);
        delete[] gobName;

        return CT.Label();
    }

    if (CT.displayMode==1)
    {
        char* figName = new char[strlen(CT.Label())+10];
        sprintf(figName,"%s.%d.fig",CT.Label(),CT.fileCounter);
        Export2XFig(figName);
        delete[] figName;

        char* commandStr = new char[strlen(CT.Label())+15];
        sprintf(commandStr,"xfig %s.%d.fig &",CT.Label(),CT.fileCounter);
        system (commandStr);
        delete[] commandStr;
    }

    if (CT.displayMode==2)
    {
        char* tkName = new char[strlen(CT.Label())+10];
        sprintf(tkName,"%s.%d.tk",CT.Label(),CT.fileCounter);
        Export2Tk(tkName);
        delete[] tkName;

        char* commandStr = new char[strlen(CT.Label())+15];
        sprintf(commandStr,"wish display %s.%d &",CT.Label(),CT.fileCounter);
        system (commandStr);
        delete[] commandStr;
    }

    #endif

    return NULL;
}


void abstractMixedGraph::TextDisplay(TNode i,TNode j) throw()
{
    sprintf(CT.logBuffer,"Graph has %ld nodes and %ld arcs",n,m);
    LogEntry(MSG_TRACE,CT.logBuffer);

    if (m==0) return;
    
    TNode first = 0;
    TNode last = n-1;

    if (i!=NoNode)
    {
        first = i;

        if (j==NoNode) last = i;
        else last = j;
    }
    
    char larc = CT.ExternalIntLength(2*m)+1;
    char lnode = CT.ExternalIntLength(n)+4;
    char lfloat = CT.ExternalFloatLength(55555555555.5555555555)+1;
    char lrow = 2 + larc + lnode + lfloat;

    if (!CUCap()) lrow += lfloat;
    if (!CLCap()) lrow += lfloat;
    if (!CLength()) lrow += lfloat;
    if (pi!=NULL) lrow += lfloat;

    THandle LH = LogStart(MSG_TRACE2,"");

    for (short k=0;k<lrow;k++) LogAppend(LH,"_");

    LogEnd(LH);
    
    LH = LogStart(MSG_TRACE2,"");
    sprintf(CT.logBuffer,"  %*.*s",larc,larc,"arc");
    LogAppend(LH,CT.logBuffer);
    sprintf(CT.logBuffer,"%*.*s",lnode,lnode,"tail");
    LogAppend(LH,CT.logBuffer);

    if (!CUCap())
    {
        sprintf(CT.logBuffer,"%*.*s",lfloat,lfloat,"ucap");
        LogAppend(LH,CT.logBuffer);
    }

    sprintf(CT.logBuffer,"%*.*s",lfloat,lfloat,"subgraph");
    LogAppend(LH,CT.logBuffer);

    if (!CLCap())
    { 
        sprintf(CT.logBuffer,"%*.*s",lfloat,lfloat,"lcap");
        LogAppend(LH,CT.logBuffer);
    }

    if (!CLength())
    {
        sprintf(CT.logBuffer,"%*.*s",lfloat,lfloat,"length");
        LogAppend(LH,CT.logBuffer);
    }

    if (pi!=NULL)
    {
        sprintf(CT.logBuffer,"%*.*s",lfloat,lfloat,"redlength");
        LogAppend(LH,CT.logBuffer);
    }
    
    LogEnd(LH);
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    for (TNode v=first;v<=last;v++)
    {
        LH = LogStart(MSG_TRACE2,"");

        for (short k=0;k<lrow;k++) LogAppend(LH,"_");

        LogEnd(LH);
    
        LogEntry(MSG_TRACE2,"");
        sprintf(CT.logBuffer,"Node %ld",v);
        LH = LogStart(MSG_TRACE2,CT.logBuffer);

        if (CMax(0)!=0 || CMax(1)!=0)
        {
            sprintf(CT.logBuffer," (%g|%g)",C(v,0),C(v,1));
            LogAppend(LH,CT.logBuffer);
        }

        LogAppend(LH,":");

        bool first = true;

        if (d!=NULL)
        {
            LogAppend(LH," (");
            sprintf(CT.logBuffer,"Distance label %g",d[v]);
            LogAppend(LH,CT.logBuffer);
            first = false;
        }

        if (pi!=NULL)
        {
            if (first) LogAppend(LH," (");
            else LogAppend(LH,", ");

            sprintf(CT.logBuffer,"Node potential %g",pi[v]);
            LogAppend(LH,CT.logBuffer);
            first = false;
        }

        if (!first) LogAppend(LH,")");

        LogEnd(LH);

        LogEntry(MSG_TRACE2,"");
        
        while (I.Active(v))
        { 
            LH = LogStart(MSG_TRACE2,"");
            TArc a = I.Read(v);

            #if defined(_FAILSAVE_)

            if (StartNode(a)!=v)
            {
                sprintf(CT.logBuffer,"Incidence mismatch: %ld",a);
                Error(MSG_WARN,"TextDisplay",CT.logBuffer);
                sprintf(CT.logBuffer,"Concerning nodes %ld and %ld",v,StartNode(a));
                Error(MSG_WARN,"TextDisplay",CT.logBuffer);
            }
	    
            #endif
	    
            if (P!=NULL && P[v]==(a^1)) LogAppend(LH,"P");
            else LogAppend(LH," ");

            if (Blocking(a)) LogAppend(LH,"B");
            else LogAppend(LH," ");

            sprintf(CT.logBuffer,"%*.*g",larc,larc,double(a));
            LogAppend(LH,CT.logBuffer);
            sprintf(CT.logBuffer,"%*.*g",lnode,lnode,double(EndNode(a)));
            LogAppend(LH,CT.logBuffer);

            if (!CUCap())
            {
                sprintf(CT.logBuffer,"%*.*g",lfloat,lfloat,UCap(a));
                LogAppend(LH,CT.logBuffer);
            }

            sprintf(CT.logBuffer,"%*.*g",lfloat,lfloat,Sub(a));
            LogAppend(LH,CT.logBuffer);

            if (!CLCap())
            {
                sprintf(CT.logBuffer,"%*.*g",lfloat,lfloat,LCap(a));
                LogAppend(LH,CT.logBuffer);
            }

            if (!CLength())
            {
                sprintf(CT.logBuffer,"%*.*g",lfloat,lfloat,Length(a));
                LogAppend(LH,CT.logBuffer);
            }

            if (pi!=NULL)
            {
                sprintf(CT.logBuffer,"%*.*g",lfloat,lfloat,RedLength(a));
                LogAppend(LH,CT.logBuffer);
            }

            LogEnd(LH);
        }
    }

    LH = LogStart(MSG_TRACE2,"");

    for (short k=0;k<lrow;k++) LogAppend(LH,"_");
    
    LogEnd(LH);
    
    LogEntry(MSG_TRACE2,"");

    if (CUCap())
    {
        sprintf(CT.logBuffer,"Capacities: %g",MaxUCap());
        LogEntry(MSG_TRACE2,CT.logBuffer);
    }

    if (CLCap())
    {
        sprintf(CT.logBuffer,"Lower bounds: %g",MaxLCap());
        LogEntry(MSG_TRACE2,CT.logBuffer);
    }

    if (CLength())
    {
        sprintf(CT.logBuffer,"Length labels: %g",MaxLength());
        LogEntry(MSG_TRACE2,CT.logBuffer);
    }
    
    LH = LogStart(MSG_TRACE2,"");

    for (short k=0;k<lrow;k++) LogAppend(LH,"_");

    LogEnd(LH);
    
    Close(H);
}


void abstractMixedGraph::DisplayPath(TNode u,TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("DisplayPath",u);

    if (v>=n) NoSuchNode("DisplayPath",v);

    #endif
    
    TFloat l = 0;
    TNode count = 0;
    TNode w = v;
    
    LogEntry(LOG_RES,"Encoded path in reverse order:");
    sprintf(CT.logBuffer," (%ld",w);
    THandle LH = LogStart(LOG_RES,CT.logBuffer);

    while (w!=u || count==0)
    {
        TArc a = P[w];
        l += Length(a);
        w = StartNode(a);
        sprintf(CT.logBuffer,", %ld",w);
        LogAppend(LH,CT.logBuffer);
        count++;
	
        #if defined(_FAILSAVE_)
	
        if (count>n) Error(ERR_REJECTED,"DisplayPath","Missing start node");

        #endif
    }
    
    LogEnd(LH,")");
    sprintf(CT.logBuffer,"Total length: %g",l);
    LogEntry(LOG_RES,CT.logBuffer);
    sprintf(CT.logBuffer,"Total number of arcs: %ld",count);
    LogEntry(LOG_RES,CT.logBuffer);
}


TFloat abstractMixedGraph::CutCapacity(TFloat separator) throw(ERCheck)
{
    if (d==NULL) Error(ERR_CHECK,"CutCapacity","No edge cut specified");
    
    LogEntry(LOG_METH,"Checking the d-labels...");
    OpenFold();

    TFloat sCap = 0;                                            

    #if defined(_LOGGING_)

    THandle LH = LogStart(LOG_METH2,"Cut edges: ");

    #endif
    
    for (TArc a=0;a<2*m;a++)
    {
        if (d[StartNode(a)]<separator && d[EndNode(a)]>=separator
            && !Blocking(a) && (Blocking(a^1) || a%2==0))
        {
            #if defined(_LOGGING_)

            if (UCap(a)>0 && CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"%ld[%g] ",a,UCap(a));
                LogAppend(LH,CT.logBuffer);
            }

            #endif
    
            sCap += UCap(a);
        }
    }

    #if defined(_LOGGING_)

    LogEnd(LH);

    #endif
                 
    CloseFold();
    sprintf(CT.logBuffer,"...Total capacity: %g",sCap);
    LogEntry(LOG_RES,CT.logBuffer);

    return sCap;
}


void abstractMixedGraph::Export2XFig(const char* figName) throw(ERFile)
{
    sprintf(CT.logBuffer,"Writing xFig canvas to \"%s\"...",figName);
    LogEntry(LOG_IO,CT.logBuffer);
        
    exp2xfig E(*this,figName);
    ConfigDisplay(E.Configuration());
    E.DisplayGraph();
}


void abstractMixedGraph::Export2Tk(const char* figName) throw(ERFile)
{
    sprintf(CT.logBuffer,"Writing Tk canvas to \"%s\"...",figName);
    LogEntry(LOG_IO,CT.logBuffer);
        
    exp2tk E(*this,figName);
    ConfigDisplay(E.Configuration());
    E.DisplayGraph();
}


void abstractMixedGraph::Export2Ascii(const char* figName,TOption dummy)
    throw(ERFile)
{
    sprintf(CT.logBuffer,"Writing text form to \"%s\"...",figName);
    LogEntry(LOG_IO,CT.logBuffer);
        
    ofstream expFile(figName, ios::out);
    
    if (n==0) return;

    char larc = CT.ExternalIntLength(2*m)+1;
    char lnode = CT.ExternalIntLength(n)+4;
    char lfloat = CT.ExternalFloatLength(55555555555.5555555555)+1;
    char lrow = 2 + larc + lnode + lfloat;

    if (!CUCap()) lrow += lfloat;
    if (!CLCap()) lrow += lfloat;
    if (!CLength()) lrow += lfloat;
    if (pi!=NULL) lrow += lfloat;

    for (short k=0;k<lrow;k++) expFile << "-";

    expFile << endl;

    sprintf(CT.logBuffer,"  %*.*s",larc,larc,"arc");
    expFile << CT.logBuffer;
    sprintf(CT.logBuffer,"%*.*s",lnode,lnode,"tail");
    expFile << CT.logBuffer;

    if (!CUCap())
    {
        sprintf(CT.logBuffer,"%*.*s",lfloat,lfloat,"ucap");
        expFile << CT.logBuffer;
    }

    sprintf(CT.logBuffer,"%*.*s",lfloat,lfloat,"subgraph");
    expFile << CT.logBuffer;

    if (!CLCap())
    { 
        sprintf(CT.logBuffer,"%*.*s",lfloat,lfloat,"lcap");
        expFile << CT.logBuffer;
    }

    if (!CLength())
    {
        sprintf(CT.logBuffer,"%*.*s",lfloat,lfloat,"length");
        expFile << CT.logBuffer;
    }

    if (pi!=NULL)
    {
        sprintf(CT.logBuffer,"%*.*s",lfloat,lfloat,"redlength");
        expFile << CT.logBuffer;
    }
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    for (TNode v=0;v<n;v++)
    {
        expFile << endl;

        for (short k=0;k<lrow;k++) expFile << "-";

        sprintf(CT.logBuffer,"Node %ld",v);
        expFile << endl << CT.logBuffer;

        if (CMax(0)!=0 || CMax(1)!=0)
        {
            sprintf(CT.logBuffer," (%g|%g)",C(v,0),C(v,1));
            expFile << CT.logBuffer;
        }

        expFile << ":";

        bool first = true;

        if (d!=NULL)
        {
            expFile << " (";
            sprintf(CT.logBuffer,"Distance label %g",d[v]);
            expFile << CT.logBuffer;
            first = false;
        }

        if (pi!=NULL)
        {
            if (first) expFile << " (";
            else expFile << ", ";

            sprintf(CT.logBuffer,"Node potential %g",pi[v]);
            expFile << CT.logBuffer;
            first = false;
        }

        if (!first) expFile << ")";

        expFile << endl;
        
        while (I.Active(v))
        { 
            expFile << endl;
            TArc a = I.Read(v);

            #if defined(_FAILSAVE_)

            if (StartNode(a)!=v)
            {
                sprintf(CT.logBuffer,"Incidence mismatch: %ld",a);
                Error(MSG_WARN,"TextDisplay",CT.logBuffer);
                sprintf(CT.logBuffer,"Concerning nodes %ld and %ld",v,StartNode(a));
                Error(MSG_WARN,"TextDisplay",CT.logBuffer);
            }
	    
            #endif
	    
            if (P!=NULL && P[v]==(a^1)) expFile << "P";
            else expFile << " ";

            if (Blocking(a)) expFile << "B";
            else expFile << " ";

            sprintf(CT.logBuffer,"%*.*g",larc,larc,double(a));
            expFile << CT.logBuffer;
            sprintf(CT.logBuffer,"%*.*g",lnode,lnode,double(EndNode(a)));
            expFile << CT.logBuffer;

            if (!CUCap())
            {
                sprintf(CT.logBuffer,"%*.*g",lfloat,lfloat,UCap(a));
                expFile << CT.logBuffer;
            }

            sprintf(CT.logBuffer,"%*.*g",lfloat,lfloat,Sub(a));
            expFile << CT.logBuffer;

            if (!CLCap())
            {
                sprintf(CT.logBuffer,"%*.*g",lfloat,lfloat,LCap(a));
                expFile << CT.logBuffer;
            }

            if (!CLength())
            {
                sprintf(CT.logBuffer,"%*.*g",lfloat,lfloat,Length(a));
                expFile << CT.logBuffer;
            }

            if (pi!=NULL)
            {
                sprintf(CT.logBuffer,"%*.*g",lfloat,lfloat,RedLength(a));
                expFile << CT.logBuffer;
            }
        }
    }

    expFile << endl;

    for (short k=0;k<lrow;k++) expFile << "-";

    bool constantsFound = false;

    if (CUCap())
    {
        sprintf(CT.logBuffer,"Capacities: %g",MaxUCap());
        expFile << endl << CT.logBuffer;
        constantsFound = true;
    }

    if (CLCap())
    {
        sprintf(CT.logBuffer,"Lower bounds: %g",MaxLCap());
        expFile << endl << CT.logBuffer;
        constantsFound = true;
    }

    if (CLength())
    {
        sprintf(CT.logBuffer,"Length labels: %g",MaxLength());
        expFile << endl << CT.logBuffer;
        constantsFound = true;
    }

    if (!constantsFound) return;
    
    expFile << endl;

    for (short k=0;k<lrow;k++) expFile << "-";

    expFile << endl;
}
