
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, February 2001.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "bbscheme.h"
    

template <class TIndex,class TObj>
goblinBBNode<TIndex,TObj>::goblinBBNode(TIndex nn,goblinController &thisContext,
        goblinBBScheme<TIndex,TObj> *thisScheme) throw() :
    goblinDataObject(thisContext)
{
    n = unfixed = nn;
    depth = n;
    solved = false;
    index = NoNode;
    scheme = thisScheme;
    
    LogEntry(LOG_MEM,"...Branch node instanciated");
}


template <class TIndex,class TObj>
goblinBBNode<TIndex,TObj>::~goblinBBNode() throw()
{
    LogEntry(LOG_MEM,"...Branch node disallocated");
}


template <class TIndex,class TObj>
unsigned long goblinBBNode<TIndex,TObj>::Allocated() throw()
{
    return 0;
}


template <class TIndex,class TObj>
TObj goblinBBNode<TIndex,TObj>::Objective() throw()
{
    if (!solved)
    {
        objective = SolveRelaxation();
        solved = true;
    }
        
    return objective;
}


#if defined(_TRACING_)

goblinBBTree::goblinBBTree(goblinController &thisContext) throw() :
    goblinDataObject(thisContext),
    abstractMixedGraph(TNode(0),TArc(0)),
    diGraph(TNode(0),TOption(0),thisContext)
{
    LogEntry(LOG_MAN,"Generating branch tree...");
}
    

void goblinBBTree::ConfigDisplay(goblinController &CFG) throw()
{
    CFG.predecessors = 0;
    CFG.arcLabels = 4;
    CFG.nodeColours = 2;
    CFG.subgraph = 2;
}

#endif


template <class TIndex,class TObj>
goblinBBScheme<TIndex,TObj>::goblinBBScheme(goblinBBNode<TIndex,TObj>* root,
    TObj aPrioriBound,char mode) throw() :
    goblinDataObject(root->Context())
{
    LogEntry(LOG_METH,"Branching...");
    OpenFold(ModBranch);

    nIterations = 0;
    nActive = 0;
    nDFS = 0;
    depth = root->depth;
    firstActive = NULL;
    bestBound = aPrioriBound;
    SetUpperBound(TimerBranch,bestBound);
    bestLower = root->Objective();

    if (bestLower<=bestBound) SetLowerBound(TimerBranch,bestLower);

    feasible = !(bestBound==root->Infinity());
    exhaustive = mode;
    root->scheme = this;
    
    #if defined(_TRACING_)

    if (CT.traceLevel>1) Tree = new goblinBBTree(CT);

    #endif
    
    LogEntry(LOG_MEM,"...B&B scheme instanciated");

    #if defined(_LOGGING_)
    
    if (CT.logMeth>1 && CT.logGaps==0)
    {
        LogEntry(LOG_METH2,"");
        LogEntry(LOG_METH2,"Iteration        Objective    Free  Status     Best Bound       Best Lower  Active  Select");
        LogEntry(LOG_METH2,"------------------------------------------------------------------------------------------");
    }

    #endif

    if (Inspect(root))
    {
         if (CT.logMeth>1 && CT.logGaps==0) LogEnd(LH,"  STOP");
         delete root;
    }
    else Optimize();

    if (CT.logMeth>1 && CT.logGaps==0) LogEntry(LOG_METH2,"");
    CloseFold(ModBranch);

    if (bestLower<=bestBound-1+CT.epsilon ||
        (bestLower<=bestBound+CT.epsilon && !feasible && nActive>0))
    {
        LogEntry(LOG_METH,"...Interrupted!");
    }
    else
    {
        bestLower = bestBound;
    }
    
    if (CT.logMeth==1 || CT.logGaps>0)
    {
        sprintf(CT.logBuffer,"...Total number of branch nodes: %ld",nIterations);
        LogEntry(LOG_METH,CT.logBuffer);
    }
        
    #if defined(_TRACING_)
    
    if (CT.traceLevel==2 && nIterations>2)
    {
        Tree -> TreeLayout();
        Tree -> Display();
    }

    #endif
}


template <class TIndex,class TObj>
goblinBBScheme<TIndex,TObj>::~goblinBBScheme() throw()
{
    goblinBBNode<TIndex,TObj> *thisNode = firstActive;

    while (thisNode!=NULL)
    {
        goblinBBNode<TIndex,TObj>* predNode = thisNode;
        thisNode = thisNode->succNode;
        delete predNode;
    }

    #if defined(_TRACING_)
    
    if (CT.traceLevel>1)
    {
        int savedLevel = CT.traceLevel;
        CT.traceLevel = 1;
        delete Tree;
        CT.traceLevel = savedLevel;
    }

    #endif

    LogEntry(LOG_MEM,"...B&B scheme disallocated");
}


template <class TIndex,class TObj>
unsigned long goblinBBScheme<TIndex,TObj>::Size() throw()
{
    return
          sizeof(goblinBBScheme<TIndex,TObj>)
        + Allocated();
}


template <class TIndex,class TObj>
unsigned long goblinBBScheme<TIndex,TObj>::Allocated() throw()
{
    return 0;
}


template <class TIndex,class TObj>
void goblinBBScheme<TIndex,TObj>::Optimize() throw()
{
    goblinBBNode<TIndex,TObj> *activeNode = NULL;
    goblinBBNode<TIndex,TObj> *leftChild = NULL;
    goblinBBNode<TIndex,TObj> *rightChild = NULL;

    while (CT.SolverRunning() && nActive>0 && nActive<unsigned(CT.maxBBNodes)*100 &&
            (exhaustive || !feasible) &&
            (bestLower<=bestBound-1+CT.epsilon || (bestLower<=bestBound+CT.epsilon && !feasible)) &&
            (CT.maxBBIterations==0 || nIterations<unsigned(CT.maxBBIterations)*1000)
          )
    {
        activeNode = PullActive();

        #if defined(_TRACING_)
    
        TNode activeIndex = activeNode->index;

        #endif
        
        TIndex i = activeNode->SelectVariable();
        char dir = 0;
        if (feasible) dir = activeNode->Direction2(i);
        else dir = activeNode->Direction1(i);
        
        leftChild = activeNode;
        rightChild = activeNode->Clone();
        
        rightChild -> Raise(i);
        leftChild -> Lower(i);
        
        bool deleteLeft = false;
        bool deleteRight = false;

        if (dir)
        {
            deleteLeft = Inspect(leftChild);
            if (CT.logMeth>1 && CT.logGaps==0) LogEnd(LH);
            deleteRight = Inspect(rightChild);
        }
        else
        {
            deleteRight = Inspect(rightChild);
            if (CT.logMeth>1 && CT.logGaps==0) LogEnd(LH);
            deleteLeft = Inspect(leftChild);
        }

        if (UpperBound(TimerBranch)>=bestLower)
            SetLowerBound(TimerBranch,bestLower);
        
        #if defined(_TRACING_)
    
        if (CT.traceLevel>1)
        {
            Tree->SetColour(activeIndex,TNode(PROCESSED));
            
            TArc a = 2*(Tree->InsertArc(activeIndex,leftChild->index));
            Tree -> SetPred(leftChild->index,a);
            Tree -> SetLength(a,i);
            
            a = 2*(Tree->InsertArc(activeIndex,rightChild->index));
            Tree -> SetPred(rightChild->index,a);
            Tree -> SetLength(a,i);

            if (CT.traceLevel==3 && nIterations>1) Tree->TreeLayout();
        }

        #endif
        
        if (deleteLeft) delete leftChild;
        if (deleteRight) delete rightChild;
    }

    if (CT.logMeth>1 && CT.logGaps==0) LogEnd(LH,"  STOP");
}


template <class TIndex,class TObj>
bool goblinBBScheme<TIndex,TObj>::Inspect(goblinBBNode<TIndex,TObj> *thisNode)
    throw()
{
    TObj thisObjective = thisNode->Objective();

    #if defined(_TRACING_)
        
    if (CT.traceLevel>1) thisNode->index = Tree->InsertNode();

    #endif
    
    #if defined(_LOGGING_)
    
    if (CT.logMeth>=2 && CT.logGaps==0)
    {
        sprintf(CT.logBuffer,"%9.1ld  ",nIterations);
        LH = LogStart(LOG_METH2,CT.logBuffer);

        if (thisObjective!=thisNode->Infinity())
        {
            sprintf(CT.logBuffer,"%15.10g",thisObjective);
            LogAppend(LH,CT.logBuffer);
        }
        else LogAppend(LH,"     INFEASIBLE");

        sprintf(CT.logBuffer,"  %6.1ld  ",thisNode->Unfixed());
        LogAppend(LH,CT.logBuffer);
    }

    #endif

    bool ret = false;

    if (thisObjective<=bestBound-1+CT.epsilon ||
        (thisObjective<=bestBound+CT.epsilon && thisObjective!=thisNode->Infinity() && !feasible))
    {
        if (thisNode->Feasible())
        {
            #if defined(_LOGGING_)
            
            if (CT.logMeth>=2 && CT.logGaps==0) LogAppend(LH,"SAVED ");

            #endif
            
            #if defined(_TRACING_)
    
            if (CT.traceLevel>1) Tree -> SetColour(thisNode->index,TNode(SAVED));

            #endif

            feasible = true;
            bestBound = thisNode->LocalSearch();
            SetUpperBound(TimerBranch,bestBound);
            thisNode -> SaveSolution();
            StripQueue();
            ret = true;
            nDFS = 0;
        }
        else
        {
            PushActive(thisNode);

            #if defined(_LOGGING_)
    
            if (CT.logMeth>=2 && CT.logGaps==0) LogAppend(LH,"QUEUED");

            #endif
            
            #if defined(_TRACING_)
    
            if (CT.traceLevel>1) Tree -> SetColour(thisNode->index,TNode(QUEUED));

            #endif
        }
    }
    else
    {
        ret = true;

        #if defined(_LOGGING_)
    
        if (CT.logMeth>=2 && CT.logGaps==0) LogAppend(LH,"CUTOFF");

        #endif
            
        #if defined(_TRACING_)
    
        if (CT.traceLevel>1) Tree -> SetColour(thisNode->index,TNode(CUTOFF));

        #endif
    }

    #if defined(_LOGGING_)
    
    if (CT.logMeth>=2 && CT.logGaps==0)
    {
        if (bestBound==thisNode->Infinity())
            sprintf(CT.logBuffer,"       UNSOLVED");
        else sprintf(CT.logBuffer,"%15.10g",bestBound);

        if (bestLower==thisNode->Infinity())
            sprintf(CT.logBuffer,"%s         UNSOLVED",CT.logBuffer);
        else sprintf(CT.logBuffer,"%s  %15.10g",CT.logBuffer,bestLower);

        sprintf(CT.logBuffer,"%s  %6.1ld",CT.logBuffer,nActive);
        LogAppend(LH,CT.logBuffer);
    }

    #endif

    nIterations++;
    
    return ret;
}


template <class TIndex,class TObj>
void goblinBBScheme<TIndex,TObj>::PushActive(goblinBBNode<TIndex,TObj> *thisNode)
    throw()
{
    thisNode->succNode = firstActive;
    firstActive = thisNode;
    nActive++;
    
    if (thisNode->Objective()<bestLower)
    {
        bestLower = thisNode->Objective();
    }
}


template <class TIndex,class TObj>
char goblinBBScheme<TIndex,TObj>::Phase() throw()
{
    if (nIterations<20*depth && exhaustive<2) return 0;
    if (nActive+depth+2>TIndex(CT.maxBBNodes)) return 3;
    if (3*nActive>2*TIndex(CT.maxBBNodes)) return 2;

    return 1;
}


template <class TIndex,class TObj>
goblinBBNode<TIndex,TObj> *goblinBBScheme<TIndex,TObj>::PullActive() throw()
{
    goblinBBNode<TIndex,TObj> *retNode = firstActive;
    
    if (Phase()==0 || Phase()==3)
    {
        firstActive = firstActive->succNode;
        nDFS++;

        #if defined(_LOGGING_)
    
        if (CT.logMeth>1 && CT.logGaps==0) LogEnd(LH,"  DEPTH");
        
        #endif
    }
    else
    {
        goblinBBNode<TIndex,TObj> *thisNode = firstActive;
        goblinBBNode<TIndex,TObj> *bestPredNode = NULL;
        bestLower = firstActive->Objective();
        nDFS = 0;

        while (thisNode!=NULL)
        {
            goblinBBNode<TIndex,TObj> *predNode = thisNode;
            thisNode = thisNode->succNode;
            if (thisNode!=NULL && thisNode->Objective()<bestLower)
            {
                bestPredNode = predNode;
                bestLower = thisNode->Objective();
            }
        }
        
        if (bestPredNode == NULL) firstActive = firstActive->succNode;
        else
        {
            retNode = bestPredNode->succNode;
            bestPredNode->succNode = retNode->succNode;
        }

        #if defined(_LOGGING_)
    
        if (CT.logMeth>1 && CT.logGaps==0) LogEnd(LH,"  BEST");
        
        #endif
    }
        
    bestLower = bestBound;
    goblinBBNode<TIndex,TObj> *thisNode = firstActive;

    while (thisNode!=NULL)
    {
        if (thisNode->Objective()<bestLower)
            bestLower = thisNode->Objective();
        thisNode = thisNode->succNode;
    }

    nActive--;

    return retNode;
}


template <class TIndex,class TObj>
void goblinBBScheme<TIndex,TObj>::StripQueue() throw()
{
    goblinBBNode<TIndex,TObj> *thisNode = firstActive;

    while (thisNode!=NULL)
    {
        goblinBBNode<TIndex,TObj> *predNode = thisNode;
        thisNode = thisNode->succNode;

        if (thisNode!=NULL && thisNode->Objective()>bestBound-1+CT.epsilon)
        {
            predNode->succNode = thisNode->succNode;
            delete thisNode;
            thisNode = predNode->succNode;
            nActive--;
        }
    }
}




template class goblinBBNode<TNode,TFloat>;
template class goblinBBScheme<TNode,TFloat>;

#if defined(_BIG_ARCS_)

template class goblinBBNode<TArc,TFloat>;
template class goblinBBScheme<TArc,TFloat>;

#endif
