
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "big2fnw.h"


big2FNW::big2FNW(abstractBiGraph& GC,TCap* pLower,TCap* pCap) throw() :
    goblinDataObject(GC.Context()),
    abstractMixedGraph(GC.N()+4,GC.M()+2*GC.N()+4),
    G(GC)
{                                                   
    Init();
    
    cap = pCap;
    lower = pLower;
    
    sumLower1 = 0;
    sumLower1 = 0;

    for (TNode v=0;v<n1;v++) sumLower1 += lower[v];
    for (TNode v=n1;v<n0;v++) sumLower2 += lower[v];

    if (pCap!=NULL)
    {
        sumUpper = 0;

        for (TNode v=0;v<n0;v++)
        {
            sumUpper += cap[v];
            cap[v] -= lower[v];
        }
    }
    else sumUpper = sumLower1+sumLower2;
    
    dgl = new TFloat[2*n0+4];
    SetDegrees();

    if (CT.traceLevel==2) Display();
}


big2FNW::big2FNW(abstractBiGraph& GC,TCap cCap) throw() :
    goblinDataObject(GC.Context()),
    abstractMixedGraph(GC.N()+4,GC.M()+2*GC.N()+4),
    G(GC)
{                                                   
    Init();

    cap = NULL;
    lower = NULL;
    
    ccap = cCap;
    sumUpper = n0*ccap;
    sumLower1 = n1*ccap;
    sumLower2 = n2*ccap;

    dgl = new TFloat[2*n0+4];
    SetDegrees();

    if (CT.traceLevel==2) Display();
}


big2FNW::big2FNW(abstractBiGraph& GC) throw() :
    goblinDataObject(GC.Context()),
    abstractMixedGraph(GC.N()+4,GC.M()+2*GC.N()+4),
    G(GC)
{                                                   
    Init();
    
    cap = NULL;
    lower = new TCap[n0];
    
    sumLower1 = 0;
    sumLower2 = 0;

    for (TNode v=0;v<n1;v++)
    {
        lower[v] = G.Demand(v);
        sumLower1 += lower[v];
    }

    for (TNode v=n1;v<n0;v++)
    {
        lower[v] = G.Demand(v);
        sumLower2 += lower[v];
    }

    sumUpper = sumLower1+sumLower2;
    
    dgl = new TFloat[2*n0+4];
    SetDegrees();

    if (CT.traceLevel==2) Display();
}


void big2FNW::Init() throw()
{
    LogEntry(LOG_MEM,"Transforming into a flow network...");
    if (!CT.logMem && CT.logMan)
        LogEntry(LOG_MAN,"Transforming into a flow network...");
    
    n0 = G.N();
    n1 = G.N1();
    n2 = G.N2();
    m0 = G.M();
    CheckLimits();
    
    t2 = n-4;
    s2 = n-3;
    t1 = n-2;
    s1 = n-1;
    ret1 = 2*m-2;
    art1 = 2*m-4;
    ret2 = 2*m-6;
    art2 = 2*m-8;
    
    bool feasible = true;

    for (TNode v=0;v<G.N() && feasible;v++)
        if ((cap!=NULL && G.Deg(v)>cap[v]) || (cap==NULL && G.Deg(v)>ccap))
            feasible = false;

    if (!feasible) G.InitSubgraph();
    
    G.MakeRef();
}


void big2FNW::SetDegrees() throw()
{ 
    dgl[(art1>>1)-m0] = dgl[(ret1>>1)-m0] = 0;
    TFloat cardinality = 0;
    TFloat surplus1 = 0;
    TFloat surplus2 = 0;
    THandle H = G.Investigate();

    for (TNode v=0;v<n0;v++)
    {
        dgl[v] = 0;

        while (G.Active(H,v))
        {
            TArc a = G.Read(H,v);
            dgl[v] += G.Sub(a);
        }

        cardinality += dgl[v];                                       
            
        TCap thisCap = UCap(2*(m0+v));

        if (dgl[v]<=thisCap) dgl[v+n0] = 0;
        else
        {
            dgl[v+n0] = dgl[v]-thisCap;
            dgl[v] = thisCap;
            if (v<n1) surplus1 += dgl[v+n0];
            else surplus2 += dgl[v+n0];
        }
    }

    G.Close(H);
        
    cardinality = cardinality/2;
    dgl[(ret2>>1)-m0] = cardinality;
    dgl[(art1>>1)-m0] = cardinality-surplus2;
    dgl[(art2>>1)-m0] = cardinality-surplus1;
    dgl[(ret1>>1)-m0] = dgl[(art1>>1)-m0]+dgl[(art2>>1)-m0];
}


unsigned long big2FNW::Size() throw()
{   
    return 
          sizeof(big2FNW)
        + goblinDataObject::Allocated()
        + abstractMixedGraph::Allocated()
        + abstractDiGraph::Allocated()
        + big2FNW::Allocated();
}


unsigned long big2FNW::Allocated() throw()
{   
    unsigned long tmpSize
        = n0*sizeof(TCap)            // lower[]
        + (2*n0+4)*sizeof(TFloat);   // dgl[]

    if (cap!=NULL)    tmpSize += n0*sizeof(TFloat);

    return tmpSize;
}


bool big2FNW::Perfect() throw()
{   
    bool solved = (G.N1()==G.N2());
    THandle H = Investigate();

    while (Active(H,s1))
    {
        TArc a = Read(H,s1);
        if (ResCap(a)>0) solved = 0;
    }

    Close(H);
    
    if (CT.logRes)
    {
        if (solved)
            LogEntry(LOG_RES,"...Flow corresponds to a perfect assignment");
        else
            LogEntry(LOG_RES,"...Flow does not correspond to a perfect assignment");
    }
       
    return solved;
}


big2FNW::~big2FNW() throw()
{
    if (CT.traceLevel==2) Display();
    
    G.ReleaseRef();

    if (cap!=NULL) delete[] cap;

    delete[] lower;
    delete[] dgl;
}
    

TNode big2FNW::StartNode(TArc a) throw(ERRange)
{
    TArc d2 = a/2;
    TArc r2 = a%2;
    
    if (d2<m0) return G.StartNode(a);
    
    if (d2<m0+n1)
    {
        if (r2) return (TNode)(d2-m0);
        else return s1;
    }

    if (d2<m0+n0)
    {
        if (r2) return t1;
        else return (TNode)(d2-m0);
    }

    if (d2<m0+n0+n1)
    {
        if (r2) return (TNode)(d2-m0-n0);
        else return s2;
    }

    if (d2<m0+2*n0)
    {
        if (r2) return t2;
        else return (TNode)(d2-m0-n0);
    }

    if (d2==(ret1/2))
    {
        if (r2) return s1;
        else return t1;    
    }

    if (d2==(art1/2))
    {
        if (r2) return t2;
        else return s1;    
    }

    if (d2==(art2/2))
    {
        if (r2) return t1;
        else return s2;    
    }

    if (d2==(ret2/2))
    {
        if (r2) return s2;
        else return t2;    
    }

    #if defined(_FAILSAVE_)

    NoSuchArc("StartNode",a);

    #endif
    
    throw ERRange();
}
                                  

TNode big2FNW::EndNode(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("EndNode",a);

    #endif

    return big2FNW::StartNode(a^1);
}


TArc big2FNW::First(TNode v) throw(ERRange)
{
    if (v<n1) return 2*(m0+v)+1;
    if (v<n0) return 2*(m0+v);

    if (v==t2) return ret2;
    if (v==s2) return art2;
    if (v==t1) return ret1;
    if (v==s1) return art1;

    #if defined(_FAILSAVE_)

    NoSuchNode("First",v);

    #endif
    
    throw ERRange();
}


TArc big2FNW::Right(TArc a,TNode u) throw(ERRange)
{
    #if defined(_FAILSAVE_)
             
    if (a>=2*m) NoSuchArc("Right",a);

    if (u!=StartNode(a))
    {
        InternalError("Right","Mismatching start nodes");
    }

    #endif
             
    if (u<n1)
    {
        if (a==2*(m0+u)+1) return 2*(m0+n0+u)+1;
        
        TArc af = G.First(u);

        if (a==2*(m0+n0+u)+1)
        {
            if (af==NoArc) return 2*(m0+u)+1;
            else return af;
        }
        
        TArc ar = G.Right(a,u);

        if (ar==af) return 2*(m0+u)+1;

        return ar;
    }

    if (u<n0)
    {
        if (a==2*(m0+u)) return 2*(m0+n0+u);
        
        TArc af = G.First(u);

        if (a==2*(m0+n0+u))
        {
            if (af==NoArc) return 2*(m0+u);
            else return af;
        }
        
        TArc ar = G.Right(a,u);

        if (ar==af) return 2*(m0+u);

        return ar;
    }

    if (u==s1)
    {
        if (a==(ret1^1)) return art1;
        if (a==art1) return 2*m0;
        if (a==2*(m0+n1)-2) return (ret1^1);
        return a+2;
    }

    if (u==t1)
    {
        if (a==ret1) return (art2^1);
        if (a==(art2^1)) return 2*(m0+n1)+1;
        if (a==2*(m0+n0)-1) return ret1;
        return a+2;
    }

    if (u==s2)
    {
        if (a==(ret2^1)) return art2;
        if (a==art2) return 2*(m0+n0);
        if (a==2*(m0+n0+n1)-2) return (ret2^1);
        return a+2;
    }

    if (u==t2)
    {
        if (a==ret2) return (art1^1);
        if (a==(art1^1)) return 2*(m0+n0+n1)+1;
        if (a==2*(m0+2*n0)-1) return ret2;
        return a+2;
    }

    #if defined(_FAILSAVE_)

    NoSuchNode("Right",u);

    #endif
    
    throw ERRange();
}


TCap big2FNW::Demand(TNode v) throw(ERRange)
{
    if (v<n0+2) return 0;

    if (v==t1) return sumLower1+sumLower2;
    if (v==s1) return -sumLower1-sumLower2;

    #if defined(_FAILSAVE_)

    NoSuchNode("Demand",v);

    #endif
    
    throw ERRange();
}


TCap big2FNW::UCap(TArc a) throw(ERRange)
{ 
    TArc d2 = a/2;

    if (d2<m0) return G.UCap(a);

    if (d2<m0+n0)
    {
        if (lower==NULL) return ccap;
        else return lower[d2-m0];
    }

    if (d2<m0+2*n0)
    {
        if (cap==NULL) return 0;
        else return cap[d2-m0-n0];
    }

    if (d2==(ret2/2)) return sumUpper;
    if (d2==(art1/2)) return sumLower2;
    if (d2==(art2/2)) return sumLower1;
    if (d2==(ret1/2)) return 0; // sumLower1+sumLower2;

    #if defined(_FAILSAVE_)

    NoSuchArc("UCap",a);

    #endif
    
    throw ERRange();
}


TFloat big2FNW::Length(TArc a) throw(ERRange)
{
    if (a<2*m0)
    {
        if (a&1) return -G.Length(a);
        else return G.Length(a);
    }

    if (a<2*m) return 0;
    
    #if defined(_FAILSAVE_)

    NoSuchArc("Length",a);

    #endif
    
    throw ERRange();
}


TFloat big2FNW::C(TNode v,char i) throw(ERRange)
{
    if (i==0)
    {
        if (v<n0) return G.C(v,0)+10;
        if (v==s1) return 0;
        if (v==t1) return G.CMax(0)+20;
        if (v==s2) return G.CMax(0)+15;
        if (v==t2) return 5;
    }
    else
    {
        if (v<n0) return G.C(v,1);
        if (v==s1) return G.CMax(1)/2;
        if (v==t1) return G.CMax(1)/2;
        if (v==s2) return G.CMax(1)+5;
        if (v==t2) return G.CMax(1)+5;
    }
    
    #if defined(_FAILSAVE_)
             
    NoSuchNode("C",v);
             
    #endif
    
    throw ERRange();
}
             
             
TFloat big2FNW::CMax(char i) throw(ERRange)
{
    if (G.Dim()==0) return 0;
    if (i==0) return G.CMax(0)+20;
    if (i==1) return G.CMax(1)+5;
    
    #if defined(_FAILSAVE_)
             
    NoSuchCoordinate("CMax",i);
             
    #endif
    
    throw ERRange();
}
             
             
bool big2FNW::HiddenNode(TNode v) throw(ERRange)
{
    if (v<n0) return G.HiddenNode(v);
    if (v<n) return true;
    
    #if defined(_FAILSAVE_)
             
    NoSuchNode("HiddenNode",v);
             
    #endif
    
    throw ERRange();
}
             
             
TFloat big2FNW::Flow(TArc a) throw(ERRange)
{ 
    TArc d2 = a/2;
    
    if (d2<m0) return G.Sub(a);
    if (d2<m-1) return dgl[d2-m0];
    if (d2==m-1) return 0;

    #if defined(_FAILSAVE_)

    NoSuchArc("Flow",a);

    #endif
    
    throw ERRange();
}


void big2FNW::Push(TArc a,TFloat Lambda) throw(ERRange,ERRejected)
{
    TArc d2 = a/2;
    
    #if defined(_FAILSAVE_)

    if (Lambda<0 || Lambda>ResCap(a)) AmountOutOfRange("Push",Lambda);

    if (d2>=m) NoSuchArc("Push",a);

    #endif

    if (d2<m0)
    {
        if (a&1) G.OmitArc(a,Lambda);
        else G.AddArc(a,Lambda);
    }
    else
    {
        TArc a0 = d2-m0;

        if (a&1) dgl[a0] = dgl[a0]-Lambda;
        else dgl[a0] = dgl[a0]+Lambda;
    }
    
    AdjustDivergence(a,Lambda);
}
