
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1999.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "binheap.h"
#include "treeView.h"
    

template <class TItem,class TKey>
binaryHeap<TItem,TKey>::binaryHeap(TItem nn,goblinController &thisContext)
    throw() : goblinDataObject(thisContext)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    n = nn;
    UNDEFINED = n+1;
    
    v = new TItem[n+1];
    index = new TItem[n];
    key = new TKey[n];
    
    maxIndex = 0;
    
    for (TItem v=0;v<n;v++) index[v] = UNDEFINED;

    LogEntry(LOG_MEM,"...Binary heap instanciated");

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif
}


template <class TItem,class TKey>
void binaryHeap<TItem,TKey>::Init() throw()
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    if (maxIndex*100<n)
    {
        while (!Empty()) Delete();
    }
    else
    {
        for (TItem v=0;v<n;v++) index[v] = UNDEFINED;
        
        maxIndex = 0;
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif
}


template <class TItem,class TKey>
unsigned long binaryHeap<TItem,TKey>::Size() throw()
{
    return
          sizeof(binaryHeap<TItem,TKey>)
        + goblinDataObject::Allocated()
        + binaryHeap::Allocated();
}       


template <class TItem,class TKey>
unsigned long binaryHeap<TItem,TKey>::Allocated() throw()
{
    return
          (2*n+1)*sizeof(TItem)
        + n*sizeof(TKey);
}


static THandle LH = NoHandle;

template <class TItem,class TKey>
char* binaryHeap<TItem,TKey>::Display() throw()
{
    if (CT.displayMode>0)
    {
        if (maxIndex==0) return NULL;
    
        goblinTreeView G(n,CT);
        G.InitPredecessors();
        G.InitColours();
        
        for (TItem i=1;i<=maxIndex;i++)
        {
            G.SetColour(v[i],0);
            G.SetDist(v[i],TFloat(key[v[i]]));
            if (i>1)

            {
                TArc a = G.InsertArc(TNode(v[i/2]),TNode(v[i]));
                G.SetPred(TNode(v[i]),2*a);
            }
        }
        
        G.TreeLayout();
        G.Display();
    }
    else
    {
        LogEntry(MSG_TRACE,"Binary heap");
        LH = LogStart(MSG_TRACE2,"    ");

    	if (maxIndex>0)
        {
            Display(1);
            LogEnd(LH);
        }
    	else LogEnd(LH,"---");
    }
   
    return NULL;
}


template <class TItem,class TKey>
void binaryHeap<TItem,TKey>::Display(TItem i) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (i>maxIndex || i==0) NoSuchIndex("Display",i);

    if (i!=index[v[i]])
    {
        sprintf(CT.logBuffer,"Inconsistent index: %ld",i);
        InternalError1("Display");
    }

    #endif

    TItem left = 2*i;
    TItem right = left+1;
    
    sprintf(CT.logBuffer,"%ld[%g]",v[i],key[v[i]]);
    LogAppend(LH,CT.logBuffer);
    
    if (left<=maxIndex)
    {
        LogAppend(LH," (");
        Display(left);
        LogAppend(LH," ");

        if (right<=maxIndex) Display(right);
        else LogAppend(LH,"*");

        LogAppend(LH,")");
    }
}


template <class TItem,class TKey>
binaryHeap<TItem,TKey>::~binaryHeap() throw()
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    delete[] v;
    delete[] index;
    delete[] key;

    LogEntry(LOG_MEM,"...Binary heap disallocated");

    #if defined(_TRACING_)

    if (CT.traceData) Display();

    #endif    
}


template <class TItem,class TKey>
void binaryHeap<TItem,TKey>::UpHeap(TItem i) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (i>maxIndex || i==0) NoSuchIndex("UpHeap",i);

    #endif

    TItem w = v[i];
    TKey alpha = key[w];
    
    TItem next;
    while (i>1 && key[v[(next = i>>1)]]>alpha)
    {
        v[i] = v[next];
        index[v[i]] = i;
        i = next;
    }

    v[i] = w;
    index[w] = i;
}


template <class TItem,class TKey>
void binaryHeap<TItem,TKey>::DownHeap(TItem i) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (i>maxIndex || i==0) NoSuchIndex("DownHeap",i);

    #endif

    TItem left = 2*i;
    TItem right = left+1;
    
    if (left<=maxIndex && key[v[left]]<key[v[i]])
    {
        TItem swap = v[i];
        v[i] = v[left];
        index[v[i]] = i;
        v[left] = swap;
        index[swap] = left;
        
        DownHeap(left);
    }
    
    if (right<=maxIndex && key[v[right]]<key[v[i]])
    {
        TItem swap = v[i];
        v[i] = v[right];
        index[v[i]] = i;
        v[right] = swap;
        index[swap] = right;
        
        DownHeap(right);
    }
}


template <class TItem,class TKey>
void binaryHeap<TItem,TKey>::Insert(TItem w,TKey alpha) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (w>=n) NoSuchIndex("Insert",w);

    if (maxIndex>=n) Error(ERR_REJECTED,"Insert","Heap overflow");

    #endif
    
    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    key[w] = alpha;
    index[w] = (++maxIndex);
    v[maxIndex] = w;
    
    UpHeap(maxIndex);

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif

    #if defined(_TRACING_)

    if (CT.traceData) Display();

    #endif    
}


template <class TItem,class TKey>
void binaryHeap<TItem,TKey>::Delete(TItem w) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (w>=n || index[w]==UNDEFINED) NoSuchItem("Delete",w);

    #endif

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    TItem i = index[w];
    index[w] = UNDEFINED;
    maxIndex--;

    if (i<=maxIndex)
    {
        v[i] = v[maxIndex+1];
        DownHeap(i);
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif

    #if defined(_TRACING_)

    if (CT.traceData) Display();

    #endif    
}


template <class TItem,class TKey>
TItem binaryHeap<TItem,TKey>::Delete() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (maxIndex==0) Error(ERR_REJECTED,"Delete","Heap is empty");

    #endif

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    TItem w = v[1];
    index[w] = UNDEFINED;
    TItem x = v[maxIndex];
    maxIndex--;

    if (maxIndex>0)
    {
        v[1] = x;
        index[x] = 1;
        DownHeap(1);
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif

    #if defined(_TRACING_)

    if (CT.traceData) Display();

    #endif    

    return w;
}


template <class TItem,class TKey>
TKey binaryHeap<TItem,TKey>::Key(TItem w) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (w>=n || index[w]==UNDEFINED) NoSuchItem("Key",w);

    #endif
    
    return key[w];
}


template <class TItem,class TKey>
void binaryHeap<TItem,TKey>::ChangeKey(TItem w,TKey alpha) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (w>=n || index[w]==UNDEFINED) NoSuchItem("ChangeKey",w);

    #endif
    
    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    if (alpha>key[w])
    {
        key[w] = alpha;
        DownHeap(index[w]);
    }
    else
    {
        key[w] = alpha;
        UpHeap(index[w]);
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif

    #if defined(_TRACING_)

    if (CT.traceData) Display();

    #endif    
}


template <class TItem,class TKey>
TItem binaryHeap<TItem,TKey>::Peek() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (maxIndex==0) Error(ERR_REJECTED,"Peek","Heap is empty");

    #endif
    
    return v[1];
}


template class binaryHeap<TNode,TFloat>;

#if defined(_BIG_ARCS_)

template class binaryHeap<TArc,TFloat>;

#endif
