
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "dnsbig.h"


denseBiGraph::denseBiGraph(TNode nn1,TNode nn2,TOption options,
    goblinController& thisContext) throw() :
    goblinDataObject(thisContext),
    abstractMixedGraph(nn1+nn2,nn1*nn2),
    abstractBiGraph(nn1,nn2),
    X(*this,options)
{
    SetCDemand(1);
    if (!CT.randUCap) X.SetCUCap(1);
    
    LogEntry(LOG_MEM,"...Dense bigraph instanciated");
}
    

denseBiGraph::denseBiGraph(const char* impFileName,goblinController& thisContext)
    throw(ERFile,ERParse) :
    goblinDataObject(thisContext), abstractMixedGraph((TNode)0,(TArc)0),
    abstractBiGraph((TNode)0,(TNode)0),
    X(*this,(TOption)0)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    LogEntry(LOG_IO,"Loading dense bigraph...");
    if (!CT.logIO && CT.logMem) LogEntry(LOG_MEM,"Loading dense bigraph...");
    
    goblinImport F(impFileName,CT);

    F.Scan("dense_bigraph");
    F.Scan("definition");

    F.Scan("nodes");
    TNode *nodes = F.GetTNodeTuple(3);
    n = nodes[0];
    n1 = nodes[1];
    ni = nodes[2];
    n2 = n-n1;
    m = n1*n2;
    delete[] nodes;    
    
    CheckLimits();
    X.ReDimension(n,m);
    
    X.ReadUCap(F);
    X.ReadLCap(F);                          
    X.ReadDemand(F);
    X.ReadOrientation(F);

    F.Scan(); // definition
    
    X.ReadObjectives(F);
    X.ReadGeometry(F);
    X.ReadLayout(F);
    ReadSolutions(F);
    F.ReadConfiguration();

    if (CT.sourceNode<n) SetSourceNode(CT.sourceNode);
    else SetSourceNode(NoNode);

    if (CT.targetNode<n) SetTargetNode(CT.targetNode);
    else SetTargetNode(NoNode);

    if (CT.rootNode<n) SetRootNode(CT.rootNode);
    else SetRootNode(NoNode);
    
    F.Scan(); // dense_bigraph
    
    int l = strlen(impFileName)-4;
    char* tmpLabel = new char[l+1];
    memcpy(tmpLabel,impFileName,l);
    tmpLabel[l] = 0;
    SetLabel(tmpLabel);
    delete[] tmpLabel;

    CT.SetMaster(Handle());

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
}


denseBiGraph::denseBiGraph(abstractBiGraph &G) throw() :
    goblinDataObject(G.Context()),
    abstractMixedGraph(G.N(),G.N1()*G.N2()),
    abstractBiGraph(G.N1(),G.N2()),
    X(*this,(TOption)0)
{
    LogEntry(LOG_MAN,"Converting into dense bigraph...");

    TArc m0 = G.M();
    X.SetCUCap(0);

    for (TArc i=0;i<m0;i++)
    {
        TArc a=2*i;
        InsertArc(G.StartNode(a),G.EndNode(a),G.UCap(a),G.LCap(a),G.Length(a));
    }
    
    if (CT.traceLevel==2) Display();
}


TArc denseBiGraph::InsertArc(TNode u,TNode v,TCap uu,TCap ll,TFloat cc)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("InsertArc",u);
    
    if (v>=n) NoSuchNode("InsertArc",v);

    if ((u>=n1 && v>=n1) || (u<n1 && v<n1))
        Error(ERR_REJECTED,"InsertArc","Illegal manipulation");

    #endif
    
    TArc a = u*n2+(v-n1);

    if (v<u)
    {
        #if defined(_LOGGING_)

        Error(MSG_WARN,"InsertArc","End nodes are flipped");

        #endif

        a = v*n2+(u-n1);
    }
    
    return X.InsertArc(a,uu,ll,cc);
}


TArc denseBiGraph::InsertArc(TNode u,TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("InsertArc",u);
    
    if (v>=n) NoSuchNode("InsertArc",v);

    if ((u>=n1 && v>=n1) || (u<n1 && v<n1))
        Error(ERR_REJECTED,"InsertArc","Illegal manipulation");

    #endif
    
    TArc a = u*n2+(v-n1);

    if (v<u)
    {
        #if defined(_LOGGING_)

        Error(MSG_WARN,"InsertArc","End nodes are flipped");

        #endif

        a = v*n2+(u-n1);
    }
    
    return X.InsertArc(a);
}


unsigned long denseBiGraph::Size() throw()
{   
    return 
          sizeof(denseBiGraph)
        + goblinDataObject::Allocated()
        + abstractMixedGraph::Allocated()
        + abstractGraph::Allocated()
        + abstractBiGraph::Allocated();
}


unsigned long denseBiGraph::Allocated() throw()
{   
    return 0;
}


denseBiGraph::~denseBiGraph() throw()
{
    LogEntry(LOG_MEM,"...Dense bigraph disallocated");
    
    if (CT.traceLevel==2) Display();
}


TArc denseBiGraph::Adjacency(TNode u,TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("Adjacency",u);
    
    if (v>=n) NoSuchNode("Adjacency",v);

    #endif
    
    TArc a = NoArc;

    if (u<v)
    {
        if (v>=n1 && u<n1) a = 2*(u*n2+(v-n1));
    }
    else
    {
        if (u>=n1 && v<n1) a = 2*(v*n2+(u-n1))+1;
    }
    
    if (a!=NoArc && UCap(a)==0) a = NoArc;
    
    #if defined(_LOGGING_)

    if (CT.logRes>2)
    {
        sprintf(CT.logBuffer,
            "The nodes %ld and %ld are adjacent by the arc %ld",u,v,a);
        LogEntry(LOG_RES2,CT.logBuffer);
    }

    #endif
    
    return a;
}


TNode denseBiGraph::StartNode(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("StartNode",a);

    #endif
    
    if (a&1) return n1+(TNode)(a>>1)%n2;
    else return (TNode)(a>>1)/n2;
}


TNode denseBiGraph::EndNode(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("EndNode",a);

    #endif
    
    if (a&1) return (TNode)(a>>1)/n2;
    else return n1+(TNode)(a>>1)%n2;
}


TArc denseBiGraph::First(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=2*n) NoSuchNode("First",v);

    #endif
    
    if (v<n1) return Adjacency(v,n1);
    else return Adjacency(v,0);
}


TArc denseBiGraph::Right(TArc a,TNode u) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("StartNode",a);
    
    if (u>=n) NoSuchNode("Right",u);

    #endif
    
    TNode v = EndNode(a);
    
    if (v==n1-1) return Adjacency(u,0);
    if (v==n-1) return Adjacency(u,n1);
    else return Adjacency(u,v+1);
}
