
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "dnsdig.h"


denseDiGraph::denseDiGraph(TNode nn,TOption options,
    goblinController& thisContext) throw() :
    goblinDataObject(thisContext),
    abstractMixedGraph(nn,nn*nn),
    X(*this,options)
{
    X.SetCDemand(0);
    X.SetCOrientation(1);
    if (!CT.randUCap) X.SetCUCap(1);

    LogEntry(LOG_MEM,"...Dense digraph instanciated");
}
    

denseDiGraph::denseDiGraph(const char* impFileName, goblinController& thisContext)
    throw(ERFile,ERParse) :
    goblinDataObject(thisContext),
    abstractMixedGraph((TNode)0,(TArc)0),
    X(*this,(TOption)0)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    LogEntry(LOG_IO,"Loading graph...");
    if (!CT.logIO && CT.logMem) LogEntry(LOG_MEM,"Loading dense digraph...");
    
    goblinImport F(impFileName,CT);

    F.Scan("dense_digraph");
    F.Scan("definition");

    F.Scan("nodes");
    TNode *nodes = F.GetTNodeTuple(3);
    n = nodes[0];
    ni = nodes[2];
    m = n*n;
    delete[] nodes;    
    
    CheckLimits();
    X.ReDimension(n,m);

    X.ReadUCap(F);
    X.ReadLCap(F);                          
    X.ReadDemand(F);
    X.ReadOrientation(F);

    F.Scan(); // definition
    
    X.ReadObjectives(F);
    X.ReadGeometry(F);
    X.ReadLayout(F);
    ReadSolutions(F);
    F.ReadConfiguration();

    if (CT.sourceNode<n) SetSourceNode(CT.sourceNode);
    else SetSourceNode(NoNode);

    if (CT.targetNode<n) SetTargetNode(CT.targetNode);
    else SetTargetNode(NoNode);

    if (CT.rootNode<n) SetRootNode(CT.rootNode);
    else SetRootNode(NoNode);
    
    F.Scan(); // dense_digraph

    X.SetCOrientation(1);
    
    int l = strlen(impFileName)-4;
    char* tmpLabel = new char[l+1];
    memcpy(tmpLabel,impFileName,l);
    tmpLabel[l] = 0;
    SetLabel(tmpLabel);
    delete[] tmpLabel;

    CT.SetMaster(Handle());

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
}


denseDiGraph::denseDiGraph(abstractDiGraph& G) throw() :
    goblinDataObject(G.Context()),
    abstractMixedGraph(G.N(),G.N()*G.N()),
    X(*this,(TOption)0)
{
    X.SetCOrientation(1);
    X.SetCUCap(0);
    
    LogEntry(LOG_MAN,"Converting into dense digraph...");

    TArc m0 = G.M();

    for (TArc i=0;i<m0;i++)
    {
        TArc a=2*i;
        InsertArc(G.StartNode(a),G.EndNode(a),G.UCap(a),G.LCap(a),G.Length(a));
    }
    
    if (CT.traceLevel==2) Display();
}


TArc denseDiGraph::InsertArc(TNode u,TNode v,TCap uu,TCap ll,TFloat cc)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("InsertArc",u);

    if (v>=n) NoSuchNode("InsertArc",v);

    #endif
    
    TArc a = u*n+v;
    
    return X.InsertArc(a,uu,ll,cc);
}


TArc denseDiGraph::InsertArc(TNode u,TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("InsertArc",u);

    if (v>=n) NoSuchNode("InsertArc",v);

    #endif
    
    TArc a = u*n+v;
    
    return X.InsertArc(a);
}


unsigned long denseDiGraph::Size() throw()
{   
    return
          sizeof(denseDiGraph)
        + goblinDataObject::Allocated()
        + abstractMixedGraph::Allocated()
        + abstractDiGraph::Allocated();
}


unsigned long denseDiGraph::Allocated() throw()
{   
    return 0;
}


denseDiGraph::~denseDiGraph() throw()
{
    LogEntry(LOG_MEM,"...Dense digraph disallocated");
    
    if (CT.traceLevel==2) Display();
}


TArc denseDiGraph::Adjacency(TNode u,TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("Adjacency",u);

    if (v>=n) NoSuchNode("Adjacency",v);                 

    #endif
    
    TArc a = 2*(u*n+v);
    
    #if defined(_LOGGING_)

    if (CT.logRes>2)
    {
        sprintf(CT.logBuffer,
            "The nodes %ld and %ld are adjacent by the arc %ld",u,v,a);
        LogEntry(LOG_RES2,CT.logBuffer);
    }

    #endif

    return a;
}


TNode denseDiGraph::StartNode(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("StartNode",a);

    #endif
    
    if (a&1) return (TNode)(a>>1)%n;
    else return (TNode)(a>>1)/n;
}


TNode denseDiGraph::EndNode(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("EndNode",a);

    #endif
    
    if (a&1) return (TNode)(a>>1)/n;
    else return (TNode)(a>>1)%n;
}


TArc denseDiGraph::First(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=2*n) NoSuchNode("First",v);

    #endif
    
    return Adjacency(v,0);
}


TArc denseDiGraph::Right(TArc a,TNode u) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("Right",a);
    
    if (u>=n) NoSuchNode("Right",u);

    #endif
    
    TNode v = EndNode(a);
    
    if (a&1)
    {
        if (v<n-1) return Adjacency(u,v+1);
        else return Adjacency(u,0);
    }
    else return Adjacency(v,u)^1;
}


distanceGraph::distanceGraph(abstractMixedGraph &G) throw(ERRejected) :
    goblinDataObject(G.Context()),
    abstractMixedGraph(G.N(),(TArc)G.N()*G.N()),
    denseDiGraph(G.N(),(TOption)0,G.Context())
{                             
    LogEntry(LOG_MAN,"Generating distance graph...");

    OpenFold(ModFloydWarshall,NO_INDENT);
    
    if (G.Dim()>=2)
    {
        for (TNode v=0;v<G.N();v++)
        {
            SetC(v,0,(G.C(v,0)));
            SetC(v,1,(G.C(v,1)));
        }
    }
    
    for (TNode u=0;u<n;u++)
        for (TNode v=0;v<n;v++)
            if (u==v) SetLength(Adjacency(u,v),0);
            else SetLength(Adjacency(u,v),InfFloat);

    for (TArc a=0;a<2*G.M();a++)
    {
        TNode u = G.StartNode(a);
        TNode v = G.EndNode(a);
        TArc a2 = Adjacency(u,v);
        if (G.Length(a)<Length(a2) && !G.Blocking(a))
            SetLength(a2,G.Length(a));
    }

    for (TNode u=0;u<n;u++)
        for (TNode v=0;v<n;v++)
            for (TNode w=0;w<n;w++)
            {
                TArc a1 = Adjacency(v,w);
                TArc a2 = Adjacency(v,u);
                TArc a3 = Adjacency(u,w);
                if (Length(a2)+Length(a3)<Length(a1))
                    SetLength(a1,Length(a2)+Length(a3));
            }

    CloseFold(ModFloydWarshall,NO_INDENT);
}
