
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "dnsgra.h"


// ------------------------------------------------------


denseGraph::denseGraph(TNode nn,TOption options,goblinController& thisContext)
    throw() :
    goblinDataObject(thisContext),
    abstractMixedGraph(nn,nn*(nn+1)/2),
    X(*this,options)
{
    SetCDemand(1);
    if (!CT.randUCap) X.SetCUCap(1);
    
    LogEntry(LOG_MEM,"...Dense graph instanciated");
};
    

denseGraph::denseGraph(const char* impFileName,goblinController& thisContext)
    throw(ERFile,ERParse) :
    goblinDataObject(thisContext),
    abstractMixedGraph((TNode)0,(TArc)0),
    X(*this,(TOption)0)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    LogEntry(LOG_IO,"Loading dense graph...");
    if (!CT.logIO && CT.logMem) LogEntry(LOG_MEM,"Loading dense graph...");
    
    goblinImport F(impFileName,CT);

    F.Scan("dense_graph");
    F.Scan("definition");

    F.Scan("nodes");
    TNode *nodes = F.GetTNodeTuple(3);
    n = nodes[0];
    ni = nodes[2];
    m = n*(n+1)/2;
    delete[] nodes;    
    
    CheckLimits();
    X.ReDimension(n,m);

    X.ReadUCap(F);
    X.ReadLCap(F);                          
    X.ReadDemand(F);
    X.ReadOrientation(F);

    F.Scan(); // definition
    
    X.ReadObjectives(F);
    X.ReadGeometry(F);
    X.ReadLayout(F);
    ReadSolutions(F);
    F.ReadConfiguration();

    if (CT.sourceNode<n) SetSourceNode(CT.sourceNode);
    else SetSourceNode(NoNode);

    if (CT.targetNode<n) SetTargetNode(CT.targetNode);
    else SetTargetNode(NoNode);

    if (CT.rootNode<n) SetRootNode(CT.rootNode);
    else SetRootNode(NoNode);
    
    F.Scan(); // dense_graph
    
    int l = strlen(impFileName)-4;
    char* tmpLabel = new char[l+1];
    memcpy(tmpLabel,impFileName,l);
    tmpLabel[l] = 0;
    SetLabel(tmpLabel);
    delete[] tmpLabel;

    CT.SetMaster(Handle());

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
};


denseGraph::denseGraph(abstractGraph& G) throw() :
    goblinDataObject(G.Context()),
    abstractMixedGraph(G.N(),G.N()*(G.N()+1)/2),
    X(*this,(TOption)0)
{
    LogEntry(LOG_MAN,"Converting into dense graph...");

    TArc m0 = G.M();

    for (TArc i=0;i<m0;i++)
    {
        TArc a = 2*i;
        InsertArc(G.StartNode(a),G.EndNode(a),G.UCap(a),G.LCap(a),G.Length(a));
    }
    
    if (CT.traceLevel==2) Display();
};
                          

denseGraph::~denseGraph() throw()
{
    LogEntry(LOG_MEM,"...Dense graph disallocated");
    
    if (CT.traceLevel==2) Display();
};
    

unsigned long denseGraph::Size() throw()
{   
    return
          sizeof(denseGraph)
        + goblinDataObject::Allocated()
        + abstractMixedGraph::Allocated()
        + abstractGraph::Allocated();
};


unsigned long denseGraph::Allocated() throw()
{   
    return 0;
};


TArc denseGraph::InsertArc(TNode u,TNode v,TCap uu,TCap ll,TFloat cc)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("InsertArc",u);
    
    if (v>=n) NoSuchNode("InsertArc",v);

    #endif

    TArc a = (u*(u+1)/2)+v;

    if (v>u)
    {
        #if defined(_LOGGING_)

        Error(MSG_WARN,"InsertArc","End nodes are flipped");

        #endif

        a = (v*(v+1)/2)+u;
    }
    
    return X.InsertArc(a,uu,ll,cc);
};


TArc denseGraph::InsertArc(TNode u,TNode v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("InsertArc",u);
    
    if (v>=n) NoSuchNode("InsertArc",v);

    #endif

    TArc a = (u*(u+1)/2)+v;

    if (v>u)
    {
        #if defined(_LOGGING_)

        Error(MSG_WARN,"InsertArc","End nodes are flipped");

        #endif

        a = (v*(v+1)/2)+u;
    }
    
    return X.InsertArc(a);
};


TArc denseGraph::Adjacency(TNode u,TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("Adjacency",u);
    
    if (v>=n) NoSuchNode("Adjacency",v);

    #endif
    
    TArc a = u*(u+1)+2*v;

    if (v>u) a = v*(v+1)+2*u+1;
    
    #if defined(_LOGGING_)

    if (CT.logRes>2)
    {
        sprintf(CT.logBuffer,
            "The nodes %ld and %ld are adjacent by the arc %ld",u,v,a);
        LogEntry(LOG_RES2,CT.logBuffer);
    }

    #endif

    return a;
};


TNode denseGraph::StartNode(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("StartNode",a);

    #endif
    
    if (a&1)
    {
        TNode u = TNode(floor(sqrt(a-0.75)-0.5));
        return (TNode(a)-(u*u)-u-1)/2;
    }
    else return TNode(floor(sqrt(a+0.25)-0.5));
};


TNode denseGraph::EndNode(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("EndNode",a);

    #endif
    
    if (a&1) return TNode(floor(sqrt(a-0.75)-0.5));
    else
    {
        TNode u = TNode(floor(sqrt(a+0.25)-0.5));
        return (TNode(a)-(u*u)-u)/2;
    };
};


TArc denseGraph::First(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=2*n) NoSuchNode("First",v);

    #endif
    
    return Adjacency(v,0);
};


TArc denseGraph::Right(TArc a,TNode u) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("StartNode",a);
    
    if (u>=n) NoSuchNode("Right",u);

    #endif
    
    TNode v = EndNode(a);
    
    if (u==v && !(a&1)) return a^1;
    
    if (v<n-1) return Adjacency(u,v+1);
    else return Adjacency(u,0);
};


metricGraph::metricGraph(abstractGraph &G) throw(ERRejected) :
    goblinDataObject(G.Context()),
    abstractMixedGraph(G.N(),(TArc)G.N()*(G.N()+1)/2),
    denseGraph(G.N(),(TOption)0,G.Context())
{                             
    LogEntry(LOG_MAN,"Generating metric graph...");
    OpenFold();

    if (G.Dim()>=2)
    {
        for (TNode v=0;v<G.N();v++)
        {
            SetC(v,0,(G.C(v,0)));
            SetC(v,1,(G.C(v,1)));
        }
    }

    for (TNode u=0;u<n;u++)
    {
        G.Dijkstra(u);
        for (TNode v=0;v<=u;v++)
        {
            if (u==v) SetLength(Adjacency(u,v),InfFloat);
            else SetLength(Adjacency(u,v),G.Dist(v));
        }
    }
    
    CloseFold();
};
