
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, September 1999.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "dnsstruc.h"


denseGraphStructure::denseGraphStructure(abstractMixedGraph &GC,
        TOption options) throw() :
    goblinDataObject(GC.Context()), genericGraphStructure(GC)
{
    clcap = 0;
    
    if (options & OPT_COMPLETE) cucap = 1;
    else cucap = 0;

    sub = NULL;
    
    LogEntry(LOG_MEM,"...Dense graph structure instanciated");
}
    

TArc denseGraphStructure::InsertArc(TArc aa,TCap cc,TCap bb,TFloat ll)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (G.IsReferenced())
        Error(ERR_REJECTED,"InsertArc","Object is referenced");

    if (aa>=mMax)
        NoSuchArc("InsertArc",aa);

    #endif

    if ((Length(2*aa)!=ll) && UCap(2*aa)>0)
        Error(MSG_WARN,"InsertArc","Labelling conflict");
    
    SetLength(2*aa,ll);

    SetUCap(2*aa,UCap(2*aa)+cc);
    
    if (bb>0)
    {
        SetLCap(2*aa,LCap(2*aa)+bb);
        
        if (LCap(2*aa)>sub->Key(aa)) sub -> ChangeKey(aa,LCap(2*aa));
    }
    
    return aa;
}


TArc denseGraphStructure::InsertArc(TArc aa)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (G.IsReferenced())
        Error(ERR_REJECTED,"InsertArc","Object is referenced");

    if (aa>=mMax)
        NoSuchArc("InsertArc",aa);

    #endif

    TCap c = MaxLength();

    if (UCap(2*aa)>0) c = Length(2*aa);
    else if (CT.randLength) c = TFloat(CT.SignedRand());
    
    TFloat u = 1;
    if (CT.randUCap) u = TCap(1+CT.UnsignedRand());
    
    TFloat l = 0;
    if (CT.randLCap) l = TCap(CT.Rand(1+int(u)));
    
    SetLength(2*aa,c);

    SetUCap(2*aa,UCap(2*aa)+u);
    SetLCap(2*aa,LCap(2*aa)+l);
    
    return aa;
}


unsigned long denseGraphStructure::Size()
    throw()
{   
    return
          sizeof(denseGraphStructure)
        + goblinDataObject::Allocated()
        + genericGraphStructure::Allocated()
        + denseGraphStructure::Allocated();
}


unsigned long denseGraphStructure::Allocated()
    throw()
{   
    return 0;
}


denseGraphStructure::~denseGraphStructure()
    throw()
{
    ReleaseSubgraph();
    
    LogEntry(LOG_MEM,"...Dense graph structure disallocated");
}


void denseGraphStructure::NewSubgraph(TArc card)
    throw(ERRejected)
{
    if (sub==NULL)
    {
        sub = new goblinHashTable<TArc,TFloat>(mMax,card,0,CT);
        LogEntry(LOG_MEM,"...Sparse subgraph allocated");
    }
    else Error(ERR_REJECTED,"NewSubgraph","A subgraph is already present");
}


void denseGraphStructure::ReleaseSubgraph()
    throw()
{
    if (sub!=NULL)
    {
        delete sub;
        LogEntry(LOG_MEM,"...Sparse subgraph disallocated");
        sub = NULL;
    }
}


void denseGraphStructure::AddArc(TArc a,TFloat Lambda)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("AddArc",a);

    if (Lambda<0 || Lambda>UCap(a)-Sub(a)) AmountOutOfRange("AddArc",Lambda);

    #endif
 
    if (sub==NULL) NewSubgraph();

    sub -> ChangeKey(a>>1,sub->Key(a>>1)+Lambda);
    G.AdjustDegrees(a,Lambda);
}


void denseGraphStructure::OmitArc(TArc a,TFloat Lambda)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("OmitArc",a);

    if (Lambda<0 || Lambda>Sub(a)-LCap(a)) AmountOutOfRange("OmitArc",Lambda);

    #endif

    if (sub==NULL) NewSubgraph();

    sub -> ChangeKey(a>>1,sub->Key(a>>1)-Lambda);
    G.AdjustDegrees(a,-Lambda);
}


TFloat denseGraphStructure::Sub(TArc a)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("Sub",a);

    #endif

    if (sub==NULL) return LCap(a);
    else return LCap(a)+sub->Key(a>>1);
}
