
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, May 2000.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "fibheap.h"
#include "treeView.h"

// status[x]==3 : x is not queued
// status[x]==2 : x is a root node
// status[x]==1 : x is a marked child node
// status[x]==0 : x is an unmarked child node
    

template <class TItem,class TKey>
fibonacciHeap<TItem,TKey>::fibonacciHeap(TItem nn,goblinController &thisContext)
    throw() : goblinDataObject(thisContext)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    n = nn;
    k = nn;
    UNDEFINED = n;
    
    pred = new TItem[n];
    first = new TItem[n];
    next = new TItem[n];
    nextLink = new TItem[n];
    previous = new TItem[n];
    rank = new TItem[n];
    status = new char[n];
    bucket = new TItem[k];
    key = new TKey[n];
    
    card = 0;
    minimal = UNDEFINED;
    firstLink = UNDEFINED;
    
    for (TItem v=0;v<n;v++) status[v] = 3;
    for (TItem v=0;v<k;v++) bucket[v] = UNDEFINED;
    for (TItem v=0;v<n;v++) nextLink[v] = UNDEFINED;

    LogEntry(LOG_MEM,"...Fibonacci heap instanciated");

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif
}


template <class TItem,class TKey>
void fibonacciHeap<TItem,TKey>::Init() throw()
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    if (card*100<n)
    {
        while (!Empty()) Delete();
    }
    else
    {
        for (TItem v=0;v<n;v++) status[v] = 3;
        for (TItem v=0;v<k;v++) bucket[v] = UNDEFINED;
        for (TItem v=0;v<n;v++) nextLink[v] = UNDEFINED;
        
        card = 0;
        minimal = UNDEFINED;
        firstLink = UNDEFINED;
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif
}


template <class TItem,class TKey>
unsigned long fibonacciHeap<TItem,TKey>::Size() throw()
{
    return
          sizeof(fibonacciHeap<TItem,TKey>)
        + goblinDataObject::Allocated()
        + fibonacciHeap::Allocated();
}


template <class TItem,class TKey>
unsigned long fibonacciHeap<TItem,TKey>::Allocated() throw()
{
    return
          (6*n+k)*sizeof(TItem)
        + n*sizeof(char)
        + n*sizeof(TKey);
}


static THandle LH = NoHandle;

template <class TItem,class TKey>
char* fibonacciHeap<TItem,TKey>::Display() throw()
{
    if (CT.displayMode>0)
    {
        if (card==0) return NULL;
    
        goblinTreeView G(n,CT);
        G.InitPredecessors();
        
        for (TItem i=0;i<n;i++)
        {
            if (status[i]==3) G.SetColour(i,NoNode);
            else
            {
                G.SetColour(i,TNode(status[i]));
                G.SetDist(i,TFloat(key[i]));
    
                TItem x = first[i];

                while (x!=UNDEFINED)
                {
                    #if defined(_FAILSAVE_)

                    if (status[x]>1)
                    {
                        sprintf(CT.logBuffer,"Inconsistent item: %ld",x);
                        InternalError1("Display");
                    }

                    #endif
                            
                    TArc a = G.InsertArc(TNode(i),TNode(x));
                    G.SetPred(TNode(x),2*a);
                    x = next[x];
                }
            }
        }
        
        G.TreeLayout();
        G.Display();
    }
    else
    {
        LogEntry(MSG_TRACE,"Fibonacci heap");
        LH = LogStart(MSG_TRACE2,"    ");
    
        if (card>0)
        {
            for (TItem i=0;i<k;i++)
            {
                if (bucket[i]!=UNDEFINED)
                {
                    #if defined(_FAILSAVE_)

                    if (status[bucket[i]]!=2)
                    {
                        sprintf(CT.logBuffer,"Inconsistent bucket: %ld",i);
                        InternalError1("Display");
                    }

                    #endif
                            
                    LogAppend(LH,"  ");
                    Display(bucket[i]);
                }
            }

            LogEnd(LH);
        }
        else LogEnd(LH,"---");
    }
    
    return NULL;
}


template <class TItem,class TKey>
void fibonacciHeap<TItem,TKey>::Display(TItem v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n || status[v]==3) NoSuchItem("Display",v);

    #endif

    sprintf(CT.logBuffer,"%ld[%g]",v,key[v]);
    LogAppend(LH,CT.logBuffer);
    
    TItem x = first[v];

    if (x!=UNDEFINED)
    {
        LogAppend(LH," (");

        while (x!=UNDEFINED)
        {
            #if defined(_FAILSAVE_)

            if (status[x]>1)
            {
                sprintf(CT.logBuffer,"Inconsistent item: %ld",x);
                InternalError1("Display");
            }

            #endif
                            
            Display(x);
            x = next[x];

            if (x!=UNDEFINED) LogAppend(LH," ");
        }

        LogAppend(LH,")");
    }
}


template <class TItem,class TKey>
fibonacciHeap<TItem,TKey>::~fibonacciHeap() throw()
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    delete[] pred;
    delete[] first;
    delete[] next;
    delete[] previous;
    delete[] rank;
    delete[] status;
    delete[] bucket;
    delete[] nextLink;
    delete[] key;

    LogEntry(LOG_MEM,"...Fibonacci heap disallocated");

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif
}


template <class TItem,class TKey>
void fibonacciHeap<TItem,TKey>::Insert(TItem w,TKey alpha)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (w>=n) NoSuchItem("Insert",w);

    if (status[w]!=3)
    {
        sprintf(CT.logBuffer,"Already on queue: %ld",w);
        Error(ERR_REJECTED,"Insert",CT.logBuffer);
    }

    #endif
                                    
    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    first[w] = UNDEFINED;
    rank[w] = 0;
    status[w] = 2;
    key[w] = alpha;
    nextLink[w] = UNDEFINED;
    card++;
    Push(w);
    
    if (minimal==UNDEFINED || alpha<key[minimal]) minimal = w;

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif

    #if defined(_TRACING_)

    if (CT.traceData) Display();

    #endif    
}


// Add w to the list of root nodes

template <class TItem,class TKey>
void fibonacciHeap<TItem,TKey>::Push(TItem w) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (w>=n || status[w]!=2)
    {
        sprintf(CT.logBuffer,"Not a root: %ld",w);
        Error(ERR_REJECTED,"Push",CT.logBuffer);
    }

    #endif
    
    nextLink[w] = firstLink;
    firstLink = w;
}


// Consolidate the buckets and recompute the global minimum

template <class TItem,class TKey>
void fibonacciHeap<TItem,TKey>::Restore() throw()
{
    for (TItem i=0;i<k;i++) bucket[i] = UNDEFINED;

    minimal = UNDEFINED;

    TItem x = firstLink;

    while (x!=UNDEFINED)
    {
        if (status[x]==2)
        {
            if (bucket[rank[x]]==UNDEFINED) bucket[rank[x]] = x;
            else Link(x,bucket[rank[x]]);
        }
        x = nextLink[x];
    }

    x = firstLink;
    TItem y = UNDEFINED;

    while (x!=UNDEFINED)
    {
        if (status[x]!=2)
        {
            if (y==UNDEFINED) firstLink = nextLink[x];
            else nextLink[y] = nextLink[x];
        }
        else
        {
            y = x;

            if (minimal==UNDEFINED || key[x]<key[minimal]) minimal = x;
        }

        x = nextLink[x];
    }
}


// Merge two trees with respective roots v and w which have equal rank.
// If the bucket (rank+1) is occupied, link is called recursively.

template <class TItem,class TKey>
void fibonacciHeap<TItem,TKey>::Link(TItem v,TItem w) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchItem("Link",v);

    if (status[v]!=2)
    {
        sprintf(CT.logBuffer,"Not a root: %ld",v);
        Error(ERR_REJECTED,"Link",CT.logBuffer);
    }

    if (w>=n) NoSuchItem("Link",w);

    if (status[w]!=2)
    {
        sprintf(CT.logBuffer,"Not a root: %ld",w);
        Error(ERR_REJECTED,"Link",CT.logBuffer);
    }

    #endif
    
    if (key[v]<key[w])
    {
        TItem swap = v;
        v = w;
        w = swap;
    }
    
    if (v==bucket[rank[v]]) bucket[rank[v]] = UNDEFINED;
    if (w==bucket[rank[w]]) bucket[rank[w]] = UNDEFINED;
    
    pred[v] = w;
    status[v] = 0;
    previous[v] = UNDEFINED;
    TItem x = first[w];
    first[w] = v;

    if (x!=UNDEFINED)
    {
        previous[x] = v;
        next[v] = x;
    }
    else next[v] = UNDEFINED;
    
    rank[w]++;

    #if defined(_TRACING_)

    if (CT.traceData) Display();

    #endif    

    if (bucket[rank[w]]==UNDEFINED) bucket[rank[w]] = w;
    else Link(w,bucket[rank[w]]);
}


// Split the tree containing v which must have a predecessor x.
// The node v becomes a root.
// If status[x]==1, a recursive call Cut(x) occurs.

template <class TItem,class TKey>
void fibonacciHeap<TItem,TKey>::Cut(TItem v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchItem("Cut",v);

    if (status[v]>1)
    {
        sprintf(CT.logBuffer,"Cut node is a root: %ld",v);
        Error(ERR_REJECTED,"Cut",CT.logBuffer);
    }                                

    #endif

    #if defined(_TRACING_)

    if (CT.traceData) Display();

    #endif    

    TItem x = pred[v];
    TItem u = previous[v];
    TItem w = next[v];

    if (u==UNDEFINED) first[x] = w;
    else next[u] = w;

    if (w!=UNDEFINED) previous[w] = u;
    
    status[v] = 2;
    Push(v);

    if (minimal==UNDEFINED || key[v]<key[minimal]) minimal = v;

    #if defined(_TRACING_)

    if (CT.traceData) Display();

    #endif    
    
    if (status[x]==1) Cut(x);
    else if (status[x]==0) status[x] = 1;

    rank[x]--;
}


template <class TItem,class TKey>
void fibonacciHeap<TItem,TKey>::Delete(TItem w) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (w>=n || status[w]>2) NoSuchItem("Delete",w);

    #endif
    
    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    TItem x = first[w];

    while (x!=UNDEFINED)
    {
        TItem y = next[x];
        Cut(x);
        x = y;
    }
    
    if (status[w]<2) Cut(w);
    
    status[w] = 3;
    card--;
    Restore();

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif

    #if defined(_TRACING_)

    if (CT.traceData) Display();

    #endif    
}


template <class TItem,class TKey>
TItem fibonacciHeap<TItem,TKey>::Delete() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (card==0) Error(ERR_REJECTED,"Delete","Heap is empty");

    #endif
        
    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    TItem ret = minimal;
    Delete(minimal);

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif

    return ret;
}


template <class TItem,class TKey>
TKey fibonacciHeap<TItem,TKey>::Key(TItem w) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (w>=n || status[w]>2) NoSuchItem("Key",w);

    #endif

    return key[w];
}


template <class TItem,class TKey>
void fibonacciHeap<TItem,TKey>::ChangeKey(TItem w,TKey alpha) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (w>=n || status[w]>2) NoSuchItem("ChangeKey",w);                                

    #endif
    
    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Enable();
    
    #endif

    if (alpha>key[w])
    {
        key[w] = alpha;
        TItem x = first[w];

        while (x!=UNDEFINED)
        {
            TItem y = next[x];
            if (key[x]<key[w]) Cut(x);
            x = y;
        }
    }
    else
    {
        key[w] = alpha;

        if (status[w]<2 && pred[w]!=UNDEFINED && key[pred[w]]>alpha) Cut(w);
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerPrioQ] -> Disable();
    
    #endif

    #if defined(_TRACING_)

    if (CT.traceData) Display();

    #endif    
}


template <class TItem,class TKey>
TItem fibonacciHeap<TItem,TKey>::Peek() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (card==0) Error(ERR_REJECTED,"PEEK","Heap is empty");

    #endif
    
    return minimal;
}


template class fibonacciHeap<TNode,TFloat>;

#if defined(_BIG_ARCS_)

template class fibonacciHeap<TArc,TFloat>;

#endif
