
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "fnw2fnw.h"


FNW2FNW::FNW2FNW(abstractDiGraph &GC,TFloat *pp) throw(ERRejected) :
    goblinDataObject(GC.Context()),
    abstractMixedGraph(GC.N()+2,GC.M()+2*GC.N()+1),
    G(GC)
{                                                   
    n0 = G.N();
    m0 = G.M();
    pot = pp;
    CheckLimits();
     
    s1 = n-1;
    t1 = n-2;
    ret1 = 2*(m-1);
    
    cap = new TCap[2*n0+1];
    flow = new TFloat[2*n0+1];

    G.MakeRef();

    if (pot!=NULL)
    {
        if (CT.logMem || CT.logMan)
        {
            LogEntry(LOG_MAN,"Repairing complementary slackness...");
            if (!CT.logMan && CT.logMem)
                LogEntry(LOG_MEM,"Repairing complementary slackness...");
        }

        for (TArc i=0;i<m0;i++)
        {
            TArc a = 2*i;
            if (G.RedLength(a)<0) G.Push(a,G.ResCap(a));
            if (G.RedLength(a)>0) G.Push(a^1,G.ResCap(a^1));
        }
    }
    
    if (CT.logMem || CT.logMan)
    {
        LogEntry(LOG_MAN,"Eliminating lower capacity bounds...");
        if (!CT.logMan && CT.logMem)
            LogEntry(LOG_MEM,"Eliminating lower capacity bounds...");
    }

    sumLower = 0;
    TFloat checkSum = 0;

    for (TNode v=0;v<n0;v++)
    {
        flow[v] = flow[v+n0] = 0;
        TFloat delta = G.Demand(v);
        checkSum += delta;

        if (delta>0)
        {
            cap[v] = 0;
            cap[v+n0] = delta;
            sumLower += delta;
        }
        else
        {
            cap[v] = -delta;
            cap[v+n0] = 0;
        }
    }
    
    #if defined(_FAILSAVE_)

    if (checkSum!=0)
        Error(ERR_REJECTED,"FNW2FNW","Node demands do not resolve");

    #endif
    
    for (TArc i=0;i<m0;i++)
    {
        TArc a = 2*i;
        TNode u = G.StartNode(a);
        TNode v = G.EndNode(a);

        TCap thisLower = G.LCap(a);
                    
        cap[v] += thisLower;
        cap[u+n0] += thisLower;

        flow[u] += G.Flow(a)-thisLower;
        flow[v+n0] += G.Flow(a)-thisLower;
        
        sumLower += thisLower;
    }
    
    flow[2*n0] = cap[2*n0] = 0;

    for (TNode v=0;v<n0;v++)
    {
        TCap delta;
        TCap delta1 = cap[v]-flow[v];
        TCap delta2 = cap[v+n0]-flow[v+n0];

        if (delta1<delta2) delta = delta1;
        else delta = delta2;
        
        flow[v] += delta;
        flow[v+n0] += delta;
    }

    if (CT.traceLevel==2) Display();
}


unsigned long FNW2FNW::Size() throw()
{   
    return
          sizeof(FNW2FNW)
        + goblinDataObject::Allocated()
        + abstractMixedGraph::Allocated()
        + abstractDiGraph::Allocated()
        + FNW2FNW::Allocated();
}


unsigned long FNW2FNW::Allocated() throw()
{   
    unsigned long tmpSize
        = (2*n0+1)*sizeof(TCap)     // cap[]
        + (2*n0+1)*sizeof(TFloat);  // flow[]

    return tmpSize;
}


FNW2FNW::~FNW2FNW() throw()
{
    if (CT.traceLevel==2) Display();

    if (pot!=NULL && pi!=NULL)
    {
        LogEntry(LOG_MAN,"Updating original node potentials...");
        if (!CT.logMan && CT.logMem)
            LogEntry(LOG_MEM,"Updating original node potentials...");

        for (TNode v=0;v<n0;v++) G.SetPotential(v,G.Pi(v)+pi[v]);
    }

    G.ReleaseRef();
    
    delete[] cap;
    delete[] flow;
}
    

bool FNW2FNW::Perfect() throw()
{   
    bool solved = true;
    THandle H = Investigate();
    TNode s = Source();

    while (Active(H,s))
    {
        TArc a = Read(H,s);
        if (ResCap(a)>0) solved = 0;
    }

    Close(H);
    
    if (CT.logRes)
    {
        if (solved) LogEntry(LOG_RES,"...Flow corresponds to a circulation");
        else LogEntry(LOG_RES,"...Flow does not correspond to a circulation");
    }
       
    return solved;
}


TNode FNW2FNW::StartNode(TArc a) throw(ERRange)
{
    TArc d2 = a/2;
    TArc r2 = a%2;
    
    if (d2<m0) return G.StartNode(a);
    
    if (d2<m0+n0)
    {
        if (r2) return (TNode)(d2-m0);
        else return s1;
    }

    if (d2<m0+2*n0)
    {
        if (r2) return t1;
        else return (TNode)(d2-m0)-n0;
    }

    if (d2==m0+2*n0)
    {
        if (r2) return s1;
        else return t1;
    }
    
    #if defined(_FAILSAVE_)

    NoSuchArc("StartNode",a);

    #endif

    throw ERRange();
}


TNode FNW2FNW::EndNode(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("EndNode",a);

    #endif

    return FNW2FNW::StartNode(a^1);
}


TArc FNW2FNW::First(TNode v) throw(ERRange)
{
    if (v<n0) return G.First(v);
    if (v==t1) return ret1;
    if (v==s1) return ret1^1;
    
    #if defined(_FAILSAVE_)
             
    NoSuchNode("First",v);
             
    #endif
    
    throw ERRange();
}
             
             
TArc FNW2FNW::Right(TArc a,TNode u) throw(ERRange,ERRejected)
{            
    #if defined(_FAILSAVE_)
             
    if (u!=StartNode(a))
        Error(ERR_REJECTED,"Right","Mismatching start nodes");

    #endif
             
    TArc d2 = a/2;
    TArc r2 = a%2;
    
    if (d2<m0)
    {
        TArc ar = G.Right(a,u);

        if (ar!=G.First(u)) return ar;
        else return 2*(m0+u)+1;
    }
    
    if (d2<m0+n0)
    {
        if (r2) return 2*(d2+n0);
        else
        {
            if (d2==(m0+n0-1)) return ret1^1;
            else return a+2;
        }
    }

    if (d2<m0+2*n0)
    {
        if (!r2)
        {
            TArc af = G.First(d2-m0-n0);
            if (af==NoArc) return 2*(m0+u)+1;
            else return af;
        }
        else
        {
            if (d2==(m0+2*n0-1)) return ret1;
            else return a+2;
        }
    }
    
    if (d2==m0+2*n0)
    {
        if (r2) return 2*m0;
        else return 2*(m0+n0)+1;
    }
    
    #if defined(_FAILSAVE_)
             
    NoSuchArc("Right",a);
             
    #endif   

    throw ERRange();
}
             
             
TCap FNW2FNW::Demand(TNode v) throw(ERRange)
{
    if (v<n0) return 0;

    if (v==t1) return sumLower;
    if (v==s1) return -sumLower;

    #if defined(_FAILSAVE_)

    NoSuchNode("Demand",v);

    #endif
    
    throw ERRange();
}


TCap FNW2FNW::UCap(TArc a) throw(ERRange)
{ 
    if (a<2*m0) return G.UCap(a)-G.LCap(a);
    
    if (a<2*m) return cap[(a>>1)-m0];
    
    #if defined(_FAILSAVE_)

    NoSuchArc("UCap",a);

    #endif

    throw ERRange();
}


TFloat FNW2FNW::Flow(TArc a) throw(ERRange)
{ 
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("Flow",a);

    #endif

    if (a>=ret1) return 0;

    if (a<2*m0) return G.Flow(a)-G.LCap(a);
    else return flow[(a>>1)-m0];
}


TFloat FNW2FNW::Length(TArc a) throw(ERRange)
{
    if (a<2*m0) return G.RedLength(a);
    if (a<2*m) return 0;

    #if defined(_FAILSAVE_)
             
    NoSuchArc("Length",a);
             
    #endif
    
    throw ERRange();
}


TFloat FNW2FNW::C(TNode v,char i) throw(ERRange)
{
    if (i==0)
    {
        if (v<n0) return G.C(v,0)+10;
        if (v==s1) return 0;
        if (v==t1) return G.CMax(0)+20;
    }
    else
    {
        if (v<n0) return G.C(v,1);
        if (v<n) return G.CMax(1)/2;
    }
    
    #if defined(_FAILSAVE_)
             
    NoSuchNode("C",v);
             
    #endif
    
    throw ERRange();
}
             
             
TFloat FNW2FNW::CMax(char i) throw(ERRange)
{
    if (G.Dim()==0) return 0;
    if (i==0) return G.CMax(0)+20;
    if (i==1) return G.CMax(1);
    
    #if defined(_FAILSAVE_)
             
    NoSuchCoordinate("CMax",i);
             
    #endif
    
    throw ERRange();
}
             
             
bool FNW2FNW::HiddenNode(TNode v) throw(ERRange)
{
    if (v<n0) return G.HiddenNode(v);
    if (v<n) return true;
    
    #if defined(_FAILSAVE_)
             
    NoSuchNode("HiddenNode",v);
             
    #endif
    
    throw ERRange();
}           
             
             
void FNW2FNW::Push(TArc a,TFloat Lambda) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("Push",a);

    if (Lambda<0 || Lambda>ResCap(a)) AmountOutOfRange("Push",Lambda);

    #endif

    if (a<2*m0) G.Push(a,Lambda);
    else
    {
        TArc a0;
        a0 = (a>>1)-m0;

        if (a&1) flow[a0] = flow[a0]-Lambda;
        else flow[a0] = flow[a0]+Lambda;
    }
    
    AdjustDivergence(a,Lambda);
}


bool abstractDiGraph::AdmissibleCirculation() throw()
{
    OpenFold(ModMaxFlow,NO_INDENT);

    LogEntry(LOG_METH,"Computing admissible circulation...");
    
    TCap checkSum = 0;

    for (TNode v=0;v<n;v++) checkSum += Demand(v);
    
    if (checkSum!=0)
    {
        LogEntry(LOG_RES,"...Node demands do not resolve");
    
        CloseFold(ModMaxFlow,NO_INDENT);

        return 0;
    }
    
    OpenFold();

    FNW2FNW G(*this);
    G.MaxFlow(G.Source(),G.Target());

    CloseFold();
    CT.Ping(Handle(),n*m);
    
    bool ret = G.Perfect();
    
    CloseFold(ModMaxFlow,NO_INDENT);
    
    return ret;
}
