
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, September 1999.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "genstruc.h"


// ------------------------------------------------------


genericGraphStructure::genericGraphStructure(abstractMixedGraph &GC) throw() :
     goblinDataObject(GC.Context()), G(GC)
{
    nAct = G.N()+G.NI();
    nMax = G.N()+G.NI();
    mMax = mAct = G.M();
    
    cucap = 1;
    if (CT.randUCap)
    {
        ucap = new TCap[mMax];

        for (TArc a=0;a<mMax;a++)
        {
            ucap[a] = 1+TCap(CT.UnsignedRand());
            if (ucap[a]>cucap) cucap = ucap[a];
        }

        LogEntry(LOG_MEM,"...Capacity bounds allocated");
    }
    else ucap = NULL;

    clength = 1;
    if (CT.randLength)
    {
        length = new TFloat[mMax];

        for (TArc a=0;a<mMax;a++)
        {
            length[a] = TFloat(CT.SignedRand());
            if (length[a]>clength) clength = length[a];
        }

        LogEntry(LOG_MEM,"...Length labels allocated");
    }
    else length = NULL;

    clcap = 0;
    if (CT.randLCap)
    {
        lcap = new TCap[mMax];

        for (TArc a=0;a<mMax;a++)
        {
            lcap[a] = TCap(CT.Rand(1+int(ucap[a])));
            if (lcap[a]>clcap) clcap = lcap[a];
        }

        LogEntry(LOG_MEM,"...Lower capacity bounds allocated");
    }
    else lcap = NULL;
    
    orientation = NULL;
    cOrientation = 0;

    demand = NULL;
    cDemand = 0;

    dim = 0;
    maxX = maxY = 0;

    if (CT.randGeometry)
    {
        dim = 2;
        x = new TFloat[nMax];
        y = new TFloat[nMax];

        for (TNode v=0;v<nMax;v++)
        {
            x[v] = TFloat(CT.SignedRand());
            if (x[v]>maxX) maxX = x[v];

            y[v] = TFloat(CT.SignedRand());
            if (y[v]>maxY) maxY = y[v];
        }

        LogEntry(LOG_MEM,"...Nodes embedded into plane");
    }
    else x = y = NULL;

    LogEntry(LOG_MEM,"...Generic graph allocated");
}


void genericGraphStructure::ReDimension(TNode nn,TArc mm)
    throw(ERRejected)
{   
    #if defined(_FAILSAVE_)

    if (nMax>0 || mMax>0)
        Error(ERR_REJECTED,"ReDimension","Graph structure must be initial");

    if (nn>=CT.MaxNode())
    {
        sprintf(CT.logBuffer,"Number of nodes is out of range: %ld",nn);
        Error(ERR_REJECTED,"ReDimension",CT.logBuffer);
    }

    if (2*mm>CT.MaxArc()-2)
    {
        sprintf(CT.logBuffer,"Number of arcs is out of range: %ld",mm);
        Error(ERR_REJECTED,"ReDimension",CT.logBuffer);
    }

    #endif
    
    nMax = nAct = nn;
    mMax = mAct = mm;
}


void genericGraphStructure::ReadUCap(goblinImport &F)
    throw(ERParse)
{
    F.Scan("ucap");
    ucap = F.GetTCapTuple(mAct);

    if (F.Constant())
    {
        cucap = ucap[0];
        delete[] ucap;
        ucap = NULL;
    }
    else
    {
        cucap = 0;
        for (TArc a=0;a<mAct;a++)
            if (ucap[a]>cucap) cucap = ucap[a];
            
        LogEntry(LOG_MEM,"...Upper capacity bounds allocated");
    }
}


void genericGraphStructure::ReadLCap(goblinImport &F)
    throw(ERParse)
{
    F.Scan("lcap");
    lcap = F.GetTCapTuple(mAct);

    if (F.Constant())
    {
        clcap = lcap[0];
        delete[] lcap;
        lcap = NULL;
    }
    else
    {
        clcap = 0;
        for (TArc a=0;a<mAct;a++)
            if (lcap[a]>clcap) clcap = lcap[a];
            
        LogEntry(LOG_MEM,"...Lower capacity bounds allocated");
    }
}


void genericGraphStructure::ReadDemand(goblinImport &F)
    throw(ERParse)
{
    F.Scan("demand");
    demand = F.GetTCapTuple(nAct);

    if (F.Constant())
    {
        cDemand = demand[0];
        delete[] demand;
        demand = NULL;
    }
    else
    {
        cDemand = 0;
        for (TNode v=0;v<nAct;v++)
            if (demand[v]>cDemand) cDemand = demand[v];
            
        LogEntry(LOG_MEM,"...Node demands allocated");
    }
}


void genericGraphStructure::ReadOrientation(goblinImport &F)
    throw(ERParse)
{
    F.Scan("directed");
    orientation = F.GetCharTuple(mAct);

    if (F.Constant())
    {
        cOrientation = orientation[0];
        delete[] orientation;
        orientation = NULL;
    }
    else
    {
        cOrientation = 0;
        LogEntry(LOG_MEM,"...Arc orientations allocated");
    }
}


void genericGraphStructure::ReadObjectives(goblinImport &F)
    throw(ERParse)
{
    F.Scan("objectives");
    
    F.Scan("commodities");
    char* commodities = F.GetCharTuple(1);
    char comm = commodities[0];
    delete[] commodities;
    
    #if defined(_FAILSAVE_)

    if (comm!=1)
        Error(ERR_PARSE,"ReadObjectives","Multiple commodities are not supported");

    #endif
    
    F.Scan("bound");
    TFloat* bound = F.GetTFloatTuple(comm);
    delete[] bound;
    
    F.Scan("length");
    
    for (char i=0;i<comm;i++)
    {
        F.Scan();
    
        length = F.GetTFloatTuple(mAct);
        if (F.Constant())
        {
            clength = length[0];
            delete[] length;
            length = NULL;
        }
        else
        {
            clength = 0;

            for (TArc a=0;a<mAct;a++)
            {
                if (length[a]>clength) clength = length[a];
                if (-length[a]>clength) clength = -length[a];
            }
            
            LogEntry(LOG_MEM,"...Length labels allocated");
        }
    }
    
    F.Scan(); // length
    
    F.Scan(); // objectives
}


void genericGraphStructure::ReadGeometry(goblinImport &F)
    throw(ERParse)
{
    F.Scan("geometry");
    
    F.Scan("metrics");
    char* metrics = F.GetCharTuple(1);
    CT.methGeometry = metrics[0];
    delete[] metrics;
    
    F.Scan("dim");
    char* dimension = F.GetCharTuple(1);
    dim = dimension[0];
    delete[] dimension;
    
    #if defined(_FAILSAVE_)

    if (dim!=0 && dim!=2)
        Error(ERR_PARSE,"ReadGeometry","Mismatching dimensions");

    #endif
    
    if (dim==2)
    {
        F.Scan("coordinates");
    
        F.Scan();
        x = F.GetTFloatTuple(nAct);

        if (F.Constant())
        {
            maxX = 0;
            delete[] x;
            x = NULL;
        }
        else
        {
            maxX = 0;

            for (TNode v=0;v<nAct;v++)
            {
                if (x[v]>maxX) maxX = x[v];
            }
            
            LogEntry(LOG_MEM,"...x-coordinates allocated");
        }
    
        F.Scan();
        y = F.GetTFloatTuple(nAct);

        if (F.Constant())
        {
            maxY = 0;
            delete[] y;
            y = NULL;
        }
        else
        {
            maxY = 0;

            for (TNode v=0;v<nAct;v++)
            {
                if (y[v]>maxY) maxY = y[v];
            }

            LogEntry(LOG_MEM,"...y-coordinates allocated");
        }
    
        F.Scan(); // coordinates
    }
    else
    {
        maxX = 0;
        maxY = 0;
        
        x = NULL;
        y = NULL;
    }
    
    F.Scan(); // geometry
}


void genericGraphStructure::ReadLayout(goblinImport& F)
    throw(ERParse)
{
    F.Scan("layout");
    
    F.Scan("style");
    char* sty = F.GetCharTuple(1);

    if (sty[0]<=2) CT.arcStyle = sty[0];
    else
    {
        sprintf(CT.logBuffer,"Unknown arc style option: %d",sty[0]);
        Error(ERR_PARSE,"ReadLayout",CT.logBuffer);
    }

    delete[] sty;
    
    F.Scan("align");
    delete[] F.GetTNodeTuple(mAct);
    if (!F.Constant()) Error(ERR_PARSE,"ReadLayout","Nontrivial layout");
    
    F.Scan("interpolate");
    delete[] F.GetTNodeTuple(nAct);
    if (!F.Constant()) Error(ERR_PARSE,"ReadLayout","Nontrivial layout");
    
    F.Scan(); // layout
}


unsigned long genericGraphStructure::Allocated()
    throw()
{   
    unsigned long tmpSize = 0;
         
    if (ucap!=NULL)         tmpSize += mMax*sizeof(TCap);
    if (length!=NULL)       tmpSize += mMax*sizeof(TFloat);
    if (lcap!=NULL)         tmpSize += mMax*sizeof(TCap);
    if (orientation!=NULL)  tmpSize += mMax*sizeof(char);
    if (demand!=NULL)       tmpSize += nMax*sizeof(TCap);
    if (x!=NULL)            tmpSize += nMax*sizeof(TFloat);
    if (y!=NULL)            tmpSize += nMax*sizeof(TFloat);

    return tmpSize;
}


genericGraphStructure::~genericGraphStructure()
    throw()
{
    delete[] ucap;
    delete[] length;
    delete[] lcap;
    delete[] orientation;
    delete[] demand;
    delete[] x;
    delete[] y;

    LogEntry(LOG_MEM,"...Generic graph disallocated");
}


TCap genericGraphStructure::UCap(TArc a)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct)  NoSuchArc("UCap",a);

    #endif

    if (CUCap()) return cucap;
    else return ucap[a>>1];
}


void genericGraphStructure::SetUCap(TArc a,TCap cc)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("SetUCap",a);

    if (cc<0)
    {
        sprintf(CT.logBuffer,"Illegal assignment: %g",cc);
        Error(ERR_RANGE,"SetUCap",CT.logBuffer);
    }

    #endif

    if (Sub(a)>cc) OmitArc(a,Sub(a)-cc);

    if ((ucap==NULL) && (cc != cucap))
    {
        ucap = new TCap[mMax];
        for (TArc i=0;i<mAct;i++) ucap[i] = cucap;
        LogEntry(LOG_MEM,"...Capacity bounds allocated");
    }
    
    if (ucap!=NULL)
    {
        if (cc>=cucap) ucap[a>>1] = cucap = cc;
        else
        {
            if (ucap[a>>1]==cucap)
            {
                cucap = 0;

                for (TArc i=0;i<mAct && cucap<InfCap;i++)
                    if (ucap[i]>cucap) cucap = ucap[i];
            }

            ucap[a>>1] = cc;
        }
    }
}


void genericGraphStructure::SetCUCap(TCap cc)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (cc<0)
    {
        sprintf(CT.logBuffer,"Illegal assignment: %g",cc);
        Error(ERR_RANGE,"SetCUCap",CT.logBuffer);
    }

    #endif

    G.InitSubgraph();

    if (ucap!=NULL)
    {
        delete[] ucap;
        ucap = NULL;
    }
    
    cucap = cc;
}


TCap genericGraphStructure::LCap(TArc a)
    throw(ERRange)
{ 
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("LCap",a);

    #endif
    
    if (lcap!=NULL) return lcap[a>>1];
    else return clcap;
}


void genericGraphStructure::SetLCap(TArc a,TCap bb)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("SetLCap",a);

    if (bb<0)
    {
        sprintf(CT.logBuffer,"Illegal assignment: %g",bb);
        Error(ERR_RANGE,"SetLCap",CT.logBuffer);
    }

    #endif

    if (Sub(a)<bb) AddArc(a,bb-Sub(a));

    if ((lcap==NULL) && (bb != clcap))
    {
        lcap = new TCap[mMax];
        for (TArc i=0;i<mAct;i++) lcap[i] = clcap;
        LogEntry(LOG_MEM,"...Lower capacity bounds allocated");
    }
    
    if (lcap!=NULL)
    {
        if (bb>=clcap) lcap[a>>1] = clcap = bb;
        else
        {
            if (lcap[a>>1]==clcap)
            {
                clcap = 0;

                for (TArc i=0;i<mAct;i++)
                    if (lcap[i]>clcap) clcap = lcap[i];
            }

            lcap[a>>1] = bb;
        }
    }
}


void genericGraphStructure::SetCLCap(TCap bb)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (bb<0)
    {
        sprintf(CT.logBuffer,"Illegal assignment: %g",bb);
        Error(ERR_RANGE,"SetCLCap",CT.logBuffer);
    }
    
    #endif

    G.InitSubgraph();

    if (lcap!=NULL)
    {
        delete[] lcap;
        lcap = NULL;
    }
    
    clcap = bb;
}


TCap genericGraphStructure::Demand(TNode v)
    throw(ERRange)
{ 
    #if defined(_FAILSAVE_)

    if (v>=nAct) NoSuchNode("Demand",v);

    #endif
    
    if (demand!=NULL) return demand[v];
    else return cDemand;
}


void genericGraphStructure::SetDemand(TNode v,TCap bb)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=nAct) NoSuchNode("SetDemand",v);

    #endif

    if ((demand==NULL) && (bb != cDemand))
    {
        demand = new TCap[nMax];
        for (TNode i=0;i<nAct;i++) demand[i] = cDemand;
        LogEntry(LOG_MEM,"...Node demands allocated");
    };
    
    if (demand!=NULL)
    {
        if (bb>=cDemand) demand[v] = cDemand = bb;
	    else
        {
            if (demand[v]==cDemand)
            {
                cDemand = 0;

                for (TNode i=0;i<nAct && cDemand<InfCap;i++)
                    if (demand[i]>cDemand) cDemand = demand[i];
            }

            demand[v] = bb;
        }
    }
}


void genericGraphStructure::SetCDemand(TCap bb)
    throw()
{
    if (demand!=NULL)
    {
        delete[] demand;
        demand = NULL;
    }
    
    cDemand = bb;
}


TFloat genericGraphStructure::Length(TArc a)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("Length",a);

    #endif

    if (CT.methGeometry==0 || length!=NULL)
    {
        if (length==NULL) return clength;
        else return length[a>>1];
    }
    
    TNode v1 = G.StartNode(a);
    TNode v2 = G.EndNode(a);
    
    if (v1==v2) return InfFloat;

    if  (CT.methGeometry==4)
    {
        double pi = 3.141592;
        double radius = 6378.388;

        double deg = int(x[v1]);
        double min = x[v1]-deg;
        double latitude1 = pi*(deg+5.0*min/3.0)/180.0;
        deg = int(y[v1]);
        min = y[v1]-deg;
        double longitude1 = pi*(deg+5.0*min/3.0)/180.0;

        deg = int(x[v2]);
        min = x[v2]-deg;
        double latitude2 = pi*(deg+5.0*min/3.0)/180.0;
        deg = int(y[v2]);
        min = y[v2]-deg;
        double longitude2 = pi*(deg+5.0*min/3.0)/180.0;

        double q1 = cos(longitude1-longitude2);
        double q2 = cos(latitude1-latitude2);
        double q3 = cos(latitude1+latitude2);
        return int(radius*acos(0.5*((1.0+q1)*q2-(1.0-q1)*q3))+1.0);
    }
    
    TFloat dx = 0;
    if (x!=NULL) dx = x[v1]-x[v2];
    if (dx<0) dx = -dx;
    
    TFloat dy = 0;
    if (y!=NULL) dy = y[v1]-y[v2];
    if (dy<0) dy = -dy;
    
    if  (CT.methGeometry==1) return floor(dx+dy+0.5);
    if  (CT.methGeometry==2) return floor(sqrt(dx*dx+dy*dy)+0.5);

    if (dx>dy) return floor(dx+0.5);
    else return floor(dy+0.5);
}


void genericGraphStructure::SetLength(TArc a,TFloat ll)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("SetLength",a);

    #endif

    if ((length==NULL) && (ll != clength))
    {
        length = new TFloat[mMax];
        for (TArc i=0;i<mAct;i++) length[i] = clength;
        LogEntry(LOG_MEM,"...Length labels allocated");
    }
    
    if (length!=NULL)
    {
        if (ll>=clength) length[a>>1] = clength = ll;
	    else
        {
            if (length[a>>1]==clength)
            {
                clength = 0;

                for (TArc i=0;i<mAct && clength<InfFloat;i++)
                    if (length[i]>clength) clength = length[i];
            }

            length[a>>1] = ll;
        }
    }
}


void genericGraphStructure::SetCLength(TFloat ll)
    throw()
{
    if (length!=NULL)
    {
        delete[] length;
        length = NULL;
    }
    
    clength = ll;
}


char genericGraphStructure::Orientation(TArc a)
    throw(ERRange)
{ 
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("Orientation",a);

    #endif
    
    if (orientation!=NULL) return orientation[a>>1];
    else return cOrientation;
}


void genericGraphStructure::SetOrientation(TArc a,char oo)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("SetOrientation",a);

    if (oo>=4)
        Error(ERR_RANGE,"SetOrientation","Illegal assignment");

    #endif

    if ((orientation==NULL) && (oo != cOrientation))
    {
        orientation = new char[mMax];
        for (TArc i=0;i<mAct;i++) orientation[i] = cOrientation;
        LogEntry(LOG_MEM,"...Arc orientations allocated");
    }
    
    if (orientation!=NULL) orientation[a>>1] = oo;
}


void genericGraphStructure::SetCOrientation(char oo)
    throw()
{
    if (orientation!=NULL)
    {
        delete[] orientation;
        orientation = NULL;
    }
    
    cOrientation = oo;
}


bool genericGraphStructure::Blocking(TArc a)
    throw(ERRange)
{ 
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("Blocking",a);

    #endif
    
    if (Orientation(a)==1) return (a&1);
    if (Orientation(a)==2) return !(a&1);

    return false;
}


TFloat genericGraphStructure::C(TNode v,char i)
    throw(ERRange)
{ 
    #if defined(_FAILSAVE_)

    if (v>=nAct) NoSuchNode("C",v);        

    #endif
    
    if (i==0)
    {
        if (x!=NULL) return x[v];
        else return 0;
    }
    
    if (i==1)
    {
        if (y!=NULL) return y[v];
        else return 0;
    }
    
    return 0;
}


void genericGraphStructure::SetC(TNode v,char j,TFloat cc)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=nAct) NoSuchNode("SetC",v);

    if (j>=2) NoSuchCoordinate("SetC",j);

    #endif

    if (j==0)
    {
        if ((x==NULL) && (cc != maxX))
        {
            x = new TFloat[nMax];
            for (TNode i=0;i<nAct;i++) x[i] = maxX;
            dim = 2;
            LogEntry(LOG_MEM,"...x-coordinates allocated");
        }
    
        if (x!=NULL)
        {
            if (cc>=maxX) x[v] = maxX = cc;
	        else
            {
                if (x[v]==maxX)
                {
                    x[v] = cc;
                    maxX = 0;
                    for (TNode i=0;i<nAct;i++)
    	            if (x[i]>maxX) maxX = x[i];
                }

                x[v] = cc;
            }
        }
    }
    else
    {
        if ((y==NULL) && (cc != maxY))
        {
            y = new TFloat[nMax];
            for (TNode i=0;i<nAct;i++) y[i] = maxY;
            dim = 2;
            LogEntry(LOG_MEM,"...y-coordinates allocated");
        }
    
        if (y!=NULL)
        {
            if (cc>=maxY) y[v] = maxY = cc;
    	    else
            {
                if (y[v]==maxY)
                {
                    y[v] = cc;
                    maxY = 0;

                    for (TNode i=0;i<nAct;i++)
                        if (y[i]>maxY) maxY = y[i];
                }

                y[v] = cc;
            }
        }
    }
}


TFloat genericGraphStructure::CMax(char i)
    throw()
{ 
    if (i==0)
    {
        if (x!=NULL) return maxX;
        else return 0;
    }

    if (i==1)
    {
        if (y!=NULL) return maxY;
        else return 0;
    }
    
    return 0;
}


TFloat genericGraphStructure::Flow(TArc a)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("Flow",a);

    #endif

    return Sub(a);
}


void genericGraphStructure::Push(TArc a,TFloat Lambda)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("Push",a);

    #endif

    if (a&1) OmitArc(a,Lambda);
    else AddArc(a,Lambda);
}
