
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "import.h"


goblinImport::goblinImport(const char* impFileName,goblinController& thisContext)
    throw(ERFile) :
    impFile(impFileName, ios::in), CT(thisContext)
{
    if (!impFile)
    {
        sprintf(CT.logBuffer,"Could not open import file %s, io_state %d",
            impFileName,impFile.rdstate());
        CT.Error(ERR_FILE,NoHandle,"goblinImport",CT.logBuffer);
    }
    
    impFile.flags(impFile.flags() | ios::right);
    currentLevel = 0;
    head = tail = false;
    complain = true;
    buffer = new char[256];
}


char* goblinImport::Scan(char* token,TOwnership tp)
    throw(ERParse)
{
    impFile.width(255);
    impFile >> ws >> buffer;

    if (buffer[0] == '(')
    {
        if (strlen(buffer)==1)
            CT.Error(ERR_PARSE,NoHandle,"Scan",
                "Misplaced white space behind opening parenthesis");

        head = true;

        if (buffer[strlen(buffer)-1] == ')')
        {              
            tail = true;
            buffer[strlen(buffer)-1] = 0;
        }
        else
        {
            tail = false;
            currentLevel++;
        }
        
        if (token!=NULL && strcmp(token,buffer+1))
        {
            sprintf(CT.logBuffer,"Unexpected token: %s",(buffer+1));
            CT.Error(ERR_PARSE,NoHandle,"Scan",CT.logBuffer);
        }

        if (tp==OWNED_BY_RECEIVER) return buffer+1;
        else
        {
            char* ret = new char[strlen(buffer+1)+1];
            sprintf(ret,"%s",buffer+1);
            return ret;
        }
    }
    else
    {
        head = false;

        if (buffer[0]=='\"')
        {
            // String handling
        
            unsigned i = 1;

            while (buffer[i]!='\"' && buffer[i]!=0) i++;

            if (buffer[i]==0)
            {
                char buffer2[256];

                impFile.get(buffer2,255-i,'\"');

                sprintf(buffer,"%s%s",buffer+1,buffer2);

                impFile.get(buffer2[0]);
            }
            else
            {
                buffer[i] = 0;
                sprintf(buffer,"%s",buffer+1);
            }
        }

        if (buffer[strlen(buffer)-1] == ')')
        {              
            tail = true;

            if (currentLevel == 0)
                CT.Error(ERR_PARSE,NoHandle,"Scan","Parenthesis mismatch");

            currentLevel--;
            buffer[strlen(buffer)-1] = 0;
        }
        else tail = false;

        if (tp==OWNED_BY_RECEIVER) return buffer;
        else
        {
            char* ret = new char[strlen(buffer)+1];
            sprintf(ret,"%s",buffer);
            return ret;
        }
    }
}


bool goblinImport::Seek(char* token)
    throw(ERParse)
{
    bool found = false;
    char minLevel = currentLevel;
    char* thisToken = Scan();
    
    while (!found)
    {
        if (strcmp(thisToken,token)) found = true;
        else thisToken = Scan();
	
        if (minLevel>currentLevel)
        {
            sprintf(CT.logBuffer,"Not in this scope: %s",token);
            CT.Error(ERR_PARSE,NoHandle,"Seek",CT.logBuffer);
        }
    }
    
    return found;
}


bool goblinImport::Head()
    throw()
{                  
    return head;
}


bool goblinImport::Tail()
    throw()
{               
    return tail;
}


bool goblinImport::Eof()
    throw()
{               
    return (tail && (currentLevel == 0));
}


TNode* goblinImport::GetTNodeTuple(unsigned long k)
    throw(ERParse)
{               
    TNode* tuple = new TNode[k];

    if (k==0) tuple = new TNode[1];

    length = 0;

    while (!tail)
    {
        char* label = NULL;

        try {label = Scan();} catch(ERParse)
        {
            delete[] tuple;
            throw ERParse();
        }

        if (strlen(label)>0)
        {
            if (length==k && length>0)
            {
                delete[] tuple;
                CT.Error(ERR_PARSE,NoHandle,"GetTNodeTuple","Length mismatch");
            }
        
            #if defined(_BIG_NODES_)

            if (strcmp(label,"*")) tuple[length++] = atol(label);

            #else

            if (strcmp(label,"*")) tuple[length++] = atoi(label);
        
            #endif

            else tuple[length++] = NoNode;
        }
    }
    
    if (length!=1 && length!=k && k!=0)
    {
        delete[] tuple;
        CT.Error(ERR_PARSE,NoHandle,"GetTNodeTuple","Length mismatch");
    };
                                     
    return tuple;
}


TArc* goblinImport::GetTArcTuple(unsigned long k)
    throw(ERParse)
{               
    TArc* tuple = new TArc[k];

    if (k==0) tuple = new TArc[1];

    length = 0;

    while (!tail)
    {
        char* label = NULL;

        try {label = Scan();} catch(ERParse)
        {
            delete[] tuple;
            throw ERParse();
        }

        if (strlen(label)>0)
        {
            if (length==k && length>0)
            {
                delete[] tuple;
                CT.Error(ERR_PARSE,NoHandle,"GetTArcTuple","Length mismatch");
            }
        
            if (k==0) tuple = (TArc *)GoblinRealloc(tuple,sizeof(TArc)*(int)(length+1));
            
            #if defined(_SMALL_ARCS_)

            if (strcmp(label,"*")) tuple[length++] = atoi(label);

            #else

            if (strcmp(label,"*")) tuple[length++] = atol(label);
        
            #endif

            else tuple[length++] = NoArc;
        }
    }
    
    if (length!=1 && length!=k && k!=0)
    {
        delete[] tuple;
        CT.Error(ERR_PARSE,NoHandle,"GetTArcTuple","Length mismatch");
    }
                                     
    return tuple;
}


TCap* goblinImport::GetTCapTuple(unsigned long k)
    throw(ERParse)
{               
    TCap* tuple = new TCap[k];

    if (k==0) tuple = new TCap[1];

    length = 0;

    while (!tail)
    {
        char* label = NULL;

        try {label = Scan();} catch(ERParse)
        {
            delete[] tuple;
            throw ERParse();
        }

        if (strlen(label)>0)
        {
            if (length==k && length>0)
            {
                delete[] tuple;
                CT.Error(ERR_PARSE,NoHandle,"GetTCapTuple","Length mismatch");
            }
        
            if (strcmp(label,"*")) tuple[length++] = (TCap)atof(label);
            else tuple[length++] = InfCap;
        }
    }
    
    if (length!=1 && length!=k && k!=0)
    {
        delete[] tuple;
        CT.Error(ERR_PARSE,NoHandle,"GetTCapTuple","Length mismatch");
    }
                                     
    return tuple;
}


TFloat* goblinImport::GetTFloatTuple(unsigned long k)
    throw(ERParse)
{               
    TFloat* tuple = new TFloat[k];                          

    if (k==0) tuple = new TFloat[1];

    length = 0;

    while (!tail)
    {
        char* label = NULL;

        try {label = Scan();} catch(ERParse)
        {
            delete[] tuple;
            throw ERParse();
        }

        if (strlen(label)>0)
        {
            if (length==k && length>0)
            {
                delete[] tuple;
                CT.Error(ERR_PARSE,NoHandle,"GetTFloatTuple","Length mismatch");
            }
        
            if (strcmp(label,"*")) tuple[length++] = atof(label);
            else tuple[length++] = InfFloat;
        }
    }   
    
    if (length!=1 && length!=k && k!=0)
    {
        delete[] tuple;
        CT.Error(ERR_PARSE,NoHandle,"GetTFloatTuple","Length mismatch");
    }
                                     
    return tuple;
}


TIndex* goblinImport::GetTIndexTuple(unsigned long k)
    throw(ERParse)
{               
    TIndex* tuple = new TIndex[k];                          

    if (k==0) tuple = new TIndex[1];

    length = 0;

    while (!tail)
    {
        char* label = NULL;
        try {label = Scan();} catch(ERParse)
        {
            delete[] tuple;
            throw ERParse();
        }

        if (strlen(label)>0)
        {
            if (length==k && length>0)
            {
                delete[] tuple;
                CT.Error(ERR_PARSE,NoHandle,"GetTIndexTuple","Length mismatch");
            }
        
            if (strcmp(label,"*")) tuple[length++] = atol(label);
            else tuple[length++] = NoIndex;
        }
    }
    
    if (length!=1 && length!=k && k!=0)
    {
        delete[] tuple;
        CT.Error(ERR_PARSE,NoHandle,"GetTIndexTuple","Length mismatch");
    }
                                     
    return tuple;
}


char* goblinImport::GetCharTuple(unsigned long k)
    throw(ERParse)
{               
    char* tuple = new char[k];

    if (k==0) tuple = new char[1];

    length = 0;

    while (!tail)
    {
        char* label = NULL;
        try {label = Scan();} catch(ERParse)
        {
            delete[] tuple;
            throw ERParse();
        }

        if (strlen(label)>0)
        {
            if (length==k && length>0)
            {
                delete[] tuple;
                CT.Error(ERR_PARSE,NoHandle,"GetCharTuple","Length mismatch");
            }
        
            tuple[length++] = (char)atoi(label);
        }
    }
    
    if (length!=1 && length!=k && k!=0)
    {
        delete[] tuple;
        CT.Error(ERR_PARSE,NoHandle,"GetCharTuple","Length mismatch");
    }
                                     
    return tuple;
}


void goblinImport::ReadConfiguration()
    throw(ERParse)
{               
    Scan("configure");
    
    int ParamCount = 0;
    char** ParamStr = new char*[500];
    unsigned offset = 0;
    char* ParamBuffer = new char[5000];
    bool truncated = false;

    try
    {
        while (!Tail())
        {
            ParamCount++;
            char* token = Scan();

            if (ParamCount==500 || strlen(token)+1>2000-offset)
            {
                truncated = true;
                break;
            }

            ParamStr[ParamCount] = &ParamBuffer[offset];
            strcpy(ParamStr[ParamCount],token);
            offset += strlen(token)+1;
        }
        
        while (!Tail()) Scan();
    }
    catch (ERParse)
    {
        delete[] ParamStr;
        delete[] ParamBuffer;
    }
    
    CT.Configure(ParamCount+1,(const char**)ParamStr);
    
    delete[] ParamStr;
    delete[] ParamBuffer;

    if (truncated)
        CT.Error(MSG_WARN,NoHandle,"ReadConfiguration",
            "Buffer overflow: Configuration truncated");
}


bool goblinImport::Constant()
    throw()
{               
    return (length == 1);
}


unsigned long goblinImport::Length()
    throw()
{               
    return length;
}


void goblinImport::DontComplain()
    throw()
{               
    complain = false;
}


goblinImport::~goblinImport()
    throw(ERParse)
{
    impFile.close();
    
    delete[] buffer;

    if (currentLevel>0 && complain)
        CT.Error(MSG_WARN,NoHandle,"goblinImport","Parenthesis mismatch");
}
