
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, November 2002.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "absmix.h"
#include "statstaq.h"
#include "matrix.h"


// ------------------------------------------------------


void abstractMixedGraph::StripEmbedding() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (CT.methGeometry!=0)
        Error(ERR_REJECTED,"StripEmbedding","Coordinates are fixed");

    #endif
    
    if (n+ni == 0) return;
    
    TFloat minX = C(0,0);
    TFloat minY = C(0,1);
    
    for (TNode v=0;v<n+ni;v++)
    {
        if (C(v,0)<minX) minX = C(v,0);
        if (C(v,1)<minY) minY = C(v,1);
    }
    
    if (CT.methGeometry==4)
    {
        minX = floor(minX);
        minY = floor(minY);
    }
    
    for (TNode v=0;v<n+ni;v++)
    {
        SetC(v,0,C(v,0)-minX);
        SetC(v,1,C(v,1)-minY);
    }
}

    
void abstractMixedGraph::ScaleEmbedding(TFloat dx,TFloat dy) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (CT.methGeometry!=0)
        Error(ERR_REJECTED,"ScaleEmbedding","Coordinates are fixed");

    #endif
    
    for (TNode v=0;v<n+ni;v++)
    {
        SetC(v,0,C(v,0)*dx);
        SetC(v,1,C(v,1)*dy);
    }
}

    
void abstractMixedGraph::AutoArcAlignment(TFloat sep,TOption opt) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (CT.methGeometry!=0 && IsDense())
        Error(ERR_REJECTED,"AutoArcAlignment","Coordinates are fixed");

    #endif
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    staticStack<TArc,TFloat> S(2*m,CT);
    
    if (sep<=0) sep = CT.nodeSep/2;

    for (TArc a=0;a<m;a++) ReleaseBendNodes(2*a);
    
    for (TNode u=0;u<n;u++)
    {
        for (TNode v=u;v<n;v++)
        {
            S.Init();

            while (I.Active(u))
            {
                TArc a = I.Read(u);
                if (EndNode(a)==v && (u!=v || (a%2))) S.Insert(a);
            }

            I.Reset(u);
            
            if (v!=u)
            {
                if (S.Cardinality()>1)
                {
                    TFloat dx = C(v,0)-C(u,0);
                    TFloat dy = C(v,1)-C(u,1);
                    TFloat norm = sqrt(dx*dx+dy*dy);
                    TFloat ox = dy/norm;
                    TFloat oy = -dx/norm;
                    TFloat cx = (C(v,0)+C(u,0))/2-norm*0.2*(S.Cardinality()-1)/2*ox;
                    TFloat cy = (C(v,1)+C(u,1))/2-norm*0.2*(S.Cardinality()-1)/2*oy;

                    while (!S.Empty())
                    {
                        TArc a = S.Delete();
                        TNode w = InsertAlignmentPoint(a);
                        SetC(w,0,cx);
                        SetC(w,1,cy);
                        w = InsertBendNode(w);
                        SetC(w,0,cx);
                        SetC(w,1,cy);
                        cx += norm*0.2*ox;
                        cy += norm*0.2*oy;
                    }
                }
            }
            else
            {
                if (!opt)
                {
                    TFloat cx = C(v,0);
                    TFloat cy = C(v,1)+sep;

                    while (!S.Empty())
                    {
                        TArc a = S.Delete();
                        TNode w = InsertAlignmentPoint(a);
                        SetC(w,0,cx);
                        SetC(w,1,cy);
                        w = InsertBendNode(w);
                        SetC(w,0,cx+sep/4);
                        SetC(w,1,cy);
                        w = InsertBendNode(w);
                        SetC(w,0,cx-sep/4);
                        SetC(w,1,cy);
                        cy += sep;
                    }
                }
            }
        }
    }

    Close(H);
}

    
void abstractMixedGraph::CircleLayout(TFloat sep) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (CT.methGeometry!=0 && IsDense())
        Error(ERR_REJECTED,"TreeLayout","Coordinates are fixed");

    #endif
    
    if (sep==0) sep = CT.nodeSep;

    TFloat rad = sep*n/2/3.1415;

    for (TNode v=0;v<n;v++)
    {
        SetC(v,0,rad*(1+cos(v*2*3.1415/n)));
        SetC(v,1,rad*(1+sin(v*2*3.1415/n)));
    }
    
    if (IsSparse()) AutoArcAlignment();
}

    
void abstractMixedGraph::GridLayout(TFloat sep) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (CT.methGeometry!=0 && IsDense())
        Error(ERR_REJECTED,"TreeLayout","Coordinates are fixed");

    #endif
    
    if (sep==0) sep = CT.nodeSep;

    TNode k = TNode(ceil(sqrt(TFloat(n))));

    for (TNode i=0;i<k;i++)
    {
        for (TNode j=0;j<k;j++)
        {
            if (i*k+j<n)
            {
                SetC(i*k+j,1,i*sep);
                SetC(i*k+j,0,j*sep);
            }
        }
    }
    
    if (IsSparse()) AutoArcAlignment();
}

    
void abstractMixedGraph::ColourLayout(TFloat dx,TFloat dy) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (colour==NULL)
        Error(ERR_REJECTED,"ColourLayout","Missing colours");

    if (CT.methGeometry!=0 && IsDense())
        Error(ERR_REJECTED,"TreeLayout","Coordinates are fixed");

    #endif
    
    if (dx==0) dx = CT.nodeSep;
    if (dy==0) dy = CT.nodeSep;

    TNode maxCardinality = 0;

    for (TNode k=0;k<=n;k++)
    {
        TNode cardinality = 0;

        for (TNode v=0;v<n;v++)
            if ((k<n && colour[v]==k) || colour[v]>=n) cardinality++;

        if (cardinality>maxCardinality) maxCardinality = cardinality;
    }

    for (TNode k=0;k<=n;k++)
    {
        TNode cardinality = 0;

        for (TNode v=0;v<n;v++)
            if ((k<n && colour[v]==k) || colour[v]>=n) cardinality++;

        TNode i = 0;

        for (TNode v=0;v<n;v++)
        {
            if ((k<n && colour[v]==k) || colour[v]>=n)
            {
                SetC(v,0,dx*k);

                if (cardinality==1) SetC(v,1,0);
                else SetC(v,1,dy*i*(maxCardinality-1)/(cardinality-1));

                i++;
            }
        }
    }
    
    if (IsSparse()) AutoArcAlignment();
}

    
void abstractMixedGraph::TreeLayout(TFloat dx,TFloat dy) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (P==NULL)
        Error(ERR_REJECTED,"TreeLayout","Missing predecessors");

    if (CT.methGeometry!=0 && IsDense())
        Error(ERR_REJECTED,"TreeLayout","Coordinates are fixed");

    #endif
    
    if (dx<=0) dx = CT.nodeSep;
    if (dy<=0) dy = CT.nodeSep;
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    staticStack<TNode,TFloat> S(n,CT);

    TNode* tmpAlign = new TNode[n];
    TNode* leftAlign = new TNode[n];
    TNode* rightmost = new TNode[n];
    TNode* depth = new TNode[n];
    TNode* traversed = new TNode[n];
    TNode* nr = new TNode[n];

    TNode maxDepth = 0;
    rightmost[0] = 0;
    TNode cNr = 0;

    for (TNode v=0;v<n;v++)
        tmpAlign[v] = leftAlign[v] = rightmost[v] = NoNode;

    for (TNode r=0;r<n;r++)
    {
        if (P[r]==NoArc)
        {
            SetC(r,0,-InfFloat);
            SetC(r,1,-InfFloat);
        }

        if (P[r]==NoArc && (colour==NULL || colour[r]!=NoNode))
        {
            depth[r] = 0;
            traversed[cNr] = r;
            cNr++;

            TNode u = r;
            TNode cDepth = 0;
            
            while (true)
            {
                if (I.Active(u))
                {
                    TArc a = I.Read(u);
                    TNode v = EndNode(a);

                    if (P[v]==a)
                    {
                        S.Insert(u);
                        cDepth++;
                        depth[v] = cDepth;

                        if (cDepth>maxDepth) maxDepth = cDepth;

                        traversed[cNr] = v;
                        nr[v] = cNr;
                        cNr++;
                        u = v;
                    }
                    
                    continue;
                }

                // Backtracking

                TNode pos = rightmost[depth[u]];
                if (leftAlign[u]==NoNode)
                {
                    if (pos==NoNode) tmpAlign[u] = 0;
                    else tmpAlign[u] = pos+1;
                }

                if (pos!=NoNode && pos>=tmpAlign[u])
                {
                    // Shift partial tree rooted at u

                    TNode shift = pos-tmpAlign[u]+1;

                    for (TNode i=nr[u];i<cNr;i++)
                    {
                        TNode x = traversed[i];

                        if (tmpAlign[x]==rightmost[depth[x]])
                            rightmost[depth[x]] += shift;

                        tmpAlign[x] += shift;
                    }
                }

                rightmost[depth[u]] = tmpAlign[u];

                if (S.Empty()) break;

                TNode v = S.Delete();

                if (leftAlign[v] == NoNode)
                    leftAlign[v] = tmpAlign[v] = tmpAlign[u];
                else tmpAlign[v] = (leftAlign[v]+tmpAlign[u])/2;

                u = v;
                cDepth --;
            }
        }
    }

    for (TNode i=0;i<cNr;i++)
    {
        TNode u = traversed[i];
        SetC(u,0,tmpAlign[u]*dx);
        SetC(u,1,depth[u]*dy);
    }
    
    delete[] tmpAlign;
    delete[] leftAlign;
    delete[] rightmost;
    delete[] depth;
    delete[] traversed;
    delete[] nr;

    Close(H);
    
    if (IsSparse()) AutoArcAlignment((dx+dy)/4);
}

    
void abstractMixedGraph::SpringEmbedding(TFloat gamma,TFloat delta)
    throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (CT.methGeometry!=0 && IsDense())
        Error(ERR_REJECTED,"SpringEmbedding","Coordinates are fixed");

    #endif
    
    LogEntry(LOG_METH,"Computing spring embedding...");
    OpenFold();
    
    for (TArc a=0;a<m;a++) ReleaseBendNodes(2*a);
    
    if (gamma==0) gamma = 0.5; // 0.5;
    if (delta==0) delta = 5.0/n; // 0.5;

    denseMatrix<TIndex,TFloat> J1(2*n-2,2*n-2);
    denseMatrix<TIndex,TFloat> J2(2*n-2,2*n-2);
    denseMatrix<TIndex,TFloat> F(2*n-2,1);
    denseMatrix<TIndex,TFloat> F2(2*n-2,1);

    TFloat          stepLength  = InfFloat;
    unsigned long   stepCount   = 1;
    bool            stepValid   = true;
    TFloat          redFactor   = 1;

    while (CT.SolverRunning() && stepLength>0.1*log(double(n)))
    {
        // Compute Jacobian matrix J1 and reset J2 to identity
        
        for (TNode i=0;i<n-1 && stepValid;i++)
        {
            F.SetCoeff(2*i,0,0);
            F.SetCoeff(2*i+1,0,0);

            J1.SetCoeff(2*i   ,2*i   ,0);
            J1.SetCoeff(2*i   ,2*i+1 ,0);
            J1.SetCoeff(2*i+1 ,2*i   ,0);
            J1.SetCoeff(2*i+1 ,2*i+1 ,0);

            for (TNode j=0;j<n;j++)
            {
                if (j<n-1)
                {
                    J2.SetCoeff(2*i   ,2*j+1, 0);
                    J2.SetCoeff(2*i+1 ,2*j,   0);

                    if (i==j)
                    {
                        J2.SetCoeff(2*i+1 ,2*j+1, 1);
                        J2.SetCoeff(2*i   ,2*j,   1);
                        continue;
                    }
                    else
                    {
                        J2.SetCoeff(2*i+1 ,2*j+1, 0);
                        J2.SetCoeff(2*i   ,2*j,   0);
                    }
                }
                
                TFloat diffx = C(i,0)-C(j,0);
                TFloat diffy = C(i,1)-C(j,1);
                TFloat diff2 = delta*(diffx*diffx+diffy*diffy);
                
                if (diff2<CT.epsilon)
                {
                    sprintf(CT.logBuffer,"diff2 = %g",diff2);
                    LogEntry(LOG_METH2,CT.logBuffer);
                    stepValid = false;
                    break;
                }

                TFloat coeff_x_x = 2*diffx*diffx/(diff2*diff2)-1/diff2;
                TFloat coeff_x_y = 2*diffx*diffy/(diff2*diff2);
                TFloat coeff_y_y = 2*diffy*diffy/(diff2*diff2)-1/diff2;
                
                F.SetCoeff(2*i  ,0,F.Coeff(2*i  ,0)+diffx/diff2);
                F.SetCoeff(2*i+1,0,F.Coeff(2*i+1,0)+diffy/diff2);
                
                // Ignore arc orientations

                if (Adjacency(i,j)!=NoArc || Adjacency(j,i)!=NoArc)
                {
                    coeff_x_x += gamma;
                    coeff_y_y += gamma;

                    F.SetCoeff(2*i  ,0,F.Coeff(2*i  ,0)-diffx*gamma);
                    F.SetCoeff(2*i+1,0,F.Coeff(2*i+1,0)-diffy*gamma);
                }
                
                if (j<n-1)
                {
                    J1.SetCoeff(2*i   ,2*j   ,coeff_x_x);
                    J1.SetCoeff(2*i   ,2*j+1 ,coeff_x_y);
                    J1.SetCoeff(2*i+1 ,2*j   ,coeff_x_y);
                    J1.SetCoeff(2*i+1 ,2*j+1 ,coeff_y_y);
                }
                
                J1.SetCoeff(2*i   ,2*i   ,J1.Coeff(2*i  ,2*i  )-coeff_x_x);
                J1.SetCoeff(2*i   ,2*i+1 ,J1.Coeff(2*i  ,2*i+1)-coeff_x_y);
                J1.SetCoeff(2*i+1 ,2*i   ,J1.Coeff(2*i+1,2*i  )-coeff_x_y);
                J1.SetCoeff(2*i+1 ,2*i+1 ,J1.Coeff(2*i+1,2*i+1)-coeff_y_y);
            }
        }
        
        try
        {
            // Invert J1

            J1.GaussElim(J2,0.00000001);
        }
        catch (ERRejected)
        {
            stepValid = false;
        }
        
        if (stepValid)
        {
            // Do Newton step x = x + J^(-1)*F
        
            F2.Product(J2,F);

            stepLength = 0;

            for (TNode i=0;i<n-1;i++)
            {
                SetC(i,0,C(i,0)-redFactor*F2.Coeff(2*i,0));
                SetC(i,1,C(i,1)-redFactor*F2.Coeff(2*i+1,0));
                stepLength += redFactor*(fabs(F2.Coeff(2*i,0))+fabs(F2.Coeff(2*i+1,0)));
            }
            
            // redFactor *= 0.99;
        
            sprintf(CT.logBuffer,"...Step %ld has length %g",stepCount,stepLength);
            LogEntry(LOG_METH2,CT.logBuffer);
        }
        else
        {
            LogEntry(LOG_METH2,"...Jacobian is undefined or singular");

            if (stepCount>1)
            {
                for (TNode i=0;i<n-1;i++)
                {
                    SetC(i,0,C(i,0)-0.1*F2.Coeff(2*i,0));
                    SetC(i,1,C(i,1)-0.1*F2.Coeff(2*i+1,0));
                }
        
                LogEntry(MSG_APPEND," (step repeated)");
            }
            else
            {
                stepLength = 0;
        
                LogEntry(MSG_APPEND," (start with different embedding)");
            }

            stepValid = true;
        }

        stepCount++;

        AutoArcAlignment();
        StripEmbedding();
        CT.Ping(Handle(),n*n);
    }
    
    CloseFold();
    
    if (CT.logMeth==1)
    {
        sprintf(CT.logBuffer,"...%ld Newton iterations in total",stepCount-1);
        LogEntry(LOG_METH,CT.logBuffer);
    }
}
