
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, October 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             

#include "setfam.h"
#include "treeView.h"


template <class TItem>
goblinSetFamily<TItem>::goblinSetFamily(TItem nn,goblinController &thisContext)
    throw() : goblinDataObject(thisContext)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    n = nn;
    UNDEFINED = n;
    
    B = new TItem[n];
    rank = new TItem[n];
    Init();

    LogEntry(LOG_MEM,"...Disjoint set family allocated");

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif
}


template <class TItem>
void goblinSetFamily<TItem>::Init() throw()
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    for (TNode v=0;v<n;v++) B[v]=UNDEFINED;

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif
}


template <class TItem>
unsigned long goblinSetFamily<TItem>::Size() throw()
{
    return
          sizeof(goblinSetFamily<TItem>)
        + goblinDataObject::Allocated()
        + goblinSetFamily::Allocated();
}


template <class TItem>
unsigned long goblinSetFamily<TItem>::Allocated() throw()
{
    return 2*n*(int)sizeof(TItem);
}


template <class TItem>
char* goblinSetFamily<TItem>::Display() throw()
{
    if (CT.displayMode>0)
    {
        goblinTreeView G(n,CT);
        G.InitPredecessors();
        
        bool voidStructure = true;
        for (TItem i=0;i<n;i++)
        {
            if (B[i]==UNDEFINED) G.SetColour(i,NoNode);
            else
            {
                G.SetColour(i,TNode(rank[i]));
                G.SetDist(i,TFloat(i));
                voidStructure = false;
    
                if (B[i]!=i)
                {
                    TArc a = G.InsertArc(TNode(B[i]),TNode(i));
                    G.SetPred(TNode(i),2*a);
                }
            }
        }
        
        if (!voidStructure)
        {
            G.TreeLayout();
            G.Display();
        }
    }
    else
    {
        LogEntry(MSG_TRACE,"Disjoint set family");
        THandle LH = LogStart(MSG_TRACE2,"  ");
        TNode counter = 0;

        for (TNode v=0;v<n;v++)
            if (B[v]!=UNDEFINED)
            {
                if ((++counter)%10==0)
                {
                    LogEnd(LH);
                    LH = LogStart(MSG_TRACE2,"  ");
                }
                sprintf(CT.logBuffer," %ld->%ld",v,B[v]);
                LogAppend(LH,CT.logBuffer);
                counter++;
            }
        
        LogEnd(LH);
    }
    
    return NULL;
}


template <class TItem>
void goblinSetFamily<TItem>::Bud(TItem v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (CT.logWarn && v>=n) NoSuchItem("Find",v);

    #endif

    B[v] = v;
    rank[v] = 1;
}


template <class TItem>
void goblinSetFamily<TItem>::Merge(TItem u,TItem v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (u>=n || B[u]==UNDEFINED) NoSuchItem("Find",u);

    if (v>=n || B[v]==UNDEFINED) NoSuchItem("Find",v);

    #endif

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    u = Find(u);
    v = Find(v);

    if (rank[u]<rank[v]) B[u] = v;
    else         
    {                    
        B[v] = u;
        if ((rank[v]==rank[u])&&(u!=v)) rank[u]++;
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif

    if (CT.traceData) Display();
}


template <class TItem>
TItem goblinSetFamily<TItem>::Find(TItem v) throw(ERRange)
{                    
    #if defined(_FAILSAVE_)

    if (CT.logWarn && (v>=n || B[v]==UNDEFINED)) NoSuchItem("Find",v);

    #endif

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    TNode w = v;

    if (B[v]!=v) w = Find(B[v]);

    if (CT.methDSU==1 && B[v]!=w)
    {
        B[v] = w;
        if (CT.traceData) Display();
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif

    return w;
}


template <class TItem>
goblinSetFamily<TItem>::~goblinSetFamily() throw()
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    delete[] B;
    delete[] rank;

    LogEntry(LOG_MEM,"...Disjoint set family disallocated");

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif
}


template class goblinSetFamily<TNode>;

#if defined(_BIG_ARCS_)

template class goblinSetFamily<TArc>;

#endif

