
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, January 2000.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             

#include "shrfam.h"
#include "treeView.h"


template <class TItem>
shrinkingFamily<TItem>::shrinkingFamily(TItem nn,TItem mm,
    goblinController &thisContext) throw() : goblinDataObject(thisContext)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    n = nn;
    m = mm;
    compress = CT.methDSU;
    UNDEFINED = n+m;
    
    B = new TItem[n+m];
    depth = new TItem[n+m];
    set = new TItem[n+m];
    canonical = new TItem[m];
    first = new TItem[m];
    next = new TItem[n+m];
    
    Init();

    LogEntry(LOG_MEM,"...Shrinking family allocated");

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif
}


template <class TItem>
void shrinkingFamily<TItem>::Init() throw()
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    for (TNode v=0;v<n+m;v++) B[v]=UNDEFINED;
    for (TNode v=0;v<n;v++) Bud(v);

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif
}


template <class TItem>
unsigned long shrinkingFamily<TItem>::Size() throw()
{
    return
          sizeof(shrinkingFamily<TItem>)
        + goblinDataObject::Allocated()
        + shrinkingFamily::Allocated();
}


template <class TItem>
unsigned long shrinkingFamily<TItem>::Allocated() throw()
{
    return (4*n+6*m)*(int)sizeof(TItem);
}


template <class TItem>
shrinkingFamily<TItem>::~shrinkingFamily() throw()
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    delete[] B;
    delete[] depth;
    delete[] set;
    delete[] canonical;
    delete[] first;
    delete[] next;

    LogEntry(LOG_MEM,"...Shrinking family disallocated");

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif
}


static THandle LH = NoHandle;
 
template <class TItem>
char* shrinkingFamily<TItem>::Display() throw()
{
    if (CT.displayMode>0)
    {
        goblinTreeView G(n+m,CT);
        G.InitPredecessors();
        
        bool voidStructure = true;

        for (TItem i=0;i<n+m;i++)
        {
            if (B[i]==UNDEFINED || (i<n && B[i]==i))
                G.SetColour(i,NoNode);
            else
            {
                G.SetColour(i,TNode(depth[i]));
                G.SetDist(i,TFloat(i));
                voidStructure = false;
    
                if (B[i]!=i)
                {
                    TArc a = G.InsertArc(TNode(B[i]),TNode(i));
                    G.SetPred(TNode(i),2*a);
                }
            }
        }
        
        if (!voidStructure)
        {
            G.TreeLayout();
            G.Display();
        }
    }
    else
    {
        LogEntry(MSG_TRACE,"Shrinking family");
    
        for (TNode v=0;v<n+m;v++)
            if (B[v]!=UNDEFINED && Top(v))
            {
                LH = LogStart(MSG_TRACE2,"   ");
                Display(v);
                LogEnd(LH);
            }
    }
    
    return NULL;
}


template <class TItem>
void shrinkingFamily<TItem>::Display(TItem v) throw(ERRange)
{
    if (v<n)
    {
        sprintf(CT.logBuffer,"%ld",v);
        LogAppend(LH,CT.logBuffer);
    }
    else
    {
        sprintf(CT.logBuffer,"(%ld",v);
        LogAppend(LH,CT.logBuffer);

        TNode u = UNDEFINED;
        TNode w = first[v-n];

        while (u!=w)
        {
            u = w;
            LogAppend(LH," ");
            Display(u);
            w = next[w];
        }

        LogAppend(LH,")");
    }
}


template <class TItem>
void shrinkingFamily<TItem>::Bud(TItem v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchItem("Bud",v);

    if (B[v]!=UNDEFINED)
    {
        sprintf(CT.logBuffer,"Already present: %ld",v);
        Error(ERR_REJECTED,"Bud",CT.logBuffer);
    }

    #endif
    
    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    B[v] = v;
    depth[v] = 1;
    next[v] = UNDEFINED;
    set[v] = v;

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif
}


template <class TItem>
TItem shrinkingFamily<TItem>::MakeSet() throw(ERRejected)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    for (TNode u=n;u<n+m;u++)
    {
        if (B[u]==UNDEFINED)
        {
            B[u] = u;
            depth[u] = 1;
            first[u-n] = UNDEFINED;
            next[u] = UNDEFINED;
            set[u] = u;
            canonical[u-n] = UNDEFINED;

            #if defined(_TIMERS_)

            CT.globalTimer[TimerUnionFind] -> Disable();

            #endif
            
            return u;
        }
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif
    
    #if defined(_FAILSAVE_)

    Error(ERR_REJECTED,"MakeSet","No more sets available");

    #endif

    throw ERRejected();
}


template <class TItem>
void shrinkingFamily<TItem>::Merge(TItem s,TItem v) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (s<n || s>=n+m)
    {
        sprintf(CT.logBuffer,"Not a set: %ld",s);
        Error(ERR_RANGE,"Merge",CT.logBuffer);
    }

    if (canonical[s-n]!=UNDEFINED)
    {
        sprintf(CT.logBuffer,"Set is already fixed: %ld",s);
        Error(ERR_REJECTED,"Merge",CT.logBuffer);
    }

    if (v>=n && canonical[v-n]==UNDEFINED)
    {
        sprintf(CT.logBuffer,"Item has not been fixed: %ld",v);
        Error(ERR_REJECTED,"Merge",CT.logBuffer);
    }

    if (!Top(v))
    {
        sprintf(CT.logBuffer,"Item is already shrunk: %ld",v);
        Error(ERR_REJECTED,"Merge",CT.logBuffer);
    }

    #endif

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    TNode w = Find(v);
    TNode u = Find(s);

    if (depth[u]<depth[w])
    {
        B[u] = w;
        set[w] = s;
    }
    else
    {                    
        B[w] = u;
        set[u] = s;

        if (depth[w]==depth[u]) depth[u]++;
    }

    if (first[s-n]!=UNDEFINED)
    {
        next[v] = first[s-n];
        first[s-n] = v;
    }
    else
    {
        first[s-n] = v;
        next[v] = v;
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif
}


template <class TItem>
void shrinkingFamily<TItem>::FixSet(TItem s) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (s<n || s>=n+m)
    {
        sprintf(CT.logBuffer,"Not a set: %ld",s);
        Error(ERR_RANGE,"FixSet",CT.logBuffer);
    }

    if (canonical[s-n]!=UNDEFINED)
    {
        sprintf(CT.logBuffer,"Set is already fixed: %ld",s);
        Error(ERR_REJECTED,"FixSet",CT.logBuffer);
    }

    if (first[s-n]==UNDEFINED)
    {
        sprintf(CT.logBuffer,"Set is empty: %ld",s);
        Error(ERR_REJECTED,"FixSet",CT.logBuffer);
    }

    #endif

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    TNode u = Find(s);
    canonical[s-n] = u;

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif

    if (CT.traceData) Display();
}


template <class TItem>
bool shrinkingFamily<TItem>::Top(TItem v) throw(ERRange)
{                    
    #if defined(_FAILSAVE_)

    if (v>=n+m) NoSuchItem("Top",v);

    #endif

    if (B[v]==UNDEFINED)
    {
        #if defined(_FAILSAVE_)

        if (CT.logMeth>1 && CT.logWarn)
        {
            sprintf(CT.logBuffer,"No such item: %ld",v);
            Error(MSG_WARN,"Top",CT.logBuffer);
        }

        #endif

      	return 1;
    }
    
    return (next[v]==UNDEFINED);
}


template <class TItem>
TItem shrinkingFamily<TItem>::Find(TItem v) throw(ERRange)
{                    
    #if defined(_FAILSAVE_)

    if (v>=n+m) NoSuchItem("Find",v);

    #endif

    if (B[v]==UNDEFINED)
    {
        #if defined(_FAILSAVE_)

        if (CT.logMeth>1 && CT.logWarn)
        {
            sprintf(CT.logBuffer,"No such item: %ld",v);
            Error(MSG_WARN,"Find",CT.logBuffer);
        }

        #endif

      	return UNDEFINED;
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    TNode w = v;

    if (B[v]!=v) w = Find(B[v]);
    if (compress) B[v] = w;

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif

    return w;
}


template <class TItem>
TItem shrinkingFamily<TItem>::Set(TItem v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n+m) NoSuchItem("Set",v);

    #endif

    if (B[v]==UNDEFINED)
    {
        #if defined(_FAILSAVE_)

        if (CT.logMeth>1 && CT.logWarn)
        {
            sprintf(CT.logBuffer,"No such item: %ld",v);
            Error(MSG_WARN,"Set",CT.logBuffer);
        }

        #endif

        return UNDEFINED;
    }

    return set[Find(v)];
}


template <class TItem>
TItem shrinkingFamily<TItem>::First(TItem v) throw(ERRange)
{                    
    #if defined(_FAILSAVE_)

    if (v>=n+m || v<n)
    {
        sprintf(CT.logBuffer,"Not a set: %ld",v);
        Error(ERR_RANGE,"First",CT.logBuffer);
    }

    #endif

    if (B[v]==UNDEFINED)
    {
        #if defined(_FAILSAVE_)

        if (CT.logMeth>1 && CT.logWarn)
        {
            sprintf(CT.logBuffer,"No such item: %ld",v);
            Error(MSG_WARN,"First",CT.logBuffer);
        }

        #endif

      	return UNDEFINED;
    }

    if (B[v]==UNDEFINED)
    {
        #if defined(_FAILSAVE_)

        if (CT.logMeth>1 && CT.logWarn)
        {
            sprintf(CT.logBuffer,"Empty set: %ld",v);
            Error(MSG_WARN,"First",CT.logBuffer);
        }

        #endif

      	return UNDEFINED;
    }

    return first[v-n];
}


template <class TItem>
TItem shrinkingFamily<TItem>::Next(TItem v) throw(ERRange,ERRejected)
{                    
    #if defined(_FAILSAVE_)

    if (v>=n+m) NoSuchItem("Next",v);

    if (B[v]==UNDEFINED)
    {
        sprintf(CT.logBuffer,"No such item: %ld",v);
        Error(ERR_REJECTED,"Next",CT.logBuffer);
    }

    if (next[v]==UNDEFINED)
    {
        sprintf(CT.logBuffer,"Toplevel item: %ld",v);
        Error(ERR_REJECTED,"Next",CT.logBuffer);
    }

    #endif

    return next[v];
}


template <class TItem>
void shrinkingFamily<TItem>::Adjust(TItem s,TItem b) throw(ERRange)
{                    
    #if defined(_FAILSAVE_)

    if (s>=n+m || s<n)
    {
        sprintf(CT.logBuffer,"Not a set: %ld",s);
        Error(ERR_RANGE,"Adjust",CT.logBuffer);
    }

    if (b>=n+m) NoSuchItem("Adjust",b);

    #endif
    
    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    TNode u = UNDEFINED;
    TNode w = first[s-n];

    while (u!=w)
    {
        u = w;
        B[u]=b;
        if (u>=n) Adjust(u,b);
        w = next[w];
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif
}


template <class TItem>
void shrinkingFamily<TItem>::Split(TItem s) throw(ERRange,ERRejected)
{                    
    #if defined(_FAILSAVE_)

    if (s>=n+m || s<n || B[s]==UNDEFINED)
    {
        sprintf(CT.logBuffer,"Not a set: %ld",s);
        Error(ERR_RANGE,"Split",CT.logBuffer);
    }

    if (first[s-n]==UNDEFINED)
    {
        sprintf(CT.logBuffer,"Empty set: %ld",s);
        Error(ERR_REJECTED,"Split",CT.logBuffer);
    }

    TNode u = Find(s);

    if (set[u]!=s)
    {
        sprintf(CT.logBuffer,"Not a toplevel set: %ld",s);
        Error(ERR_REJECTED,"Split",CT.logBuffer);
    }

    if (canonical[s-n]==UNDEFINED)
    {
        sprintf(CT.logBuffer,"Set has not been fixed: %ld",s);
        Error(ERR_REJECTED,"Split",CT.logBuffer);
    }

    #endif
    
    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    TNode v = UNDEFINED;
    TNode w = first[s-n];

    while (v!=w)
    {
        v = w;
        
        if (v<n)
        {
            B[v] = v;
            if (compress) set[v] = v;
        }
        else
        {
            TNode x = canonical[v-n];
            B[v] = B[x] = x;
            if (compress) Adjust(v,x);
            set[x] = v;
        }
        
        w = next[w];
        next[v] = UNDEFINED;
    }
    
    B[s] = UNDEFINED;

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif

    if (CT.traceData) Display();
}


template <class TItem>
void shrinkingFamily<TItem>::Block(TItem v) throw(ERRange,ERRejected)
{                    
    #if defined(_FAILSAVE_)

    if (v>=n+m) NoSuchItem("Block",v);

    if (B[v]==UNDEFINED)
    {
        sprintf(CT.logBuffer,"No such item: %ld",v);
        Error(ERR_REJECTED,"Block",CT.logBuffer);
    }

    #endif
    
    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    TNode u = UNDEFINED;
    TNode w = first[v-n];

    while (u!=w)
    {
        u = w;
        
        if (u<n)
        {
            B[u] = u;
            if (compress) set[u] = u;
        }
        else
        {
            TNode x = canonical[u-n];
            B[u] = B[x] = x;
            if (compress) Adjust(u,x);
            set[x] = u;
        }
        
        w = next[w];
    }

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif
}


template <class TItem>
void shrinkingFamily<TItem>::UnBlock(TItem v) throw(ERRange,ERRejected)
{                    
    #if defined(_FAILSAVE_)

    if (v>=n+m) NoSuchItem("Block",v);

    if (B[v]==UNDEFINED)
    {
        sprintf(CT.logBuffer,"No such item: %ld",v);
        Error(ERR_REJECTED,"UnBlock",CT.logBuffer);
    }

    #endif
    
    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Enable();
    
    #endif

    TNode u = UNDEFINED;
    TNode w = first[v-n];
    TNode x = canonical[v-n];

    while (u!=w)
    {
        u = w;
        if (u<n) B[u] = x;
        else B[canonical[u-n]] = x;
        
        w = next[w];
    }

    set[x] = v;    

    #if defined(_TIMERS_)

    CT.globalTimer[TimerUnionFind] -> Disable();
    
    #endif
}


template class shrinkingFamily<TNode>;
