
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, September 2001.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "bbatsp.h"
#include "spsdig.h"
#include "dnsdig.h"
#include "binheap.h"
    

BBATSP::BBATSP(abstractMixedGraph& GC,TNode r) throw() :
    goblinBBNode<TArc,TFloat>(GC.M(),GC.Context()), G(GC)
{
    if (CT.methCandidates>=0 && G.IsDense()) SetCandidateGraph();
    else X = new diGraph(G,OPT_CLONE);

    H = X->Investigate();
    selected = 0;
    root = r;
    depth = TArc(X->N()*log(double(X->N())*0.1));

    for (TNode v=0;v<G.N();v++) X->SetPotential(v,G.Pi(v));

    for (TArc a=0;a<n;a++)
        if (X->StartNode(2*a)==X->EndNode(2*a)) Lower(a);

    for (TNode v=0;v<G.N();v++) CheckNode(v);
    
    LogEntry(LOG_MEM,"(asymmetric TSP)");
}


BBATSP::BBATSP(BBATSP& Node) throw() :
    goblinBBNode<TArc,TFloat>(Node.G.M(),Node.Context(),Node.scheme), G(Node.G)
{
    X = new diGraph(*Node.X,OPT_CLONE);
    H = X->Investigate();
    unfixed = Node.Unfixed();
    selected = Node.selected;
    root = Node.root;
    depth = TArc(X->N()*log(double(X->N())*0.1));
    
    for (TNode v=0;v<G.N();v++) X->SetPotential(v,G.Pi(v));

    for (TArc a=0;a<X->M();a++) X->SetSub(2*a,Node.X -> Sub(2*a));

    objective = Node.Objective();
    solved = 1;

    LogEntry(LOG_MEM,"(asymmetric TSP)");
}


BBATSP::~BBATSP() throw()
{
    X -> Close(H);
    delete X;
    
    LogEntry(LOG_MEM,"(asymmetric TSP)");
}


void BBATSP::SetCandidateGraph() throw()
{
    LogEntry(LOG_METH,"Constructing candidate graph...");

    CT.SuppressLogging();
    diGraph* Y = new diGraph(G,OPT_CLONE);
    Y -> SetCUCap(1);
    Y -> InitSubgraph();
    CT.RestoreLogging();

    for (TNode v=0;v<G.N();v++)
        if (G.Pred(v)!=NoArc)
        {
            TNode u = G.StartNode(G.Pred(v));
            TNode w = G.EndNode(G.Pred(v));
            Y -> AddArc(Y->Adjacency(u,w),1);
        }
        
    for (int i=0;i<20;i++)
    {
        CT.SuppressLogging();
        TFloat ret = Y->TSPHeuristicsRandom();
        CT.RestoreLogging();
        if (ret<InfFloat)
        {    
            for (TNode v=0;v<G.N();v++)
                if (Y->Sub(Y->Pred(v))==0) Y->AddArc(Y->Pred(v),1);

            #if defined(_LOGGING_)

            if (CT.logMeth>=2)
            {
                sprintf(CT.logBuffer,"Tour of length %g added",ret);
                LogEntry(LOG_METH2,CT.logBuffer);
            }
            
            #endif
        }
    }

    binaryHeap<TArc,TFloat> Q(2*Y->M(),CT);
    H = Y->Investigate();
    for (TNode v=0;v<G.N();v++)
    {
        while (Y->Active(H,v))
        {
            TArc a = Y->Read(H,v);
            Q.Insert(a,Y->Length(a));
        }
        
        int i = 0;
        while (!Q.Empty())
        {
            TArc a = Q.Delete();       
            if (Y->Sub(a)==0 && !G.Blocking(a) &&
                (i<CT.methCandidates || G.Sub(a)>0))
            {
                Y->AddArc(a,1);
                i++;
            }
        }
        
        Y -> Reset(H,v);
        while (Y->Active(H,v))
        {
            TArc a = Y->Read(H,v);
            Q.Insert(a,Y->Length(a));
        }
        
        i = 0;
        while (!Q.Empty())
        {
            TArc a = Q.Delete();       
            if (Y->Sub(a)==0 && !G.Blocking(a^1) &&
                (i<CT.methCandidates || G.Sub(a)>0))
            {
                Y->AddArc(a^1,1);
                i++;
            }
        }
    }
    Y->Close(H);
    
    X = new diGraph(*Y,OPT_SUB);
    X -> SetCUCap(1);
    n = unfixed = X->M();
        
    if (CT.traceLevel==3) CT.Ping(Y->Handle(),X->M());
    
    CT.SuppressLogging();
    delete Y;
    CT.RestoreLogging();
}


unsigned long BBATSP::Size() throw()
{
    return
          sizeof(BBATSP)
        + goblinDataObject::Allocated()
        + goblinBBNode<TArc,TFloat>::Allocated()
        + Allocated();
}


unsigned long BBATSP::Allocated() throw()
{
    return 0;
}


TArc BBATSP::SelectVariable() throw()
{
    TArc ret0 = NoArc;
    TArc ret1 = NoArc;

    for (TNode v=0;v<X->N() && ret1==NoArc;v++)
    {
        bool fixedIn = false;
        bool fixedOut = false;
        X -> Reset(H,v);
        TArc a0 = NoArc;
        while (X->Active(H,v) && (fixedIn==0 || fixedOut==0))
        {
            TArc a = X->Read(H,v);
            if (X->Sub(a)==1 && X->LCap(a)==0 &&
                (a0==NoArc || X->Length(a)<X->Length(a0))) a0 = a;
            if (X->LCap(a)==1 && !X->Blocking(a^1)) fixedIn = 1;
            if (X->LCap(a)==1 && !X->Blocking(a)) fixedOut = 1;
        }
        
        if (fixedIn==0 && fixedOut==0 && a0!=NoArc) ret0 = a0;
        if ((fixedIn==1 || fixedOut==1) && a0!=NoArc) ret1 = a0;
    }
    
    if (ret1!=NoArc) return ret1>>1;
    if (ret0!=NoArc) return ret0>>1;

    #if defined(_FAILSAVE_)
    
    InternalError("SelectVariable","Solution is fixed");

    #endif

    throw ERInternal();
}


char BBATSP::Direction1(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=n) NoSuchArc("Direction1",a);

    #endif
    
    return 1;
}


char BBATSP::Direction2(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=n) NoSuchArc("Direction2",a);

    #endif
    
    return 1;
}


goblinBBNode<TNode,TFloat> *BBATSP::Clone() throw()
{
    return new BBATSP(*this);
}


void BBATSP::Raise(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=n) NoSuchArc("Raise",a);

    #endif
    
    if (X->Sub(2*a)==0)
    {
        X -> AddArc(2*a,1);
        solved = 0;
    }

    X -> SetLCap(2*a,1);
    unfixed--;
    selected++;
    
    CheckNode(X->StartNode(2*a));
    CheckNode(X->EndNode(2*a));
    
    if (unfixed==0) solved = 0;
}


void BBATSP::Lower(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=n) NoSuchArc("Lower",a);

    #endif
    
    if (X->Sub(2*a)==1)
    {
        X -> OmitArc(2*a,1);
        solved = 0;
    }

    X -> SetUCap(2*a,0);
    unfixed--;
    
    CheckNode(X->StartNode(2*a));
    CheckNode(X->EndNode(2*a));
    
    if (unfixed==0) solved = 0;
}


void BBATSP::CheckNode(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=X->N()) NoSuchNode("CheckNode",v);

    #endif
    
    bool fixedIn = false;
    bool fixedOut = false;
    X -> Reset(H,v);
    while (X->Active(H,v) && (fixedIn==0 || fixedOut==0))
    {
        TArc a = X->Read(H,v);
        if (X->LCap(a)==1 && !(X->Blocking(a^1))) fixedIn = 1;
        if (X->LCap(a)==1 && !(X->Blocking(a))) fixedOut = 1;
    }

    if (fixedIn)
    {
        X -> Reset(H,v);
        while (X->Active(H,v))
        {
            TArc a2 = X->Read(H,v);
            if (X->LCap(a2)==0 && X->UCap(a2)==1 && !(X->Blocking(a2^1)))
                Lower(a2>>1);
        }
        return;
    }

    if (fixedOut)
    {
        X -> Reset(H,v);
        while (X->Active(H,v))
        {
            TArc a2 = X->Read(H,v);
            if (X->LCap(a2)==0 && X->UCap(a2)==1 && !(X->Blocking(a2)))
                Lower(a2>>1);
        }
        return;
    }

    TArc aIn = NoArc;
    TArc aOut = NoArc;
    TArc incidIn = 0;
    TArc incidOut = 0;
    X -> Reset(H,v);
    while ((incidIn<2 || incidOut<2) && X->Active(H,v))
    {
        TArc a = X->Read(H,v);
        if (X->UCap(a)==1)
        {
            if (!(X->Blocking(a^1)))
            {
                incidIn++;
                if (X->LCap(a)==0) aIn = a;
            }

            if (!(X->Blocking(a)))
            {
                incidOut++;
                if (X->LCap(a)==0) aOut = a;
            }
        }
    }
    
    if (incidIn==1) Raise(aIn>>1);
    if (incidOut==1) Raise(aOut>>1);
}


TFloat BBATSP::SolveRelaxation() throw()
{
    for (TNode v=0;v<X->N();v++)
    {
        bool fixedIn = false;
        bool fixedOut = false;
        X -> Reset(H,v);
        while (X->Active(H,v))
        {
            TArc a = X->Read(H,v);
            if (X->LCap(a)==1 && !(X->Blocking(a^1)))
            {
                if (!fixedIn) fixedIn = 1;
                else return InfFloat;
            }

            if (X->LCap(a)==1 && !(X->Blocking(a)))
            {
                if (!fixedOut) fixedOut = 1;
                else return InfFloat;
            }
        }
    }

    CT.SuppressLogging();
    
    if (X->CutNodes() || !(X->StronglyConnected()))
    {
        CT.RestoreLogging();
        return InfFloat;
    }
    
    TFloat objective = InfFloat;
    try
    {
        if (scheme!=NULL && scheme->Phase()==2)
        {
            X -> InitSubgraph();
            X -> ReleasePredecessors();
            objective = X->TSPSubOpt1Tree(root,InfFloat,CT.methRelaxTSP2|4);
        }
        else
        {
            X -> InitSubgraph();
            objective = X->MinTree(root,abstractMixedGraph::MTREE_1TREE);
            X -> AddToSubgraph();
            if (scheme!=NULL && objective<scheme->bestLower && scheme->Phase()>0)
            {
                X -> InitSubgraph();
                X -> ReleasePredecessors();
                objective = X->TSPSubOpt1Tree(root,InfFloat,CT.methRelaxTSP2|4);
            }
        }
    }
    catch (ERRejected)
    {
        CT.RestoreLogging();
        return InfFloat;
    }
    
    CT.RestoreLogging();

    if (unfixed==0 && !Feasible()) return InfFloat;
        
    return objective;
}


bool BBATSP::Feasible() throw()
{
    try
    {
        CT.SuppressLogging();
        X->ExtractCycles();
        CT.RestoreLogging();
    }
    catch (ERCheck)
    {
        X -> ReleaseDegrees();
        CT.RestoreLogging();
        return 0;
    }

    return 1;
}


TFloat BBATSP::LocalSearch() throw()
{
    G.InitPredecessors();

    for (TNode v=0;v<G.N();v++)
    {
        TArc a = X->Pred(v);
        TNode u = X->StartNode(a);
        TArc a2 = G.Adjacency(u,v);
        G.SetPred(v,a2);
    }
    
    CT.SuppressLogging();
    objective = G.TSPLocalSearch();
    CT.RestoreLogging();
    SetUpperBound(TimerTsp,objective);

    if (CT.traceLevel==3) G.Display();
    
    return objective;
};


void BBATSP::SaveSolution() throw()
{
}


TFloat abstractMixedGraph::TSPBranchAndBound(TNode r,TFloat upperBound) throw(ERRejected)
{
    LogEntry(LOG_METH,"ATSP Branch and Bound...");
    OpenFold(ModTSP);
        
    BBATSP *rootNode = new BBATSP(*this,r);

    if (P==NULL) upperBound = InfFloat;
    char mode = 1;
    if (CT.methCandidates<0) mode = 2;
    goblinBBScheme<TNode,TFloat> scheme(rootNode,upperBound,mode);

    CloseFold(ModTSP);

    if (scheme.bestBound!=InfFloat)
    {
        SetUpperBound(TimerTsp,scheme.bestBound);

        if (CT.logRes)
        {
            sprintf(CT.logBuffer,"...Optimal tour has length %g",scheme.bestBound);
            LogEntry(LOG_RES,CT.logBuffer);
        }
        return scheme.bestBound;
    }
    
    LogEntry(LOG_RES,"...Problem is infeasible");

    return InfFloat;
}


TFloat abstractMixedGraph::TSP(TNode r) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (r>=n && r!=NoNode) NoSuchNode("TSP",r);

    #endif
    
    LogEntry(LOG_METH,"Starting TSP solver...");
    OpenFold(ModTSP);
    
    TArc* savedTour = NULL;
    TFloat savedLength = 0;

    if (P!=NULL)
    {
        TNode i = 0;
        TNode u = NoNode;
        TArc a = P[0];
        while (i<n+1 && u!=0 && a!=NoArc)
        {
            savedLength += Length(a);
            i++;
            u = StartNode(a);
            a = P[u];
        }
        
        if (i==n)
        {
            savedTour = P;
            P = NULL;

            if (CT.logMeth)
            {
                sprintf(CT.logBuffer,"...Initial tour has length %g",savedLength);
                LogEntry(LOG_RES,CT.logBuffer);
            }

            SetUpperBound(TimerTsp,savedLength);
        }
    }
    
    TFloat lower = InfFloat;

    if (!IsDense())
    {
        LogEntry(LOG_METH,"Checking for feasibility...");
    
        OpenFold();
        lower = MinTree(r,abstractMixedGraph::MTREE_1TREE);
        CloseFold();
        if (lower==InfFloat)
        {
            LogEntry(LOG_RES,"...Graph is non-Hamiltonian");
            CloseFold(ModTSP);
            throw ERRejected();
        }
        else LogEntry(LOG_RES,"...Check passed successfully");
    }
            
    TFloat bestUpper = InfFloat;
    try
    {
        if (CLCap() && CUCap())
        {
            bestUpper = TSPHeuristics(r);
            SetUpperBound(TimerTsp,bestUpper);
        }
    }
    catch (ERRejected) {};
    
    if (savedTour!=NULL)
    {
        if (savedLength<bestUpper)
        {
            if (P!=NULL) ReleasePredecessors();
            P = savedTour;
            bestUpper = savedLength;
            SetUpperBound(TimerTsp,bestUpper);
        }
        else delete[] savedTour;
    }
    
    if (CT.methSolve)
    {
        LogEntry(LOG_METH,"Computing lower bound...");
    
        try
        {
            if (r==NoNode) lower = TSPSubOpt1Tree(0,bestUpper,CT.methRelaxTSP1);
            else lower = TSPSubOpt1Tree(r,bestUpper,CT.methRelaxTSP1);
            SetLowerBound(TimerTsp,lower);

            if (P!=NULL)
            {
                TArc a = P[0];
                bestUpper = RedLength(a);
                TNode x = StartNode(a);
                while (x!=0)
                {
                    a = P[x];
                    bestUpper += RedLength(a);
                    x = StartNode(a);
                }

                SetUpperBound(TimerTsp,bestUpper);
            }
        }
        catch (ERRejected) {};
        
        if (lower==InfFloat)
        {
            CloseFold(ModTSP);
            return InfFloat;
        }
    }
        
    if (CT.methSolve>=2 && lower<bestUpper)
    {
        if (r==NoNode) bestUpper = TSPBranchAndBound(0,bestUpper);
        else bestUpper = TSPBranchAndBound(r,bestUpper);
        SetUpperBound(TimerTsp,bestUpper);
    }
        
    CloseFold(ModTSP);

    if (bestUpper==InfFloat) throw ERRejected();
    
    return bestUpper;
}


TFloat abstractMixedGraph::TSPSubOpt1Tree(TNode r,TFloat bestUpper,TOption level) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (r>=n && r!=NoNode) NoSuchNode("TSPSubOpt1Tree",r);

    #endif

    OpenFold(ModSubgradOptTSP,SHOW_TITLE);

    TArc *bestTour = P;
    P = NULL;

    TFloat resolve = 0;
    if (pi==NULL) InitPotentials();
    else
    {
        LogEntry(LOG_METH,"Using existing potentials..");
        resolve = 1;
    }

    TFloat *bestPi = pi;
    pi = NULL;
    InitPotentials();
    TFloat *oldPi = pi;
    pi = NULL;
    InitPotentials();

    for (TNode v=0;v<n;v++) pi[v] = bestPi[v];

    TFloat *oldDegOut = new TFloat[n];

    TFloat bestLower = -InfFloat;
    TFloat lastBound = -InfFloat;
    TFloat gap = InfFloat;
    TNode dev = n;
    TNode bestDev = n;
    unsigned int step = 0;

    sprintf(CT.logBuffer,"Computing minimum %ld-tree...",r);
    LogEntry(LOG_METH,CT.logBuffer);
    CT.SuppressLogging();
    TFloat thisBound = MinTree(r,abstractMixedGraph::MTREE_1TREE);
    CT.RestoreLogging();

    if ((level&3)==0)
    {
        CloseFold(ModSubgradOptTSP);
        InitSubgraph();
        AddToSubgraph();
        ReleasePredecessors();
        P = bestTour;
        ReleasePotentials();
        pi = oldPi;
        ReleasePotentials();
        pi = bestPi;

        if (CT.logRes && bestUpper!=InfFloat && thisBound!=-InfFloat)
        {
            sprintf(CT.logBuffer,"Found gap is [%g,%g] or %g percent",
                ceil(thisBound),bestUpper,(bestUpper/ceil(thisBound)-1)*100);
            LogEntry(LOG_RES,CT.logBuffer);
        }

        return ceil(thisBound);
    }

    // Compute initial step length

    TFloat t0 = thisBound/n;
    if (t0<0) t0 *= -1;
    if (resolve && t0>200 && (level&3)==1) t0 = 200;
    TFloat t = t0;

    unsigned int count = 0;
    while (bestUpper>=ceil(bestLower-CT.epsilon)+1
           && t>0.02 && dev>0 && step<5000
           && CT.SolverRunning())
    {
        step++;

        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"Step number %d",step);
            LogEntry(LOG_METH2,CT.logBuffer);
            sprintf(CT.logBuffer,"Step length: %g",t);
            LogEntry(LOG_METH2,CT.logBuffer);
        }

        #endif

        InitSubgraph();
        if (step>1)
        {
            sprintf(CT.logBuffer,"Computing minimum %ld-tree...",r);
            LogEntry(LOG_METH,CT.logBuffer);
            CT.SuppressLogging();
            InitSubgraph();
            thisBound = MinTree(r,abstractMixedGraph::MTREE_1TREE);
            CT.RestoreLogging();

            if (CT.logRes>=2)
            {
                sprintf(CT.logBuffer,"...Tree length is %g",thisBound);
                LogEntry(LOG_RES2,CT.logBuffer);
            }
        }

        AddToSubgraph();

        TFloat *swapPi = oldPi;
        dev = 0;
        for (TNode v=0;v<n;v++)
        {
            if (DegOut(v)!=1) dev++;

            if (step==1) swapPi[v] = pi[v]+t*(DegOut(v)-1);
            else swapPi[v] =
                pi[v]+t*(0.7*DegOut(v)+0.3*oldDegOut[v]-1);
        }

        #if defined(_LOGGING_)

        if (CT.logMeth>=2)
        {
            sprintf(CT.logBuffer,"Nodes with incorrect degrees: %ld",dev);
            LogEntry(LOG_METH2,CT.logBuffer);
        }

        #endif

        bool newGap = false;
        if (thisBound>bestLower+CT.epsilon)
        {
            bestLower = thisBound;

            // Do not export the lower bound if the procedure was called
            // for a branch&bound subproblem
            if (!(level&4)) SetLowerBound(TimerTsp,ceil(bestLower));

            newGap = true;
            for (TNode v=0;v<n;v++) bestPi[v] = pi[v];
        }

        if (CLCap() && CUCap() && (step==1 || dev<bestDev || dev<5+n/100))
        {
            bestDev = dev;
            TFloat thisUpper = InfFloat;
            try
            {
                sprintf(CT.logBuffer,"Transforming to Hamiltonian cycle...");
                LogEntry(LOG_METH,CT.logBuffer);
                thisUpper = TSPHeuristicsTree(r);
            }
            catch (ERRejected) {};

            if (thisUpper<bestUpper)
            {
                #if defined(_LOGGING_)

                LogEntry(LOG_METH,"Saving tour...");

                #endif

                TArc *swapTour = bestTour;
                bestTour = P;
                P = swapTour;

                bestUpper = thisUpper;
                newGap = true;
            }
        }

        if (newGap)
        {
            gap = bestUpper/ceil(bestLower-CT.epsilon)-1;

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"Gap decreases to %g percent",gap*100);
                LogEntry(LOG_METH2,CT.logBuffer);
            }

            #endif
        }

        oldPi = pi;
        pi = swapPi;

        TFloat *swapDegOut = oldDegOut;
        oldDegOut = sDegOut;
        sDegOut = swapDegOut;

        if (!newGap && ((level&3)==1 || thisBound<=lastBound))
        {
            if ((level&3)==1)
            {
                t *= 0.9;
                count = 0;
            }
            else t *= 0.98;
        }
        else
        {
            count++;
            if (count>3 && newGap) t *= 1.15;
        }
        
        lastBound = thisBound;
    }

    ReleasePotentials();
    pi = oldPi;
    ReleasePotentials();
    pi = bestPi;

    delete[] sDegOut;
    sDegOut = oldDegOut;

    ReleasePredecessors();
    if (bestTour!=NULL) P = bestTour;
    if (bestUpper==InfFloat) ReleasePredecessors();

    if (CT.logMeth==1)
    {
        sprintf(CT.logBuffer,"Total number of iterations: %d",step);
        LogEntry(LOG_METH,CT.logBuffer);
    }

    CloseFold(ModTSP);

    if (CT.logRes && bestUpper!=InfFloat && bestLower!=-InfFloat)
    {
        sprintf(CT.logBuffer,"...Final gap is [%g,%g] or %g percent",
            ceil(bestLower),bestUpper,gap*100);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    return bestLower;
}


TFloat denseDiGraph::TSPHeuristics(TNode r) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (r>=n && r!=NoNode) NoSuchNode("TSPHeuristics",r);

    #endif
    
    OpenFold(ModTSP,NO_INDENT);
    
    LogEntry(LOG_METH,"Computing heuristic tour...");
    
    TFloat ret = InfFloat;

    try
    {
        switch (CT.methTSP)
        {
            case 0:
            {
                ret = TSPHeuristicsRandom();
                break;
            }
            case 1:
            {
                if (r==NoNode)
                    ret = TSPHeuristicsInsert();
                else
                    ret = TSPHeuristicsInsert(r);
                break;
            }
            case 2:
            {
                if (r==NoNode)
                    ret = TSPHeuristicsTree();
                else
                    ret = TSPHeuristicsTree(r);
                break;
            }
            default:
            {
                UnknownOption("TSPHeuristics",CT.methTSP);
            }
        }
    }
    catch (ERRejected) {};

    CloseFold(ModTSP,NO_INDENT);

    if (ret==InfFloat) throw ERRejected();
    else return ret;
}


TFloat abstractMixedGraph::TSPHeuristics(TNode r) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (!CLCap() || MaxLCap()!=0)
        Error(ERR_REJECTED,"TSPHeuristics","Non-trivial lower bounds");
    
    #endif

    OpenFold(ModTSP,NO_INDENT);
    
    LogEntry(LOG_METH,"Transforming to dense digraph...");
    
    denseDiGraph G(n,0,CT);

    for (TArc i=0;i<G.M();i++)
    {
        TArc a = Adjacency(G.StartNode(2*i),G.EndNode(2*i));
        if (CLength())
        {
            if (a!=NoArc) G.SetLength(2*i,1);
            else G.SetLength(2*i,2);
        } 
        else
        {
            if (a!=NoArc) G.SetLength(2*i,Length(a));
            else G.SetLength(2*i,(MaxLength()+1)*n);
        }
    }
    
    if (Dim()>0)
    {
        for (TNode v=0;v<n;v++)
        {
            G.SetC(v,0,C(v,0));
            G.SetC(v,1,C(v,1));
        }
    }

    int swapSolve = CT.methSolve;
    int swapSubgraph = CT.subgraph;
    CT.methSolve = 1;
    CT.subgraph = 0;
    
    G.TSP(r);
    
    InitPredecessors();
    TFloat length = 0;

    if (r==NoNode) r = 0;
    TNode v = r;

    while (true)
    {
        TNode u = G.StartNode(G.Pred(v));
        TArc a = Adjacency(u,v);
        
        if (a!=NoArc)
        {
            P[v] = a;
            length += Length(a);
        }
        else
        {
            CT.methSolve = swapSolve;
            CT.subgraph = swapSubgraph;
            ReleasePredecessors();
            throw ERRejected();
        }
        
        v = u;
        if (v==r) break;
    }
    
    CT.methSolve = swapSolve;
    CT.subgraph = swapSubgraph;
    
    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"Tour has Length %g",length);
        LogEntry(LOG_RES,CT.logBuffer);
    }
        
    CloseFold(ModTSP,NO_INDENT);
    
    return length;
}


TFloat abstractMixedGraph::TSPHeuristicsRandom() throw(ERRejected)
{
    OpenFold(ModRandomTour,SHOW_TITLE);
    
    InitPredecessors();
    TFloat sum = 0;
    
    TNode x = 1+(TNode)CT.Rand(n-1);
    P[x] = Adjacency(0,x);
        
    if (P[x]==NoArc)
    {
        #if defined(_LOGGING_)

        Error(MSG_WARN,"TSPHeuristicsRandom","Missing arc");

        #endif
    
        CloseFold(ModTSP);
        throw ERRejected();
    }
        
    sum += Length(P[x]);
    
    #if defined(_LOGGING_)

    THandle LH = NoHandle;
    
    if (CT.logMeth>1)
    {
        sprintf(CT.logBuffer,"Constructed tour: 0,%ld",x);
        LH = LogStart(LOG_METH2,CT.logBuffer);
    }

    #endif
    
    for (TNode i=2;i<n;i++)
    {
        TNode y = x;
        while (P[y]!=NoArc) y = 1+(TNode)CT.Rand(n-1);
        P[y] = Adjacency(x,y);
        
        if (P[y]==NoArc)
        {
            #if defined(_LOGGING_)

            Error(MSG_WARN,"TSPHeuristicsRandom","Missing arc");

            #endif
    
            CloseFold(ModRandomTour);
            throw ERRejected();
        }
                
        sum += Length(P[y]);
        x = y;
        
        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,",%ld",x);
            LogAppend(LH,CT.logBuffer);
        }

        #endif
    }
            
    #if defined(_LOGGING_)

    LogEnd(LH);

    #endif
    
    P[0] = Adjacency(x,0);
    sum += Length(P[0]);
    SetUpperBound(TimerTsp,sum);

    CT.Ping(Handle(),n);
    
    CloseFold(ModRandomTour);
    
    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Tour has length %g",sum);
        LogEntry(LOG_RES,CT.logBuffer);
    }
    
    if (CT.methLocal) sum = TSPLocalSearch();

    return sum;
}


TFloat abstractMixedGraph::TSPHeuristicsInsert() throw(ERRejected)
{
    OpenFold(ModTSP,NO_INDENT);

    TFloat bestBound = InfFloat;
    TArc* bestTour = P;
    P = NULL;
    InitPredecessors();
            
    for (TNode v=0;v<n && CT.SolverRunning();v++)
    {
        TFloat thisBound;
        try {thisBound = TSPHeuristicsInsert(v);}
        catch (ERRejected) {thisBound = InfFloat;};
        
        if (thisBound<bestBound)
        {
            bestBound = thisBound;
            TArc* swap = bestTour;
            bestTour = P;
            P = swap;    
        }
    }

    ReleasePredecessors();
    P = bestTour;
    
    CloseFold(ModTSP,NO_INDENT);

    if (bestBound==InfFloat) throw ERRejected();
    else return bestBound;
}


TFloat abstractMixedGraph::TSPHeuristicsInsert(TNode r) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (r>=n) NoSuchNode("TSPHeuristicsInsert",r);

    #endif
    
    OpenFold(ModFarthestInsert,SHOW_TITLE);

    #if defined(_LOGGING_)

    THandle LH = NoHandle;

    if (CT.logMeth>1)
    {
        sprintf(CT.logBuffer,"Root node: %ld",r);
        LogEntry(LOG_METH2,CT.logBuffer);
        LH = LogStart(LOG_METH2,"Inserted nodes: ");
    }

    #endif
    
    InitPredecessors();
    
    TNode u = r;
    TArc insArc1 = NoArc;
    TArc insArc2 = NoArc;
    
    TNode maxNode = NoNode;
    TFloat maxDist = -InfFloat;

    for (TNode v=0;v<n;v++)
    {
        if (v!=r)
        {
            TArc a1 = Adjacency(r,v);
            TArc a2 = Adjacency(v,r);
            if (a1!=NoArc && a2!=NoArc && Length(a1)+Length(a2)>maxDist)
            {
                maxDist = Length(a1)+Length(a2);
                maxNode = v;
                insArc1 = a1;
                insArc2 = a2;
            }
        }
    }
    
    TFloat sum = 0;
    
    for (TNode i=0;i<n-1;i++)
    {
        if (maxNode==NoNode)
        {
            #if defined(_LOGGING_)

            Error(MSG_WARN,"TSPHeuristicsInsert","Graph is disconnected");

            #endif
    
            CloseFold(ModFarthestInsert);
            throw ERRejected();
        }
        
        P[EndNode(insArc1)] = insArc1;
        P[EndNode(insArc2)] = insArc2;
        sum += maxDist;
        
        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"%ld,",maxNode);
            LogAppend(LH,CT.logBuffer);
        }

        #endif
    
        CT.Ping(Handle(),n);
        
        maxDist = -InfFloat;
        maxNode = NoNode;
        
        for (TNode v=0;v<n;v++)
        {
            if (P[v]==NoArc)
            {
                TNode w = r;
                TFloat minDelta = InfFloat;
                TArc minArc1 = NoArc;
                TArc minArc2 = NoArc;
                
                while (true)
                {
                    u = StartNode(P[w]);
                    TArc a1 = NoArc;
                    TArc a2 = NoArc;
                    TFloat delta = InfFloat;
                    a1 = Adjacency(u,v);
                    a2 = Adjacency(v,w);
                    if (a1!=NoArc && a2!=NoArc)
                        delta = Length(a1)+Length(a2)-Length(P[w]);
                
                    if (delta<minDelta)
                    {
                        minArc1 = a1;
                        minArc2 = a2;
                        minDelta = delta;
                    }
                
                    w = u;
                    if (w==r) break;
                }
            
                if (minDelta==InfFloat)
                {
                    #if defined(_LOGGING_)

                    Error(MSG_WARN,"TSPHeuristicsInsert","Missing arc(s)");

                    #endif
    
                    CloseFold(ModFarthestInsert);
                    throw ERRejected();
                }
                
                if (minDelta>maxDist)
                {
                    maxNode = v;
                    maxDist = minDelta;
                    insArc1 = minArc1;
                    insArc2 = minArc2;
                }
            }
        }
    }
    
    #if defined(_LOGGING_)

    if (CT.logMeth>1) LogEnd(LH);

    #endif

    SetUpperBound(TimerTsp,sum);
    CT.Ping(Handle(),n*m);
        
    CloseFold(ModFarthestInsert);

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Tour has length %g",sum);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    if (CT.methLocal) sum = TSPLocalSearch();
    
    return sum;
}


TFloat abstractMixedGraph::TSPHeuristicsTree() throw(ERRejected)
{
    OpenFold(ModTreeApproxTSP,SHOW_TITLE);

    TFloat bestBound = InfFloat;
    TArc *bestTour = P;
    P = NULL;
    InitPredecessors();
            
    for (TNode v=0;v<n && CT.SolverRunning();v++)
    {
        TFloat thisBound = InfFloat;

        try
        {
            thisBound = TSPHeuristicsTree(v);
        }
        catch (ERRejected)
        {
            thisBound = InfFloat;
        }
        
        if (thisBound<bestBound)
        {
            bestBound = thisBound;
            TArc *swap = bestTour;
            bestTour = P;
            P = swap;    
        }
    }

    ReleasePredecessors();
    P = bestTour;

    CloseFold(ModTreeApproxTSP);

    if (bestBound==InfFloat) throw ERRejected();
    else return bestBound;
}


TFloat abstractMixedGraph::TSPHeuristicsTree(TNode r) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (r>=n) NoSuchNode("TSPHeuristicsTree",r);

    #endif
    
    OpenFold(ModTreeApproxTSP,SHOW_TITLE);

    try
    {
        ExtractTree(r,abstractMixedGraph::MTREE_1TREE);
    }
    catch (ERCheck)
    {
        try
        {
            MinTree(r,abstractMixedGraph::MTREE_1TREE);
            ExtractTree(r,abstractMixedGraph::MTREE_1TREE);
        }
        catch (ERCheck)
        {
            CloseFold(ModTSP);
            throw ERRejected();
        }
    }

    const char leave = 1;
    const char reached = 2;

    char* state = new char[n];

    for (TNode v=0;v<n;v++) state[v] = leave;
    
    try
    {
        for (TNode v=0;v<n;v++)
            if (v!=r) state[StartNode(P[v])] = 0;
    }
    catch (ERRange)
    {
        delete[] state;
                        
        #if defined(_LOGGING_)

        Error(MSG_WARN,"TSPHeuristicsTree","Graph is disconnected");

        #endif
        
        CloseFold(ModTreeApproxTSP);
        throw ERRejected();
    }
    
    TNode prevLeave = NoNode;
    TFloat sum = 0;

    try
    {
        for (TNode v=0;v<n;v++)
        {
            if (state[v]&leave)
            {
                TNode w = v;
                TArc a = P[v];
                TNode u = StartNode(a);
            
                while (!(state[u]&reached) && u!=r)
                {
                    sum += RedLength(a);
                    state[u] = state[u] | reached;
                    w = u;
                    a = P[w];
                    u = StartNode(a);
                }
            
                if (prevLeave!=NoNode)
                {
                    a = Adjacency(prevLeave,w);
                    if (a==NoArc) throw ERRejected();
                    else P[w] = a;
                }
            
                sum += RedLength(a);
                prevLeave = v;
                
                CT.Ping(Handle(),1);
            }
        }

        TArc a = Adjacency(prevLeave,r);
        if (a==NoArc) throw ERRejected();
        else P[r] = a;
        sum += RedLength(a);
    }
    catch (ERRejected)
    {
        delete[] state;
                        
        #if defined(_LOGGING_)

        Error(MSG_WARN,"TSPHeuristicsTree","Missing arc(s)");

        #endif
    
        CloseFold(ModTreeApproxTSP);
        throw ERRejected();
    }
    
    delete[] state;

    SetUpperBound(TimerTsp,sum);

    CT.Ping(Handle(),n);

    CloseFold(ModTreeApproxTSP);

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Tour has length %g",sum);
        LogEntry(LOG_RES,CT.logBuffer);
    }
    
    if (CT.methLocal) sum = TSPLocalSearch();
    
    return sum;
}


TFloat abstractMixedGraph::TSPLocalSearch() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (P==NULL) Error(ERR_REJECTED,"TSPLocalSearch","Missing tour");

    #endif
    
    OpenFold(ModTSP,NO_INDENT);

    LogEntry(LOG_METH,"Searching for local optimality...");

    while (CT.SolverRunning() && TSPNodeExchange(-MaxLength())) {};
    while (CT.SolverRunning() && TSPNodeExchange(0)) {};
    
    TArc a = P[0];
    TFloat l = Length(a);
    TNode x = StartNode(a);
    while (x!=0)
    {
        a = P[x];
        l += Length(a);
        x = StartNode(a);
    }
    
    SetUpperBound(TimerTsp,l);

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Tour has length %g",l);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    CloseFold(ModTSP,NO_INDENT);

    return l;
}


bool abstractMixedGraph::TSPNodeExchange(TFloat limit) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (P==NULL) Error(ERR_REJECTED,"TSPNodeExchange","Missing tour");

    #endif
    
    OpenFold(ModNodeInsert);

    TNode r = CT.Rand(n);
    TNode v1 = r;
    TArc c1 = NoArc;
        
    bool unchanged = true;
    while (unchanged && (v1!=r || c1==NoArc))
    {
        c1 = P[v1];
        TNode x = StartNode(c1);
        TArc b1 = P[x];
        TNode u1 = StartNode(b1);
        TArc a1 = Adjacency(u1,v1);

        TNode v2 = u1;
        while (unchanged && v2!=v1 && a1!=NoArc)
        {
            TArc a2 = P[v2];
            TNode u2 = StartNode(a2);
            TArc b2 = Adjacency(u2,x);
            TArc c2 = Adjacency(x,v2);
            
            TFloat diff = InfFloat;
            if (b2!=NoArc && c2!=NoArc)
                diff = Length(a1)+Length(b2)+Length(c2)
                     - Length(a2)-Length(b1)-Length(c1);
                
            if (diff<limit)
            {
                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,
                        "Local improvement (%g units, node insertion)",-diff);
                    LogEntry(LOG_METH2,CT.logBuffer);
                    sprintf(CT.logBuffer,
                        "New tour: ... %ld %ld ... %ld %ld %ld ...",
                        u1,v1,u2,x,v2);
                    LogEntry(LOG_METH2,CT.logBuffer);
                }

                #endif
                    
                P[x] = b2;
                P[v2] = c2;
                P[v1] = a1;
                unchanged = false;
                    
                CT.Ping(Handle(),n*n);
            }
                
            v2 = u2;
        }
                
        v1 = x;
    }

    CloseFold(ModNodeInsert);

    return !unchanged;
}
