
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, November 2000.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "bbcolour.h"
#include "spsgra.h"
#include "statq.h"


BBColouring::BBColouring(abstractMixedGraph& GC,TNode kk) throw():
    goblinBBNode<TNode,TFloat>(GC.N(),GC.Context()), G(GC)
{
    n = G.N();
    nActive = n;
    nDominated = 0;
    nColoured = 0;
    m = G.M();
    master = 1;
    
    k = kk;
    if (k<2) k = 2;
    depth = n*(k-1)+1;
    
    DOMINATED = NoNode;

    colour = new TNode[n];
    active = new bool[n];
    conflicts = new TNode[n];
    
    for (TNode w=0;w<n;w++)
    {
        colour[w] = 0;
        active[w] = true;
        conflicts[w] = 0;
    }
    
    for (TArc a=0;a<2*m;a++)
    {
        TNode w = G.StartNode(a);
        conflicts[w]++;
    }
    
    kMax = 0;
    TNode vMax = 0;
    for (TNode w=0;w<n;w++)
    {
        if (conflicts[w]>kMax)
        {
            kMax = conflicts[w];
            vMax = w;
        }
    }
    
    neighbours = new TNode*[n];
    for (TNode w=0;w<n;w++)
    {
        neighbours[w] = new TNode[k];
        for (TNode c=0;c<k;c++) neighbours[w][c] = 0;
    }
    
    Dominated = new staticStack<TNode,TFloat>(n,CT);
    
    I = G.NewIterator();
    
    LogEntry(LOG_MEM,"...Partial colouring generated");
    
    bool isClique = true;
    for (TNode w=0;w<n && isClique;w++)
        for (TNode v=w+1;G.Colour(w)>0 && v<n && isClique;v++)
            if (G.Colour(v)>0 && G.Adjacency(v,w)==NoArc) isClique = false;

    if (isClique)
    {
        TNode i = 0;
        for (TNode w=0;w<n;w++)
            if (G.Colour(w)>0)
            {
                if (active[w]) SetColour(w,i);
                i++;
            }

        if (CT.logMeth)
        {
            sprintf(CT.logBuffer,"Starting with %ld-clique",i);
            LogEntry(LOG_METH,CT.logBuffer);
        }

        SetLowerBound(TimerColour,i);
    }
    else
    {        
        SetColour(vMax,0);
        I -> Reset(vMax);
        TNode v1 = NoNode;
        while (I->Active(vMax))
        {
            TNode w = G.EndNode(I->Read(vMax));
            if (active[w] && (v1==NoNode || conflicts[w]>conflicts[v1])) v1 = w;
        }

        if (v1!=NoNode) SetColour(v1,1);

        SetLowerBound(TimerColour,2);
    }
    
    Reduce();
}
    

BBColouring::BBColouring(BBColouring& node) throw():
    goblinBBNode<TNode,TFloat>(node.G.N(),node.Context(),node.scheme), G(node.G)
{
    n = G.N();
    nActive = node.nActive;
    nDominated = node.nDominated;
    nColoured = node.nColoured;
    m = G.M();
    k = node.k;
    kMax = node.kMax;
    depth = n*(k-1)+1;
    master = 0;
    
    DOMINATED = NoNode;

    colour = new TNode[n];
    active = new bool[n];
    conflicts = new TNode[n];
    neighbours = new TNode*[n];
    
    for (TNode w=0;w<n;w++)
    {
        colour[w] = node.colour[w];
        active[w] = node.active[w];
        conflicts[w] = node.conflicts[w];
        neighbours[w] = new TNode[k];

        for (TNode c=0;c<k;c++)
            neighbours[w][c] = node.neighbours[w][c];
    }
    
    Dominated = new staticStack<TNode,TFloat>(n,CT);
    staticStack<TNode,TFloat> S(n,CT);

    while (!node.Dominated->Empty())
        S.Insert(node.Dominated->Delete());

    while (!S.Empty())
    {
        TNode w = S.Delete();
        node.Dominated->Insert(w);
        Dominated -> Insert(w);
    }
    
    I = G.NewIterator();
    
    LogEntry(LOG_MEM,"...Partial colouring generated");
}
    

void BBColouring::Show() throw()
{
    graph H(G,OPT_CLONE);

    for (TNode v=0;v<n;v++)
        if (active[v]) H.SetColour(v,n);
        else H.SetColour(v,colour[v]);

    H.Display();
}
    

unsigned long BBColouring::Size() throw()
{   
    return
          sizeof(BBColouring)
        + goblinDataObject::Allocated()
        + goblinBBNode<TNode,TFloat>::Allocated()
        + BBColouring::Allocated();
}


unsigned long BBColouring::Allocated() throw()
{   
    unsigned long tmpSize
        = n*sizeof(TNode)             // colour[]
        + n*sizeof(bool)              // active[]
        + n*sizeof(TNode)             // conflicts[]
        + n*sizeof(TNode*)            // neighbours[]
        + n*k*sizeof(TNode);          // neighbours[][]

    return tmpSize;
}


BBColouring::~BBColouring() throw()
{
    for (TNode w=0;w<n;w++) delete[] neighbours[w];
    delete[] neighbours;

    delete[] conflicts;
    delete[] active;
    delete[] colour;
    
    delete I;
    delete Dominated;
    
    LogEntry(LOG_MEM,"...Partial colouring disallocated");
}


void BBColouring::SetColour(TNode v,TNode c) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("SetColour",v);

    if (c>=k)
    {
        sprintf(CT.logBuffer,"No such colour: %ld",c);
        Error(ERR_RANGE,"SetColour",CT.logBuffer);
    }

    if (!active[v])
    {
        sprintf(CT.logBuffer,"Node is coloured or dominated: %ld",v);
        Error(ERR_REJECTED,"SetColour",CT.logBuffer);
    }

    #endif

    colour[v] = c;
    nColoured++;
    active[v] = false;
    nActive--;
    
    if (CT.traceLevel==3) Show();
    
    I -> Reset(v);
    while (I->Active(v))
    {
        TNode w = G.EndNode(I->Read(v));
        
        if (neighbours[w][c]>0) conflicts[w]--;
        neighbours[w][c]++;
        if (conflicts[w]<k && active[w]) Reduce(w);
    }
    
    unfixed = nActive;
}


void BBColouring::Reduce(TNode w) throw(ERRange,ERRejected)
{
    staticQueue<TNode,TFloat> Q(n,CT);
    
    if (w==NoNode)
    {
        for (TNode v=0;v<n;v++)
            if (conflicts[v]<k && active[v])
        {
            Q.Insert(v);
            colour[v] = DOMINATED;
            nDominated++;
            active[v] = false;
            nActive--;
    
            if (CT.traceLevel==3) Show();
        }
    }
    else
    {
        #if defined(_FAILSAVE_)

        if (w>=n) NoSuchNode("Reduce",w);

        if (conflicts[w]>=k || !active[w])
        {
            sprintf(CT.logBuffer,"Inappropriate node: %ld",w);
            Error(ERR_REJECTED,"Reduce",CT.logBuffer);
        }

        #endif
        
        Q.Insert(w);
        colour[w] = DOMINATED;
        nDominated++;
        active[w] = false;
        nActive--;
    
        if (CT.traceLevel==3) Show();
    }
    
    TNode j = 0;
    while (!Q.Empty())
    {
        TNode v = Q.Delete();
        Dominated -> Insert(v);
        j++;
        
        I -> Reset(v);
        while (I->Active(v))
        {
            TNode u = G.EndNode(I->Read(v));
            conflicts[u]--;
            if (conflicts[u]<k && active[u])
            {
                Q.Insert(u);
                colour[u] = DOMINATED;
                nDominated++;
                active[u] = false;
                nActive--;
    
                if (CT.traceLevel==3) Show();
            }
        }
    }
    
    unfixed = nActive;
}


void BBColouring::PlanarComplete() throw(ERCheck)
{
    staticQueue<TNode,TFloat> Q(n,CT);
    
    for (TNode v=0;v<n;v++)
    {
        if (active[v]) Q.Insert(v);

        G.SetColour(v,colour[v]);
    }
    
    while (!(Dominated->Empty() && Q.Empty()))
    {
        TNode v = NoNode;
        if (!Q.Empty()) v = Q.Delete();
        else v = Dominated->Delete();
        
        if (colour[v]!=0 && colour[v]!=DOMINATED) continue;
        
        TNode c = 0;
        for (;c<5 && neighbours[v][c]>0;c++) {};

        if (c<5)
        {
            active[v] = true;
            nActive++;
            nDominated++;
            SetColour(v,c);
            G.SetColour(v,c);
        }
        else
        {
            TArc a1 = G.First(v);
            TArc a2 = G.Right(a1,v);
            TArc a3 = G.Right(a2,v);
            TArc a4 = G.Right(a3,v);
            TNode x1 = G.EndNode(a1);
            TNode x2 = G.EndNode(a2);
            TNode x3 = G.EndNode(a3);
            TNode x4 = G.EndNode(a4);
            TNode c1 = G.Colour(x1);

            if (G.NCKempeExchange(x1,x3))
            {
                active[v] = true;
                nActive++;
                nDominated++;
                SetColour(v,c1);
                G.SetColour(v,c1);
            }
            else
            {
                TNode c2 = G.Colour(x2);
                if (G.NCKempeExchange(x2,x4))
                {
                    active[v] = true;
                    nActive++;
                    nDominated++;
                    SetColour(v,c2);
                    G.SetColour(v,c2);
                }
                else Error(ERR_CHECK,"PlanarComplete","Not a planar graph");
            }
        }
    }
}


bool BBColouring::Complete() throw(ERRejected)
{
    for (TNode v=0;v<n;v++)
    {
        if (active[v])
        {
            TNode c = colour[v];
            for (;c<k && neighbours[v][c]>0;c++) {};
            
            if (c==k)
            {
                sprintf(CT.logBuffer,"Got stuck at node %ld",v);
                Error(ERR_REJECTED,"Complete",CT.logBuffer);
            }
            
            SetColour(v,c);
        }
    }
    
    while (!Dominated->Empty())
    {
        TNode v = Dominated->Delete();
        
        if (colour[v]==DOMINATED)
        {
            TNode c = 0;
            for (;c<k && neighbours[v][c]>0;c++) {};
            
            if (c==k)
            {
                sprintf(CT.logBuffer,"Got stuck at node %ld",v);
                Error(ERR_REJECTED,"Complete",CT.logBuffer);
            }
            
            active[v] = 1;
            nActive++;
            nDominated++;
            SetColour(v,c);
        }
    }
    
    return true;
}


TNode BBColouring::SelectVariable() throw()
{
    TNode retNode = NoNode;
    TNode minColour = NoNode;
    
    for (TNode v=0;v<n;v++)
        if (active[v] && colour[v]<minColour)
        {
            retNode = v;
            minColour = colour[v];
        }
    
    return retNode;
}


char BBColouring::Direction1(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Direction1",v);

    #endif
    
    return 0;
}


char BBColouring::Direction2(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Direction2",v);

    #endif
    
    return Direction1(v);
}


void BBColouring::Raise(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Raise",v);

    #endif
    
    colour[v]++;
    
    solved = 0;
}


void BBColouring::Lower(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Lower",v);

    #endif
    
    SetColour(v,colour[v]);
    
    solved = 0;
}


TFloat BBColouring::SolveRelaxation() throw()
{
    bool feasible = true;
    bool reduction = true;
    while (reduction && feasible)
    {
        reduction = false;

        for (TNode w=0;w<n && feasible;w++)
        {
            if (active[w])
            {
                TNode freeColour = NoNode;
                TNode nFree = 0;

                for (TNode c=colour[w];c<k;c++)
                    if (neighbours[w][c]==0)
                    {
                        if (nFree==0) freeColour = c;
                        nFree++;
                    }

                if (nFree>0) colour[w] = freeColour;
                else feasible = false;
            
                if (nFree==1)
                {
                    SetColour(w,freeColour);
                    reduction = true;
                }
            }
        }
    }
    
    if (!feasible) return InfFloat;
    
    if (Feasible()) Complete();
    TNode maxColour = 0;
    for (TNode w=0;w<n && feasible;w++)
        if (colour[w]!=DOMINATED && colour[w]>maxColour) maxColour = colour[w];

    return maxColour+1;
}


goblinBBNode<TNode,TFloat> *BBColouring::Clone() throw()
{
    return new BBColouring(*this);
}


void BBColouring::SaveSolution() throw()
{
    bool isUncoloured = false;
    for (TArc a=0;a<m && !isUncoloured;a++)
        if (G.Colour(G.StartNode(2*a))==G.Colour(G.EndNode(2*a)))
            isUncoloured = true;

    TNode maxColourNew = 0;
    TNode maxColourOld = 0;
    for (TNode v=0;v<n;v++)
    {
        if (G.Colour(v)>maxColourOld) maxColourOld = G.Colour(v);
        if (colour[v]>maxColourNew) maxColourNew = colour[v];
    }
    
    if (maxColourOld>maxColourNew || isUncoloured)
    {        
        for (TNode v=0;v<n;v++) G.SetColour(v,colour[v]);
        SetUpperBound(TimerColour,maxColourNew+1);
    }
}


TNode abstractMixedGraph::NodeColouring(TNode k) throw()
{
    if (k>=5 && m<=3*n-6)
    {
        TNode maxColour = PlanarColouring();

        if (CT.methColour==0 || maxColour<=2) return maxColour;
    }

    // Check for present colouring

    bool isUncoloured = false;
    for (TArc a=0;a<m && !isUncoloured;a++)
        if (Colour(StartNode(2*a))==Colour(EndNode(2*a))) isUncoloured = true;

    TFloat maxColour = 0;
    for (TNode v=0;v<n;v++)
        if (Colour(v)>maxColour) maxColour = Colour(v);
    maxColour++;
    
    if (isUncoloured || CT.methColour==0) maxColour = InfFloat;
    else
    {
        sprintf(CT.logBuffer,
            "Starting with %ld-node colouring...",TNode(maxColour));
        LogEntry(LOG_METH,CT.logBuffer);
        SetUpperBound(TimerColour,maxColour);
    }

    TNode k0 = k;
    if (k==NoNode) k0 = n;
    
    if (CT.logMeth)
    {
        sprintf(CT.logBuffer,"Searching for %ld-node colouring...",k0);
        LogEntry(LOG_METH,CT.logBuffer);
    }
    OpenFold(ModColour);
        
    BBColouring *rootNode = new BBColouring(*this,k0);

    goblinBBScheme<TNode,TFloat> *scheme
        = new goblinBBScheme<TNode,TFloat>(rootNode,maxColour,CT.methColour);
    
    if (!scheme->feasible)
    {
        delete scheme;
        sprintf(CT.logBuffer,"...No %ld-node colouring exists",k0);
        LogEntry(LOG_RES,CT.logBuffer);
        CloseFold(ModColour);
        return 0;
    }
    
    maxColour = 0;
    for (TNode v=0;v<n;v++)
        if (colour[v]>maxColour) maxColour = colour[v];
    maxColour++;

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...%ld-node colouring found",TNode(maxColour));
        LogEntry(LOG_RES,CT.logBuffer);
    }

    while (maxColour<=k0 && k0>2 && k==NoNode && CT.SolverRunning() &&
           LowerBound(TimerColour)<UpperBound(TimerColour)) 
    {
        delete scheme;

        k0--;
    
        if (CT.logMeth)
        {
            sprintf(CT.logBuffer,"Searching for %ld-node colouring...",k0);
            LogEntry(LOG_METH,CT.logBuffer);
        }

        OpenFold();
    
        BBColouring *rootNode = new BBColouring(*this,k0);

        if (CT.methColour==0) maxColour = InfFloat;
        scheme = new goblinBBScheme<TNode,TFloat>(rootNode,maxColour,CT.methColour);

        maxColour = 0;
        for (TNode v=0;v<n;v++)
            if (colour[v]>maxColour) maxColour = colour[v];
        maxColour++;

        CloseFold();

        if (CT.logRes)
        {
            sprintf(CT.logBuffer,"...%ld-node colouring found",TNode(maxColour));
            LogEntry(LOG_RES,CT.logBuffer);
        }
    }
    
    delete scheme;

    if (CT.methLocal && LowerBound(TimerColour)<UpperBound(TimerColour))
        maxColour = TFloat(NCLocalSearch());
    
    CloseFold(ModColour);

    return TNode(maxColour);
}


TNode abstractMixedGraph::PlanarColouring() throw(ERRejected)
{
    LogEntry(LOG_METH,"Searching for planar 5-node colouring...");
    
    if (m==0)
    {
        LogEntry(LOG_RES,"...Graph is empty");
        InitColours(0);
        return 1;
    }
    
    OpenFold(ModColour);
        
    BBColouring rootNode(*this,5);
    try
    {
        rootNode.PlanarComplete();
    }
    catch (ERCheck)
    {
        if (CT.logRes)
        {
            sprintf(CT.logBuffer,"...Planar colouring failed");
            LogEntry(LOG_RES,CT.logBuffer);
        }
        CloseFold(ModColour);
        return NoNode;
    }

    TNode maxColour = 0;
    for (TNode v=0;v<n;v++)
        if (colour[v]>maxColour) maxColour = colour[v];
    maxColour++;

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...%ld-node colouring found",maxColour);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    if (CT.methLocal && LowerBound(TimerColour)<UpperBound(TimerColour))
        maxColour = NCLocalSearch();
    
    CloseFold(ModColour);

    return maxColour;
}


TNode abstractMixedGraph::NCLocalSearch() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (colour==NULL)
        Error(ERR_REJECTED,"NCKempeExchange","Missing node colours");

    #endif

    LogEntry(LOG_METH,"Searching for local optimality...");
    OpenFold(ModColour);

    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    TNode count = 0;
    while (count<500*n && CT.SolverRunning())
    {
        TArc a1 = CT.Rand(2*m);
        TNode v = StartNode(a1);
        TNode w = EndNode(a1);
        TNode c = colour[w];
        count++;

        if (colour[v]<=c)
        {
            a1 = a1^1;
            v = StartNode(a1);
            w = EndNode(a1);
            c = colour[w];
        }

        bool unique = true;
        I.Reset(v);
	    while (I.Active(v) && unique)
	    {
            TNode u = EndNode(I.Read(v));
            if (colour[u]==c && u!=w) unique = false;
        }

        if (!unique) continue;
 
        TNode c2 = CT.Rand(colour[v]);
        if (c==c2) continue;

        unique = true;
        TNode w2 = NoNode;
        I.Reset(v);
	    while (I.Active(v) && unique)
	    {
            TNode u = EndNode(I.Read(v));
            if (colour[u]==c2)
            {
                if (w2!=NoNode) unique = false;
                else w2 = u;
            }
        }

        if (!unique) continue;

        if (w2==NoNode)
        {
            colour[v] = c2;
            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,
                    "Colour of node %ld is changed to %ld",v,c2);
                LogEntry(LOG_METH2,CT.logBuffer);
            }
                
            TNode count = 0;
            for (TNode x=0;x<n;x++) if (colour[x]==c2) count++;
            
            if (count==0)
            {
                for (TNode x=0;x<n;x++)
                    if (colour[x]>c2) colour[x]--;
            }
        }
        else if (NCKempeExchange(w,w2))
        {
            colour[v] = c;
            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,
                    "Colour of node %ld is changed to %ld",v,c);
                LogEntry(LOG_METH2,CT.logBuffer);
            };
                
            TNode count = 0;
            for (TNode x=0;x<n;x++) if (colour[x]==c) count++;
            
            if (count==0)
            {
                for (TNode x=0;x<n;x++)
                    if (colour[x]>c) colour[x]--;
            }
        }

        CT.Ping(Handle(),n*m); 
    }

    Close(H);
    CloseFold(ModColour);

    TNode maxColour = 0;
    for (TNode v=0;v<n;v++)
        if (colour[v]>maxColour) maxColour = colour[v];

    SetUpperBound(TimerColour,maxColour+1);
        
    return maxColour+1;
}


bool abstractMixedGraph::NCKempeExchange(TNode r,TNode x) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (r>=n) NoSuchNode("NCKempeExchange",r);

    if (x>=n) NoSuchNode("NCKempeExchange",x);

    if (colour==NULL)
        Error(ERR_REJECTED,"NCKempeExchange","Missing node colours");

    #endif
    
    TNode colour1 = colour[r];
    TNode colour2 = colour[x];

    bool* marked = new bool[n];
    for (TNode w=0;w<n;w++) marked[w] = false;

    staticQueue<TNode,TFloat> Q(n,CT);
    Q.Insert(r);
    colour[r] = colour2;

    if (CT.logMeth>1)
    {
        sprintf(CT.logBuffer,
             "Alternating colours %ld,%ld...",colour1,colour2);
        LogEntry(LOG_METH2,CT.logBuffer);
    }

    #if defined(_LOGGING_)

    THandle LH = LogStart(LOG_METH2,"Expanded nodes: ");

    #endif
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    while (!Q.Empty())
    {
        TNode u = Q.Delete();

        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"%ld ",u);
            LogAppend(LH,CT.logBuffer);
        }

        #endif
        
        while (I.Active(u))
        {
            TArc a = I.Read(u);
            TNode v = EndNode(a);

            if (colour[u]==colour[v] && !marked[v])
            {               
                marked[v] = true;

                if (colour[v]==colour1) colour[v] = colour2;
                else colour[v] = colour1;

                Q.Insert(v);
            }
        }
    }

    #if defined(_LOGGING_)

    LogEnd(LH);

    #endif
    
    Close(H);

    delete[] marked;

    return colour[x]==colour2;
}


TNode abstractMixedGraph::CliqueCover(TNode k) throw()
{
    if (CT.logMeth)
    {
        if (k==NoNode) sprintf(CT.logBuffer,"Computing minimum clique cover...");
        else sprintf(CT.logBuffer,"Computing %ld-clique cover...",k);
        LogEntry(LOG_METH,CT.logBuffer);
    }
    OpenFold(ModColour);
    
    complementaryGraph *G = new complementaryGraph(*this,(TOption)0);

    if (colour!=NULL)
        for (TNode v=0;v<n;v++) G->SetColour(v,colour[v]);

    TNode chi = G->NodeColouring(k);
    
    for (TNode v=0;v<n;v++) SetColour(v,G->Colour(v));
    
    delete G;
    
    CloseFold(ModColour);
    
    if (CT.logRes) 
    {
        if (chi>0) sprintf(CT.logBuffer,"...%ld-clique cover found",chi);
        else sprintf(CT.logBuffer,"...No %ld-clique cover found",k);
        LogEntry(LOG_RES,CT.logBuffer);
    }
    
    return chi;
}


TNode abstractMixedGraph::EdgeColouring(TArc k) throw(ERRange)
{
    if (CT.logMeth)
    {
        if (k==NoNode) sprintf(CT.logBuffer,"Computing minimum edge colouring...");
        else sprintf(CT.logBuffer,"Computing %ld-edge colouring...",k);
        LogEntry(LOG_METH,CT.logBuffer);
    }
    OpenFold(ModColour);

    lineGraph *G = new lineGraph(*this,(TOption)0);
    
    THandle H = Investigate();
    goblinIterator& I = Iterator(H);
    goblinQueue<TArc,TFloat>* Q = NewArcHeap();
    goblinQueue<TArc,TFloat>* Q2 = NewArcHeap();

    for (TArc a=0;a<m;a++)
    {
        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"Colouring arc %ld...",2*a);
            LogEntry(LOG_METH2,CT.logBuffer);
        }

        TArc aCurrent = 2*a;
        TNode u = StartNode(aCurrent);
        TNode v1 = EndNode(aCurrent);

        // Compute the smallest colour free at node u
                
        I.Reset(u);
        Q -> Init();
        TArc thisDeg = 0;
        while (I.Active(u))
        {
            TArc a = I.Read(u);
            Q->Insert(a>>1,G->Colour(TNode(a>>1)));
            thisDeg++;
        }

        SetLowerBound(TimerColour,thisDeg);
        
        if (thisDeg>k)
        {
            delete G;
            
            CloseFold(ModColour);
            sprintf(CT.logBuffer,"...No %ld-edge colouring exists",k);
            LogEntry(LOG_RES,CT.logBuffer);

            return NoNode;
        }

        TNode alpha = NoNode;
        TArc c2 = NoNode;
        while (!Q->Empty() && alpha==NoNode)
        {
            TArc a2 = Q->Delete();
            TArc c3 = G->Colour(TNode(a2));
            if (c2==NoNode && c3>0) alpha = 0;
            if (c2!=NoNode && c2<c3-1) alpha = c2+1;
            c2 = c3;
        }

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"...Available colour is %ld",alpha);
            LogEntry(LOG_METH2,CT.logBuffer);
        }
        
        // Check if the colour alpha is free at node v1

        I.Reset(v1);
        TArc aNext = NoArc;
        while (I.Active(v1) && aNext==NoArc)
        {
            TArc a = I.Read(v1);
            if (G->Colour(TNode(a>>1))==alpha) aNext = a;
        }
        
        bool searching = true;
        if (aNext==NoArc)
        {
            G->SetColour(TNode(aCurrent>>1),alpha);
            searching = false;

            LogEntry(LOG_METH2,"...Arc is coloured");
            CT.Ping(G->Handle(),m);
        }
        
        TNode beta = NoNode;
        while (searching)
        {
            // Compute the smallest colour free at node v1
                
            I.Reset(v1);
            Q2 -> Init();
            while (I.Active(v1))
            {
                TArc a = I.Read(v1);
                Q2->Insert(a>>1,G->Colour(TNode(a>>1)));
            }

            TNode betaOld = beta;
            beta = NoNode;
            TArc c2 = NoNode;
            while (!Q2->Empty() && beta==NoNode)
            {
                TArc a2 = Q2->Delete();
                TArc c3 = G->Colour(TNode(a2));
                if (c2==NoNode && c3>0) 
                {
                    if (betaOld!=0) beta = 0;
                    else if (c3>1) beta = 1;
                }

                if (c2!=NoNode && c2<c3-1)
                {
                    if (betaOld!=c2+1) beta = c2+1;
                    else if (c2<c3-2)  beta = c2+2;
                }
                c2 = c3;
            }

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"...Available colour is %ld",beta);
                LogEntry(LOG_METH2,CT.logBuffer);
            }

            // Check if the colour beta is free at node u

            I.Reset(u);
            TArc aNext = NoArc;
            while (I.Active(u) && aNext==NoArc)
            {
                TArc a2 = I.Read(u);
                if (G->Colour(TNode(a2>>1))==beta) aNext = a2;
            }
        
            G->SetColour(TNode(aCurrent>>1),beta);

            if (aNext==NoArc)
            {
                searching = false;
                LogEntry(LOG_METH2,"...arc is coloured");
                CT.Ping(G->Handle(),m);
                break;
            }

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"...Next arc is %ld",aNext);
                LogEntry(LOG_METH2,CT.logBuffer);
            }

            G->SetColour(TNode(aNext>>1),NoNode);

            TNode v2 = EndNode(aNext);

            // Check if alpha is free at node v2

            I.Reset(v2);
            TArc aAux = NoArc;
            while (I.Active(v2) && aAux==NoArc)
            {
                TArc a2 = I.Read(v2);
                if (G->Colour(TNode(a2>>1))==alpha) aAux = a2;
            }

            if (aAux==NoArc ||
                G->NCKempeExchange(TNode(aAux>>1),TNode(aCurrent>>1)))
            {
                G->SetColour(TNode(aNext>>1),alpha);
                searching = false;
                LogEntry(LOG_METH2,"...Arc is coloured");
            }
            else
            {
                G->NCKempeExchange(TNode(aAux>>1),TNode(aCurrent>>1));

                v1 = v2;
                aCurrent = aNext;
            }

            CT.Ping(G->Handle(),m);
        }
    }

    TNode chi = 0;
    for (TArc a=0;a<m;a++)
        if (chi<=G->Colour(a)) chi = G->Colour(a)+1;

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...%ld-edge colouring found",chi);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    SetUpperBound(TimerColour,chi);

    // Apply local search if a k = chi-1 colouring was requested

    if (CT.methLocal && k<chi)
        chi = G->NCLocalSearch();

    // Fall back to enumeration if a k = chi-1 colouring was requested

    if (chi>k) chi = G->NodeColouring(k);
        
    Close(H);
    delete Q;
    delete Q2;
    
    for (TArc a=0;a<m;a++)
    {
        SetEdgeColour(2*a,G->Colour(a));

        if (chi<=G->Colour(a)) chi = G->Colour(a)+1;
    }
    
    delete G;

    CloseFold(ModColour);
    
    if (CT.logRes)
    {
        if (chi>0) sprintf(CT.logBuffer,"...%ld-edge colouring found",chi);
        else sprintf(CT.logBuffer,"...No %ld-edge colouring found",k);

        LogEntry(LOG_RES,CT.logBuffer);
    }
    
    return chi;
}
