
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, July 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "absmix.h"
#include "spsdig.h"
#include "statstaq.h"


bool abstractMixedGraph::StronglyConnected() throw()
{
    OpenFold(ModStrongConnectivity,NO_INDENT);

    LogEntry(LOG_METH,"Computing strongly connected components...");
        
    #if defined(_LOGGING_)
    
    OpenFold();
    OpenFold();
    
    #endif
    
    TNode i = 0;
    
    InitColours();
    InitPredecessors();
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    staticStack<TNode,TFloat> S(n,CT);
        // Ordering of the reverse DFS operations
    
    #if defined(_TRACING_)
    
    staticStack<TNode,TFloat> Recovery(n,CT);
        // Needed for reinitialization of predecessors
            
    #endif
    
    bool* marked = new bool[n];
    for (TNode v=0;v<n;v++) marked[v] = false;
    
    for (TNode r=0;r<n;r++)
    {
        if (colour[r]==NoNode)
        {
            #if defined(_LOGGING_)
    
            CloseFold();
            LogEntry(LOG_METH2,"Generating forward DFS tree...");
            OpenFold();
            
            #endif
    
            TNode u = r;
        
            while (true)
            {
                if (I.Active(u))
                {
                    TArc a = I.Read(u);
                    TNode v = EndNode(a);

                    if (UCap(a)>0 && !Blocking(a) && P[v]==NoArc && colour[v]==NoNode && v!=r)
                    {
                        P[v] = a;
                        u = v;
                    }
                }
                else
                {
                    // Backtracking
                    
                    S.Insert(u);

                    #if defined(_TRACING_)
    
                    Recovery.Insert(u);
            
                    #endif
    
                    I.Reset(u);
                    marked[u] = true;
                    
                    if (u==r) break;
                    else u = StartNode(P[u]);
                }
            }

            CT.Ping(Handle(),m);
            
            #if defined(_TRACING_)
    
            while (!Recovery.Empty())
            {
                u = Recovery.Delete();
                P[u] = NoArc;
            }
            
            #endif
            
            while (!S.Empty())
            {
                TNode s = S.Delete();
                if (colour[s]==NoNode && marked[s])
                {
                    u = s;
        
                    #if defined(_LOGGING_)

                    THandle LH = NoHandle;

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,"Component %ld: %ld",i,u);
                        LH = LogStart(LOG_METH2,CT.logBuffer);
                    }
        
                    #endif
        
                    while (true)
                    {
                        if (I.Active(u))
                        {
                            TArc a = I.Read(u);
                    
                            TNode v = EndNode(a);
                            if (UCap(a^1)>0 && !Blocking(a^1) && colour[v]==NoNode && marked[v] && v!=s)
                            {
                                P[v] = a;
                                u = v;
                                marked[u] = false;
        
                                #if defined(_LOGGING_)
    
                                if (CT.logMeth>1)
                                {
                                    sprintf(CT.logBuffer,",%ld",u);
                                    LogAppend(LH,CT.logBuffer);
                                }

                                #endif
                            }
                        }
                        else
                        {
                            // Backtracking

                            #if defined(_TRACING_)
    
                            Recovery.Insert(u);
        
                            #endif

                            I.Reset(u);
                            colour[u] = i;
                    
                            if (u==s) break;
                            else u = StartNode(P[u]);
                        }
                    }

                    #if defined(_LOGGING_)

                    if (CT.logMeth>1) LogEnd(LH);

                    #endif
        
                    CT.Ping(Handle(),m);
            
                    #if defined(_TRACING_)
    
                    while (!Recovery.Empty())
                    {
                        u = Recovery.Delete();
                        P[u] = NoArc;
                    }
        
                    #endif
                
                    i++;
                }
            }
        }
    }

    Close(H);
    delete[] marked;
        
    #if defined(_LOGGING_)
    
    CloseFold();
    CloseFold();
        
    #endif
    
    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Graph has %ld strongly connected components",i);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    SetBounds(TimerConnectivity,i,i);

    CloseFold(ModStrongConnectivity,NO_INDENT);
    
    return (i<=1);
}


bool abstractMixedGraph::Connected() throw()
{                       
    OpenFold(ModConnectivity,NO_INDENT);
    
    LogEntry(LOG_METH,"Computing connected components...");
        
    #if defined(_LOGGING_)
    
    OpenFold();
    
    #endif
    
    TNode i = 0;
    
    InitPredecessors();
    InitColours();
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    
    for (TNode r=0;r<n;r++)
    {
        if (colour[r]==NoNode)
        {
            TNode u = r;
        
            #if defined(_LOGGING_)

            THandle LH = NoHandle;

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"Component %ld: %ld",i,u);
                LH = LogStart(LOG_METH2,CT.logBuffer);
            }
        
            #endif
    
            while (true)
            {
                if (I.Active(u))
                {
                    TArc a = I.Read(u);
                    
                    TNode v = EndNode(a);
                    if (UCap(a)>0 && P[v]==NoArc && v!=r)
                    {
                        P[v] = a;
                        u = v;
        
                        #if defined(_LOGGING_)
    
                        if (CT.logMeth>1)
                        {
                            sprintf(CT.logBuffer,",%ld",u);
                            LogAppend(LH,CT.logBuffer);
                        }
        
                        #endif
                    }
                }
                else
                {
                    // Backtracking
                    
                    colour[u] = i;
                    
                    if (u==r) break;
                    else u = StartNode(P[u]);
                }
            }
            
            #if defined(_LOGGING_)

            if (CT.logMeth>1) LogEnd(LH);

            #endif

            i++;
        }
    }

    Close(H);

    CT.Ping(Handle(),m);
        
    #if defined(_LOGGING_)
    
    CloseFold();
        
    #endif
    
    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Graph has %ld connected components",i);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    SetBounds(TimerConnectivity,i,i);

    CloseFold(ModConnectivity,NO_INDENT);
    
    return (i<=1);
}


bool abstractMixedGraph::CutNodes(
    TNode* order,TArc* lowArc,TNode* thread,TArc retArc) throw()
{                       
    OpenFold(Mod2Connectivity,NO_INDENT);
    
    LogEntry(LOG_METH,"Computing blocks and cut nodes... ");
        
    #if defined(_LOGGING_)
    
    OpenFold();
    
    #endif
    
    TNode i = 1; // Current ordinal number
    TNode k = 0; // Current edge colour index
    TNode c = 0; // Current node colour index
    TNode prevThreadNode = 0;
    
    InitPredecessors();
    InitColours();
    InitSubgraph();
    
    bool newOrder = (order==NULL);
    if (newOrder) order = new TNode[n];

    bool newLow = (lowArc==NULL);
    if (newLow) lowArc = new TArc[n];
    
    for (TNode v=0;v<n;v++) lowArc[v] = NoArc;

    TNode* low = new TNode[n];
    
    staticStack<TNode,TFloat> Q(n,CT);
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    
    for (TNode r=0;r<n;r++)
    {
        if (colour[r]!=NoNode) continue;

        TNode u = r;

        if (retArc<2*m)
        {
            // Force return arc to be the first DFS tree arc
            // (this is needed for the st-numbering algorithm)
        
            TNode t = StartNode(retArc);
            TNode s = EndNode(retArc);
            
            low[t] = order[t] = i++;
            low[s] = order[s] = i++;
            lowArc[t] = lowArc[s] = NoArc;
            P[s] = retArc;
            
            r = t;
            u = s;
        
            Q.Insert(t);
            Q.Insert(s);
        }
        else
        {
            low[r] = order[r] = i++;
            lowArc[r] = NoArc;
        
            Q.Insert(r);
        }

        if (thread)
        {
            thread[prevThreadNode] = r;
            prevThreadNode = r;
        }

        while (true)
        {
            if (I.Active(u))
            {
                TArc a = I.Read(u);

                TNode v = EndNode(a);

                if (UCap(a)>0)
                {
                    if (P[v]==NoArc && v!=r)
                    {
                        // Tree edge
                        
                        P[v] = a;
                        order[v] = i++;
                        low[v] = order[v];
                        lowArc[v] = NoArc;
                        Q.Insert(v);
                        u = v;
                    }
                    else if (P[u]!=(a^1) && low[u]>order[v])
                    {
                        // Backward edge

                        low[u] = order[v];
                        lowArc[u] = a;
                    }
                }
            }
            else
            {
                // Backtracking

                if (u==r) break;

                TNode w = StartNode(P[u]);
                if (low[u]<order[w])
                {
                    if (low[w]>low[u])
                    {
                        low[w] = low[u];
                        lowArc[w] = lowArc[u];
                    }
                }
                else
                {
                    // A block has just been completed
                    // w is the root of the new block
                
                    k++;

                    #if defined(_LOGGING_)

                    THandle LH = NoHandle;

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,"Block %ld: %ld,",k,w);
                        LH = LogStart(LOG_METH2,CT.logBuffer);
                    }

                    #endif

                    // Colour the nodes and arcs of the block
                    // Check if the block forms a leave in the cut tree
                    
                    TNode v = Q.Delete();
                    bool leave = true;
                    
                    while (v!=u)
                    {
                        if (colour[v]==NoNode) colour[v] = k;
                        else leave = false;

                        I.Reset(v);
                        while (I.Active(v))
                        {
                            TArc a = I.Read(v);
                            if (UCap(a)>n && Sub(a)==0) SetEdgeColour(a,k);
                        }

                        #if defined(_LOGGING_)

                        if (CT.logMeth>1)
                        {
                            sprintf(CT.logBuffer,"%ld,",v);
                            LogAppend(LH,CT.logBuffer);
                        }

                        #endif

                        v = Q.Delete();    
                    }

                    if (colour[u]==NoNode) colour[u] = k;
                    else leave = false;

                    I.Reset(u);
                    while (I.Active(u))
                    {
                        TArc a = I.Read(u);
                        if (UCap(a)>n && Sub(a)==0) SetEdgeColour(a,k);
                    }

                    #if defined(_LOGGING_)

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,"%ld",u);
                        LogEnd(LH,CT.logBuffer);
                    }

                    #endif

                    if (w==r && colour[r]==NoNode) colour[r] = k;
                    else
                    {
                        // Mark w as a cut node
                    
                        colour[w] = 0;
                        c++;
                    }
                    
                    // If the new block is a leave, save u as a thread node.
                    // Adding the complete thread makes the graph 2-connected.
                    // Since u is a neighbor of w, both occur on the unbounded
                    // region in some plane embedding of the block -> Adding
                    // the thread preserves the planarity of the compound graph.
                        
                    if (thread && leave)
                    {
                        thread[prevThreadNode] = u;
                        thread[u] = 0;
                        prevThreadNode = u;
                    }

                    CT.Ping(Handle(),m);
                }

                u = w;
            }
        }

        if (retArc<2*m) break;
    }

    Close(H);
    
    delete[] low;

    if (newLow) delete[] lowArc;
    if (newOrder) delete[] order;

    #if defined(_LOGGING_)
    
    CloseFold();
        
    #endif

    if (retArc<2*m)
    {
        if (i>n && k==1) LogEntry(LOG_RES,"...Graph is 2-connected");
        else LogEntry(LOG_RES,"...Graph is not 2-connected");

        CloseFold(Mod2Connectivity,NO_INDENT);
    
        return (i<=n || k>1);
    }

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Graph has %ld blocks and %ld cut nodes",k,c);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    if (newLow) SetBounds(TimerConnectivity,k,k);

    CloseFold(Mod2Connectivity,NO_INDENT);
    
    return (k!=1);
}


bool abstractMixedGraph::TwoConnected() throw()
{
    OpenFold(Mod2Connectivity,NO_INDENT);
    
    LogEntry(LOG_METH,"Computing 2-connected components...");
        
    #if defined(_LOGGING_)
    
    OpenFold();
    
    #endif
    
    TNode* order = new TNode[n];
    TArc* low = new TArc[n];

    CutNodes(order,low);
    
    InitPartition();
    for (TNode u=0;u<n;u++) Bud(u);
    
    TNode ncomponents = n;
    SetBounds(TimerConnectivity,1,n);
    for (TArc a=0;a<m;a++)
    {
        TNode v = StartNode(2*a);
        TNode w = EndNode(2*a);
        if ((colour[v]==colour[w] && colour[v]!=0) ||
            (P[w]!=2*a && P[v]!=2*a+1) ||
            (P[w]==2*a && low[w]!=NoArc && order[EndNode(low[w])]<order[w]) ||
            (P[v]==2*a+1 && low[v]!=NoArc && order[EndNode(low[v])]<order[v]))
        {
            if (Find(v)!=Find(w))
            {
                ncomponents--;
                Merge(w,v);
                SetUpperBound(TimerConnectivity,ncomponents);
            }
        }
    }

    for (TNode u=0;u<n;u++) colour[u]=colour[Find(u)];

    CT.Ping(Handle(),m);
        
    #if defined(_LOGGING_)
    
    CloseFold();
    
    #endif
    
    delete[] order;
    delete[] low;

    SetLowerBound(TimerConnectivity,ncomponents);
    
    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Graph has %ld 2-edge connected components",
            ncomponents);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    CloseFold(Mod2Connectivity,NO_INDENT);
    
    return (ncomponents==1);
}


bool abstractMixedGraph::STNumbering(TArc returnArc) throw()
{
    OpenFold(Mod2Connectivity,NO_INDENT);
    
    LogEntry(LOG_METH,"Computing st-numbering...");
        
    #if defined(_LOGGING_)
    
    OpenFold();
    
    #endif
    
    TNode* order = new TNode[n];
    TArc* low = new TArc[n];

    if (returnArc>=2*m) returnArc = 0;

    bool connected = !CutNodes(order,low,NULL,returnArc);

    if (connected)
    {
        LogEntry(LOG_METH,"Computing ear decomposition...");

        #if defined(_LOGGING_)

        OpenFold();

        #endif

        InitEdgeColours();
        SetEdgeColour(returnArc,0);
        
        TArc* Q = new TArc[n];
        for (TNode v=0;v<n;v++) Q[v] = NoArc;
        
        Q[StartNode(returnArc)] = returnArc^1;
        Q[EndNode(returnArc)] = returnArc;

        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"Path 0 : %ld[%ld]%ld",
                StartNode(returnArc),returnArc,EndNode(returnArc));
            LogEntry(LOG_METH2,CT.logBuffer);
        }

        #endif

        CT.Ping(Handle(),n);
    
        TNode nrVertex = 0;
        TArc nrPath = 1;
        
        staticStack<TNode,TFloat> S1(n,CT);
        
        S1.Insert(EndNode(returnArc));
        
        InitColours();
    
        THandle H = Investigate();
        goblinIterator &I = Iterator(H);
    
        while (!S1.Empty())
        {
            TNode u = S1.Delete();
            
            while (I.Active(u))
            {
                TArc a = I.Read(u);

                if (P[u]==(a^1) || EdgeColour(a)<NoArc) continue;

                // Construct path Q from u to another visited node v

                TNode v = EndNode(a);

                TNode w = v;

                if (P[v]==a)
                {
                    w = EndNode(low[v]);
                    Q[w] = low[v];

                    v = StartNode(low[v]);

                    while (v!=u)
                    {
                        Q[v] = P[v];
                        v = StartNode(P[v]);
                    }
                }
                else if (order[v]>order[u])
                {
                    while (Q[w]==NoArc) w = StartNode(P[w]);

                    Q[v] = a;

                    while (v!=w)
                    {
                        TArc a2 = P[v];

                        v = StartNode(a2);
                        Q[v] = a2^1;
                    }
                }
                else Q[w] = a;

                // Push path nodes onto stack S1

                #if defined(_LOGGING_)

                THandle LH = NoHandle;

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,"Path %ld : %ld",nrPath,w);
                    LH = LogStart(LOG_METH2,CT.logBuffer);
                }

                #endif

                while (w!=u)
                {
                    TArc a2 = Q[w];
                    SetEdgeColour(a2,nrPath);
                    w = StartNode(a2);
                    if (w!=u) S1.Insert(w);

                    #if defined(_LOGGING_)

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,"[%ld]%ld",a2^1,w);
                        LogAppend(LH,CT.logBuffer);
                    }

                    #endif
                }

                #if defined(_LOGGING_)

                if (CT.logMeth>1) LogEnd(LH);

                #endif

                nrPath++;

                CT.Ping(Handle(),n);
            }

            colour[u] = nrVertex;
            nrVertex++;

        }
        
        colour[StartNode(returnArc)] = nrVertex;

        Close(H);
        
        delete[] Q;

        #if defined(_LOGGING_)

        CloseFold();

        #endif
        
        LogEntry(LOG_RES,"...Graph has been split into paths");
    }

    #if defined(_LOGGING_)
    
    CloseFold();
    
    #endif
    
    delete[] order;
    delete[] low;

    if (connected) LogEntry(LOG_RES,"...st-numbering found");
    else LogEntry(LOG_RES,"...no st-numbering exists");

    CloseFold(Mod2Connectivity,NO_INDENT);
    
    return connected;
}


bool abstractMixedGraph::MarkComponent(TArc bridge,TNode r)
    throw ()
{
    // This procedure determines the effect of contracting the bridge.
    // If 2-connectivity is maintained, the return value is false.
    // Otherwise, a 2-connected component containing r is returned by
    // the node colours.

    OpenFold(ModConnectivity,NO_INDENT);
    
    TNode t = StartNode(bridge);
    TNode s = EndNode(bridge);
    
    sprintf(CT.logBuffer,"Computing (%ld,%ld)-component...",t,s);
    LogEntry(LOG_METH,CT.logBuffer);
        
    InitPredecessors();
    InitColours(2);
    colour[t] = colour[s] = 0;
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    TNode u = r;
    TNode i = 3;
    
    #if defined(_LOGGING_)

    OpenFold();

    THandle LH = NoHandle;

    if (CT.logMeth>1)
    {
        sprintf(CT.logBuffer,"Component: %ld",u);
        LH = LogStart(LOG_METH2,CT.logBuffer);
    }

    #endif
        
    while (true)
    {
        if (I.Active(u))
        {
            TArc a = I.Read(u);
            TNode v = EndNode(a);

            if (UCap(a)>0 && P[v]==NoArc && v!=r && v!=t && v!=s)
            {
                P[v] = a;
                u = v;
                i++;

                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,",%ld",u);
                    LogAppend(LH,CT.logBuffer);
                }

                #endif
            }
        }
        else
        {
            // Backtracking

            colour[u] = 1;

            if (u==r) break;
            else u = StartNode(P[u]);
        }
    }
    
    Close(H);
        
    CT.Ping(Handle(),m);

    #if defined(_LOGGING_)

    if (CT.logMeth>1) LogEnd(LH);

    CloseFold();
        
    #endif

    CloseFold(ModConnectivity,NO_INDENT);
    
    return (i<n);
}


bool abstractMixedGraph::Connected(TCap k) throw()
{
    if (k==1) return Connected();
    if (k==2) return !CutNodes();

    OpenFold(ModConnectivity,NO_INDENT);

    ReleasePredecessors();

    TCap kAct = Connectivity();
    
    CloseFold(ModConnectivity,NO_INDENT);

    return (kAct>=k);
}


bool abstractMixedGraph::EdgeConnected(TCap k) throw()
{
    if (k==1) return Connected();
    if (k==2) return TwoConnected();

    ReleasePredecessors();
    InitColours(0);

    if (EdgeConnectivity()>=k)
    {
        SetBounds(TimerConnectivity,1,1);
        InitColours(0);
        return 1;
    }
        
    OpenFold(ModConnectivity,NO_INDENT);
    
    sprintf(CT.logBuffer,"Computing %g-edge connected components...",k);
    LogEntry(LOG_METH,CT.logBuffer);
    OpenFold();
        
    SetBounds(TimerConnectivity,2,n);

    TNode cNext = 2;
    TNode cCurrent = 0;
 
    bool searching = true;
    while (cCurrent<cNext)
    {
        if (!searching)
        {
            TNode nMerged = 0;
            for (TNode v=0;v<n;v++)
                if (colour[v]<=cCurrent) nMerged++;
            cCurrent++;
            SetUpperBound(TimerConnectivity,n-nMerged+cCurrent);
        }
        
        searching = false;
        for (TNode v=0;v<n;v++)
        {
            if (colour[v]!=cCurrent) continue;

            TNode w = (v+1)%n;
            for (;w!=v && colour[w]!=cCurrent;w=(w+1)%n) {};

            if (w==v) continue;

            sprintf(CT.logBuffer,"Trying to separate %ld and %ld...",v,w);
            LogEntry(LOG_METH2,CT.logBuffer);

            diGraph G(n,TOption(0),CT);
            for (TArc a=0;a<2*m;a++)
            {
                TNode x = StartNode(a);
                TNode y = EndNode(a);
                if (colour[x]==cCurrent && colour[y]==cCurrent)
                    G.InsertArc(x,y,UCap(a),0,0);
            }

            #if defined(_TRACING_)

            for (TNode u=0;u<n;u++)
            {
                if (Dim()>0)
                {
                    G.SetC(u,0,C(u,0));
                    G.SetC(u,1,C(u,1));
                }
            }

            #endif

            CT.SuppressLogging();
            G.InitSubgraph();
            G.MaxFlow(v,w);
            TCap val = G.FlowValue(v,w);
            CT.RestoreLogging();

            if (val<k)
            {
                sprintf(CT.logBuffer,"...Minimum cut has capacity %g",val);
                LogEntry(LOG_RES2,CT.logBuffer);

                for (TNode u=0;u<n;u++)
                    if (colour[u]==colour[v] && G.Dist(u)==InfFloat)
                        colour[u] = cNext;
                cNext++;
                SetLowerBound(TimerConnectivity,cNext);
                CT.Ping(Handle(),n*n*m);
                searching = true;
            }
        }
    }
    
    CloseFold();

    SetUpperBound(TimerConnectivity,cNext);

    sprintf(CT.logBuffer,"...Graph has %ld %g-edge connected components",cNext,k);
    LogEntry(LOG_RES,CT.logBuffer);

    CloseFold(ModConnectivity,NO_INDENT);
    
    return cNext==1;
}


bool abstractMixedGraph::StronglyConnected(TCap k) throw()
{
    if (k==1) return StronglyConnected();
        
    OpenFold(ModConnectivity,NO_INDENT);

    ReleasePredecessors();

    TCap kAct = StrongConnectivity();
    
    CloseFold(ModConnectivity,NO_INDENT);

    return (kAct>=k);
}


bool abstractMixedGraph::StronglyEdgeConnected(TCap k) throw()
{
    if (k==1) return StronglyConnected();
    
    ReleasePredecessors();
    InitColours(0);

    if (StrongEdgeConnectivity()>=k)
    {
        SetBounds(TimerConnectivity,1,1);
        return true;
    }
        
    OpenFold(ModConnectivity,NO_INDENT);
    
    sprintf(CT.logBuffer,"Computing strong %g-edge connected components...",k);
    LogEntry(LOG_METH,CT.logBuffer);
    OpenFold();
        
    SetBounds(TimerConnectivity,2,n);

    TNode cNext = 1;
    TNode cCurrent = 0;
 
    bool searching = true;
    while (cCurrent<cNext)
    {
        if (!searching)
        {
            TNode nMerged = 0;
            for (TNode v=0;v<n;v++)
                if (colour[v]<=cCurrent) nMerged++;
            cCurrent++;
            SetUpperBound(TimerConnectivity,n-nMerged+cCurrent);
        }
        
        searching = false;
        for (TNode v=0;v<n;v++)
        {
            if (colour[v]!=cCurrent) continue;

            TNode w = (v+1)%n;
            for (;w!=v && colour[w]!=cCurrent;w=(w+1)%n) {};

            if (w==v) continue;

            sprintf(CT.logBuffer,"Trying to separate %ld and %ld...",v,w);
            LogEntry(LOG_METH2,CT.logBuffer);

            diGraph G(n,TOption(0),CT);
            for (TArc a=0;a<2*m;a++)
            {
                TNode x = StartNode(a);
                TNode y = EndNode(a);
                if (colour[x]==cCurrent && colour[y]==cCurrent && !Blocking(a))
                    G.InsertArc(x,y,UCap(a),0,0);
            }

            #if defined(_TRACING_)

            for (TNode u=0;u<n;u++)
            {
                if (Dim()>0)
                {
                    G.SetC(u,0,C(u,0));
                    G.SetC(u,1,C(u,1));
                }
            }

            #endif

            CT.SuppressLogging();
            G.InitSubgraph();
            G.MaxFlow(v,w);
            TCap val = G.FlowValue(v,w);
            CT.RestoreLogging();

            if (val<k)
            {
                sprintf(CT.logBuffer,"...Minimum cut has capacity %g",val);
                LogEntry(LOG_RES2,CT.logBuffer);

                for (TNode u=0;u<n;u++)
                    if (colour[u]==colour[v] && G.Dist(u)==InfFloat)
                        colour[u] = cNext;
                CT.Ping(Handle(),n*n*m);
                searching = true;
                cNext++;
                SetLowerBound(TimerConnectivity,cNext);
            }
        }
    }

    CloseFold();

    SetUpperBound(TimerConnectivity,cNext);
    
    sprintf(CT.logBuffer,"...Graph has %ld %g-edge connected components",cNext,k);
    LogEntry(LOG_RES,CT.logBuffer);

    CloseFold(ModConnectivity,NO_INDENT);
    
    return cNext==1;
}
