
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, July 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "absdig.h"
#include "statq.h"
#include "shrfam.h"
#include "fnw2fnw.h"
#include "lpSolve.h"
#include "networkSimplex.h"


TFloat abstractDiGraph::MinCostFlow(TNode s,TNode t,TFloat offset)
    throw(ERRange,ERRejected)
{
    OpenFold(ModMinCFlow,NO_INDENT);

    if (CT.logMeth)
    {
        sprintf(CT.logBuffer,"Computing minimum cost (%ld,%ld)-flow...",s,t);
        LogEntry(LOG_METH,CT.logBuffer);
    }

    TFloat ret = InfFloat;

    #if defined(_FAILSAVE_)

    if (!CLCap()) Error(ERR_REJECTED,"MinCostFlow","Non-trivial lower bounds");

    #endif
    
    if (CT.methMinCFlow==0) ret = EdmondsKarp2(s,t,offset);
    if (CT.methMinCFlow==1) ret = BusackerGowen(s,t,offset);
    if (CT.methMinCFlow==2) 
    {
        #if defined(_FAILSAVE_)

        if (CT.methMinCCirc == 4)
            Error(ERR_REJECTED,"MinCostFlow","Incompatible methMinCCirc option");

        #endif
    
        MaxFlow(s,t);
        ret = MinCostCirculation();
    }
    
    #if defined(_FAILSAVE_)

    if (CT.methMinCFlow>2) UnknownOption("MinCostFlow",CT.methMinCFlow);
    
    if (CT.methFailSave==1 && !Compatible())
    {
        InternalError("MinCostFlow","Solutions are corrupted");
    }
    
    try
    {
        FlowValue(s,t);
    }
    catch (...)
    {
        InternalError("MaxFlow","Solutions are corrupted");
    }

    #endif
    
    CloseFold(ModMinCFlow,NO_INDENT);
 
    return ret;
}


TFloat abstractDiGraph::BusackerGowen(TNode s,TNode t,TFloat offset)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("BusackerGowen",s);

    if (t>=n) NoSuchNode("BusackerGowen",t);

    #endif
    
    OpenFold(ModBusackerGowen,SHOW_TITLE);

    TFloat w = offset;
    SetLowerBound(TimerMinCFlow,w);

    while (true)
    {
        LogEntry(LOG_METH,"Computing shortest augmenting path...");

        ShortestPath(s,t);
         
        if (d[t]==InfFloat) break;

        TFloat Lambda = FindCap(s,t);

        if (Lambda==InfCap)
        {
            LogEntry(LOG_RES,"...Problem is unbounded");
            SetBounds(TimerMinCFlow,-InfFloat,-InfFloat);
            CT.Ping(Handle(),m);
            delete nHeap;
            nHeap = NULL;
            CloseFold(ModBusackerGowen);
            return -InfFloat;
        }

        Augment(s,t,Lambda);
            
        w += Lambda*d[t];
        SetLowerBound(TimerMinCFlow,w);

        CT.Ping(Handle(),m);
    }
    
    ReleasePotentials();

    LogEntry(LOG_METH,"Computing optimal node potentials...");

    NegativeCycle();
    UpdatePotentials(InfFloat);

    SetUpperBound(TimerMinCFlow,w);

    CloseFold(ModBusackerGowen);

    return w;
}


TFloat abstractDiGraph::EdmondsKarp2(TNode s,TNode t,TFloat offset)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("EdmondsKarp2",s);

    if (t>=n) NoSuchNode("EdmondsKarp2",t);

    #endif
    
    OpenFold(ModShortPath2,SHOW_TITLE);

    nHeap = NewNodeHeap();
    
    TFloat w = offset;
    SetLowerBound(TimerMinCFlow,w);

    while (true)
    {
        LogEntry(LOG_METH,"Computing shortest augmenting path...");

        Dijkstra(s,t);
       
        if (d[t]==InfFloat) break;

        TFloat Lambda = FindCap(s,t);

        if (Lambda==InfCap)
        {
            LogEntry(LOG_RES,"...Problem is unbounded");
            SetBounds(TimerMinCFlow,-InfFloat,-InfFloat);
            CT.Ping(Handle(),m);
            delete nHeap;
            nHeap = NULL;
            CloseFold(ModShortPath2);
            return -InfFloat;
        }

        Augment(s,t,Lambda);
        UpdatePotentials(d[t]);

        w += Lambda*pi[t];
        SetLowerBound(TimerMinCFlow,w);

        CT.Ping(Handle(),m);
    }
    
    delete nHeap;
    nHeap = NULL;

    SetUpperBound(TimerMinCFlow,w);

    CloseFold(ModShortPath2);
    
    return w;
}


TFloat abstractDiGraph::MinCostCirculation() throw(ERRejected)
{
    OpenFold(ModMinCFlow,NO_INDENT);

    LogEntry(LOG_METH,"Computing minimum cost circulation...");

    #if defined(_FAILSAVE_)
 
    TCap checkSum = 0;
    for (TNode v=0;v<n;v++) checkSum += Demand(v);
    
    if (checkSum!=0)
    {
        CloseFold(ModMinCFlow,NO_INDENT);
        Error(ERR_REJECTED,"MinCostCirculation","Node demands do not resolve");
    }

    #endif

    TFloat ret = InfFloat;
    
    switch (CT.methMinCCirc)
    {
        case 0:
        {
            ret = CycleCanceling();
            break;
        }
        case 1:
        case 2:
        {
            ret = CostScaling();
            break;
        }
        case 3:
        {
            ret = MinMeanCycleCanceling();
            break;
        }
        case 4:
        {
            #if defined(_FAILSAVE_)

            if (CT.methMinCFlow == 4)
            {
                CloseFold(ModMinCFlow,NO_INDENT);
                Error(ERR_REJECTED,"MinCostCirculation",
                    "Incompatible methMinCFlow option");
            }

            #endif

            ret = ShortestAugmentingPath();
            break;
        }
        case 5:
        {
            ret = NWSimplex();
            break;
        }
        case 6:
        {
            goblinILPModule *ModLP =
                (goblinILPModule*)goblinController::pLPModule;
            goblinILPWrapper *XLP  =
                ModLP->NewInstance(n,m,2*m,goblinILPWrapper::MINIMIZE,CT);

            for (TNode v=0;v<n;v++)
            {
                XLP -> AddRestr(-Demand(v),-Demand(v));
            }

            TRestr rowIndex[2] = {NoRestr,NoRestr};
            TFloat rowCoeff[2] = {1,-1};

            for (TArc a=0;a<m;a++)
            {
                XLP -> AddVar(LCap(2*a),UCap(2*a),Length(2*a),
                                goblinILPWrapper::VAR_FLOAT);
                
                rowIndex[0] = StartNode(2*a);
                rowIndex[1] = EndNode(2*a);
                XLP -> SetColumn(a,2,rowIndex,rowCoeff);
            }

            if (m>0)
            {
                XLP -> ResetBasis();
                XLP -> SolveLP();
            }

            InitPotentials();
            for (TNode v=0;v<n;v++)
                pi[v] = -XLP->Y(v,goblinILPWrapper::LOWER)
                        -XLP->Y(v,goblinILPWrapper::UPPER);

            for (TArc a=0;a<m;a++) SetSub(2*a,XLP->X(a));

            ret = XLP -> ObjVal();

            delete XLP;

            SetBounds(TimerMinCFlow,ret,ret);

            break;
        }
        default:
        {
            UnknownOption("MinCostCirculation",CT.methMinCCirc);
        }
    }
    
    #if defined(_FAILSAVE_)

    if (CT.methFailSave==1 && !Compatible())
    {
        CloseFold(ModMinCFlow,NO_INDENT);
        InternalError("MinCostCirculation","Solutions are corrupted");
    }

    for (TNode v=0;v<n && CT.methFailSave==1;v++)
    {
        if (Divergence(v)!=Demand(v))
        {
            #if defined(_LOGGING_)

            CloseFold(ModMinCFlow,NO_INDENT);
            InternalError("FlowValue","Not a legal b-flow");

            #endif
        }
    }

    #endif
 
    CloseFold(ModMinCFlow,NO_INDENT);

    return ret;
}


TFloat abstractDiGraph::CycleCanceling() throw()
{
    OpenFold(ModKleinCanceling,SHOW_TITLE);

    TFloat w = Weight();
    SetUpperBound(TimerMinCFlow,w);

    InitPotentials();

    while (CT.SolverRunning())
    {
        LogEntry(LOG_METH,"Searching for negative length augmenting cycle...");

        TNode x = NegativeCycle();

        if (x==NoNode) break;

        TFloat Lambda = FindCap(x,x);

        if (Lambda==InfCap)
        {
            LogEntry(LOG_RES,"...Problem is unbounded");
            SetBounds(TimerMinCFlow,-InfFloat,-InfFloat);
            CT.Ping(Handle(),n*m);
            CloseFold(ModKleinCanceling);
            return -InfFloat;
        }

        Augment(x,x,Lambda);

        CT.Ping(Handle(),n*m);
    }
    
    UpdatePotentials(0);

    w = Weight();
    SetBounds(TimerMinCFlow,w,w);

    CloseFold(ModKleinCanceling);

    return w;
}


TFloat abstractDiGraph::MinMeanCycleCanceling() throw()
{
    OpenFold(ModMeanCycleCanceling,SHOW_TITLE);

    TFloat w = Weight();
    SetUpperBound(TimerMinCFlow,w);

    while (CT.SolverRunning())
    {
        LogEntry(LOG_METH,"Searching for minimum mean augmenting cycle...");

        TNode x = MinimumMeanCycle();

        if (x==NoNode || mu>=0) break;

        CT.Ping(Handle(),n*m);
        TFloat Lambda = FindCap(x,x);

        if (Lambda==InfCap)
        {
            LogEntry(LOG_RES,"...Problem is unbounded");
            SetBounds(TimerMinCFlow,-InfFloat,-InfFloat);
            CT.Ping(Handle(),n*m);
            CloseFold(ModMeanCycleCanceling);
            return -InfFloat;
        }

        Augment(x,x,Lambda);
    }
    
    LogEntry(LOG_METH,"Computing optimum node potentials...");

    mu = 0;
    NegativeCycle();

    UpdatePotentials(0);

    w = Weight();
    SetBounds(TimerMinCFlow,w,w);

    CloseFold(ModMeanCycleCanceling);

    return w;
}


TFloat abstractDiGraph::CostScaling() throw()
{
    OpenFold(ModCostScaling,SHOW_TITLE);

    TFloat w = Weight();
    SetUpperBound(TimerMinCFlow,w);

    InitLabels();

    if (pi==NULL) InitPotentials();

    InitColours(); 
    // Colours maintain a topological ordering of the admissible graph
    
    // A queue for the active nodes
    
    goblinQueue<TNode,TFloat> *Q = NewNodeHeap();
    
    // Additional incidence structure which is initialized by relabel(u)
    // operations. Admissible[u] contains of all arcs a=uv such that
    // rescap(a)>0 and d[u]=d[v]+1.
    
    staticQueue<TNode,TFloat> **Admissible = new staticQueue<TNode,TFloat>*[n];
    Admissible[0] = new staticQueue<TArc,TFloat>(2*m,CT);

    for (TNode i=1;i<n;i++)
        Admissible[i] = new staticQueue<TArc,TFloat>(*Admissible[0]);
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    TFloat epsilon = InfFloat;
        
    while (CT.SolverRunning())
    {
        bool searching = true;

        if (CT.methMinCCirc==2)
        {
            LogEntry(LOG_METH,"Checking for tightness...");
            MinimumMeanCycle();
            epsilon = -mu;

            if (mu>=0)
            {
                searching = false;
                mu = 0;
            }

            ReleasePotentials();
            NegativeCycle();
            UpdatePotentials(0);
        }
        else
        {
            epsilon = 0;
            for (TArc a=0;a<2*m;a++)
                if (ResCap(a)>0 && RedLength(a)<(-epsilon))
                    epsilon = -RedLength(a);
                    
            if (n*epsilon<CT.tolerance)
            {
                searching = false;
                ReleasePotentials();
                NegativeCycle();
                UpdatePotentials(0);
            }
        }
        
        if (!searching) break;

        if (CT.logMeth)
        {
            sprintf(CT.logBuffer,
                "Starting scaling phase with epsilon = %.3f...",epsilon);
            LogEntry(LOG_METH,CT.logBuffer);
        }

        OpenFold();

        CT.Ping(Handle(),n*n*n);

        for (TArc a=0;a<2*m;a++)
        {
            if (ResCap(a)>0 && RedLength(a)<0)
            {
                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,
                        "Pushed %g flow units from node %ld to node %ld",
                        ResCap(a),StartNode(a),EndNode(a));
                    LogEntry(LOG_METH2,CT.logBuffer);
                }

                #endif

                Push(a,ResCap(a));
            }
        }

        for (TNode v=0;v<n;v++)
        {
            colour[v] = 0;
            if (Divergence(v)>Demand(v))
            {
                Q -> Insert(v,colour[v]);

                I.Reset(v);
                while (I.Active(v))
                {
                    TArc a = I.Read(v);
                    if (ResCap(a)>0 && RedLength(a)<0)
                        Admissible[v]->Insert(a);
                }
            }
        }

        while (!Q->Empty())
        {
            TNode v = Q->Delete();
            I.Reset(v);

            while (Divergence(v)>Demand(v) && !Admissible[v]->Empty())
            {
                TArc a = Admissible[v]->Delete();

                if (ResCap(a)>0 && RedLength(a)<0)
                {
                    TNode w = EndNode(a);
                    TFloat Lambda = ResCap(a);

                    if (Lambda>Divergence(v)-Demand(v))
                        Lambda = Divergence(v)-Demand(v);

                    Push(a,Lambda);

                    if (ResCap(a)>0) Admissible[v]->Insert(a);

                    #if defined(_LOGGING_)

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,
                            "Pushed %g flow units from node %ld to node %ld",
                            Lambda,v,w);
                        LogEntry(LOG_METH2,CT.logBuffer);
                    }

                    #endif

                    if (Divergence(w)>Demand(w)
                        && Divergence(w)-Demand(w)<=Lambda)
                        Q -> Insert(w,colour[w]);
                }
            }

            if (Divergence(v)>Demand(v))
            {
                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,"Relabelled node %ld: %g -> %g",
                        v,pi[v],pi[v]-epsilon/2);
                    LogEntry(LOG_METH2,CT.logBuffer);
                }

                #endif

                pi[v] -= epsilon/2;

                I.Reset(v);

                while (I.Active(v))
                {
                    TArc a = I.Read(v);
                    TNode w = EndNode(a);

                    if (ResCap(a)>0 && RedLength(a)<0)
                        Admissible[v]->Insert(a);

                    if (ResCap(a^1)>0 && RedLength(a^1)<0)
                        Admissible[w]->Insert(a);
                }

                for (TNode w=0;w<n;w++)
                    if (colour[w]<colour[v])
                    {
                        colour[w]--;

                        if (Divergence(w)>Demand(w))
                            Q->ChangeKey(w,colour[w]); 
                    }

                colour[v] = 0;
                Q -> Insert(v,colour[v]);
            }
        }

        CloseFold();

        TFloat w = Weight();
        SetUpperBound(TimerMinCFlow,w);

        if (CT.logRes>1)
        {
            sprintf(CT.logBuffer,
                "...Feasible flow of weight %.3f found",w);
            LogEntry(LOG_RES2,CT.logBuffer);
        }
    }

    Close(H);
    
    for (TNode i=1;i<n;i++) delete Admissible[i];

    delete Admissible[0];
    delete[] Admissible;
    delete Q;
    ReleaseColours();

    w = Weight();
    SetBounds(TimerMinCFlow,w,w);

    CloseFold(ModCostScaling);

    return w;
}


TFloat abstractDiGraph::ShortestAugmentingPath() throw(ERRejected)
{
    if (pi==NULL) InitPotentials();

    FNW2FNW G(*this,pi);

    TFloat w = G.MinCostFlow(G.Source(),G.Target(),Weight());
    
    CT.Ping(Handle(),n*m);
    
    if (!G.Perfect())
        Error(ERR_REJECTED,"ShortestAugmentingPath","No b-flow exists");
    
    return w;
}


TFloat abstractDiGraph::NWSimplex() throw()
{
    OpenFold(ModNetworkSimplex,SHOW_TITLE);

    // Compute basic solution
    NWSimplexCancelFree();
    
    // Set up basis structure
    NWSimplexStrongTree();

    networkSimplex NWSimplexData(*this);
    
    NWSimplexData.InitThreadIndex();
    NWSimplexData.ComputePotentials();

    // Main loop
    LogEntry(LOG_METH,"Pivoting...");
    CT.Ping(Handle(),m);
    
    TArc a = NWSimplexData.PivotArc();
    unsigned long count = 0;
    bool unbounded = false;

    while (CT.SolverRunning() && a!=NoArc)
    {
        unbounded = NWSimplexData.PivotOperation(a);
        CT.Ping(Handle(),m);
        
        if (unbounded) a = NoArc;
        else a = NWSimplexData.PivotArc();

        count++;
    }
    
    sprintf(CT.logBuffer,"...%ld pivots in total",count);
    LogEntry(LOG_METH,CT.logBuffer);

    TFloat w = -InfFloat;

    if (!unbounded) w = Weight();

    SetBounds(TimerMinCFlow,w,w);
    
    CloseFold(ModNetworkSimplex);

    return w;
}


void abstractDiGraph::NWSimplexCancelFree() throw()
{
    LogEntry(LOG_METH,"Cancelling free cycles...");
    OpenFold();

    InitPredecessors();
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    unsigned long k = 0;

    for (TNode r=0;r<n;r++)
    {
        TNode v = r;

        while (v!=r || I.Active(v))
        {
            if (I.Active(v))
            {
                TArc a = I.Peek(v);
                TNode u = EndNode(a);
            
                if (ResCap(a)>0 && ResCap(a^1)>0 && P[v]!=(a^1))
                {
                    if (P[u]!=NoArc || u==r)
                    {
                        CT.Ping(Handle(),m);
                        TArc savedPu = P[u];
                        P[u] = a;
                        CT.Ping(Handle(),m);

                        TFloat length = 0;
                        TNode w = u;

                        while (true)
                        {
                            a = P[w];
    
                            length += Length(a);
                            w = StartNode(a);
                            if (w==u) break;
                        }
    
                        if (length>0)
                        {
                            TArc a = P[u];
                            TArc final = a;
                            w = StartNode(a);

                            while (P[w]!=final)
                            {
                                TArc a2 = P[w];
                                P[w] = (a^1);
                                w = StartNode(a2);
                                a = a2;
                            }

                            P[w] = (a^1);
                        }

                        TFloat Lambda = FindCap(u,u);
                        Augment(u,u,Lambda);
                        P[u] = savedPu;
                        v = u;
                        k++;
                    }
                    else
                    {
                        P[u] = a;
                        v = u;
                    }
                }
                else I.Skip(v);
            }
            else
            {
                TArc a = P[v];
                P[v] = NoArc;
                v = StartNode(a);
                I.Skip(v);
            }
        }
    }
    
    Close(H);
    ReleasePredecessors();

    CloseFold();
    sprintf(CT.logBuffer,"...%ld cycles eliminated",k);
    LogEntry(LOG_METH,CT.logBuffer);
}

 
void abstractDiGraph::NWSimplexStrongTree() throw()
{
    LogEntry(LOG_METH,"Computing strongly feasible spanning tree...");
    OpenFold();
    
    TNode nTotal = 2*n;
    shrinkingFamily<TNode> S(n,nTotal,CT);
    
    TArc* inArc = new TArc[nTotal];

    for (TNode v=0;v<nTotal;v++) inArc[v] = NoArc;
    
    staticQueue<TArc,TFloat> Q(2*m,CT);

    for (TArc a=0;a<2*m;a++)
        if (ResCap(a^1)>0 && ResCap(a)>0) Q.Insert(a);

    for (TArc a=0;a<2*m;a++)
        if (ResCap(a^1)>0 && ResCap(a)==0) Q.Insert(a);
    
    LogEntry(LOG_METH2,"Shrinking cycles...");

    OpenFold();

    TNode rank = 0;
    TNode nCycles = 0;

    while (true)
    {
        TArc minArc = NoArc;

        while (!Q.Empty() && minArc==NoArc)
        {
            TArc a = Q.Delete();
            TNode u = S.Set(StartNode(a));
            TNode v = S.Set(EndNode(a));
            
            if (u!=v && inArc[v]==NoArc) minArc = a;
        }
        
        if (minArc==NoArc) break;
        else
        {
            TNode u = S.Set(StartNode(minArc));
            TNode v = S.Set(EndNode(minArc));
            
            inArc[v] = minArc;

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"inArc[%ld] = %ld",v,minArc);
                LogEntry(LOG_METH2,CT.logBuffer);
            }
            
            #endif

            TNode w = u;

            while (w!=v && inArc[w]!=NoArc) w = S.Set(StartNode(inArc[w]));

            if (w==v)
            {
                #if defined(_LOGGING_)

                THandle LH = NoHandle;

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,"Merging cycle (%ld",u);
                    LH = LogStart(LOG_METH2,CT.logBuffer);
                }

                #endif
        
                TNode cycle = S.MakeSet();
                S.Merge(cycle,u);
                w = u;

                while (w!=v)
                {
                    w = S.Set(StartNode(inArc[w]));
                    S.Merge(cycle,w);
                    
                    #if defined(_LOGGING_)

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,",%ld",w);
                        LogAppend(LH,CT.logBuffer);
                    }
                    
                    #endif
                }
                S.FixSet(cycle);
                
                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,",%ld)",u);
                    LogEnd(LH,CT.logBuffer);
                }

                #endif

                nCycles++;
            }
            else  rank++;
        }
    }

    CloseFold();
    
    LogEntry(LOG_METH2,"Expanding cycles...");

    OpenFold();

    for (TNode v=n+nCycles-1;v>=n;v--)
    {
        TArc a = inArc[v];
        if (a==NoArc)
        {
            TNode u = S.First(v);
            S.Split(v);
            
            TNode thisRoot = u;
            TNode w = S.Set(StartNode(inArc[u]));

            while (w!=u)
            {
                if (ResCap(inArc[w])==0)
                    thisRoot = w;
                w = S.Set(StartNode(inArc[w]));
            }
            
            inArc[thisRoot] = NoArc;
            
            #if defined(_LOGGING_)

            if (CT.logMeth>1 && IsDirected())
            {
                sprintf(CT.logBuffer,"inArc[%ld] = *",thisRoot);
                LogEntry(LOG_METH2,CT.logBuffer);
            }
            
            #endif
        }
        else
        {
            S.Split(v);

            TNode u = S.Set(EndNode(a));
            inArc[u] = a;
            
            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"inArc[%ld] = %ld",u,a);
                LogEntry(LOG_METH2,CT.logBuffer);
            }
            
            #endif
        }
    }

    CloseFold();
    
    InitPredecessors();

    for (TNode v=0;v<n;v++) P[v] = inArc[v];

    delete[] inArc;

    CloseFold();
    
    if (CT.logRes && n>rank+1)
    {
        sprintf(CT.logBuffer,"...Network splits into %ld independent problems",n-rank);
        LogEntry(LOG_METH,CT.logBuffer);
    }
}
