
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, July 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsgra.h"
#include "spsbig.h"
#include "gra2bal.h"
#include "big2fnw.h"
#include "binheap.h"


bool abstractGraph::MaximumMatching() throw()
{
    LogEntry(LOG_METH,"Computing maximum matching...");

    OpenFold(ModMatching);
    
    gra2bal G(*this);
    G.MaxBalFlow(G.Source());
    CT.Ping(Handle(),n*m);
    
    CloseFold(ModMatching);

    return (G.Perfect());
};


bool abstractGraph::MaximumMatching(TCap cDeg) throw()
{
    LogEntry(LOG_METH,"Computing maximum matching...");

    OpenFold(ModMatching);
    
    gra2bal G(*this,cDeg);
    G.MaxBalFlow(G.Source());
    CT.Ping(Handle(),n*m);
    
    CloseFold(ModMatching);

    return (G.Perfect());
};


bool abstractGraph::MaximumMatching(TCap *pLower,TCap *pDeg) throw()
{
    LogEntry(LOG_METH,"Computing maximum matching...");

    OpenFold(ModMatching);
    
    gra2bal G(*this,pLower,pDeg);
    G.MaxBalFlow(G.Source());
    CT.Ping(Handle(),n*m);
    
    CloseFold(ModMatching);

    return (G.Perfect());
};


bool abstractGraph::MinCMatching() throw()
{
    LogEntry(LOG_METH,"Computing minimum cost matching...");

    OpenFold(ModMatching);
    
    if (IsDense() && CT.methCandidates>=0)
    {
        CloseFold(ModMatching);

        return PMHeuristicsCandidates();
    }

    InitSubgraph();
    bool ret = false;
    try
    {
        gra2bal G(*this);
        G.MinCBalFlow(G.Source());
    
        ret = G.Perfect();
    }
    catch (ERRejected) {}
    
    CloseFold(ModMatching);

    return ret;
};


bool abstractGraph::MinCMatching(TCap cDeg) throw()
{
    LogEntry(LOG_METH,"Computing minimum cost matching...");

    OpenFold(ModMatching);
    
    InitSubgraph();
    bool ret = false;
    try
    {
        gra2bal G(*this,cDeg);
        G.MinCBalFlow(G.Source());
        ret = G.Perfect();
    }
    catch (ERRejected) {}
    
    CloseFold(ModMatching);
    return ret;
};


bool abstractGraph::MinCMatching(TCap *pLower,TCap *pDeg) throw()
{
    LogEntry(LOG_METH,"Computing minimum cost matching...");

    OpenFold(ModMatching);
    
    InitSubgraph();
    bool ret = false;
    try
    {
        gra2bal G(*this,pLower,pDeg);
        G.MinCBalFlow(G.Source());
        ret = G.Perfect();
    }
    catch (ERRejected) {}
    
    CloseFold(ModMatching);

    return ret;
};


bool abstractGraph::PMHeuristicsCandidates() throw()
{
    LogEntry(LOG_METH,"(Candidate Subgraph Heuristics)");
    OpenFold();

    graph G(n,0,CT);
    int savedMethAdjacency = CT.methAdjacency;
    CT.methAdjacency = 0;

    int i = 0;
    while (i<10)
    {
        LogEntry(LOG_METH,"Searching for candidate matching...");
        TFloat ret = PMHeuristicsRandom();
        if (ret<InfFloat)
        {
            for (TArc a=0;a<m;a++)
            {
                TNode u = StartNode(2*a);
                TNode v = EndNode(2*a);

                if (Sub(2*a)>0 && G.Adjacency(u,v)==NoArc)
                    G.InsertArc(u,v,UCap(2*a),0,Length(2*a));
            }
            
            i++;
        }
    }

    binaryHeap<TArc,TFloat> Q(n,CT);
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    for (TNode v=0;v<n;v++)
    {
        G.SetDemand(v,Demand(v)); 
        G.SetC(v,0,C(v,0)); 
        G.SetC(v,1,C(v,1)); 

        while (I.Active(v))
        {
            TArc a = I.Read(v);
            if (EndNode(a)!=v) 
                Q.Insert(EndNode(a),Length(a));
        }
        
        for (int i=0;!Q.Empty();i++)
        {
            TNode w = Q.Delete();
            TArc a = Adjacency(v,w);       

            if (i<CT.methCandidates && G.Adjacency(v,w)==NoArc)
                G.InsertArc(v,w,UCap(a),0,Length(a));
        }
    }

    Close(H);
    CT.methAdjacency = savedMethAdjacency;
    
    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...candidate subgraph has %ld arcs",G.M());
        LogEntry(LOG_RES,CT.logBuffer);
    }

    if (CT.traceLevel==3) CT.Ping(G.Handle(),21*m);
    
    bool ret = false;
    if (CT.methSolve==0)
    {
        // Candidate heuristics
    
        ret = G.MinCMatching();
    
        if (ret)
        {
            InitSubgraph();

            for (TArc a=0;a<G.M();a++)
            {
                TNode u = StartNode(2*a);
                TNode v = EndNode(2*a);

                if (G.Sub(2*a)>0)
                    AddArc(Adjacency(u,v),G.Sub(2*a));
            }
        }
    }
    else
    {
        // Exact method
    
        gra2bal H1(G);
        H1.MinCostFlow(H1.Source(),H1.Target());
        H1.Symmetrize();
    
        InitSubgraph();

        for (TArc a=0;a<G.M();a++)
        {
                TNode u = StartNode(2*a);
                TNode v = EndNode(2*a);

            if (G.Sub(2*a)>0)
                AddArc(Adjacency(u,v),G.Sub(2*a));
        }

        gra2bal H2(*this);

        for (TNode v=0;v<H1.N();v++) H2.SetPotential(v,H1.Pi(v));

        H2.ShortestAugmentingPath();
        H2.CancelEven();
        H2.CancelPD();
    
        ret = H2.Perfect();
    }
            
    CloseFold();
    return ret;
};


TFloat abstractGraph::PMHeuristicsRandom() throw()
{
    OpenFold();
    LogEntry(LOG_METH,"(Random Heuristics)");

    InitSubgraph();
    TFloat weight = 0;
    
    goblinQueue<TNode,TFloat> *Q = NewNodeHeap();
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    
    for (TNode i=0;i<n;i++)
    {
        TNode v = TNode(CT.Rand(n));
        TNode shift = 0;

        while (Deg(v)==Demand(v) && shift<n)
        {
            v = (v+1)%n;
            shift++;
        }

        if (shift==n) break;
        
        I.Reset(v);
        while (I.Active(v))
        {
            TArc a = I.Read(v);
            if (EndNode(a)!=v) 
                Q -> Insert(EndNode(a),Length(a));
        }
        
        while (Deg(v)<Demand(v))
        {
            if (Q->Empty())
            {
                LogEntry(LOG_RES,"...no matching found");
                CloseFold();
                Close(H);
                delete Q;
                return InfFloat;
            }
            
            TNode w = Q->Delete();
            TArc a = Adjacency(v,w);
            TFloat lambda = UCap(a)-Sub(a);
            TFloat lambda2 = Demand(v)-Deg(v);

            if (v==w) lambda2 = floor(lambda2/2);
            else if (Demand(w)-Deg(w)<lambda2)
                lambda2 = Demand(w)-Deg(w);

            if (lambda2<lambda) lambda = lambda2;
            
            if (lambda>0)
            {
                AddArc(a,lambda);
                weight += lambda*Length(a);
                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,
                        "Adding arc %ld(%ld)%ld with multiplicity %.3f",
                        v,a,w,lambda);
                    LogEntry(LOG_METH2,CT.logBuffer);
                }
            }
        }
        
        Q -> Init();
    }
    
    Close(H);
    delete Q;
    
    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Matching of weight %.3f found",weight);
        LogEntry(LOG_RES,CT.logBuffer);
    }
    CloseFold();

    return weight;
};


bool abstractBiGraph::MaximumAssignment() throw()
{
    LogEntry(LOG_METH,"Computing maximum assignment...");

    OpenFold(ModMatching);
    
    big2FNW G(*this);
    G.MaxFlow(G.Source(),G.Target());
    CT.Ping(Handle(),n*m);
    
    CloseFold(ModMatching);

    return (G.Perfect());
};


bool abstractBiGraph::MaximumAssignment(TCap cDeg) throw()
{
    LogEntry(LOG_METH,"Computing maximum assignment...");

    OpenFold(ModMatching);
    
    big2FNW G(*this,cDeg);
    G.MaxFlow(G.Source(),G.Target());
    CT.Ping(Handle(),n*m);
    
    CloseFold(ModMatching);

    return (G.Perfect());
};


bool abstractBiGraph::MaximumAssignment(TCap *pLower,TCap *pDeg) throw()
{
    LogEntry(LOG_METH,"Computing maximum assignment...");

    OpenFold(ModMatching);
    
    big2FNW G(*this,pLower,pDeg);
    G.MaxFlow(G.Source(),G.Target());
    CT.Ping(Handle(),n*m);
    
    CloseFold(ModMatching);

    return (G.Perfect());
};


bool abstractBiGraph::MinCAssignment() throw()
{
    LogEntry(LOG_METH,"Computing optimal assignment...");

    OpenFold(ModMatching);
    
    if (IsDense() && CT.methCandidates>=0)
    {
        CloseFold();
        return PMHeuristicsCandidates();
    }

    InitSubgraph();
    bool ret = false;

    try
    {
        big2FNW G(*this);
        G.MinCostFlow(G.Source(),G.Target());
        ret = G.Perfect();
    }
    catch (ERRejected) {}

    CT.Ping(Handle(),n*m);
    
    CloseFold(ModMatching);

    return (ret);
};


bool abstractBiGraph::MinCAssignment(TCap cDeg) throw()
{
    LogEntry(LOG_METH,"Computing optimal assignment...");

    OpenFold(ModMatching);
    
    InitSubgraph();
    bool ret = false;

    try
    {
        big2FNW G(*this,cDeg);
        G.MinCostFlow(G.Source(),G.Target());
        ret = G.Perfect();
    }
    catch (ERRejected) {CloseFold();};
    CT.Ping(Handle(),n*m);
    
    CloseFold(ModMatching);

    return (ret);
};


bool abstractBiGraph::MinCAssignment(TCap *pLower,TCap *pDeg) throw()
{
    LogEntry(LOG_METH,"Computing optimal assignment...");

    OpenFold(ModMatching);
    
    InitSubgraph();
    bool ret = false;

    try
    {
        big2FNW G(*this,pLower,pDeg);
        G.MinCostFlow(G.Source(),G.Target());
        ret = G.Perfect();
    }
    catch (ERRejected) {CloseFold();};
    CT.Ping(Handle(),n*m);

    CloseFold(ModMatching);

    return (ret);
};


bool abstractBiGraph::PMHeuristicsCandidates() throw()
{
    LogEntry(LOG_METH,"(Candidate Subgraph Heuristics)");
    OpenFold();

    biGraph G(n1,n2,0,CT);
    int savedMethAdjacency = CT.methAdjacency;
    CT.methAdjacency = 0;

    int i = 0;
    while (i<10)
    {
        LogEntry(LOG_METH,"Searching for candidate matching...");
        TFloat ret = PMHeuristicsRandom();
        if (ret<InfFloat)
        {
            for (TArc a=0;a<m;a++)
                if (Sub(2*a)>0 && G.Adjacency(StartNode(2*a),EndNode(2*a))==NoArc)
                    G.InsertArc(StartNode(2*a),EndNode(2*a),UCap(2*a),0,Length(2*a));

            i++;
        }
    }

    binaryHeap<TArc,TFloat> Q(n,CT);
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    for (TNode v=0;v<n;v++)
    {
        G.SetDemand(v,Demand(v)); 
        G.SetC(v,0,C(v,0)); 
        G.SetC(v,1,C(v,1)); 

        while (I.Active(v))
        {
            TArc a = I.Read(v);
            if (EndNode(a)!=v) 
                Q.Insert(EndNode(a),Length(a));
        }
        
        int i = 0;

        while (!Q.Empty())
        {
            TNode w = Q.Delete();
            TArc a = Adjacency(v,w);       

            if (i<CT.methCandidates && G.Adjacency(v,w)==NoArc)
            {
                if (v<n1) G.InsertArc(v,w,UCap(a),0,Length(a));
                else  G.InsertArc(w,v,UCap(a),0,Length(a));
            }

            i++;
        }
    }

    Close(H);
    CT.methAdjacency = savedMethAdjacency;
    
    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...candidate subgraph has %ld arcs",G.M());
        LogEntry(LOG_RES,CT.logBuffer);
    }

    if (CT.traceLevel==3) CT.Ping(G.Handle(),11*m);
    
    bool ret = false;

    if (CT.methSolve==0)
    {
        // Candidate heuristics
        
        G.InitSubgraph();
        ret = G.MinCAssignment();
    
        if (ret)
        {
            InitSubgraph();

            for (TArc a=0;a<G.M();a++)
                if (G.Sub(2*a)>0)
                    AddArc(Adjacency(G.StartNode(2*a),G.EndNode(2*a)),G.Sub(2*a));
        }
    }
    else
    {
        // Exact method
    
        big2FNW H1(G);
        H1.MinCostFlow(H1.Source(),H1.Target());
    
        InitSubgraph();

        for (TArc a=0;a<G.M();a++)
            if (G.Sub(2*a)>0)
                AddArc(Adjacency(G.StartNode(2*a),G.EndNode(2*a)),G.Sub(2*a));

        big2FNW H2(*this);

        for (TNode v=0;v<H1.N();v++) H2.SetPotential(v,H1.Pi(v)); 

        H2.ShortestAugmentingPath();
    
        ret = H2.Perfect();
    }
            
    CloseFold();

    return ret;
};
