
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, July 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "branchMaxCut.h"
#include "spsgra.h"
#include "statq.h"


branchMaxCut::branchMaxCut(abstractMixedGraph &GC,TNode s,TNode t) throw() :
    goblinBBNode<TNode,TFloat>(GC.N(),GC.Context()), G(GC)
{
    currentVar = 0;

    totalWeight = 0;
    selectedWeight = 0;
    dismissedWeight = 0;

    colour = new char[n];
    leftWeight = new TFloat[n];
    rightWeight = new TFloat[n];

    for (TNode v=0;v<n;v++)
    {
        colour[v] = 1;
        leftWeight[v] = rightWeight[v] = 0;
    }
    
    // Compute all node capacities
    
    bool negWeights = false;
    for (TArc a=0;a<G.M();a++)
    {
        if (G.StartNode(2*a)==G.EndNode(2*a)) continue;

        leftWeight[G.StartNode(2*a)] += G.UCap(2*a)*G.Length(2*a);
        leftWeight[G.EndNode(2*a)] += G.UCap(2*a)*G.Length(2*a);

        if (G.Length(2*a)>=0)
            totalWeight += G.UCap(2*a)*G.Length(2*a);
        else negWeights = true;
    }
    
    if (negWeights) totalWeight = InfFloat;
    
    // Determine a node with maximum capacity

    TNode x = 0;
    TFloat maxNodeWeight = leftWeight[0];

    for (TNode v=1;v<n;v++)
    {
        if (leftWeight[v]>maxNodeWeight)
        {
            x = v;
            maxNodeWeight = leftWeight[v];
        }

        leftWeight[v] = 0;
    }

    if (t==NoNode)
    {

        if (s!=NoNode || G.IsUndirected())
        {
            if (s!=NoNode) x = s;

            // Mark x as a left-hand side node

            colour[x] = 0;
            unfixed--;

            // Update leftWeight with the arcs incident with x

            TArc a = G.First(x);
            do
            {
                TNode u = G.EndNode(a);

                if (!G.Blocking(a) && u!=x)
                    leftWeight[u] += G.UCap(a)*G.Length(a&(a^1));

                a = G.Right(a,x);
            }
            while (a!=G.First(x));
        }
    }
    else
    {
        // Mark t as a right-hand side node

        colour[t] = 2;
        unfixed--;

        // Update leftWeight with the arcs incident with x

        TArc a = G.First(t);
        do
        {
            TNode u = G.EndNode(a);

            if (!G.Blocking(a^1) && u!=t)
                rightWeight[G.EndNode(a)] += G.UCap(a)*G.Length(a&(a^1));

            a = G.Right(a,t);
        }
        while (a!=G.First(t));

        if (s!=NoNode) Lower(s);
    }
    
    source = s;
    target = t;
    
    LogEntry(LOG_MEM,"(maximum cut)");
}


branchMaxCut::branchMaxCut(branchMaxCut &Node) throw() :
    goblinBBNode<TNode,TFloat>(Node.G.N(),Node.Context(),Node.scheme), G(Node.G)
{
    colour = new char[n];
    leftWeight = new TFloat[n];
    rightWeight = new TFloat[n];

    currentVar = Node.currentVar;

    totalWeight = Node.totalWeight;
    selectedWeight = Node.selectedWeight;
    dismissedWeight = Node.dismissedWeight;

    for (TNode v=0;v<n;v++)
    {
        colour[v] = Node.colour[v];
        if (colour[v]!=1) unfixed--;

        leftWeight[v] = Node.leftWeight[v];
        rightWeight[v] = Node.rightWeight[v];
    }
    
    source = Node.source;
    target = Node.target;
    
    LogEntry(LOG_MEM,"(maximum cut)");
}


branchMaxCut::~branchMaxCut() throw()
{
    delete[] colour;
    delete[] leftWeight;
    delete[] rightWeight;
    
    LogEntry(LOG_MEM,"(maximum cut)");
}


unsigned long branchMaxCut::Size() throw()
{
    return
          sizeof(branchMaxCut)
        + goblinDataObject::Allocated()
        + goblinBBNode<TNode,TFloat>::Allocated()
        + Allocated();
}


unsigned long branchMaxCut::Allocated() throw()
{
    return n*(sizeof(char)+2*sizeof(TFloat));
}


TFloat branchMaxCut::SumWeight(TNode v) throw()
{
    return leftWeight[v]+rightWeight[v];
}


TFloat branchMaxCut::MinWeight(TNode v) throw()
{
    if (leftWeight[v]>rightWeight[v])
        return rightWeight[v];
    else return leftWeight[v];
}


TFloat branchMaxCut::MaxWeight(TNode v) throw()
{
    if (leftWeight[v]<rightWeight[v])
        return rightWeight[v];
    else return leftWeight[v];
}


TNode branchMaxCut::SelectVariable() throw()
{
    TNode maxNode = NoNode;
    TFloat maxWeight = -InfFloat;

    for (TNode v=0;v<n;v++)
    {
        if (colour[v]!=1) continue;
        
        TFloat thisWeight = SumWeight(v);
        
        if (thisWeight>maxWeight)
        {
            maxNode = v;
            maxWeight = thisWeight;
        }
    }
    
    #if defined(_FAILSAVE_)

    if (maxNode==NoNode)
    {
        sprintf(CT.logBuffer,"All variables are fixed: %ld",unfixed);
        InternalError1("Raise");
    }

    #endif

    return maxNode;
}


char branchMaxCut::Direction1(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Direction1",v);

    #endif
    
    if (rightWeight[v]<=leftWeight[v]) return 0;
    else return 1;
}


char branchMaxCut::Direction2(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Direction2",v);

    #endif
    
    return Direction1(v);
}


goblinBBNode<TNode,TFloat> *branchMaxCut::Clone() throw()
{
    return new branchMaxCut(*this);
}


void branchMaxCut::Raise(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Raise",v);

    #endif
    
    // Mark v as a right-hand side node
    
    colour[v] = 2;
    unfixed--;
    solved = false;
    
    // Update rightWeight with the incidences of v
    
    TArc a = G.First(v);
    do
    {
        TNode w = G.EndNode(a);
        
        if (w!=v)
        {
            switch (colour[w])
            {
                case 0:
                {
                    if (!G.Blocking(a^1))
                        selectedWeight += G.UCap(a)*G.Length(a&(a^1));

                    break;
                }
                case 1:
                {
                    if (G.Blocking(a^1))
                        dismissedWeight += G.UCap(a)*G.Length(a&(a^1));
                    else rightWeight[w] += G.UCap(a)*G.Length(a&(a^1));

                    break;
                }
                case 2:
                {
                    if (!G.Blocking(a))
                        dismissedWeight += G.UCap(a)*G.Length(a&(a^1));

                    break;
                }
            }
        }

        a = G.Right(a,v);
    }
    while (a!=G.First(v));
    
    if (unfixed==1)
    {
        TNode uncoloured = NoNode;
        bool leftNodes = false;
        bool rightNodes = false;
        
        for (TNode v=0;v<n;v++)
        {
            if (colour[v]==0) leftNodes = true;
            if (colour[v]==1) uncoloured = v;
            if (colour[v]==2) rightNodes = true;
        }
        
        if (!leftNodes) Lower(uncoloured);
        if (!rightNodes) Raise(uncoloured);
    }
}


void branchMaxCut::Lower(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Lower",v);

    #endif
    
    // Mark v as a left-hand side node
    
    colour[v] = 0;
    unfixed--;
    solved = false;
    
    // Update leftWeight with the incidences of v
    
    TArc a = G.First(v);
    do
    {
        TNode w = G.EndNode(a);
        
        if (w!=v)
        {
            switch (colour[w])
            {
                case 0:
                {
                    if (!G.Blocking(a^1))
                        dismissedWeight += G.UCap(a)*G.Length(a&(a^1));

                    break;
                }
                case 1:
                {
                    if (G.Blocking(a))
                        dismissedWeight += G.UCap(a)*G.Length(a&(a^1));
                    else leftWeight[w] += G.UCap(a)*G.Length(a&(a^1));

                    break;
                }
                case 2:
                {
                    if (!G.Blocking(a))
                        selectedWeight += G.UCap(a)*G.Length(a&(a^1));

                    break;
                }
            }
        }

        a = G.Right(a,v);
    }
    while (a!=G.First(v));
    
    if (unfixed==1)
    {
        TNode uncoloured = NoNode;
        bool leftNodes = false;
        bool rightNodes = false;
        
        for (TNode v=0;v<n;v++)
        {
            if (colour[v]==0) leftNodes = true;
            if (colour[v]==1) uncoloured = v;
            if (colour[v]==2) rightNodes = true;
        }
        
        if (!leftNodes) Lower(uncoloured);
        if (!rightNodes) Raise(uncoloured);
    }
}


TFloat branchMaxCut::SolveRelaxation() throw()
{
    // All arc weights are non-negative

    if (totalWeight!=InfFloat) return dismissedWeight-totalWeight;


    TFloat thisWeight = selectedWeight;

    for (TArc a=0;a<G.M();a++)
    {
        TNode v = G.StartNode(a);
        TNode w = G.EndNode(a);

        if (colour[v]!=1 && colour[w]!=1) continue;

        if (colour[v]==2 && G.Blocking(2*a+1)) continue;

        if (colour[w]==0 && G.Blocking(2*a+1)) continue;
        
        if (G.Length(2*a)>0) thisWeight += G.UCap(2*a)*G.Length(2*a);
    }

    return -thisWeight;
}


void branchMaxCut::SaveSolution() throw()
{
    G.InitColours();

    for (TNode v=0;v<n;v++)
    {
        if (colour[v]==0) G.SetColour(v,0);
        if (colour[v]==1) G.SetColour(v,NoNode);
        if (colour[v]==2) G.SetColour(v,1);
    }
}


TFloat branchMaxCut::LocalSearch() throw()
{
    G.InitColours();

    for (TNode v=0;v<n;v++) G.SetColour(v,colour[v]/2);

    CT.SuppressLogging();
    objective = -G.MaxCutLocalSearch(source,target);
    CT.RestoreLogging();

    if (CT.traceLevel==3) G.Display();

    return objective;
}


TFloat abstractMixedGraph::MaxCutBranchAndBound(TNode s,TNode t,TFloat lowerBound)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (s>=n && s!=NoNode) NoSuchNode("MaxCut",s);
    
    if (t>=n && t!=NoNode) NoSuchNode("MaxCut",t);

    #endif
    
    LogEntry(LOG_METH,"Max-Cut branch and bound...");
    OpenFold(ModMaxCut);
        
    branchMaxCut* rootNode = new branchMaxCut(*this,s,t);

    goblinBBScheme<TNode,TFloat> scheme(rootNode,-lowerBound);
    
    TFloat ret = -scheme.bestBound;

    SetBounds(TimerMaxCut,ret,-scheme.bestLower);
    CloseFold(ModMaxCut);

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Maximum cut has weight %g",
            TFloat(-scheme.bestBound));
        LogEntry(LOG_RES,CT.logBuffer);
    }
    
    return ret;
}


TFloat abstractMixedGraph::MaxCut(TNode s,TNode t) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (s>=n && s!=NoNode) NoSuchNode("MaxCut",s);
    if (t>=n && t!=NoNode) NoSuchNode("MaxCut",t);

    #endif
    
    LogEntry(LOG_METH,"Computing maximum edge cut...");
    OpenFold(ModMaxCut);

    TFloat totalWeight = 0;
    bool negWeights = false;

    for (TArc a=0;a<m;a++)
        if (UCap(2*a)>0 && Length(2*a)>0)
            totalWeight += UCap(2*a)*Length(2*a);
        else if (Length(2*a)<0) negWeights = true;

    TNode* savedCut = NULL;
    TFloat savedWeight = 0;
    if (colour!=NULL)
    {
        if (s!=NoNode && t!=NoNode && colour[s]==colour[t])
            savedWeight = -InfFloat;

        for (TArc a=0;a<2*m && savedWeight!=-InfFloat;a++)
        {
            TNode u = StartNode(a);
            TNode v = EndNode(a);
            
            if (colour[u]==0 && colour[v]==1 && !Blocking(a))
            {
                savedWeight += UCap(a)*Length(a&(a^1));
            }
            else if (colour[u]>1 || colour[v]>1) savedWeight = -InfFloat;
        }
        
        if (savedWeight != -InfFloat)
        {
            if ((s!=NoNode && colour[s]==1) || (t!=NoNode && colour[t]==0))
            {
                for (TNode v=0;v<n;v++) colour[v] = 1-colour[v];
            }
        
            savedCut = colour;
            colour = NULL;

            if (CT.logMeth)
            {
                sprintf(CT.logBuffer,"...Initial cut has weight %g",savedWeight);
                LogEntry(LOG_RES,CT.logBuffer);
            }

            SetBounds(TimerMaxCut,savedWeight,totalWeight);
        }
    }
    else SetUpperBound(TimerMaxCut,totalWeight);


    TFloat ret = MaxCutHeuristics(s,t);

    if (savedCut!=NULL)
    {
        if (savedWeight>ret)
        {
            if (colour!=NULL) ReleaseColours();
            colour = savedCut;
            ret = savedWeight;
        }
        else delete[] savedCut;
    }

    if (CT.methSolve>=2 /* && !negWeights */)
    {
        ret = MaxCutBranchAndBound(s,t,ret);
    }

    CloseFold(ModMaxCut);

    return ret;
}


TFloat abstractMixedGraph::MaxCutHeuristics(TNode s,TNode t)
    throw(ERRange,ERRejected)
{
    LogEntry(LOG_METH,"Computing heuristic maximum cut...");

    return MaxCutHeuristicsGRASP(s,t);
}


TFloat abstractMixedGraph::MaxCutHeuristicsGRASP(TNode s,TNode t)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (s>=n && s!=NoNode) NoSuchNode("MaxCutHeuristicsGRASP",s);

    if (t>=n && t!=NoNode) NoSuchNode("MaxCutHeuristicsGRASP",t);

    #endif
    
    OpenFold(ModMaxCutGRASP,SHOW_TITLE);

    goblinQueue<TNode,TFloat>* Q = nHeap;

    if (Q!=NULL) Q -> Init();
    else Q = NewNodeHeap();
    
    branchMaxCut thisNode(*this,s,t);
    
    thisNode.SaveSolution();

    #if defined(_LOGGING_)

    THandle LH = LogStart(LOG_METH2,"Expanded nodes: ");

    #endif
    
    while (!thisNode.Feasible())
    {
        Q -> Init();
    
        for (TNode v=0;v<n;v++)
            if (colour[v]==NoNode) Q->Insert(v,thisNode.MinWeight(v));
            
        TNode u = Q->Delete();

        TNode draw = CT.Rand(2);
        if (draw >= Q->Cardinality()) draw = Q->Cardinality();
        
        for (;draw>0;draw--) u = Q->Delete();
        
        char dir = thisNode.Direction1(u);

        if (dir)
        {
            thisNode.Lower(u);
            colour[u] = 0;
        }
        else
        {
            thisNode.Raise(u);
            colour[u] = 1;
        }

        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            if (dir) sprintf(CT.logBuffer,"%ld[S] ",u);
            else sprintf(CT.logBuffer,"%ld[T] ",u);
            
            LogAppend(LH,CT.logBuffer);
        }

        #endif
    
        CT.Ping(Handle(),n);
    }

    thisNode.SaveSolution();

    #if defined(_LOGGING_)

    if (CT.logMeth>1) LogEnd(LH);

    #endif

    if (nHeap==NULL) delete Q;


    TFloat weight = 0;

    for (TArc a=0;a<2*m;a++)
    {
        TNode u = StartNode(a);
        TNode w = EndNode(a);

        if (colour[u]==0 && colour[w]==1 && !Blocking(a))
            weight += UCap(a)*Length(a&(a^1));
    }

    SetLowerBound(TimerMaxCut,weight);
    CT.Ping(Handle(),m);

    CloseFold(ModMaxCutGRASP);

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Cut has weight %g",weight);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    if (CT.methLocal) weight = MaxCutLocalSearch(s,t);
    
    return weight;
}


TFloat abstractMixedGraph::MaxCutLocalSearch(TNode s,TNode t)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (s>=n && s!=NoNode) NoSuchNode("MaxCutLocalSearch",s);

    if (t>=n && t!=NoNode) NoSuchNode("MaxCutLocalSearch",t);

    if (colour==NULL)
        Error(ERR_REJECTED,"MaxCutLocalSearch","Missing cut");

    #endif
    
    LogEntry(LOG_METH,"Searching for local optimality...");

    OpenFold(ModMaxCut);
    
    TNode card = 0;
    for (TNode v=0;v<n;v++)
        if (colour[v]==1) card++;

    TFloat weight = 0;
    for (TArc a=0;a<2*m;a++)
    {
        TNode u = StartNode(a);
        TNode w = EndNode(a);

        if (colour[u]==0 && colour[w]==1 && !Blocking(a))
            weight += UCap(a)*Length(a&(a^1));
    }

    TFloat* profit = new TFloat[n];

    bool searching = true;
    while (searching)
    {
        CT.Ping(Handle(),m);
        
        for (TNode v=0;v<n;v++) profit[v] = 0;

        for (TArc a=0;a<m;a++)
        {
            TNode u = StartNode(2*a);
            TNode w = EndNode(2*a);
            
            if (u==w) continue;

            if (colour[u]==1 && colour[w]==1)
            {
                profit[u] += UCap(2*a)*Length(2*a);

                if (!Blocking(2*a+1))
                    profit[w] += UCap(2*a)*Length(2*a); 
            }

            if (colour[u]==0 && colour[w]==0)
            {
                profit[w] += UCap(2*a)*Length(2*a);

                if (!Blocking(2*a+1))
                    profit[u] += UCap(2*a)*Length(2*a); 
            }

            if (colour[u]==1 && colour[w]==0 && !Blocking(2*a+1))
            {
                profit[w] -= UCap(2*a)*Length(2*a);
                profit[u] -= UCap(2*a)*Length(2*a);
            }

            if (colour[u]==0 && colour[w]==1)
            {
                profit[w] -= UCap(2*a)*Length(2*a);
                profit[u] -= UCap(2*a)*Length(2*a);
            }
        }
        
        TNode maxNode = NoNode;
        for (TNode v=0;v<n;v++)
            if (v!=s && v!=t && (maxNode==NoNode || profit[v]>profit[maxNode]) &&
                (card>1 || colour[v]==0) && (card<n-1 || colour[v]==1)) 
                maxNode = v;
        
        if (profit[maxNode]>0)
        {
            colour[maxNode] = 1-colour[maxNode];
            weight += profit[maxNode];

            if (colour[maxNode]==1) card++; else card--;
                    
            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,
                    "Node %ld moves to component %c",
                    maxNode,char('S'+colour[maxNode]));
                LogEntry(LOG_METH2,CT.logBuffer);
            }

            #endif
            
            SetLowerBound(TimerMaxCut,weight);
        }
        else searching = false;
    }
    
    delete[] profit;

    CloseFold(ModMaxCut);

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Cut has weight: %g",weight);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    return weight;
}


TFloat abstractGraph::MaxCutHeuristics(TNode s,TNode t)
    throw(ERRange,ERRejected)
{
    OpenFold(ModMaxCut,NO_INDENT);

    TFloat ret = InfFloat;

    switch (CT.methMaxCut)
    {
        case 2:
        {
            if (true || s==NoNode || t==NoNode)
            {
                LogEntry(LOG_METH,"Applying planar max-cut method...");
                
                bool solved = true;
                try
                {
                    ret = MaxCutDualTJoin(s);
                }
                catch (ERRejected)
                {
                    solved = false;
                }
                
                if (solved && (t==NoNode || colour[t]==1))
                {
                    SetUpperBound(TimerSPTree,ret);
                    break;
                }
                
                if (solved && t!=NoNode && colour[t]==0)
                    LogEntry(LOG_RES,"...Target node has not be separated");
            }
        }
        case 0:
        case 1:
        {
            LogEntry(LOG_METH,"Computing heuristic maximum cut...");
            
            if (CT.methMaxCut==1)
            {
                try
                {
                    ret = MaxCutHeuristicsTree(s,t);
                }
                catch (ERRejected)
                {
                    ret = MaxCutHeuristicsGRASP(s,t);
                }
            }
            else ret = MaxCutHeuristicsGRASP(s,t);

            break;
        }
        default:
        {
            CloseFold(ModMaxCut,NO_INDENT);
            UnknownOption("MaxCut",CT.methMaxCut);
        }
    }

    CloseFold(ModMaxCut,NO_INDENT);

    return ret;
}


TFloat abstractGraph::MaxCutDualTJoin(TNode s) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (s>=n && s!=NoNode) NoSuchNode("MaxCutDualTJoin",s);

    #endif
    
    OpenFold(ModMaxCutTJoin,SHOW_TITLE);

    dualGraph* D = NULL;
    
    try
    {
        D = new dualGraph(*this);
    }
    catch (ERRejected)
    {
        CloseFold(ModMaxCutTJoin);
        Error(ERR_REJECTED,"MaxCutDualTJoin","No planar embedding found");
    }
    
    for (TNode v=0;v<D->N();v++) D->SetDemand(v,0);
    
    bool negWeights = false;

    for (TArc a=0;a<m;a++)
    {
        TFloat thisWeight = Length(2*a)*UCap(2*a);
        D -> SetLength(2*a,-thisWeight);

        if (thisWeight<0) negWeights = true;
    }
    
    if (negWeights)
    {
        CloseFold(ModMaxCutTJoin);
        Error(ERR_REJECTED,"MaxCutDualTJoin",
            "Negative edge weights are not supported");
    }


    D -> MinCTJoin();

    LogEntry(LOG_METH,"Extracting cut from T-join...");
    OpenFold();
    
    InitColours(NoNode);

    staticQueue<TNode,TFloat> Q(n,CT);
    
    if (s!=NoNode)
    {
        Q.Insert(s);
        colour[s] = 0;
    }
    else
    {
        Q.Insert(0);
        colour[0] = 0;
    }

    TNode nodeCount = 1;
    TFloat weight = 0;

    #if defined(_LOGGING_)

    THandle LH = LogStart(LOG_METH2,"Expanded nodes: ");

    #endif
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    while (!(Q.Empty()))
    {
        CT.Ping(Q.Handle(),1); 
        TNode u = Q.Delete();

        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"%ld ",u);
            LogAppend(LH,CT.logBuffer);
        }

        #endif
        
        while (I.Active(u))
        {
            TArc a = I.Read(u);
            TNode v = EndNode(a);

            TNode reqColour = colour[u];

            if (D -> Sub(a)>0)
            {
                reqColour = 1-colour[u];
                
                if (u<v) weight += Length(a)*UCap(a);
            }

            if (colour[v]==NoNode)
            {               
                colour[v] = reqColour;
                CT.Ping(Q.Handle(),1); 
                Q.Insert(v);
                nodeCount++;
            }
            else if (colour[v]!=reqColour)
            {
                InternalError("MaxCutDualTJoin","Dual Subgraph is not a T-Join");
            }
        }
    }

    #if defined(_LOGGING_)

    if (CT.logMeth>1) LogEnd(LH);

    #endif

    Close(H);

    CloseFold();
    
    delete D;

    #if defined(_FAILSAVE_)

    if (nodeCount<n)
    {
        CloseFold(ModMaxCutTJoin);
        Error(ERR_REJECTED,"MaxCutDualTJoin","Graph is disconnected");
    }

    #endif
    
    SetLowerBound(TimerSPTree,weight);
    CT.Ping(Handle(),m); 

    CloseFold(ModMaxCutTJoin);

    sprintf(CT.logBuffer,"...Cut has weight %g",weight);
    LogEntry(LOG_RES,CT.logBuffer);

    return weight;
}


TFloat abstractGraph::MaxCutHeuristicsTree(TNode s,TNode t)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (s>=n && s!=NoNode) NoSuchNode("MaxCutHeuristicsTree",s);
    if (t>=n && t!=NoNode) NoSuchNode("MaxCutHeuristicsTree",t);

    #endif
    
    OpenFold(ModMaxCut);

    LogEntry(LOG_METH,"(Tree Heuristics)");

    InitColours();

    TNode r = s;
    if (r==NoNode) r = TNode(CT.Rand(n));

    for (TNode v=0;v<n;v++) colour[v] = NoNode;

    colour[r] = 0;

    if (t!=NoNode) colour[t] = 1;

    graph G(*this,OPT_CLONE);

    for (TArc a=0;a<m;a++)
        G.SetLength(2*a,-UCap(2*a)*Length(2*a));

    G.MinTree(r);
    
    staticQueue<TNode,TFloat> S(n,CT);

    for (TNode v=0;v<n;v++)
        if (colour[v]==NoNode) S.Insert(v);
    
    while (!S.Empty())
    {
        TNode v = S.Delete();
        TArc a = G.Pred(v);
        
        if (a!=NoArc)
        {
            TNode u = G.StartNode(a);

            if (colour[u]==NoNode) S.Insert(v);
            else colour[v] = 1-colour[u];
        }
        else
        {
            CloseFold(ModMaxCut);
            Error(ERR_REJECTED,"MaxCutHeuristicsTree","Graph is disconnected");
        }
    }

    TFloat weight = 0;
    for (TArc a=0;a<m;a++)
    {
        TNode u = StartNode(2*a);
        TNode w = EndNode(2*a);
        if (colour[u]!=colour[w]) weight += UCap(2*a)*Length(2*a);
    }

    SetLowerBound(TimerMaxCut,weight);
    CT.Ping(Handle(),m);

    CloseFold(ModMaxCut);

    sprintf(CT.logBuffer,"...Cut has weight %g",weight);
    LogEntry(LOG_RES,CT.logBuffer);

    if (CT.methLocal) weight = MaxCutLocalSearch(s,t);

    return weight;
}
