
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, July 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "absdig.h"
#include "statq.h"
#include "auxnet.h"


TFloat abstractDiGraph::MaxFlow(TNode s,TNode t) throw(ERRange,ERRejected)
{
    OpenFold(ModMaxFlow,NO_INDENT);

    if (CT.logMeth)
    {
        sprintf(CT.logBuffer,"Computing maximum (%ld,%ld)-flow...",s,t);
        LogEntry(LOG_METH,CT.logBuffer);
    }

    TFloat ret = InfCap;
    
    switch (CT.methMaxFlow)
    {
        case 0:
        {
            ret = EdmondsKarp(s,t);
            break;
        }
        case 1:
        {
            ret = Dinic(s,t);
            break;
        }
        case 2:
        case 3:
        {
            ret = GoldbergTarjan(s,t);
            BFS(s,t);
            break;
        }
        case 4:
        {
            ret = CapacityScaling(s,t);
            break;
        }
        default:
        {
            CloseFold(ModMaxFlow,NO_INDENT);
            UnknownOption("MaxFlow",CT.methMaxFlow);
        }
    }
    
    #if defined(_FAILSAVE_)

    try
    {
        if (ret!=InfCap && CT.methFailSave==1 && CutCapacity()!=FlowValue(s,t))
        {
            InternalError("MaxFlow","Solutions are corrupted");
        }
    }
    catch (...)
    {
        InternalError("MaxFlow","Solutions are corrupted");
    }

    #endif
    
    CloseFold(ModMaxFlow,NO_INDENT);
    
    return ret;
}


TFloat abstractDiGraph::EdmondsKarp(TNode s,TNode t) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("EdmondsKarp",s);

    if (t>=n) NoSuchNode("EdmondsKarp",t);

    #endif
    
    OpenFold(ModEdmondsKarp,SHOW_TITLE);

    TFloat val = 0;
    TArc a = First(s);
    do
    {
        if (!Blocking(a)) val += Flow(a);
        a = Right(a,s);
    }
    while (a!=First(s));

    SetLowerBound(TimerMaxFlow,val);
    
    while (BFS(s,t))
    {
        TCap Lambda = FindCap(s,t);
        if (Lambda==InfCap)
        {
            CloseFold(ModMaxFlow);
            LogEntry(LOG_RES,"...Problem is unbounded");
            CT.Ping(Handle(),m);
            return InfCap;
        }

        Augment(s,t,Lambda);
        val += Lambda;
        SetLowerBound(TimerMaxFlow,val);
        CT.Ping(Handle(),m);
    }
    
    ReleasePredecessors();

    SetUpperBound(TimerMaxFlow,val);
    CloseFold(ModEdmondsKarp);
    
    return val;
}


TFloat abstractDiGraph::CapacityScaling(TNode s,TNode t) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("CapacityScaling",s);

    if (t>=n) NoSuchNode("CapacityScaling",t);

    #endif
    
    OpenFold(ModMaxFlow);

    LogEntry(LOG_METH,"(Capacity scaling method)");
    
    TFloat val = 0;
    TArc a = First(s);
    do
    {
        if (!Blocking(a)) val += Flow(a);
        a = Right(a,s);
    }
    while (a!=First(s));

    SetLowerBound(TimerMaxFlow,val);

    delta = 0;

    for (TArc a=0;a<m;a++)
    {
        TCap thisCap = UCap(2*a);
        if (thisCap!=InfCap && thisCap>delta) delta = thisCap;
    }
    if (delta>0) delta -= 1;
    
    if (CT.logMeth)
    {
        sprintf(CT.logBuffer,"Starting with delta = %g",delta);
        LogEntry(LOG_METH,CT.logBuffer);
    }
    
    while (true)
    {
        if (CT.logMeth)
        {
            sprintf(CT.logBuffer,"Next scaling phase, delta = %.0f",delta);
            LogEntry(LOG_METH,CT.logBuffer);
        }

        while (BFS(s,t))
        {
            TCap Lambda = FindCap(s,t);
            if (Lambda==InfCap)
            {
                CloseFold(ModMaxFlow);
                LogEntry(LOG_RES,"...Problem is unbounded");
                CT.Ping(Handle(),m);
                return InfCap;
            }

            Augment(s,t,Lambda);
            val += Lambda;
            SetLowerBound(TimerMaxFlow,val);
            CT.Ping(Handle(),m);
        }

        if (delta==0) break;
        
        delta = floor(delta/2);
    }
        
    ReleasePredecessors();
    
    SetUpperBound(TimerMaxFlow,val);
    CloseFold(ModMaxFlow);
    
    return val;
}


TFloat abstractDiGraph::Dinic(TNode s,TNode t) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("Dinic",s);

    if (t>=n) NoSuchNode("Dinic",t);

    #endif
    
    OpenFold(ModDinic,SHOW_TITLE);

    InitPredecessors();
    InitLabels();
    layeredAuxNetwork Aux(*this,P,s);
    
    TFloat val = 0;
    TArc a = First(s);
    do
    {
        if (!Blocking(a)) val += Flow(a);
        a = Right(a,s);
    }
    while (a!=First(s));

    SetLowerBound(TimerMaxFlow,val);

    while (true)
    {
        InitLabels();
        d[s] = 0;
        staticQueue<TNode,TFloat> Q(n,CT);
        Q.Insert(s);

        LogEntry(LOG_METH,"Graph is searched...");

        #if defined(_LOGGING_)

        OpenFold();
        THandle LH = LogStart(LOG_METH2,"Found props: ");

        #endif

        THandle H = Investigate();
        goblinIterator &I = Iterator(H);

        while (!(Q.Empty()))
        {
            TNode u = Q.Delete();
            if (d[u]==d[t]) break;

            while (I.Active(u))
            {
                TArc a = I.Read(u);
                TNode v = EndNode(a);
                if (d[v]>=d[u]+1 && ResCap(a)>0)
                {               
                    if (d[v]==InfFloat)
                    {
                        d[v] = d[u]+1;
                        Q.Insert(v);
                    }
                    Aux.InsertProp(a);

                    #if defined(_LOGGING_)

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,"%ld(%ld)%ld ",u,a,v);
                        LogAppend(LH,CT.logBuffer);
                    }

                    #endif
                }
            }
        }

        Close(H);

        #if defined(_LOGGING_)

        if (CT.logMeth>1) LogEnd(LH);
        CloseFold();

        #endif
        
        if (d[t]==InfFloat) break;

        CT.Ping(Handle(),m);
        if (CT.logMeth)
        {
            sprintf(CT.logBuffer,"Phase %g augmenting...",d[t]);
            LogEntry(LOG_METH,CT.logBuffer);
        }
        OpenFold();

        Aux.Phase2();
        while (!(Aux.Blocked(t)))
        {
            CT.Ping(Aux.Handle(),n);

            TFloat Lambda = Aux.FindPath(t);
            if (Lambda==InfCap)
            {
                LogEntry(LOG_RES,"...Problem is unbounded");
                CT.Ping(Handle(),n*m);
                CloseFold(ModDinic);
                return InfCap;
            }

            val += Lambda;

            TArc a;
            TNode w = t;

            OpenFold();

            while (w!=s)
            {
                a = P[w];
                Push(a,Lambda);
                P[w] = NoArc;
                w = StartNode(a);

                if (ResCap(a)==0) Aux.TopErasure(a);
            }

            CloseFold();
        }

        CloseFold();
        SetLowerBound(TimerMaxFlow,val);

        Aux.Phase1();
        Aux.Init();
    }

    CT.Ping(Handle(),m);
    
    ReleasePredecessors();
    
    SetUpperBound(TimerMaxFlow,val);
    CloseFold(ModDinic);

    return val;
}


TFloat abstractDiGraph::GoldbergTarjan(TNode s,TNode t) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("GoldbergTarjan",s);

    if (t>=n) NoSuchNode("GoldbergTarjan",t);

    #endif
    
    OpenFold(ModPushRelabel,SHOW_TITLE);

    // A queue for the active nodes
    
    goblinQueue<TNode,TFloat> *Q = NULL;
    if (CT.methMaxFlow!=2) Q = NewNodeHeap();
    else Q = new staticQueue<TNode,TFloat>(n,CT);
    
    // Additional incidence structure which is updated by relabel(u)
    // operations. Admissible[u] contains of all arcs a=uv such that
    // rescap(a)>0 and d[u]=d[v]+1.
    
    staticQueue<TNode,TFloat> **Admissible = new staticQueue<TNode,TFloat>*[n];
    Admissible[0] = new staticQueue<TArc,TFloat>(2*m,CT);
    for (TNode i=1;i<n;i++)
        Admissible[i] = new staticQueue<TArc,TFloat>(*Admissible[0]);
    
    // Temporary incidence lists used in relabel operations
    
    staticQueue<TArc,TFloat> S1(2*m,CT);
    staticQueue<TArc,TFloat> S2(S1);
    
    InitLabels();
    for (TNode i=0;i<n;i++) d[i]=0;
    d[s] = n;
    
    // Push all flow away from the root node
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    bool unbounded = false;
    while (I.Active(s))
    {
        TArc a = I.Read(s);
        TNode v = EndNode(a);
        if (Divergence(v)==0 && v!=t && v!=s && ResCap(a)>0)
            Q -> Insert(v,-d[v]);
        if (UCap(a)!=InfCap) Push(a,ResCap(a));
        else unbounded = true;
    }
    
    SetBounds(TimerMaxFlow,Divergence(t),-Divergence(s));
    CT.Ping(Handle(),1);

    while (!(Q->Empty()) && !unbounded)
    {
        TNode u = Q->Delete();
        
        while (!Admissible[u]->Empty() && Divergence(u)>0)
        {
            TArc a = Admissible[u]->Delete();
            TNode v = EndNode(a);

            TFloat lambda = ResCap(a);
            if (lambda>Divergence(u)) lambda = Divergence(u);
            
            if (d[u]==d[v]+1)
            {
                if (Divergence(v)==0 && v!=t && v!=s) Q->Insert(v,-d[v]);

                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,
                        "Pushed %g flow units from node %ld to node %ld",lambda,u,v);
                    LogEntry(LOG_METH2,CT.logBuffer);
                }

                #endif
    
                Push(a,lambda);
                
                if (v==s) SetUpperBound(TimerMaxFlow,-Divergence(s));
                if (v==t) SetLowerBound(TimerMaxFlow,Divergence(t));
                CT.Ping(Handle(),1);

                if (ResCap(a)>0) Admissible[u]->Insert(a);
            }
        }
        
        if (Divergence(u)>0)
        {
            I.Reset(u);
            TFloat dMin = NoNode;
            while (I.Active(u))
            {
                TArc a = I.Read(u);
                TNode v = EndNode(a);
                if (ResCap(a)>0 && d[v]<=dMin)
                {
                    dMin = d[v];
                    S1.Insert(a);
                }
                
                if (ResCap(a^1)>0 && dMin+2>=d[v]) S2.Insert(a^1);
            }
            
            while (!S1.Empty())
            {
                TArc a = S1.Delete();
                TNode v = EndNode(a);
                
                if (dMin==d[v]) Admissible[u]->Insert(a);
            }
            
            while (!S2.Empty())
            {
                TArc a = S2.Delete();
                TNode v = StartNode(a);
                
                if (dMin+2==d[v]) Admissible[v]->Insert(a);
            }
            
            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"Relabelled node %ld: %g -> %g",u,d[u],dMin+1);
                LogEntry(LOG_METH2,CT.logBuffer);
            }

            #endif
    
            d[u] = dMin+1;
            CT.Ping(Handle(),n);
            Q->Insert(u,-d[u]);
        }
    }

    Close(H);
    
    delete Q;
    
    for (TNode i=1;i<n;i++) delete Admissible[i];
    delete Admissible[0];
    delete[] Admissible;
    
    ReleasePredecessors();
    
    if (unbounded) Error(ERR_RANGE,"GoldbergTarjan",
        "Arcs emanating from source must have finite capacities");
    
    CloseFold(ModPushRelabel);

    return Divergence(t);
}
