
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, July 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsdig.h"
#include "statq.h"
#include "spsgra.h"


TCap abstractMixedGraph::Connectivity() throw()
{
    OpenFold(ModMinCut,NO_INDENT);

    LogEntry(LOG_METH,"Computing graph connectivity...");
    OpenFold();
        
    InitColours();
    nodeSplitting G(*this,1);
    
    TCap lambda = InfFloat;
    TCap k = 0;
    for (TNode v=0;v<n && CT.SolverRunning();v++)
    {
        for (TNode w=v+1;w<n;w++)
        {
            if (Adjacency(v,w)!=NoArc) continue;
            CT.SuppressLogging();
            G.InitSubgraph();
            G.MaxFlow(2*v+1,2*w);
            TCap val = G.FlowValue(2*v+1,2*w);
            CT.RestoreLogging();
            
            if (val>=lambda) continue;

            #if defined(_LOGGING_)

            if (CT.logRes>=2)
            {
                sprintf(CT.logBuffer,
                    "Minimum (%ld,%ld)-cut has capacity %g",v,w,val);
                LogEntry(LOG_RES2,CT.logBuffer);
            }

            #endif 

            SetUpperBound(TimerMinCut,val);

            lambda = val;

            for (TNode u=0;u<n;u++)
            {
                if (G.Dist(2*u)==InfFloat && G.Dist(2*u+1)==InfFloat) colour[u] = 2;
                if (G.Dist(2*u)<InfFloat && G.Dist(2*u+1)<InfFloat) colour[u] = 1;
                if (G.Dist(2*u)<InfFloat && G.Dist(2*u+1)==InfFloat) colour[u] = 0;
            }
            colour[v] = 1;

            CT.Ping(Handle(),n*n*m);
        }
        k++;
    }
    
    if (CT.SolverRunning()) SetLowerBound(TimerMinCut,lambda);
    
    CloseFold();

    sprintf(CT.logBuffer,"...Graph connectivity is %g",lambda);
    LogEntry(LOG_RES,CT.logBuffer);
        
    for (TArc a=0;a<2*m;a++)
    {
        TNode x = StartNode(a);
        TNode y = EndNode(a);
        if (colour[x]==1 && colour[y]==2 && Demand(y)==1 && !Blocking(a))
            colour[y] = 0;
    }
        
    CloseFold(ModMinCut,NO_INDENT);

    return lambda;
}


TCap abstractMixedGraph::EdgeConnectivity() throw()
{
    OpenFold(ModMinCut,NO_INDENT);

    LogEntry(LOG_METH,"Computing edge connectivity...");
    
    TCap lambda = InfCap;
    if (CT.methMinCut>2) lambda = MinCutNodeIdentification();
    else
    {
        InitColours();
        diGraph G(n,TOption(0),CT);

        for (TArc a=0;a<2*m;a++)
        {
            TNode x = StartNode(a);
            TNode y = EndNode(a);
            G.InsertArc(x,y,UCap(a),0,0);
        }

        #if defined(_TRACING_)

        for (TNode u=0;u<n;u++)
        {
            if (Dim()>0)
            {
                G.SetC(u,0,C(u,0));
                G.SetC(u,1,C(u,1));
            }
        }

        #endif

        if (CT.methMinCut>0)
        {
            lambda = G.MinCutHaoOrlin(0);

            for (TNode u=0;u<n;u++) colour[u] = G.Colour(u);
        }
        else
        {
            OpenFold();

            for (TNode v=0;v<n && CT.SolverRunning();v++)
            {
                TNode w = (v+1)%n;
                CT.SuppressLogging();
                G.InitSubgraph();
                G.MaxFlow(v,w);
                TCap val = G.FlowValue(v,w);
                CT.RestoreLogging();

                if (val>=lambda) continue;

                #if defined(_LOGGING_)

                if (CT.logRes>=2)
                {
                    sprintf(CT.logBuffer,
                        "Minimum (%ld,%ld)-cut has capacity %g",v,w,val);
                    LogEntry(LOG_RES2,CT.logBuffer);
                }

                #endif

                SetUpperBound(TimerMinCut,val);

                lambda = val;

                for (TNode u=0;u<n;u++)
                    if (G.Dist(u)<InfFloat) colour[u] = 1;
                    else colour[u] = 0;

                CT.Ping(Handle(),n*n*m);
            }

            if (CT.SolverRunning()) SetLowerBound(TimerMinCut,lambda);
    
            CloseFold();

            sprintf(CT.logBuffer,"...Edge connectivity is %g",lambda);
            LogEntry(LOG_RES,CT.logBuffer);
        }
    }
        
    CloseFold(ModMinCut,NO_INDENT);

    return lambda;
}


TCap abstractMixedGraph::StrongConnectivity() throw()
{
    OpenFold(ModMinCut,NO_INDENT);

    LogEntry(LOG_METH,"Computing strong connectivity...");

    OpenFold();
        
    InitColours();
    nodeSplitting G(*this,0);
    
    TCap lambda = InfFloat;
    for (TNode v=0;v<n && CT.SolverRunning();v++)
    {
        for (TNode w=0;w<n && CT.SolverRunning();w++)
        {
            if (v==w) continue;
            CT.SuppressLogging();
            G.InitSubgraph();
            G.MaxFlow(2*v+1,2*w);
            TCap val = G.FlowValue(2*v+1,2*w);
            CT.RestoreLogging();
            if (val>=lambda) continue;

            CT.RestoreLogging();

            #if defined(_LOGGING_)

            if (CT.logRes>=2)
            {
                sprintf(CT.logBuffer,
                    "Minimum (%ld,%ld)-cut has capacity %g",v,w,val);
                LogEntry(LOG_RES2,CT.logBuffer);
            }

            #endif

            SetUpperBound(TimerMinCut,val);
            CT.SuppressLogging();

            lambda = val;

            for (TNode u=0;u<n;u++)
            {
                if (G.Dist(2*u)==InfFloat && G.Dist(2*u+1)==InfFloat) colour[u] = 2;
                if (G.Dist(2*u)<InfFloat && G.Dist(2*u+1)<InfFloat) colour[u] = 1;
                if (G.Dist(2*u)<InfFloat && G.Dist(2*u+1)==InfFloat) colour[u] = 0;
            }
            colour[v] = 1;

            CT.Ping(Handle(),n*n*m);
        }
    }
    
    if (CT.SolverRunning()) SetLowerBound(TimerMinCut,lambda);
    
    CloseFold();
    
    sprintf(CT.logBuffer,"...Strong connectivity is %g",lambda);
    LogEntry(LOG_RES,CT.logBuffer);
        
    TArc a = 0;
    for (;a<2*m;a++)
    {
        TNode x = StartNode(a);
        TNode y = EndNode(a);
        if (colour[x]==1 && colour[y]==2 && Demand(y)==1 && !Blocking(a))
            colour[y] = 0;
    }
        
    CloseFold(ModMinCut,NO_INDENT);
        
    return lambda;
}


TCap abstractMixedGraph::StrongEdgeConnectivity() throw()
{
    OpenFold(ModSchnorr,NO_INDENT);

    LogEntry(LOG_METH,"Computing strong edge connectivity...");

    OpenFold();
        
    InitColours();
    diGraph G(*this);
    
    #if defined(_LOGGING_)

    THandle LH = LogStart(LOG_METH2,"  Minimum of ");

    int swapMeth = CT.logMeth;

    #endif
    
    CT.SuppressLogging();

    TCap lambda = InfFloat;

    for (TNode v=0;v<n && CT.SolverRunning();v++)
    {
        TNode w = (v+1)%n;
        G.InitSubgraph();
        G.MaxFlow(v,w);
        TCap val = G.FlowValue(v,w);
        if (val<lambda)
        {
            SetUpperBound(TimerMinCut,val);
            lambda = val;
                
            TNode u = 0;
            for (;u<n;u++)
                if (G.Dist(u)<InfFloat) colour[u] = 1;
                else colour[u] = 0;
            CT.Ping(Handle(),n*n*m);
        }

        #if defined(_LOGGING_)

        if (swapMeth>1)
        {
            CT.RestoreLogging();
            sprintf(CT.logBuffer,"%g(%ld,%ld) ",val,v,w);
            LogAppend(LH,CT.logBuffer);
            CT.SuppressLogging();
        }

        #endif
    }

    #if defined(_LOGGING_)

    if (CT.logMeth>1) LogEnd(LH);

    #endif

    CT.RestoreLogging();
    
    if (CT.SolverRunning()) SetLowerBound(TimerMinCut,lambda);
    
    CloseFold();
    
    sprintf(CT.logBuffer,"...Strong edge connectivity is %g",lambda);
    LogEntry(LOG_RES,CT.logBuffer);
         
    CloseFold(ModSchnorr,NO_INDENT);
       
    return lambda;
}


TCap abstractMixedGraph::StrongEdgeConnectivity(TNode r) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (r>=n && r!=NoNode) NoSuchNode("StrongEdgeConnectivity",r);

    #endif
    
    OpenFold(ModMinCut,NO_INDENT);

    sprintf(CT.logBuffer,"Computing strong edge %ld-connectivity...",r);
    LogEntry(LOG_METH,CT.logBuffer);

    InitColours();
    diGraph G(*this);

    TCap lambda = InfFloat;
    if (CT.methMinCut>0)
    {
        lambda = G.MinCutHaoOrlin(r);

        for (TNode u=0;u<n;u++) colour[u] = G.Colour(u);
    }
    else
    {
        OpenFold();
        
        #if defined(_LOGGING_)

        THandle LH = LogStart(LOG_METH2,"Minimum of ");

        int swapMeth = CT.logMeth;

        #endif

        CT.SuppressLogging();

        for (TNode v=0;v<n && CT.SolverRunning();v++)
        {
            if (v==r) continue;

            G.InitSubgraph();
            G.MaxFlow(r,v);
            TCap val = G.FlowValue(r,v);
            if (val<lambda)
            {
                lambda = val;
                SetUpperBound(TimerMinCut,val);

                for (TNode u=0;u<n;u++)
                    if (G.Dist(u)<InfFloat) colour[u] = 1;
                    else colour[u] = 0;
                CT.Ping(Handle(),n*n*m);
            }

            #if defined(_LOGGING_)

            if (swapMeth==2)
            {
                CT.RestoreLogging();
                sprintf(CT.logBuffer,"%g(%ld) ",val,v);
                LogAppend(LH,CT.logBuffer);
                CT.SuppressLogging();
            }

            #endif
        }

        #if defined(_LOGGING_)

        if (CT.logMeth>1) LogEnd(LH);

        #endif

        CT.RestoreLogging();

        if (CT.SolverRunning()) SetLowerBound(TimerMinCut,lambda);
    
        CloseFold();

        sprintf(CT.logBuffer,"...Strong edge %ld-connectivity is %g",r,lambda);
        LogEntry(LOG_RES,CT.logBuffer);
    }
        
    CloseFold(ModMinCut,NO_INDENT);
        
    return lambda;
}


TCap abstractMixedGraph::StrongEdgeConnectivity(TNode s,TNode t) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (s>=n && s!=NoNode) NoSuchNode("StrongEdgeConnectivity",s);

    if (t>=n && t!=NoNode) NoSuchNode("StrongEdgeConnectivity",t);

    #endif
    
    OpenFold(ModMinCut,NO_INDENT);

    sprintf(CT.logBuffer,"Computing strong edge (%ld,%ld)-connectivity...",s,t);
    LogEntry(LOG_METH,CT.logBuffer);

    OpenFold();
        
    InitColours();
    diGraph G(*this);
    
    CT.SuppressLogging();
    G.InitSubgraph();
    G.MaxFlow(s,t);
    TCap lambda = G.FlowValue(s,t);
    CT.RestoreLogging();
    for (TNode u=0;u<n;u++)
        if (G.Dist(u)<InfFloat) colour[u] = 1;
        else colour[u] = 0;
    
    SetBounds(TimerMinCut,lambda,lambda);
    
    CloseFold();

    sprintf(CT.logBuffer,"...Strong edge (%ld,%ld)-connectivity is %g",s,t,lambda);
    LogEntry(LOG_RES,CT.logBuffer);
        
    CloseFold(ModMinCut,NO_INDENT);
        
    return lambda;
}

 
TCap abstractDiGraph::MinCutHaoOrlin(TNode r) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (r>=n) NoSuchNode("MinCutHaoOrlin",r);

    #endif
    
    OpenFold(ModHaoOrlin,SHOW_TITLE);
    
    // A queue for the active nodes
    
    goblinQueue<TNode,TFloat> *Q = NULL;
    if (CT.methMinCut!=1) Q = NewNodeHeap();
    else Q = new staticQueue<TNode,TFloat>(n,CT);
    
    // Additional incidence structure which is updated by relabel(u)
    // operations. Admissible[u] contains of all arcs a=uv such that
    // rescap(a)>0 and d[u]=d[v]+1.
    
    staticQueue<TNode,TFloat> **Admissible = new staticQueue<TNode,TFloat>*[n];
    Admissible[0] = new staticQueue<TArc,TFloat>(2*m,CT);
    for (TNode i=1;i<n;i++)
        Admissible[i] = new staticQueue<TArc,TFloat>(*Admissible[0]);
    
    // Temporary incidence lists used in relabel operations
    
    staticQueue<TArc,TFloat> S1(2*m,CT);
    staticQueue<TArc,TFloat> S2(S1);
    
    // The sink node, changed in every main loop iteration

    TNode t = 0;
    if (r==0) t = 1;

    InitLabels(1);
    d[t] = 0;
    
    // Store the minimum cut and its capacity

    InitColours(0);
    TNode *cut = colour;
    colour = NULL;
    TCap minCap = InfCap;
    
    // Keep the dormient nodes in a tracable data structure

    InitColours(NoNode);
    TNode *dorm = colour;
    dorm[r] = 0;
    TNode dormMax = 0;
    
    // Push all flow away from the root node
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    bool unbounded = false;
    while (I.Active(r))
    {
        TArc a = I.Read(r);
        TNode v = EndNode(a);
        if (Divergence(v)==0 && v!=r && v!=t && ResCap(a)>0)
            Q -> Insert(v,-d[v]);
        if (UCap(a)!=InfCap) Push(a,ResCap(a));
        else unbounded = true;
    }

    CT.Ping(Handle(),1);

    // Initialize the incidence lists

    I.Reset(t);
    while (I.Active(t))
    {
        TArc a = I.Read(t);
        TNode v = EndNode(a);
        if (ResCap(a^1)>0 && v!=r) Admissible[v]->Insert(a^1);
    }

    for (TNode itc=1;itc<n && !unbounded;itc++)
    {
        sprintf(CT.logBuffer,"New sink node is %ld",t);
        LogEntry(LOG_METH,CT.logBuffer);

        while (!(Q->Empty()) && !unbounded)
        {
            TNode u = Q->Delete();
            if (dorm[u]!=NoNode || u==t) continue;
            
            while (!Admissible[u]->Empty() && Divergence(u)>0)
            {
                TArc a = Admissible[u]->Delete();
                TNode v = EndNode(a);

                TFloat Lambda = ResCap(a);
                if (ResCap(a)>Divergence(u)) Lambda = Divergence(u);

                if (d[u]==d[v]+1 && dorm[v]==NoNode)
                {
                    if (Divergence(v)==0 && v!=t) Q->Insert(v,-d[v]);

                    #if defined(_LOGGING_)

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,
                            "Pushed %g flow units from node %ld to node %ld",Lambda,u,v);
                        LogEntry(LOG_METH2,CT.logBuffer);
                    }

                    #endif

                    Push(a,Lambda);
                    CT.Ping(Handle(),1);
                    if (ResCap(a)>0)
                        try {Admissible[u]->Insert(a);} catch(ERCheck) {};
                }
            }

            if (Divergence(u)>0)
            {
                // Check if there are other awake nodes with the same
                // distance label
                
                bool isUnique = true;
                for (TNode v=0;v<n && isUnique;v++)
                    if (v!=u && d[v]==d[u] && dorm[v]==NoNode)
                        isUnique = false;
                
                if (isUnique)
                {
                    dormMax++;
                    for (TNode v=0;v<n;v++)
                        if (d[v]>=d[u] && dorm[v]==NoNode)
                        {
                            dorm[v] = dormMax;

                            #if defined(_LOGGING_)

                            if (CT.logMeth>1)
                            {
                                sprintf(CT.logBuffer,"dorm[%ld] = %ld",v,dormMax);
                                LogEntry(LOG_METH2,CT.logBuffer);
                            }

                            #endif
                        }
                }
                else
                {
                    I.Reset(u);
                    TFloat dMin = NoNode;
                    while (I.Active(u))
                    {
                        TArc a = I.Read(u);
                        TNode v = EndNode(a);
                        
                        if (dorm[v]!=NoNode || v==t) continue;
                        
                        if (ResCap(a)>0 && d[v]<=dMin)
                        {
                            dMin = d[v];
                            try {S1.Insert(a);} catch(ERCheck) {};
                        }

                        if (ResCap(a^1)>0 && dMin+2>=d[v])
                        {
                            try {S2.Insert(a^1);} catch(ERCheck) {};
                        }
                    }
                    
                    if (dMin<NoNode)
                    {
                        while (!S1.Empty())
                        {
                            TArc a = S1.Delete();
                            TNode v = EndNode(a);

                            if (dMin==d[v])
                                try
                                {
                                    Admissible[u]->Insert(a);
                                }
                                catch(ERCheck) {};
                        }

                        while (!S2.Empty())
                        {
                            TArc a = S2.Delete();
                            TNode v = StartNode(a);

                            if (dMin+2==d[v])
                                try
                                {
                                    Admissible[v]->Insert(a);
                                }
                                catch(ERCheck) {};
                        }

                        #if defined(_LOGGING_)

                        if (CT.logMeth>1)
                        {
                            sprintf(CT.logBuffer,
                                "Relabelled node %ld: %g -> %g",u,d[u],dMin+1);
                            LogEntry(LOG_METH2,CT.logBuffer);
                        }

                        #endif

                        d[u] = dMin+1;
                        CT.Ping(Handle(),n);
                        Q->Insert(u,-d[u]);
                    }
                    else
                    {
                        dormMax++;
                        dorm[u] = dormMax;

                        #if defined(_LOGGING_)

                        if (CT.logMeth>1)
                        {
                            sprintf(CT.logBuffer,"dorm[%ld] = %ld",u,dormMax);
                            LogEntry(LOG_METH2,CT.logBuffer);
                        }

                        #endif
                    }
                }
            }
        }
        
        // Check if the new cut improves the bound
        
        TCap thisCap = 0;
        for (TArc a=0;a<m;a++)
        {
            if (dorm[StartNode(2*a)]!=NoNode && dorm[EndNode(2*a)]==NoNode)
                thisCap += UCap(2*a);
        }

        #if defined(_LOGGING_)

        if (CT.logRes>1)
        {
            sprintf(CT.logBuffer,"Found cut with capacity %g",thisCap);
            LogEntry(LOG_RES2,CT.logBuffer);
        }

        #endif
            
        if (thisCap<minCap)
        {
            minCap = thisCap;
            for (TNode i=0;i<n;i++)
                if (dorm[i]!=NoNode) cut[i] = 0;
                else cut[i] = 1;
            SetUpperBound(TimerMinCut,minCap);
        }
        
        // Push all flow towards the right hand side

        dorm[t] = 0;
        I.Reset(t);
        while (I.Active(t))
        {
            TArc a = I.Read(t);
            TNode v = EndNode(a);
            if (UCap(a)!=InfCap)
            {
                if (ResCap(a)>0 && dorm[v]!=0)
                {
                    if (Divergence(v)==0 && dorm[v]==NoNode)
                        Q -> Insert(v,-d[v]);
                    Push(a,ResCap(a));
                };
            }
            else unbounded = true;
        }

        CT.Ping(Handle(),1);

        // Select a new sink node

        t = NoNode;
        for (TNode i=1;i<n;i++)
            if (dorm[i]==NoNode && (t==NoNode || d[i]<d[t])) t = i;
        
        if (t==NoNode && itc<n-1)
        {
            #if defined(_LOGGING_)

            if (CT.logMeth>1) LogEntry(LOG_METH2,"Awaked nodes:");

            #endif
         
            for (TNode i=1;i<n;i++)
            {
                if (dorm[i]==dormMax)
                {
                    if (t==NoNode || d[i]<d[t]) t = i;
                    dorm[i] = NoNode;
                    if (Divergence(i)>0) Q -> Insert(i,-d[i]);

                    I.Reset(i);
                    while (I.Active(i))
                    {
                        TArc a = I.Read(i);
                        TNode v = EndNode(a);
                        if (dorm[v]!=NoNode) continue;
                        
                        if (ResCap(a)>0 && d[v]+1==d[i])
                            try {Admissible[i]->Insert(a);} catch(ERCheck) {};
                        
                        if (ResCap(a^1)>0 && d[v]==d[i]+1)
                            try {Admissible[v]->Insert(a^1);} catch(ERCheck) {};
                    }

                    #if defined(_LOGGING_)

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer," %ld",i);
                        LogEntry(MSG_APPEND,CT.logBuffer);
                    }

                    #endif
                }
            }
            
            dormMax--;
        }
    }

    Close(H);
    
    delete Q;
    
    for (TNode i=1;i<n;i++) delete Admissible[i];
    delete Admissible[0];
    delete[] Admissible;
    
    ReleasePredecessors();
    
    // Move the minimum cut into an exported data structure

    ReleaseColours();
    colour = cut;
    
    if (unbounded) Error(ERR_RANGE,"MinCutHaoOrlin",
        "Arcs emanating from source must have finite capacities");
    
    SetLowerBound(TimerMinCut,minCap);
        
    CloseFold(ModHaoOrlin);

    sprintf(CT.logBuffer,"...Strong edge %ld-connectivity is %g",r,minCap);
    LogEntry(LOG_RES,CT.logBuffer);
    
    return minCap;
}


TCap abstractMixedGraph::MinCutLegalOrdering(TNode r,TNode &x,TNode &y)
    throw(ERRange,ERRejected)
{                       
    #if defined(_FAILSAVE_)

    if (r>=n) NoSuchNode("MinCutLegalOrdering",r);

    #endif
    
    OpenFold(ModNodeIdentification,NO_INDENT);
    
    LogEntry(LOG_METH,"Computing legal ordering...");
    
    InitLabels(InfFloat);
    InitColours(NoNode);

    goblinQueue<TNode,TFloat> *Q = NULL;

    if (nHeap!=NULL)
    {
        Q = nHeap;
        Q -> Init();
    }
    else Q = NewNodeHeap();
    
    Q -> Insert(r,0);
    colour[r] = 0;
    
    OpenFold();
    
    #if defined(_LOGGING_)
    
    THandle LH = LogStart(LOG_METH2,"Expanded nodes: ");
    
    #endif
    
    TNode count = 0;
    TFloat ret = InfCap;
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    while (!(Q->Empty()))
    {
        TNode u = Q->Delete();
        
        x = y;
        y = u;
        ret = d[u];

        colour[u] = count;
        d[u] = -InfFloat;

        #if defined(_LOGGING_)

        if (CT.logMeth>1 && I.Active(u))
        {
            sprintf(CT.logBuffer,"%ld ",u);
            LogAppend(LH,CT.logBuffer);
        }

        #endif

        while (I.Active(u))
        {
            TArc a = I.Read(u);
            TFloat l = UCap(a);
            TNode v = EndNode(a);
            if (colour[v]==NoNode)
            {
                TFloat dd = d[v]+l;
                if (d[v]==InfFloat)
                {
                    dd = l;
                    Q->Insert(v,-dd);
                }
                else Q->ChangeKey(v,-dd);
                d[v] = dd;
            }
        }

        CT.Ping(Handle(),n);

        count++;
    }

    #if defined(_LOGGING_)

    if (CT.logMeth>1) LogEnd(LH);

    #endif

    Close(H);

    if (nHeap==NULL) delete Q;
    
    CloseFold();

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...delta(%ld)-cut has capacity: %g",x,ret);
        LogEntry(LOG_RES2,CT.logBuffer);
    }
        
    CloseFold(ModNodeIdentification,NO_INDENT);
    
    return TCap(ret);
}


TCap abstractMixedGraph::MinCutNodeIdentification()
    throw(ERRejected)
{
    OpenFold(ModNodeIdentification,SHOW_TITLE);
    
    graph G(*this);

    #if defined(_TRACING_)

    for (TNode u=0;u<n;u++)
    {
        if (Dim()>0)
        {
            G.SetC(u,0,C(u,0));
            G.SetC(u,1,C(u,1));
        }
    }

    #endif

    TCap lambda = InfCap;
    InitColours();
    
    InitPartition();
    for (TNode i=0;i<n;i++) Bud(i);
    
    for (TNode i=1;CT.SolverRunning() && i<n;i++)
    {
        TNode x = NoNode;
        TNode y = NoNode;
        
        TCap thisCap = G.MinCutLegalOrdering(Find(0),x,y);
        if (thisCap<lambda)
        {
            // Update minimum cut
        
            lambda = thisCap;

            for (TNode j=0;j<n;j++)
            {
                if (Find(j)==Find(y)) colour[j] = 1;
                else colour[j] = 0;
            }
        
            CT.Ping(Handle(),n*m);
        }
    
        Merge(x,y);
        G.Struct().IdentifyNodes(x,y);

        if ((i%5)==0)
        {
            TArc a = 0;
            while (a<G.M())
            {
                if (G.StartNode(2*a)==G.EndNode(2*a)) G.DeleteArc(2*a);
                else a++;
            }
        }

        #if defined(_TRACING_)

        G.AutoArcAlignment(0,1);

        #endif
    }

    CloseFold(ModNodeIdentification);

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Edge connectivity is %g",lambda);
        LogEntry(LOG_RES,CT.logBuffer);
    }
    
    return lambda;                      
}
