
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, July 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsmix.h"
#include "shrfam.h"
#include "absgra.h"
#include "queue.h"


TFloat abstractMixedGraph::MinTree(TNode r,TMinTree options)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (r>=n) NoSuchNode("MinTree",r);

    #endif
    
    OpenFold(ModMinTree,NO_INDENT);

    if (CT.logMeth)
    {
        if (r==NoNode)
            LogEntry(LOG_METH,"Computing minimum arborescence...");
        else
        {
            sprintf(CT.logBuffer,"Computing minimum %ld-arborescence...",r);
            LogEntry(LOG_METH,CT.logBuffer);
        }
    }

    OpenFold(ModEdmondsArb,SHOW_TITLE);
    
    TNode nTotal = 2*n;
    shrinkingFamily<TNode> S(n,nTotal,CT);
    
    TArc* inArc = new TArc[nTotal];
    bool* support = new bool[nTotal];
    for (TNode v=0;v<nTotal;v++)
    {
        inArc[v] = NoArc;
        support[v] = false;
    }
    
    TFloat* modLength = new TFloat[2*m];
    for (TArc a=0;a<2*m;a++)
        if (LCap(a)==0) modLength[a] = RedLength(a);
        else modLength[a] = -InfFloat;
    
    LogEntry(LOG_METH,"Shrinking cycles...");
    OpenFold();
    
    TNode rank = 0;
    TNode nCycles = 0;
    bool shrinking = true;
    while (shrinking)
    {
        TArc minArc = NoArc;
        for (TArc a=0;a<2*m;a++)
        {
            TNode u = S.Set(StartNode(a));
            TNode v = S.Set(EndNode(a));
            
            if (!Blocking(a) && u!=v && v!=r && UCap(a)>0 && inArc[v]==NoArc
                && (minArc==NoArc || modLength[minArc]>modLength[a]))
                minArc = a;
        }
        
        if (minArc==NoArc || modLength[minArc]==InfFloat) shrinking = false;
        else
        {
            TNode u = S.Set(StartNode(minArc));
            TNode v = S.Set(EndNode(minArc));
            
            inArc[v] = minArc;
            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"inArc[%ld] = %ld",v,minArc);
                LogEntry(LOG_METH2,CT.logBuffer);
            };
            
            #if defined(_TRACING_)
    
            if (CT.traceLevel==3 && IsDirected())
            {
                explicitSurfaceGraph G(*this,S,modLength,inArc);
                
                CT.Ping(G.Handle(),m);
            };
                    
            #endif

            TNode w = u;
            while (w!=v && inArc[w]!=NoArc) w = S.Set(StartNode(inArc[w]));

            if (w==v)
            {
                #if defined(_LOGGING_)

                THandle LH = NoHandle;

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,"Merging cycle (%ld",u);
                    LH = LogStart(LOG_METH2,CT.logBuffer);
                }

                #endif
        
                support[u] = true;
                w = u;
                while (w!=v)
                {
                    w = S.Set(StartNode(inArc[w]));
                    support[w] = true;
                    
                    #if defined(_LOGGING_)

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,",%ld",w);
                        LogAppend(LH,CT.logBuffer);
                    }
                    
                    #endif
                }
                
                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,",%ld)",u);
                    LogEnd(LH,CT.logBuffer);
                }

                #endif

                for (TArc a=0;a<2*m;a++)
                {
                    TNode u = S.Set(StartNode(a));
                    TNode v = S.Set(EndNode(a));
            
                    if (support[v] && !support[u] && modLength[a]!=-InfFloat)
                    {
                        if (modLength[inArc[v]]==-InfFloat)
                            modLength[a] = InfFloat;
                        else modLength[a] -= modLength[inArc[v]];
                    }
                }
                
                TNode cycle = S.MakeSet();
                support[u] = false;
                S.Merge(cycle,u);
                w = u;
                while (w!=v)
                {
                    w = S.Set(StartNode(inArc[w]));
                    support[w] = false;
                    S.Merge(cycle,w);
                }
                S.FixSet(cycle);
                        
                nCycles++;
            
                #if defined(_TRACING_)

                if (CT.traceLevel==3 && IsDirected())
                {
                    explicitSurfaceGraph G(*this,S,modLength,inArc);
                
                    CT.Ping(G.Handle(),m);
                }
                    
                #endif
            }
            else  rank++;
        }
    }
    
    CloseFold();
    LogEntry(LOG_METH,"Expanding cycles...");
    OpenFold();

    for (TNode v=n+nCycles-1;v>=n;v--)
    {
        TArc a = inArc[v];
        if (a==NoArc)
        {
            TNode u = S.First(v);
            S.Split(v);
            
            TNode thisRoot = u;
            TNode w = S.Set(StartNode(inArc[u]));
            while (w!=u)
            {
                if (modLength[inArc[w]]>modLength[inArc[thisRoot]])
                    thisRoot = w;
                w = S.Set(StartNode(inArc[w]));
            }
            
            inArc[thisRoot] = NoArc;
            
            #if defined(_LOGGING_)

            if (CT.logMeth>1 && IsDirected())
            {
                sprintf(CT.logBuffer,"inArc[%ld] = *",thisRoot);
                LogEntry(LOG_METH2,CT.logBuffer);
            }
            
            #endif
        }
        else
        {
            S.Split(v);

            TNode u = S.Set(EndNode(a));
            inArc[u] = a;
            
            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"inArc[%ld] = %ld",u,a);
                LogEntry(LOG_METH2,CT.logBuffer);
            }
            
            #endif
        }

        #if defined(_TRACING_)
    
        if (CT.traceLevel==3)
        {
            explicitSurfaceGraph G(*this,S,modLength,inArc);
                
            CT.Ping(G.Handle(),m);
        }
                    
        #endif
    }
    
    if ((options & MTREE_1TREE) && r!=NoNode)
    {
        TArc minArc = NoArc;
        THandle H = Investigate();
        goblinIterator &I = Iterator(H);
        while (I.Active(r))
        {
            TArc a = I.Read(r);
            if (EndNode(a)!=r && !Blocking(a^1) && UCap(a)>0
                    && (minArc==NoArc || modLength[minArc]>modLength[a^1]))
                minArc = a^1;
        }
        Close(H);
        
        if (minArc!=NoArc)
        {
            inArc[r] = minArc;
            rank++;
            
            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"inArc[%ld] = %ld",r,minArc);
                LogEntry(LOG_METH2,CT.logBuffer);
            }
            
            #endif
        }
    }
    
    TFloat sum = 0;
    InitPredecessors();
    for (TNode v=0;v<n;v++)
    {
        P[v] = inArc[v];
        if (inArc[v]!=NoArc) sum += RedLength(inArc[v]);
    }

    delete[] inArc;
    delete[] support;
    delete[] modLength;

    CloseFold();

    SetBounds(TimerMinTree,sum,sum);

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Graph rank is %ld",rank);
        LogEntry(LOG_RES,CT.logBuffer);
    }
    
    CloseFold(ModEdmondsArb);

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Minimum spanning tree has weight %g",sum);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    CloseFold(ModMinTree,NO_INDENT);

    if ((!(options & MTREE_1TREE) && rank!=n-1) ||
        ((options & MTREE_1TREE) && rank!=n)
       )
        return InfFloat;
    else return sum;
}


TFloat abstractGraph::MinTree(TNode r,TMinTree options)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (r>=n && r!=NoNode) NoSuchNode("MinTree",r);

    #endif
    
    OpenFold(ModMinTree,NO_INDENT);
    LogEntry(LOG_METH,"Computing minimum spanning tree... ");

    TFloat totalWeight = InfFloat;
    
    if (CT.methMinTree==0)
    {
        if (r!=NoNode)
        {
            totalWeight = Prim(r,options);
            InitSubgraph();

            for (TNode w=0;w<n;w++)
            {
                TArc a = P[w];
                if (a!=NoArc && LCap(a)==0) AddArc(a,1);
            }
        }
        else totalWeight = Prim(NoNode,options);
    }
    
    if (CT.methMinTree==1)
    {
        if (r!=NoNode)
        {
            totalWeight = Prim2(r,options);
            InitSubgraph();
            for (TNode w=0;w<n;w++)
            {
                TArc a = P[w];
                if (a!=NoArc && LCap(a)==0) AddArc(a,1);
            }
        }
        else totalWeight = Prim2(NoNode,options);
    }
    
    if (CT.methMinTree==2)
    {
        if (r!=NoNode)
        {
            totalWeight = Kruskal(r,options);
            try {ExtractTree(r,options);}
            catch (ERCheck)
            {
                CloseFold(ModMinTree,NO_INDENT);
                throw ERRejected();
            }
        }
        else totalWeight = Kruskal(NoNode,options);
    }

    CloseFold(ModMinTree,NO_INDENT);

    #if defined(_FAILSAVE_)

    if (CT.methMinTree>2) UnknownOption("MinTree",CT.methMinTree);

    #endif
    
    return totalWeight;
}


TFloat abstractGraph::Prim(TNode r,TMinTree options)
    throw(ERRange,ERRejected)
{
    if (r==NoNode) r = 0;

    #if defined(_FAILSAVE_)

    if (r>=n) NoSuchNode("Prim",r);

    if (!CLCap())
        Error(ERR_REJECTED,"Prim","Non-trivial lower bounds");

    #endif

    OpenFold(ModPrim,SHOW_TITLE);
    
    InitLabels();
    InitPredecessors();

    for (TNode i=0;i<n;i++) d[i] = InfFloat;
    
    TFloat sum = 0;
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    #if defined(_LOGGING_)

    THandle LH = LogStart(LOG_METH2,"Expanded nodes: ");

    #endif
    
    TNode r2 = r;

    if (options & MTREE_1TREE)   // Compute 1-tree with r and r2 on the cycle.
    {
        TArc a1 = NoArc;
        TArc a2 = NoArc;
        
        while (I.Active(r))
        {
            TArc a = I.Read(r);

            if (EndNode(a)!=r)
            {
                if (a1==NoArc || LCap(a)>0 || (RedLength(a)<RedLength(a1) && LCap(a1)==0))
                {
                    a2 = a1;
                    a1 = a;
                }
                else if (a2==NoArc || (RedLength(a)<RedLength(a2) && LCap(a2)==0)) a2 = a;
            }
        }
        
        if (a2==NoArc)
        {
            #if defined(_LOGGING_)

            sprintf(CT.logBuffer,"There is no cycle through node %ld",r);
            Error(MSG_WARN,"Prim",CT.logBuffer);

            #endif
    
            Close(H);
            CloseFold(ModPrim);
            throw ERRejected();
        }
        
        r2 = EndNode(a1);
        P[r2] = a1;
        P[r] = a2^1;
        d[r] = -InfFloat;
        sum += RedLength(a1)+RedLength(a2);

        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"%ld ",r);
            LogAppend(LH,CT.logBuffer);
        }

        #endif
    
        CT.Ping(Handle(),1);
    }

    TNode u = r2;

    d[u] = 0;

    while (u!=NoNode)
    {
        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"%ld ",u);
            LogAppend(LH,CT.logBuffer);
        }

        #endif
    
        CT.Ping(Handle(),1);
        sum += d[u];
        d[u] = -InfFloat;
        while (I.Active(u))
        {
            TArc a = I.Read(u);
            TFloat l = RedLength(a);
            TNode v = EndNode(a);
        
            if (d[v]>l && UCap(a)>0)
            {
                d[v] = RedLength(a);
                P[v] = a;
            }
        }
        
        u = NoNode;
        for (TNode v=0;v<n;v++)
            if (d[v]>-InfFloat && (u==NoNode || d[v]<d[u])) u = v;
    }

    #if defined(_LOGGING_)

    if (CT.logMeth>1) LogEnd(LH);

    #endif

    Close(H);

    SetBounds(TimerMinTree,sum,sum);

    CT.Ping(Handle(),n*n);

    CloseFold(ModPrim);

    sprintf(CT.logBuffer,"...Minimum spanning tree has weight %g",sum);
    LogEntry(LOG_RES,CT.logBuffer);

    return sum;
}


TFloat abstractGraph::Prim2(TNode r,TMinTree options)
    throw(ERRange,ERRejected)
{
    if (r==NoNode) r = 0;

    #if defined(_FAILSAVE_)

    if (r>=n) NoSuchNode("Prim2",r);

    #endif

    OpenFold(ModPrim2,SHOW_TITLE);
    
    InitLabels();
    InitPredecessors();

    for (TNode i=0;i<n;i++) d[i] = InfFloat;
    
    TFloat sum = 0;
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    #if defined(_LOGGING_)

    THandle LH = LogStart(LOG_METH2,"Expanded nodes: ");

    #endif
    
    TNode r2 = r;

    if (options & MTREE_1TREE)   // Compute 1-tree with r and r2 on the cycle.
    {
        TArc a1 = NoArc;
        TArc a2 = NoArc;
        
        while (I.Active(r))
        {
            TArc a = I.Read(r);

            if (EndNode(a)!=r && UCap(a)>0)
            {
                if (a1==NoArc || LCap(a)>0 ||
                    (RedLength(a)<RedLength(a1) && LCap(a1)==0))
                {
                    a2 = a1;
                    a1 = a;
                }
                else if (a2==NoArc || (RedLength(a)<RedLength(a2) && LCap(a2)==0)) a2 = a;
            }
        }
        
        if (a2==NoArc)
        {
            #if defined(_LOGGING_)

            sprintf(CT.logBuffer,"There is no cycle through node %ld",r);
            Error(MSG_WARN,"Prim",CT.logBuffer);

            #endif
    
            Close(H);
            CloseFold(ModPrim2,NO_INDENT);
        }
        
        r2 = EndNode(a1);
        P[r2] = a1;
        P[r] = a2^1;
        d[r] = -InfFloat;
        sum += RedLength(a1)+RedLength(a2);

        #if defined(_LOGGING_)

        if (CT.logMeth>=2)
        {
            sprintf(CT.logBuffer,"%ld ",r);
            LogAppend(LH,CT.logBuffer);
        }

        #endif
    
        CT.Ping(Handle(),1);
    }

    TNode u = r2;

    goblinQueue<TNode,TFloat>* Q = nHeap;

    if (Q!=NULL) Q -> Init();
    else Q = NewNodeHeap();
    
    Q -> Insert(u,0);
    d[u] = 0;
    
    bool searching = true;
    while (searching)
    {
        CT.Ping(Handle(),1); 
    
        if (Q->Empty()) searching = false;
        else
        {
            u = Q->Delete();
            if (u!=r2) sum += RedLength(P[u]);
            d[u] = -InfFloat;
            
            #if defined(_LOGGING_)
            
            if (CT.logMeth>=2)
            {
                sprintf(CT.logBuffer,"%ld ",u);
                LogAppend(LH,CT.logBuffer);
            }
            
            #endif
    
            while (I.Active(u))
            {
                TArc a = I.Read(u);
                TFloat l = -InfFloat;
                if (LCap(a)==0) l = RedLength(a);
                TNode v = EndNode(a);
                if (d[v]>l && UCap(a)>0)
                {               
                    if (d[v]==InfFloat) Q->Insert(v,l);
                    else Q->ChangeKey(v,l);
                    d[v] = l;
                    P[v] = a;
                }
            }
        }
    }

    #if defined(_LOGGING_)

    if (CT.logMeth>1) LogEnd(LH);

    #endif

    Close(H);

    if (nHeap==NULL) delete Q;

    SetBounds(TimerMinTree,sum,sum);

    CT.Ping(Handle(),m); 

    CloseFold(ModPrim2);

    sprintf(CT.logBuffer,"...Minimum spanning tree has weight %g",sum);
    LogEntry(LOG_RES,CT.logBuffer);

    return sum;
}


TFloat abstractGraph::Kruskal(TNode r,TMinTree options)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (r>=n && (r!=NoNode || (options & MTREE_1TREE)))
        NoSuchNode("Kruskal",r);

    #endif
    
    OpenFold(ModKruskal,SHOW_TITLE);
    
    InitPartition();
    InitSubgraph();

    for (TNode i=0;i<n;i++) Bud(i);
    
    TFloat sum = 0;
    TNode rank = 0;

    if ((options & MTREE_1TREE) && !CLCap())
    {
        for (TArc a=0;a<m;a++)
        {
            TNode u = StartNode(2*a);
            TNode v = EndNode(2*a);
            if (LCap(2*a)==1 && u!=r && v!=r)
            {
                if (Find(u)==Find(v))
                {
                    CloseFold(ModKruskal);
                    Error(ERR_REJECTED,"Kruskal","Fixed subtour detected");
                }
                
                Merge(u,v);
                sum += RedLength(2*a);
                rank++;

                #if defined(_LOGGING_)

                sprintf(CT.logBuffer,
                    "Forced arc %ld with end nodes %ld and %ld, arc length: %g",
                    2*a,u,v,RedLength(2*a));
                LogEntry(LOG_METH2,CT.logBuffer);
                
                #endif
            }
        }
    }
    
    goblinQueue<TArc,TFloat>* Q = NewArcHeap();
    
    for (TArc a=0;a<m;a++)
        if (Length(2*a)!=InfFloat) Q->Insert(a,RedLength(2*a));
    
    while (!Q->Empty() && (rank<n-1))
    {
        TArc a = 2*(Q->Delete());
        
        TNode u = StartNode(a);
        TNode v = EndNode(a);
        
        TNode Bu = Find(u);
        TNode Bv = Find(v);
        
        if ((!(options & MTREE_1TREE) && Bu!=Bv) ||
            ((options & MTREE_1TREE) && Bu!=Bv && Bu!=r && Bv!=r && UCap(a)>0))
        {
            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,
                    "Merging nodes %ld and %ld, arc length: %g",
                    Bu,Bv,RedLength(a));
                LogEntry(LOG_METH2,CT.logBuffer);
            }

            #endif
    
            CT.Ping(Handle(),1);
            Merge(u,v);
            AddArc(a,1);
            sum += RedLength(a);
            rank++;
        }
    }
    
    if (options & MTREE_1TREE)   // Compute 1-tree with node r on the cycle.
    {
        if (rank<n-2)
        {
            delete Q;
            CloseFold(ModKruskal);
            sprintf(CT.logBuffer,"%ld is a cut node",r);
            Error(ERR_REJECTED,"Kruskal",CT.logBuffer);
        }
        
        TArc a1 = NoArc;
        TArc a2 = NoArc;
        
        THandle H = Investigate();
        goblinIterator &I = Iterator(H);
        while (I.Active(r))
        {
            TArc a = I.Read(r);
            if (EndNode(a)!=r && UCap(a)>0)
            {
                if (a1==NoArc || RedLength(a)<RedLength(a1))
                {
                    a2 = a1;
                    a1 = a;
                }
                else if (a2==NoArc || RedLength(a)<RedLength(a2)) a2 = a;
            }
        }
        
        TArc fixedArcs = 0;
        I.Reset(r);
        while (I.Active(r))
        {
            TArc a = I.Read(r);
            if (LCap(a)==1 && a!=a1 && a!=a2)
            {
                a2 = a1;
                a1 = a;
                fixedArcs++;
            }
        }
        Close(H);
        
        if (fixedArcs>2)
        {
            delete Q;
            CloseFold(ModKruskal);
            sprintf(CT.logBuffer,"Too much fixed arcs adjacent with %ld",r);
            Error(ERR_REJECTED,"Kruskal",CT.logBuffer);
        }
        
        if (a2==NoArc)
        {
            delete Q;
            CloseFold(ModKruskal);
            sprintf(CT.logBuffer,"There is no cycle through node %ld",r);
            Error(ERR_REJECTED,"Kruskal",CT.logBuffer);
        }
        
        #if defined(_LOGGING_)

        if (CT.logMeth>=2)
        {
            sprintf(CT.logBuffer,
                "Adding arc %ld with end nodes %ld and %ld, arc length: %g",
                a1,r,EndNode(a1),RedLength(a1));
            LogEntry(LOG_METH2,CT.logBuffer);

            sprintf(CT.logBuffer,
                "Adding arc %ld with end nodes %ld and %ld, arc length: %g",
                a2,r,EndNode(a2),RedLength(a2));
            LogEntry(LOG_METH2,CT.logBuffer);
        }

        #endif
    
        CT.Ping(Handle(),1);
        if (Sub(a1)==0) AddArc(a1,1);
        if (Sub(a2)==0) AddArc(a2,1);
        sum += RedLength(a1)+RedLength(a2);
        rank += 2;
    }
    
    SetBounds(TimerMinTree,sum,sum);

    CT.Ping(Handle(),n*n);

    delete Q;

    if (!(options & MTREE_1TREE))
    {
        sprintf(CT.logBuffer,"...Graph rank is %ld",rank);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    CloseFold(ModKruskal);

    sprintf(CT.logBuffer,"...Minimum spanning tree has weight %g",sum);
    LogEntry(LOG_RES,CT.logBuffer);

    return sum;
}
