
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Birk Eisermann, August 2003.
**  Revised by Christian Fremuth-Paeger, December 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "absmix.h"
#include "statstaq.h"
#include "dynq.h"
#include "subgraph.h"
#include "segment.h"


typedef struct {
    segmentGraph*   (*Segments);
    TNode           Count;
    TNode           StartNewSeg;
    TNode           EmbSegNo;       //fr recalc == true
} TSegData;


typedef struct {
    TArc*                       predArc;
    goblinHashTable<TArc,TArc>* succArc;
    TNode*                      incidentRegion;
    int                         Count;
    int                         MinRegions;
    int                         Region1No,Region2No;
} TRegData;

typedef segmentGraph*  PSegmentGraph;


//forward declaration
static void determineSegments(TSegData& SD, abstractSubgraph* oldSegment,
        abstractSubgraph* Embedded, abstractSubgraph* ToEmbed,
        abstractMixedGraph* G);
static void determinePossibleRegions(TSegData&,TRegData&,abstractMixedGraph*,bool);

static staticStack<TNode,TFloat>* Q;


bool abstractMixedGraph::IsPlanar() throw (ERRejected)
{
    OpenFold(ModPlanarity,NO_INDENT);

    LogEntry(LOG_METH,"Testing planarity...");
    
    // Store the regions in terms of arc predecessor lists
    // Reserve memory for n additional arcs 
    TArc* predArc = new TArc[2*(m+n)];
    
    bool ret = PlanarityDMP64(predArc);
    
    delete[] predArc;

    if (ret) LogEntry(LOG_RES, "...Graph is planar");
    else LogEntry(LOG_RES, "...Graph is non-planar");

    CloseFold(ModPlanarity,NO_INDENT);
    
    return ret;
}


bool abstractMixedGraph::Planarize() throw(ERRejected)
{
    Error(ERR_REJECTED,"Planarize","Method applies to directed graphs only");
    throw ERRejected();
}


bool abstractMixedGraph::PlanarityDMP64(TArc* predArc) throw (ERRejected)
{
    TRegData          RegData;
    TSegData          SegData;

    
    // Determine the number of arcs in the underlying simple graph
    TArc m0 = 0;
    TArc ml = 0;
    for (TArc a=0;a<m;a++)
    {
        TNode u = StartNode(2*a);
        TNode v = EndNode(2*a);
        
        if (u==v) ml++;
        else if (Adjacency(u,v)==2*a) m0++;
    }

    if (m0>3*n-6) return false;


    OpenFold(ModPlanarityDMP,SHOW_TITLE);

    RegData.predArc = predArc;
    RegData.succArc = new goblinHashTable<TArc,TArc>(n*m,2*(m+n),NoArc,CT);
    RegData.incidentRegion = new TNode[2*m+n];

    for (TArc a=0;a<2*m;a++)
    {
        RegData.predArc[a] = NoArc;
        RegData.incidentRegion[a] = NoNode;
    }
    
    graph G(*this,OPT_CLONE);

    // Queue for the nodes of G when computing connected components
    Q = new staticStack<TNode,TFloat>(n,CT);
    
    // Export the embedded subgraph to a trace image
    G.InitSubgraph();
    

    TNode* thread = new TNode[n];
    
    if (CutNodes(NULL,NULL,thread))
    {
        // Graph is not 2-connected:
        // Construct the initial cycle from the CutNodes() thread of leaves

        LogEntry(LOG_METH,"Adding the thread of leaves...");
        OpenFold();

        TNode w = 0;
        TNode v = thread[0];

        TArc a0 = G.Adjacency(v,w);
        if (a0==NoArc) a0 = 2*G.InsertArc(v,w);
        TArc a = a0;
        
        #if defined(_LOGGING_)

        THandle LH = NoHandle;
        
        if (CT.logMeth>1) LH = LogStart(LOG_METH2, "(");
        
        #endif

        while (v!=0)
        {
            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer, "%ld[%ld]",w,a^1);
                LogAppend(LH,CT.logBuffer);
            }
            
            #endif

            TNode x = thread[v];
            if (x==NoNode) x = 0;
            TArc a2 = G.Adjacency(x,v);
            if (a2==NoArc) a2 = 2*G.InsertArc(x,v);

            RegData.predArc[a] = a2;
            RegData.predArc[a2^1] = (a^1);
            RegData.incidentRegion[a] = 0;
            RegData.incidentRegion[a2^1] = 1;
            RegData.succArc -> ChangeKey(0*n+v,a);
            RegData.succArc -> ChangeKey(1*n+v,a2^1);

            a = a2;
            w = v;
            v = x;
        }

        RegData.predArc[a] = a0;
        RegData.predArc[a0^1] = (a^1);
        RegData.incidentRegion[a] = 0;
        RegData.incidentRegion[a0^1] = 1;
        RegData.succArc -> ChangeKey(0*n+v,a);
        RegData.succArc -> ChangeKey(1*n+v,a0^1);

        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer, "%ld[%ld]%ld)",w,a^1,v);
            LogEnd(LH,CT.logBuffer);
        }
        
        #endif

        CloseFold();
    }
    else
    {
        // Choose the initial cycle from the CutNodes() DFS tree rooted at node 0

        LogEntry(LOG_METH,"Extracting an initial cycle...");
        OpenFold();

        TArc a0 = First(0);
        while (a0==P[EndNode(a0)]) a0 = Right(a0,0);

        TNode v = EndNode(a0);
        TNode w = 0;
        TArc a = G.Adjacency(v,0);

        #if defined(_LOGGING_)

        THandle LH = NoHandle;
        
        if (CT.logMeth>1) LH = LogStart(LOG_METH2, "(");
        
        #endif

        while (v!=0)
        {
            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer, "%ld[%ld]",w,a^1);
                LogAppend(LH,CT.logBuffer);
            }
            
            #endif

            TArc a2 = G.Adjacency(StartNode(P[v]),v);
            RegData.predArc[a] = a2;
            RegData.predArc[a2^1] = (a^1);
            RegData.incidentRegion[a] = 0;
            RegData.incidentRegion[a2^1] = 1;
            RegData.succArc -> ChangeKey(0*n+v,a);
            RegData.succArc -> ChangeKey(1*n+v,a2^1);

            a = a2;
            w = v;
            v = StartNode(a);
        }

        RegData.predArc[a] = (a0^1);
        RegData.predArc[a0] = (a^1);
        RegData.incidentRegion[a] = 0;
        RegData.incidentRegion[a0] = 1;
        RegData.succArc -> ChangeKey(0*n+v,a);
        RegData.succArc -> ChangeKey(1*n+v,a0);

        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer, "%ld[%ld]0)",w,a^1);
            LogEnd(LH,CT.logBuffer);
        }
        
        #endif

        CloseFold();
    }

    delete[] thread;


    LogEntry(LOG_METH, "Determine the initial segments...");
    OpenFold();

    abstractSubgraph* Embedded = new subgraph(&G);
    
    for (TArc a=0;a<G.M();a++)
    {
        if (RegData.predArc[2*a]!=NoArc)
        {
            Embedded->AddArc(2*a);

            // Export the embedded subgraph to a trace image
            G.AddArc(2*a,G.UCap(2*a)-G.Sub(2*a));
        }
    }

    abstractSubgraph* ToEmbed = Embedded->ComplementarySubgraph();


    SegData.Segments = NULL;
    SegData.Count = 0;
    SegData.StartNewSeg = 0;
    SegData.EmbSegNo = 0;

    determineSegments(SegData,NULL,Embedded,ToEmbed,&G);
    
    CloseFold();

    RegData.Count = 2;
    RegData.Region1No = 0;
    RegData.Region2No = 1;

    determinePossibleRegions(SegData,RegData,&G,false);

    LogEntry(LOG_METH, "Try to embed all segments...");
    OpenFold();

    int i = 1;
    while ( (Embedded->M() < G.M()-ml) && (SegData.Count > 0) )
    {
        CT.Ping(G.Handle(),n);

        segmentGraph* Segment = SegData.Segments[SegData.EmbSegNo];

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,
                "Segment %li has a minimum number of possible regions",
                SegData.EmbSegNo);
            LogEntry(LOG_METH2,CT.logBuffer);
        }


        G.InitPredecessors();

        THandle H = G.Investigate();
        goblinIterator &I = G.Iterator(H);

        TNode u = Segment->GetContactNode(0);
        TNode w = u;

        while (w==u || !Embedded->HasNode(w))
        {
            if (I.Active(w))
            {
                TArc a = I.Read(w);
                TNode v = G.EndNode(a);

                if (G.P[v]==NoArc && Segment->HasArc(a))
                {
                    G.P[v] = a;
                    w = v;
                }
            }
            else w = G.StartNode(G.P[w]);
        }

        G.Close(H);
        

        RegData.Region1No = Segment->GetRegion(0);
        RegData.Region2No = RegData.Count;
        RegData.Count++;

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"Splitting region %i into two parts...", RegData.Region1No);
            LogEntry(LOG_METH2,CT.logBuffer);
        }

        // Determine the successor arc of w in the current region
        TArc aSuccW = RegData.succArc -> Key(RegData.Region1No*n+w);
        
        // Move uw-path from Region1 to Region2
        TArc a = aSuccW;
        while (G.StartNode(a) != u)
        {
            a = RegData.predArc[a];
            RegData.incidentRegion[a] = RegData.Region2No;
            RegData.succArc -> ChangeKey(RegData.Region2No*n+G.StartNode(a),a);
            RegData.succArc -> ChangeKey(RegData.Region1No*n+G.StartNode(a),NoArc);
        }
        
        TArc aSuccU = a;


        LogEntry(LOG_METH2,"Adding path to both regions and removing it from segment...");

        TNode v1, v1_succ;
        v1_succ = v1 = w;
        subgraph *oldSegment = new subgraph(Segment);
        Segment -> OmitNode(w);

        OpenFold();
        THandle LH = LogStart(LOG_METH2, "(");

        a = NoArc;
        TArc a2 = aSuccW;
        TArc a1 = RegData.predArc[a2];

        while (v1 != u)
        {
            a = G.P[v1];
            v1 = G.StartNode(a);

            Segment -> OmitNode(v1);
            Embedded -> AddArc(a);

            RegData.predArc[a^1] = a1;
            RegData.predArc[a2] = a;
            RegData.incidentRegion[a^1] = RegData.Region2No;
            RegData.incidentRegion[a] = RegData.Region1No;
            RegData.succArc -> ChangeKey(RegData.Region2No*n+v1_succ,a^1);
            RegData.succArc -> ChangeKey(RegData.Region1No*n+v1,a);
                        
            // Export the embedded subgraph to a trace image
            G.AddArc(a,G.UCap(a)-G.Sub(a));

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer, "%ld[%ld]",v1_succ,a^1);
                LogAppend(LH,CT.logBuffer);
            }

            v1_succ = v1;
            a1 = (a^1);
            a2 = a;
        }
        
        a2 = aSuccU;
        a1 = RegData.predArc[a2];

        RegData.predArc[a2] = (a^1);
        RegData.predArc[a] = a1;
    
        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer, "%ld)",v1_succ);
            LogEnd(LH,CT.logBuffer);
        }
        CloseFold();


        LogEntry(LOG_METH2, "Determine new segments...");
        OpenFold();
        determineSegments(SegData,oldSegment,Embedded,NULL,&G);
        delete oldSegment;
        CloseFold();


        LogEntry(LOG_METH2, "Updating possible regions of all segments...");
        determinePossibleRegions(SegData,RegData,&G,true);

        if (RegData.MinRegions <= 0)
        {
            delete[] RegData.incidentRegion;
            delete RegData.succArc;

            // This implicitly deletes ToEmbed;
            delete Embedded;

            sprintf(CT.logBuffer,"Segment %li cannot be embedded!",SegData.EmbSegNo);
            LogEntry(LOG_METH2,CT.logBuffer);
            CloseFold();
            CloseFold(ModPlanarityDMP);

            return false;
        }
        
        i++;
    }
    
    CloseFold();
        
    
    LogEntry(LOG_METH,"Postprocessing...");

    TArc mp = 0;
    for (TArc a=0;a<m;a++)
    {
        TNode v = StartNode(2*a);
        TNode w = EndNode(2*a);
        
        if (v==w) continue;

        // This can occur only if a is on the initial cycle
        if (RegData.predArc[2*a]!=NoArc) continue;
        
        TArc a2 = Adjacency(v,w);

        if (a==(a2>>1)) continue;

        TArc a3 = RegData.succArc -> Key(RegData.incidentRegion[a2]*n+w);
        RegData.predArc[a3] = 2*a;
        RegData.predArc[2*a+1] = a2;
        RegData.predArc[2*a] = RegData.predArc[a2];
        RegData.predArc[a2] = 2*a+1;

        RegData.succArc -> ChangeKey(RegData.Count*n+w,2*a+1);
        RegData.succArc -> ChangeKey(RegData.Count*n+v,a2);
        RegData.succArc -> ChangeKey(RegData.incidentRegion[a2]*n+w,a3);
        RegData.succArc -> ChangeKey(RegData.incidentRegion[a2]*n+v,2*a);

        RegData.incidentRegion[2*a] = RegData.incidentRegion[a2];
        RegData.incidentRegion[2*a+1] = RegData.Count;
        RegData.incidentRegion[a2] = RegData.Count;
        RegData.Count++;
        mp++;
    }
    
    if (mp>0 && CT.logMeth>1)
    {
        sprintf(CT.logBuffer,"...%li parallel arcs embedded",mp);
        LogEntry(LOG_METH2,CT.logBuffer);
    }

    if (ml>0)
    {
        for (TArc a=0;a<m;a++)
        {
            TNode v = StartNode(2*a);
            if (EndNode(2*a)!=v) continue;
            
            TArc a2 = G.First(v);
            while ((a2>>1)==a) a2 = G.Right(a2,v);
            
            RegData.predArc[2*a+1] = (2*a+1);
            RegData.predArc[2*a] = RegData.predArc[a2];
            RegData.predArc[a2] = 2*a;

            RegData.succArc -> ChangeKey(RegData.Count*n+v,2*a+1);
            
            RegData.incidentRegion[2*a] = RegData.incidentRegion[a2];
            RegData.incidentRegion[2*a+1] = RegData.Count;
            RegData.Count++;
        }
        
        sprintf(CT.logBuffer,"...%li loops embedded",ml);
        LogEntry(LOG_METH2,CT.logBuffer);
    }
        
    
    delete[] RegData.incidentRegion;
    delete RegData.succArc;

    // This implicitly deletes ToEmbed;
    delete Embedded;

    delete Q;

    CloseFold(ModPlanarityDMP);

    return true;
}


/*!
 * /param The data structure SD delivers the old segments
 * /param oldSegment is the segment from which a path has been embedded.
 *        The path is not deleted from the segment yet
 *        If oldSegment==NULL, the whole graph is considered.
 * /param Embedded is the embedded subgraph (including the new path)
 * /param ToEmbed is the node complement of Embedded
 * /param G is the entire graph
 */
void determineSegments(TSegData& SD, abstractSubgraph* oldSegment,
        abstractSubgraph* Embedded, abstractSubgraph* ToEmbed,
        abstractMixedGraph* G)
{
    segmentGraph*       Segment;

    goblinController& CT = G->Context();

    if (oldSegment)
    {
        ToEmbed = SD.Segments[SD.EmbSegNo];
        TArc m = oldSegment->M();

        SD.Count--;
        for (TNode c=SD.EmbSegNo; c<SD.Count; c++)
            SD.Segments[c] = SD.Segments[c+1];

        sprintf(CT.logBuffer,"Former segment has %li arcs", m);
        G->LogEntry(LOG_METH2,CT.logBuffer);

        if (m == 1)
        {
            SD.Segments = (segmentGraph **)GoblinRealloc(SD.Segments,
                SD.Count * sizeof(segmentGraph *));
            SD.StartNewSeg = SD.Count;
            delete ToEmbed;
            return;
        }
    }


    G -> LogEntry(LOG_METH2, "Computing connected components...");
    G -> OpenFold();

    G -> InitColours(NoNode);
    TNode CompCount = 0;

    THandle H = G->Investigate();
    goblinIterator &I = G->Iterator(H);

    for (TNode u=0; u<G->N(); u++)
    {
        if (Embedded->HasNode(u) || G->Colour(u)!=NoNode ||
            (oldSegment && !oldSegment->HasNode(u))
           ) continue;
        
        Q -> Insert(u);
        G -> SetColour(u,CompCount);
    
        #if defined(_LOGGING_)

        THandle LH = NoHandle;

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"Component %ld:",CompCount);
            LH = G->LogStart(LOG_METH2,CT.logBuffer);
        }

        #endif

        while (!(Q->Empty()))
        {
            TNode u = Q->Delete();

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer," %ld",u);
                G->LogAppend(LH,CT.logBuffer);
            }

            #endif

            while (I.Active(u))
            {
                TArc a = I.Read(u);
                TNode v = G->EndNode(a);

                if (G->Colour(v)==NoNode &&
                    !Embedded->HasNode(v) &&
                    ToEmbed->HasArc(a) &&
                    (!oldSegment || oldSegment->HasNode(v))
                   )
                {
                    G -> SetColour(v,CompCount);
                    try {Q->Insert(v);} catch (ERCheck) {};
                }
            }
        }
        
        #if defined(_LOGGING_)

        if (CT.logMeth>1) G->LogEnd(LH);
        
        #endif

        CompCount++;
    }

    G -> Close(H);
    G -> CloseFold();


    G->LogEntry(LOG_METH2, "Create segment objects; marking all contact points...");

    // Allocate new segments for the found connected components.
    // Later, additional segments are added which consist of a single arc.
    if (!oldSegment)
    {
        SD.Count = 0;
        SD.Segments = new PSegmentGraph[CompCount*sizeof(PSegmentGraph)];
    }
    else
    {
        SD.Segments = (segmentGraph **)GoblinRealloc(SD.Segments,
            (SD.Count+CompCount) * sizeof(segmentGraph *));
        delete ToEmbed;
    }

    for (TNode c=SD.Count; c<SD.Count+CompCount; c++)
         SD.Segments[c] = new segmentGraph(G);
    SD.StartNewSeg = SD.Count;

    for (TArc a=0; a<G->M()*2; a++)
    {
        // Consider only arcs of the former segment
        if (oldSegment && !oldSegment->HasArc(a))
            continue;

        TNode sNode = G->StartNode(a);
        TNode eNode = G->EndNode(a);
        
        // Do not consider loops
        if (sNode==eNode) continue;
        
        // Do not consider parallel arcs
        if (!oldSegment && (a>>1) != (G->Adjacency(sNode,eNode)>>1)) continue;

        if (!Embedded->HasArc(a))
        {
            if (!Embedded->HasNode(sNode) || !Embedded->HasNode(eNode))
            {
                TNode color = G->Colour(sNode);
                if (color==NoNode) color = G->Colour(eNode);

                Segment = SD.Segments[color + SD.Count];
                Segment->AddArc(a);

                if (Embedded->HasNode(sNode))
                {
                    Segment -> MarkAsContactNode(sNode);
                }

                if (Embedded->HasNode(eNode))
                {
                    Segment -> MarkAsContactNode(eNode);
                }
            }
            else
            {  
                // Check if a is already stored as a segment
                bool add = true;
                for (TNode s = 0; s < SD.Count+CompCount; s++)
                {
                    Segment = SD.Segments[s];
                    if ((Segment->GetContactNodeCount() == 2)
                        && Segment->HasArc(a))
                    {
                        add = false;
                        break;
                    }
                }

                if (add)
                {
                    CompCount++;
                    SD.Segments = (segmentGraph **)GoblinRealloc(SD.Segments,
                                   (SD.Count+CompCount)*sizeof(segmentGraph*));
                    SD.Segments[SD.Count+CompCount-1] = new segmentGraph(G);
                    Segment = SD.Segments[SD.Count+CompCount-1];
                    Segment->AddArc(a);
                    Segment->MarkAsContactNode(sNode);
                    Segment->MarkAsContactNode(eNode);
                }
            }
        }
    }


    TNode newSegCount = 0;
    for (TNode c=SD.Count; c<SD.Count+CompCount; c++)
    {
        Segment = SD.Segments[c];
        SD.Segments[SD.Count+newSegCount] = SD.Segments[c];
        newSegCount++;
    }

    SD.Count += newSegCount;
    SD.Segments = (segmentGraph **)GoblinRealloc(SD.Segments,SD.Count*sizeof(segmentGraph*));
}


void determinePossibleRegions(TSegData &SegData,TRegData &RegData,abstractMixedGraph *G,bool recalc)
{
    segmentGraph *Segment;

    #if defined(_LOGGING_)

    goblinController& CT = G->Context();
    
    #endif

    SegData.EmbSegNo = NoNode;
    RegData.MinRegions = 1000000;

    G -> OpenFold();

    // For each old segment:
    for (TNode s = 0; s<SegData.StartNewSeg && recalc; s++)
    {
        Segment = SegData.Segments[s];
        
        if (Segment->HasRegion(RegData.Region1No))
        {
            for (TNode CN = 0; CN<Segment->GetContactNodeCount(); CN++)
            {
                if (RegData.succArc -> 
                      Key(RegData.Region1No*G->N()+Segment->GetContactNode(CN)) == NoArc)
                {
                    Segment->OmitRegion(RegData.Region1No);
                    break;
                }
            }

            bool fit = true;
            for (TNode CN = 0; CN<Segment->GetContactNodeCount(); CN++)
            {
                if (RegData.succArc ->
                      Key(RegData.Region2No*G->N()+Segment->GetContactNode(CN)) == NoArc)
                {
                    fit = false;
                    break;
                }
            }

            if (fit) Segment -> AddRegion(RegData.Region2No);
            // This region is new, and no segment is incident yet
            // Hence no Segment->omitRegion are necessary
        }

        if (Segment->GetRegionsCount() < RegData.MinRegions)
        {
            SegData.EmbSegNo = s;
            RegData.MinRegions = Segment->GetRegionsCount();
        }
    }

    // For each new segment:
    for (TNode s = SegData.StartNewSeg; s<SegData.Count; s++)
    {
        Segment = SegData.Segments[s];

        for (int r=0; r<RegData.Count; r++)
        {
            bool fit = true;
            for (TNode CN = 0; CN<Segment->GetContactNodeCount(); CN++)
            {
                if (RegData.succArc ->
                      Key(r*G->N()+Segment->GetContactNode(CN)) == NoArc)
                {
                    fit = false;
                    break;
                }
            }

            if (fit) Segment->AddRegion(r);
        }

        if (Segment->GetRegionsCount() < RegData.MinRegions)
        {
            SegData.EmbSegNo = s;
            RegData.MinRegions = Segment->GetRegionsCount();
        }
    }

    #if defined(_LOGGING_)

    for (TNode s = 0; s<SegData.Count && CT.logMeth>1; s++)
    {
        Segment = SegData.Segments[s];
        
        sprintf(CT.logBuffer,"Segment %ld fits into regions",s);
        G -> LogEntry(LOG_METH2,CT.logBuffer);

        for (int r=0; r<RegData.Count; r++)
            if (Segment->HasRegion(r))
            {
                sprintf(CT.logBuffer," %d",r);
                G -> LogEntry(MSG_APPEND,CT.logBuffer);
            }

        G -> LogEntry(LOG_METH2,"Contact nodes:");

        for (TNode CN = 0; CN<Segment->GetContactNodeCount(); CN++)
        {
            sprintf(CT.logBuffer," %ld",Segment->GetContactNode(CN));
            G -> LogEntry(MSG_APPEND,CT.logBuffer);
        }
    }
    
    #endif

    G -> CloseFold();
}


TNode abstractMixedGraph::ExtractEmbedding(TNode* argFace)
    throw (ERRejected)
{
    if (face)
    {
        TNode l = 0;
        TArc a = aExterior;
        do
        {
            TNode v = EndNode(a);
            a = Right(a^1,v);
            l++;
        }
        while (a!=aExterior);
    
        sprintf(CT.logBuffer,"...Graph has %ld faces",ND());
        LogEntry(LOG_RES,CT.logBuffer);
        sprintf(CT.logBuffer,"...Exterior face %ld has length %ld",
            face[aExterior],l);
        LogEntry(LOG_RES,CT.logBuffer);

        return l;
    }

    LogEntry(LOG_METH,"Extracting faces from incidence lists...");

    face = new TNode[2*m];
    
    for (TArc i=0;i<2*m;i++) face[i] = NoNode;
    
    OpenFold();
    
    // Number of faces
    TNode k = 0;
    
    // Number of cancelled arcs
    TNode c = 0;

    // Candidate arc for aExterior and the lenght of the incident region    
    TArc retArc = NoArc;
    TNode maxLength = 0;

    for (TArc a=0;a<2*m;a++)
    {
        if (face[a]!=NoNode) continue;
        
        TNode w = StartNode(a^1);
        
        if (w==NoNode)
        {
            // Acknowledge the cancelled arc a
            c++;
            continue;
        }

        TArc regionLength = 1;
        TArc a2 = Right(a^1,w);
        face[a2] = k;

        #if defined(_LOGGING_)

        THandle LH = NoHandle;

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"Face %ld = (%ld [%ld] %ld",
                k,StartNode(a2),a2,StartNode(a2^1));
            LH = LogStart(LOG_METH2,CT.logBuffer);
        }

        #endif

        while (a2!=a)
        {
            regionLength++;
            a2 = Right(a2^1,StartNode(a2^1));
            face[a2] = k;

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer," [%ld] %ld",a2,StartNode(a2^1));
                LogAppend(LH,CT.logBuffer);
            }

            #endif
        }

        #if defined(_LOGGING_)

        if (CT.logMeth>1) LogEnd(LH,")");

        #endif
        
        if (regionLength>maxLength)
        {
            retArc = a;
            maxLength = regionLength;
        }
        
        k++;
    }
    
    CloseFold();
    
    #if defined(_FAILSAVE_)

    if (m-(c/2)-n+2!=k)
        Error(ERR_REJECTED,"ExtractEmbedding","Not a planar embedding");

    #endif
    
    sprintf(CT.logBuffer,"...Graph has %ld faces",k);
    LogEntry(LOG_RES,CT.logBuffer);
    sprintf(CT.logBuffer,"...Exterior face %ld has length %ld",
        face[retArc],maxLength);
    LogEntry(LOG_RES,CT.logBuffer);

    for (TArc a=0;argFace && a<2*m;a++) argFace[a] = face[a];
    
    aExterior = retArc;

    return k;
}


void abstractMixedGraph::MarkExteriorFace(TArc aBasis) throw ()
{
    LogEntry(LOG_MEM,"Initializing exterior face...");

    TArc a = aBasis;
    do
    {
        TNode v = EndNode(a);
        SetFirst(v,a^1);
        a = Right(a^1,v);
    }
    while (a!=aBasis);
    
    aExterior = aBasis;
}


void abstractMixedGraph::GrowExteriorFace() throw(ERRejected)
{
    Error(ERR_REJECTED,"GrowExteriorFace","Method applies to sparse graphs only");
    throw ERRejected();
}


TNode abstractMixedGraph::LMCOrderedPartition(
    TArc* cLeft,TArc* cRight,TNode* vFirst,TNode* vRight)
    throw (ERRejected)
{
    // Prerequisite for several graph drawing methods:
    // - (Convex) straight line embedding
    // - Polyline drawings with ports and 3 bends per edge


    LogEntry(LOG_METH,"Computing LMC ordered partition...");
    OpenFold(ModLMCOrder);


    // If possible, start with a known basis arc
    TArc aBasis = aExterior;
    
    // Node deletions are performed on a copy
    graph G(*this,OPT_CLONE);
    
    // Delete all loops and parallel arcs from G
    for (TArc a=0;a<m;a++)
    {
        TNode u = G.StartNode(2*a);
        TNode v = G.EndNode(2*a);
        TArc a2 = G.Right(2*a,u);
    
        if (u==v) G.CancelArc(2*a);
        else if (G.EndNode(a2)==v)
        {
            if (aBasis==2*a) aBasis = a2;

            G.CancelArc(2*a);
        }
    }
    
    
    // Determine number of faces (dual nodes) and the left hand side face
    // for every graph arc. Stop if graph is non-planar or not embedded yet.
    
    TNode nd = 0;
    TNode* face = new TNode[2*m];

    try
    {
        nd = G.ExtractEmbedding(face);
    }
    catch (ERRejected)
    {
        delete[] face;
        CloseFold(ModPlanarity);
        throw ERRejected();
    }
    
    if (aBasis==NoArc) aBasis = G.ExteriorArc()^1;

    aExterior = aBasis;
    
    
    // The extremal components and the unbounded face

    TNode v1 = G.StartNode(aBasis);
    TNode v2 = G.EndNode(aBasis);
    TNode fOut = face[aBasis^1];
    TNode fBasis = face[aBasis];
    
    G.MarkExteriorFace(aBasis^1);
    aExterior = aBasis;

    // Choose the final node vn half the way on the exterior face
    TNode vn = G.EndNode(Right(aBasis,v1));
    TNode v = vn;
    do
    {
        v = G.EndNode(G.Right(G.First(v),v));
        
        if (v!=v2)
        {
            vn = G.EndNode(G.Right(G.First(vn),vn));
            v = G.EndNode(G.Right(G.First(v),v));
        }
    }
    while (v!=v2);
    

    // A node can be deleted only if it has been visited before
    bool* visited = new bool[n];
    for (v=0;v<n;v++) visited[v] = false;
    
    // A node can be deleted only if it is not in a separation pair.
    // fSep[] is maintained for exterior nodes only and = 0 otherwise
    TNode* fSep = new TNode[n];
    for (v=0;v<n;v++) fSep[v] = 0;

    #if defined(_TRACING_)

    if (CT.traceLevel>1) G.InitColours(0);

    #endif
    
    // Number of components
    TNode k = 0;
    
    
    // Number of outer nodes / arcs incident with a face
    TNode* vOut = new TNode[nd];
    TArc*  aOut = new TArc[nd];
    
    // In which iteration has aOut[f] or vOut[f] changed the last time?
    TNode* prevChange = new TNode[nd];
    
    for (TNode f=0;f<nd;f++)
    {
        vOut[f] = aOut[f] = 0;
        prevChange[f] = NoNode;
    }

    
    // Compute vOut and aOut for the original graph
    
    LogEntry(LOG_METH,"Computing face incidence counters...");
    OpenFold();

    TArc a = aBasis^1;
    v = v2;

    do
    {
        aOut[face[a^1]]++;
        
        TArc a2 = G.Right(a,v);
        do
        {
            vOut[face[a2]]++;
            a2 = G.Right(a2,v);
        }
        while (a2!=a);

        // Traverse the unbounded face in clockwise order
        v = G.EndNode(a);
        a = G.Right(a^1,v);
    }
    while (a!=(aBasis^1));
    

    // Partial 3-connectivity check: If this check is passed successfully,
    // then indeed fSep[v]=0 holds for every node v.

    bool connected = true;

    for (TNode f=0;f<nd;f++)
    {
        if (f==fOut) continue;

        // f shares a separation pair with the exterior face
        if (vOut[f]>aOut[f]+1 || aOut[f]>1) connected = false;

        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"vOut[%ld] = %ld, aOut[%ld] = %ld",
                f,vOut[f],f,aOut[f]);
            LogEntry(LOG_METH2,CT.logBuffer);
        }

        #endif
    }

    CloseFold();

    CT.Ping(G.Handle(),n);


    // Candidate queue for the nodes which can be deleted currently
    staticStack<TNode,TFloat> Q1(n,CT);
    
    // Candidate queue for the faces which can be deleted currently.
    // Store the arc which connects the left contact node and the first
    // node of the later component
    staticStack<TArc,TFloat> Q2(2*m,CT);
    
    // Keep all faces which have becomes a separating face in an iteration
    staticStack<TArc,TFloat> Q3(2*m,CT);

    
    // The component generated first consists of vn only
    visited[vn] = true;
    Q1.Insert(vn);

    
    // Loop until only the basis arc is left or until it turns out that
    // the graph is not 3-connected
    
    // The left most and the right most contact node of the current component
    TNode cl = NoNode;
    TNode cr = NoNode;

    // The node / region which is deleted in the current iteration
    TNode vDel = NoNode;
    TNode fDel = NoNode;


    LogEntry(LOG_METH,"Deleting nodes and faces...");
    OpenFold();

    while (connected && G.Right(aBasis,v1)!=aBasis)
    {
        // Construct the component to delete

        if (!Q1.Empty())
        {
            vDel = Q1.Delete();

            // Check if vDel is really admissible for deletion
            if (   fSep[vDel]>0
                || G.First(vDel)==NoArc
                || face[G.First(vDel)^1]!=fOut
               )
            {
                continue;
            }

            vFirst[k] = vDel;
            vRight[vDel] = NoNode;
            cLeft[k] = G.First(vDel)^1;
            cRight[k] = G.Right(G.First(vDel),vDel);

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"Component %ld (Node): %ld",k,vDel);
                LogEntry(LOG_METH2,CT.logBuffer);
            }

            #endif
            
            cl = G.StartNode(cLeft[k]);
            cr = G.EndNode(cRight[k]);

            // Check for degree 1 nodes
            if (cr==cl)
            {
                connected = false;
                continue;
            }

            // Mark all contact nodes as visited. 
            TArc a = cRight[k];
            do
            {
                TNode w = G.EndNode(a);
                visited[w] = true;

                try
                {
                    if (w!=v1 && w!=v2) Q1.Insert(w);
                }
                catch (ERCheck) {};

                a = G.Right(a,vDel);
            }
            while (a!=cRight[k]);
        }
        else if (!Q2.Empty())
        {
            TArc aDel = Q2.Delete();
            
            // We must verify that the queued arc is still the first outer arc!

            if (vOut[face[aDel^1]]==NoNode) continue;

            vDel = G.EndNode(aDel);
            fDel = face[aDel^1];
            
            // Check if fDel is really admissible for deletion
            if (   vOut[fDel]>aOut[fDel]+1
                || aOut[fDel]<2
                || (face[aDel^1]==fBasis && aOut[fBasis]!=vOut[fBasis])
               )
            {
                continue;
            }

            cLeft[k] = aDel;
            vFirst[k] = vDel;
            TNode vPrev = NoNode;

            #if defined(_LOGGING_)

            THandle LH = NoHandle;

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"Component %ld (Path): %ld",k,vDel);
                LH = LogStart(LOG_METH2,CT.logBuffer);
            }

            #endif
            
            while (face[G.Right(aDel^1,vDel)^1]==fDel)
            {
                // Check for degree 2 nodes
                if (!visited[vDel]) connected = false;

                aDel = G.Right(aDel^1,vDel);
                vPrev = vDel;
                vDel = G.EndNode(aDel);
                
                if (vDel==v2) break; // Only needed in the final iteration
                
                if (face[G.Right(aDel^1,vDel)^1]==fDel)
                {
                    vRight[vPrev] = vDel;

                    #if defined(_LOGGING_)

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,",%ld",vDel);
                        LogAppend(LH,CT.logBuffer);
                    }

                    #endif
                }
            }

            #if defined(_LOGGING_)

            if (CT.logMeth>1) LogEnd(LH);

            #endif
            
            if (!connected) continue;
            
            vRight[vPrev] = NoNode;
            cRight[k] = aDel;

            cl = G.StartNode(cLeft[k]);
            cr = G.EndNode(cRight[k]);
            

            // fDel contains a separation pair (the contact nodes cl and cr).
            // Traverse the new part of the exterior face and update fSep[]

            fSep[cr]--;

            #if defined(_TRACING_)

            if (CT.traceLevel>1) G.SetColour(cr,fSep[cr]);
            
            #endif

            TArc a = cLeft[k]^1;
            TNode v = cl;
            while (v!=cr)
            {
                fSep[v]--;

                #if defined(_TRACING_)

                if (CT.traceLevel>1) G.SetColour(v,fSep[v]);
                
                #endif

                a = G.Right(a^1,v);
                v = G.EndNode(a);
            }

            // Mark all contact nodes as visited
            visited[cl] = visited[cr] = true;
            
            // cl and cr have been visited and may become feasible for
            // deletion. The other nodes which are incident with the deleted
            // face have not been visited yet!

            if (cl!=v1 && fSep[cl]==0)
            {
                try
                {
                    Q1.Insert(cl);
                }
                catch (ERCheck) {};
            }

            if (cr!=v2 && fSep[cr]==0)
            {
                try
                {
                    Q1.Insert(cr);
                }
                catch (ERCheck) {};
            }
        }
        else
        {
            connected = false;
            continue;
        }
        

        // Cancel component from G instead of deleting it (preserve indices).

        TNode vNext = vFirst[k];
        G.CancelNode(vNext);
        
        while (vRight[vNext]!=NoNode)
        {
            vNext = vRight[vNext];
            G.CancelNode(vNext);
        }


        // Start updating vOut[], aOut[], fSep[] and collect candidates for Q2

        TArc a = G.Right(G.First(cl),cl);
        TNode u = cl;
        TNode v = G.EndNode(a);
        TNode f = face[a^1];

        // Very special case: cl and cr, but no other nodes and arcs of f
        // were exterior before this iteration

        if (vOut[f]==2 && aOut[f]==0 && v==cr)
        {
            // The face f is no longer a separating face.
            // Adjust fSep[] for all nodes incident with f

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"Not a separator face: %ld",f);
                LogEntry(LOG_METH2,CT.logBuffer);
            }

            #endif

            TArc a1 = a^1;
            do
            {
                TNode w = G.EndNode(a1);
                fSep[w]--;

                #if defined(_TRACING_)

                if (CT.traceLevel>1) G.SetColour(w,fSep[w]);
                
                #endif

                a1 = G.Right(a1^1,w);
            }
            while (a1!=(a^1));
            
            // cl and cr have been visited and may become feasible for
            // deletion. The other nodes which are incident with the face f
            // are interior and have not been visited yet!

            if (cl!=v1 && fSep[cl]==0)
            {
                try
                {
                    Q1.Insert(cl);
                }
                catch (ERCheck) {};
            }

            if (cr!=v2 && fSep[cr]==0)
            {
                try
                {
                    Q1.Insert(cr);
                }
                catch (ERCheck) {};
            }
        }


        // The left hand arc a0 on the exterior face is a candidate for Q2

        TArc a0 = G.Right(a,cl);
        f = face[a0];

        if (   G.Right(a0,cl)==a  // cl has degree 2
            && f!=face[aBasis]    // Do not delete the basis arc
            && f!=face[G.First(EndNode(a0))]
                                  // The left hand node of cl has degree >2
           )
        {
            sprintf(CT.logBuffer,"Added to Q2: %ld",a0^1);
            LogEntry(LOG_METH2,CT.logBuffer);
            try
            {
                Q2.Insert(a0^1);
            }
            catch (ERCheck) {};
        }


        // Traverse the new part of the exterior face and update all data
        
        #if defined(_LOGGING_)

        THandle LH = NoHandle;

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"New exterior path: (%ld",cl);
            LH = LogStart(LOG_METH2,CT.logBuffer);
        }

        #endif
        
        while (u!=cr)
        {
            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer," [%ld] %ld",a,v);
                LogAppend(LH,CT.logBuffer);
            }

            #endif

            if (a!=G.Right(G.Right(a,u),u))
            {
                // The start node of arc a has degree>2. After deleting
                // the component, the new exterior segment of face[a^1]
                // starts with this arc.
                Q2.Insert(a);
            }
            
            vOut[face[a]] = NoNode;
            face[a] = fOut;
            f = face[a^1];
            
            TArc aNext = G.Right(a^1,v);
            
            if (prevChange[f]!=k && vOut[f]==aOut[f]+1 && aOut[f]<2)
            {
                // f was not a separating face before the deletion of this
                // component. Adjust fSep[] after aOut[] and vOut[] have
                // been updated completely. For the time being, add to Q3
                prevChange[f] = k;
                Q3.Insert(a^1);
            }
            
            aOut[f]++;
            
            // By doing this, v is recognized as an external node later:
            G.SetFirst(v,a^1);

            TArc a2 = aNext;
            while (v!=cr && a2!=(a^1))
            {
                f = face[a2^1];

                if (   prevChange[f]!=k
                    && vOut[f]==aOut[f]+1 && aOut[f]<2
                    && G.Right(a2,v)!=(a^1)
                   )
                {
                    // f was not a separating face before the deletion of this
                    // component. Adjust fSep[] after aOut[] and vOut[] have
                    // been updated completely. For the time being, add to Q3
                    prevChange[f] = k;
                    Q3.Insert(a2^1);
                }

                vOut[f]++;

                a2 = G.Right(a2,v);
            }
            
            a = aNext;
            u = G.StartNode(a);
            v = G.EndNode(a);
        }

        #if defined(_LOGGING_)

        if (CT.logMeth>1) LogEnd(LH,")");

        #endif

        while (!Q3.Empty())
        {
            TArc a2 = Q3.Delete();
            TNode f = face[a2];

            if (vOut[f]>aOut[f]+1 || aOut[f]>1)
            {
                // The face f has become a separating face.
                // Adjust fSep[] for all nodes incident with f

                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,"New separator face: %ld",f);
                    LogEntry(LOG_METH2,CT.logBuffer);
                }
                
                #endif
                
                TArc a3 = a2;
                do
                {
                    TNode u = G.EndNode(a3);
                    fSep[u]++;

                    #if defined(_TRACING_)

                    if (CT.traceLevel>1) G.SetColour(u,fSep[u]);
                    
                    #endif

                    a3 = G.Right(a3^1,u);
                }
                while (a3!=a2);
            }
        }
        
        
        // In the final step, the face incident with aBasis must be deleted
        if (aOut[fBasis]==vOut[fBasis])
        {
            Q2.Insert(G.Right(aBasis,v1));
        }
            

        #if defined(_LOGGING_)

        for (TNode f=0;f<nd;f++)
        {
            if (f==fOut || vOut[f]==NoNode || vOut[f]==0) continue;

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"vOut[%ld] = %ld, aOut[%ld] = %ld",
                    f,vOut[f],f,aOut[f]);
                LogEntry(LOG_METH2,CT.logBuffer);
            }
        }

        for (TNode v=0;v<n;v++)
        {
            if (G.First(v)==NoArc || fSep[v]==0) continue;

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"fSep[%ld] = %ld",
                    v,fSep[v]);
                LogEntry(LOG_METH2,CT.logBuffer);
            }
        }

        #endif

        k++;

        CT.Ping(G.Handle(),n);
    }


    // Final component consists of v1 and v2
    
    vFirst[k] = v1;
    vRight[v1] = v2;
    vRight[v2] = NoNode;

    #if defined(_LOGGING_)

    if (connected && CT.logMeth>1)
    {
        sprintf(CT.logBuffer,"Component %ld (Basis): %ld,%ld",k,v1,v2);
        LogEntry(LOG_METH2,CT.logBuffer);
    }

    #endif

    k++;
    
    
    // Assign the node colours with the canonical ordering

    if (connected)
    {
        InitColours();

        TNode nr = 0;
        for (TNode l=k;l>0;l--)
        {
            TNode vNext = vFirst[l-1];
            colour[vNext] = nr;

            while (vRight[vNext]!=NoNode)
            {
                vNext = vRight[vNext];
                colour[vNext] = nr;
            }
            
            nr++;
        }
    }


    if (connected) CT.Ping(Handle(),m);

    CloseFold();

    delete[] visited;
    delete[] fSep;
    delete[] vOut;
    delete[] aOut;
    delete[] prevChange;

    CloseFold(ModLMCOrder);

    if (connected)
    {
        LogEntry(LOG_RES, "...Graph has been decomposed");
    }
    else
    {
        LogEntry(LOG_RES, "...Graph is not 3-connected");

        throw ERRejected();
    }
    
    return k;
}


void abstractMixedGraph::StraightLineDrawing(TArc aBasis) throw (ERRejected)
{
    // Place the nodes on an integer grid.
    // All edges are displayed as straight lines without crossings.
    // Input graph must be planar and triconnected

    #if defined(_FAILSAVE_)

    if (aBasis>=2*m && aBasis!=NoArc) NoSuchArc("StraightLineDrawing",aBasis);

    #endif
    
    OpenFold(ModPlanarity,NO_INDENT);

    LogEntry(LOG_METH,"Embedding the graph nodes...");
    OpenFold();
    
    if (aBasis!=NoArc) aExterior = aBasis;
    
    for (TArc a=0;a<m;a++) ReleaseBendNodes(2*a);

    // Data structures associated with the ordered partition.
    // Number k of components is not known a priori but bounded by n.
    
    // Leftmost / rightmost contact arc of each component
    TArc* cLeft  = new TArc[n]; // Directed towards the component
    TArc* cRight = new TArc[n]; // Directed away from the component

    // Leftmost node of each component
    TNode* vFirst = new TNode[n];
    
    // Right hand node of each graph node in its component
    TNode* vRight = new TNode[n];


    TNode k = 0;
    
    // Compute the shelling order. Exit if graph is not triconnected
    try
    {
        k = LMCOrderedPartition(cLeft,cRight,vFirst,vRight);
    }
    catch (ERRejected)
    {
        CloseFold();

        delete[] cLeft;
        delete[] cRight;
        delete[] vFirst;
        delete[] vRight;
        
        throw ERRejected();
    }
    
    
    LogEntry(LOG_METH,"Node placement...");
    OpenFold();
    
    TFloat sep = CT.nodeSep;
    
    // The end nodes of the basis arc
    TNode v1 = StartNode(aExterior);
    TNode v2 = EndNode(aExterior);

    // Clockwise path on the exterior face of the currently embedded subgraph
    TNode* cNext = new TNode[n];
    cNext[v1] = v2;

    // For exterior nodes, the amount of shift in the current iteration
    TFloat* shift = new TFloat[n];
    for (TNode v=0;v<n;v++) shift[v] = 0;
    
    // The embedded subgraph splits into node sets each of which includes
    // exactly one exterior node. For an interior node v, this is exterior[v]
    // and v is shifted with exterior[v]
    TNode* exterior = new TNode[n];
    for (TNode v=0;v<n;v++) exterior[v] = NoNode;

    
    // Only for display: Hide all nodes initially
    for (TNode v=0;v<n;v++) SetC(v,0,-InfFloat);

    // Initial placement of the basis nodes v1 and v2
    SetC(v1,0,0);
    SetC(v1,1,0);
    SetC(v2,0,0);
    SetC(v2,1,0);
    
    for (TNode l=k-1;l>0;l--)
    {
        // Determine the cardinality of the new component
        TNode vNext = vFirst[l-1];
        TNode q = 0;

        while (vNext!=NoNode)
        {
            vNext = vRight[vNext];
            q++;
        }


        // Set the right hand shift of the contact nodes
        TNode vl = StartNode(cLeft[l-1]);
        TNode vr = EndNode(cRight[l-1]);
        TNode v = vl;
        TArc a = cLeft[l-1]^1;
        
        do
        {
            a = Right(a^1,v);
            if (EndNode(a)==vFirst[l-1]) a = Right(a,v);
            v = EndNode(a);

            if (v==vr)
            {
                shift[v] = 2*q;
            }
            else
            {
                shift[v] = q;
                exterior[v] = vFirst[l-1];
            }
        }
        while (v!=vr);

        // Set the shift of the exterior nodes on the right-hand
        // side of the new component
        while (v!=v2)
        {
            v = cNext[v];
            shift[v] = 2*q;
        }

        // Shift the nodes where necessary
        for (v=0;v<n;v++)
        {
            TFloat thisShift = shift[v];
            
            // Nodes which were interior before this iteration
            if (exterior[v]!=NoNode && thisShift==0)
            {
                thisShift = shift[exterior[v]];

                // Path compression
                if (exterior[exterior[v]]!=NoNode)
                    exterior[v] = exterior[exterior[v]];
            }

            if (thisShift>0)
            {
                SetC(v,0,C(v,0)+thisShift*sep);

                #if defined(_LOGGING_)
    
                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,"Shift node %ld to (%g,%g)",
                        v,C(v,0),C(v,1));
                    LogEntry(LOG_METH2,CT.logBuffer);
                }

                #endif
            }
        }

        // Reset the shift for nodes which become interior by this iteration
        v = vl;
        while (cNext[v]!=vr)
        {
            v = cNext[v];
            shift[v] = 0;
        }
        

        // Place the k-th component and update cNext
        TFloat xl = C(vl,0);
        TFloat yl = C(vl,1);
        TFloat xr = C(vr,0)-2*(q-1)*sep;
        TFloat yr = C(vr,1);

        TFloat xNew = xl + (xr+yr-xl-yl)/2;
        TFloat yNew = yl + (xr+yr-xl-yl)/2;
        
        vNext = vFirst[l-1];
        cNext[vl] = vNext;
        TNode vOld = NoNode;

        while (vNext!=NoNode)
        {
            SetC(vNext,0,xNew);
            SetC(vNext,1,yNew);

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"Insert node %ld at (%g,%g)",
                    vNext,xNew,yNew);
                LogEntry(LOG_METH2,CT.logBuffer);
            }

            #endif

            vOld = vNext;
            cNext[vNext] = vRight[vNext];
            vNext = vRight[vNext];
            xNew += 2*sep;
        }
        
        cNext[vOld] = vr;

        // Reset the shift right-hand of vl on the updated exterior path
        v = vl;
        do
        {
            v = cNext[v];
            shift[v] = 0;
        }
        while (v!=v2);

        CT.Ping(Handle(),n);
    }

    CloseFold();

    sprintf(CT.logBuffer,"...Grid size: %ldx%ld",
        TNode(C(v2,0)),TNode(C(vFirst[0],1)));
    LogEntry(LOG_RES,CT.logBuffer);

    delete[] cLeft;
    delete[] cRight;
    delete[] vFirst;
    delete[] vRight;
    delete[] cNext;
    delete[] shift;
    delete[] exterior;

    if (IsSparse()) AutoArcAlignment();

    CloseFold();
    LogEntry(LOG_RES, "...Straight line line drawing found");
    CloseFold(ModPlanarity,NO_INDENT);
}
