
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, July 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsgra.h"
#include "spsdig.h"
#include "dnsbig.h"


void abstractMixedGraph::ChinesePostman() throw(ERRejected)
{
    Error(ERR_REJECTED,"ChinesePostman",
        "Method applies to sparse undirected and sparse directed graphs only");

    throw ERRejected();
}


void abstractGraph::ChinesePostman() throw()
{
    OpenFold(ModPostman,NO_INDENT);
    
    LogEntry(LOG_METH,"Computing maximum Eulerian subgraph...");
    OpenFold();
    
    LogEntry(LOG_METH,"Eliminating negative length labels...");
    
    graph G(*this,OPT_CLONE);
    
    for (TArc a=0;a<m;a++) AddArc(2*a,UCap(2*a)-Sub(2*a));
    
    for (TNode v=0;v<n;v++)
    {
    	if (((TArc)Deg(v))%2) G.SetDemand(v,1);
        else G.SetDemand(v,0);
    }
    
    for (TArc a=0;a<2*m;a++)
    {
        if (Length(a)<0)
        {
            TNode v = StartNode(a);

            if (G.Demand(v)==1) G.SetDemand(v,0);
            else G.SetDemand(v,1);

            if (a&1) G.SetLength(a,-Length(a));
        }
    }
    
    G.ComputeTJoin();
    
    LogEntry(LOG_METH,"Constructing maximal Eulerian subgraph...");
    
    for (TArc a=0;a<m;a++)
    {
        TArc a2 = 2*a;

        OmitArc(a2,Sub(a2));

        if ((G.Sub(a2)>0 && Length(a2)>=0) || (G.Sub(a2)==0 && Length(a2)<0))
            AddArc(a2,UCap(a2)-1);
        else AddArc(a2,UCap(a2));
    }
    
    CloseFold();

    CloseFold(ModPostman,NO_INDENT);
}


void graph::ChinesePostman() throw()
{
    OpenFold(ModPostman,NO_INDENT);
    
    LogEntry(LOG_METH,"Computing minimum Eulerian supergraph...");
    OpenFold();
    
    LogEntry(LOG_METH,"Eliminating negative length labels...");
    
    graph G(*this,OPT_CLONE);
    
    for (TArc a=0;a<m;a++) AddArc(2*a,UCap(2*a)-Sub(2*a));
    
    for (TNode v=0;v<n;v++)
    {
    	if (((TArc)Deg(v))%2) G.SetDemand(v,1);
        else G.SetDemand(v,0);
    }
    
    for (TArc a=0;a<2*m;a++)
    {
        if (Length(a)<0)
        {
            TNode v = StartNode(a);

            if (G.Demand(v)==1) G.SetDemand(v,0);
            else G.SetDemand(v,1);

            if (a&1) G.SetLength(a,-Length(a));
        }
    }
    
    G.ComputeTJoin();
    
    LogEntry(LOG_METH,"Adjusting arc capacities...");
    
    for (TArc a=0;a<m;a++)
    {
        TArc a2 = 2*a;

        OmitArc(a2,Sub(a2));

        if ((G.Sub(a2)>0 && Length(a2)>=0) || (G.Sub(a2)==0 && Length(a2)<0))
        {
            AddArc(a2,UCap(a2)-1);
            SetUCap(a2,UCap(a2)+1);   
        }
        else AddArc(a2,UCap(a2));
    }
    
    CloseFold();

    CloseFold(ModPostman,NO_INDENT);
}


void diGraph::ChinesePostman() throw()
{
    OpenFold(ModPostman,NO_INDENT);
    
    LogEntry(LOG_METH,"Computing minimum Eulerian supergraph...");

    OpenFold();
    
    TNode* map1 = new TNode[n];
    TNode* map2 = new TNode[n];
    
    TNode n1 = 0;
    TNode n2 = 0;
    TFloat deficiency = 0;
    
    for (TArc a=0;a<m;a++) AddArc(2*a,UCap(2*a)-Sub(2*a));
    
    LogEntry(LOG_METH,"Computing unbalanced nodes...");
    OpenFold();

    for (TNode v=0;v<n;v++)
    {
    	if (Divergence(v)>0)
        {
            map1[v] = n1++;
            deficiency += Divergence(v);

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"Divergence(%ld) = %g",v,Divergence(v));
                LogEntry(LOG_METH2,CT.logBuffer);
            }

            #endif
        }
        else map1[v] = NoNode;
    }
    
    for (TNode v=0;v<n;v++)
    {
    	if (Divergence(v)<0)
        {
            map2[v] = n1+(n2++);

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"Divergence(%ld) = %g",v,Divergence(v));
                LogEntry(LOG_METH2,CT.logBuffer);
            }

            #endif
        }
        else map2[v] = NoNode;
    }

    TNode n0 = n1+n2;
    
    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Problem has %ld unbalanced nodes",n0);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    CloseFold();
    
    if (n0==0)
    {
        delete[] map1;
        delete[] map2;
        CloseFold();
        return;
    }

    LogEntry(LOG_METH,"Transforming into matching problem...");
    OpenFold();
    
    denseBiGraph *G = new denseBiGraph(n1,n2,0,CT);
    denseGraphStructure &Y = G->Struct();
    Y.NewSubgraph((TArc)deficiency);
    
    TNode *revmap = new TNode[n0];
    TCap *dgl = new TCap[n0];
    
    for (TNode u=0;u<n;u++)
    {
        if (map2[u]!=NoNode)
        {
            if (Dim()>=2)
            {
                Y.SetC(map2[u],0,C(u,0));
                Y.SetC(map2[u],1,C(u,1));
            }
            
            revmap[map2[u]] = u;
            dgl[map2[u]] = TCap(-Divergence(u));
        }
        
        if (map1[u]!=NoNode)
        {
            if (Dim()>=2)
            {
                Y.SetC(map1[u],0,C(u,0));
                Y.SetC(map1[u],1,C(u,1));
            }
            
            revmap[map1[u]] = u;
            dgl[map1[u]] = TCap(Divergence(u));
        }
    }

    for (TArc a=0;a<m;a++) OmitArc(2*a,Sub(2*a));
    
    for (TNode u=0;u<n;u++)
    {
        if (map1[u]!=NoNode)
        {
            LogEntry(LOG_METH,"Computing shortest path tree...");

            OpenFold();

            Dijkstra(u);

            CloseFold();

            for (TNode v=0;v<n;v++)
            {
            	if (map2[v]!=NoNode && d[v]<InfFloat)
                {
                    TArc a = G->Adjacency(map1[u],map2[v]);
                    Y.SetLength(a,d[v]);
                }
            }
        }
    }

    CloseFold();
    CT.Ping(G->Handle(),n*m);

    G -> MinCAssignment(dgl);

    LogEntry(LOG_METH,"Adjusting arc capacities...");
    OpenFold();
    
    for (TArc a=0;a<G->M();a++)
    {
        if (G->Sub(2*a)>0)
        {
            TNode u = revmap[G->StartNode(2*a)];
            TNode v = revmap[G->EndNode(2*a)];

            LogEntry(LOG_METH,"Computing shortest path tree...");

            OpenFold();

            Dijkstra(u);

            CloseFold();

            if (P[v]==NoArc)
            {
                Error(MSG_WARN,"ChinesePostman","Digraph is not strongly connected");
    
                delete G;
                delete[] map1;
                delete[] map2;
                delete[] revmap;
		
                throw ERRejected();
            }
	    
            #if defined(_LOGGING_)

            THandle LH = NoHandle;

            if (CT.logMeth>1)
            {
                LogEntry(LOG_METH2,"Adding the path (in reverse order):");
                sprintf(CT.logBuffer,"(%ld",v);
                LH = LogStart(LOG_METH2,CT.logBuffer);
            }

            #endif

            TCap delta = (TCap)G->Sub(2*a);

            while (v!=u)
            {
                TArc a2 = P[v];
                X.SetUCap(a2,UCap(a2)+delta);
                v = StartNode(a2);

                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,",%ld",v);
                    LogAppend(LH,CT.logBuffer);
                }

                #endif
            }

            #if defined(_LOGGING_)

            if (CT.logMeth>1) LogEnd(LH,")");

            #endif
        }
    }
    
    CloseFold();
    
    CT.Ping(Handle(),n*m);
    delete G;
        
    delete[] map1;
    delete[] map2;
    delete[] revmap;
    
    CloseFold();

    CloseFold(ModPostman,NO_INDENT);
}


bool abstractMixedGraph::EulerCycle() throw(ERRejected)
{
    TArc* pred = new TArc[m];

    bool eulerian = EulerCycle(pred);
    
    if (eulerian)
    {
        TArc a = pred[0];
        for (TArc i=1;i<=m;i++)
        {
            SetEdgeColour(a,m-i);
            a = pred[a>>1];
        }
    }
    
    delete[] pred;
    
    return eulerian;
}


bool abstractMixedGraph::EulerCycle(TArc* pred) throw()
{
    LogEntry(LOG_METH,"Computing an Eulerian cycle...");

    OpenFold(ModHierholzer,SHOW_TITLE);

    TArc count = 0;

    for (TArc a=0;a<m;a++) pred[a] = NoArc;

    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    // Start with an arbitrary arc
    TArc a = 0;

    #if defined(_LOGGING_)

    THandle LH = NoHandle;

    if (CT.logMeth>1)
    {
        sprintf(CT.logBuffer,"Initial cycle: (%ld[%ld]",StartNode(a),a);
        LH = LogStart(LOG_METH2,CT.logBuffer);
    }

    #endif

    // Find a closed walk through arc 0
    while (true)
    {
        TNode u = EndNode(a);

        if (I.Active(u))
        {
            TArc a2 = I.Read(u);

            if (pred[a2>>1]!=NoArc || (a2>>1)==0 || Blocking(a2)) continue;
            
            pred[a2>>1] = a;
            a = a2;
            count++;

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"%ld[%ld]",u,a);
                LogAppend(LH,CT.logBuffer);
            }

            #endif
        }
        else break;
    }

    #if defined(_LOGGING_)

    if (CT.logMeth>1)
    {
        sprintf(CT.logBuffer,"%ld)",EndNode(a));
        LogEnd(LH,CT.logBuffer);
    }

    #endif
    
    if (EndNode(a)!=StartNode(0))
    {
        Close(H);
        CloseFold(ModHierholzer);
        LogEntry(LOG_RES,"...Graph is not Eulerian");
        return false;
    }
    else
    {
        pred[0] = a;
        count++;
    }
    
    // Track back on the cycle found so far.
    while (a != 0)
    {
        TNode u = StartNode(a);

        if (I.Active(u))
        {
            TArc a2 = I.Read(u);

            if (pred[a2>>1]!=NoArc || a2==0 || Blocking(a2)) continue;
            
            // Find a closed walk through arc a2

            #if defined(_LOGGING_)

            THandle LH = NoHandle;

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"Inserting before %ld: (%ld[%ld]",
                    a,StartNode(a2),a2);
                LH = LogStart(LOG_METH2,CT.logBuffer);
            }

            #endif

            pred[a2>>1] = pred[a>>1];
            count++;

            while (true)
            {
                TNode u = EndNode(a2);

                if (I.Active(u))
                {
                    TArc a3 = I.Read(u);

                    if (pred[a3>>1]!=NoArc || Blocking(a3)) continue;

                    pred[a3>>1] = a2;
                    a2 = a3;
                    count++;

                    #if defined(_LOGGING_)

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,"%ld[%ld]",u,a2);
                        LogAppend(LH,CT.logBuffer);
                    }

                    #endif
                }
                else break;
            }

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"%ld)",EndNode(a2));
                LogEnd(LH,CT.logBuffer);
            }

            #endif

            if (EndNode(a2)!=StartNode(a))
            {
                Close(H);
                CloseFold(ModHierholzer);
                LogEntry(LOG_RES,"...Graph is not Eulerian");
                return false;
            }
            else
            {
                pred[a>>1] = a2;
            }
        }
        else a = pred[a>>1];
    }
    
    Close(H);

    if (count<m)
    {
        CloseFold(ModHierholzer);
        LogEntry(LOG_RES,"...Graph is disconnected");
        return false;
    }

    CloseFold(ModHierholzer);
    LogEntry(LOG_RES,"...Graph is Eulerian");
    
    return true;
}
