
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, July 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "absdig.h"
#include "statq.h"


bool abstractMixedGraph::BFS(TNode s,TNode t) throw(ERRange)
{                       
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("BFS",s);

    if (t>=n && t!=NoNode) NoSuchNode("BFS",t);

    #endif
    
    OpenFold(ModSPTree,NO_INDENT);
    
    InitLabels();
    InitPredecessors();

    d[s] = 0;
    staticQueue<TNode,TFloat> Q(n,CT);
    Q.Insert(s);

    LogEntry(LOG_METH,"Graph is searched...");
    OpenFold();
    
    #if defined(_LOGGING_)

    THandle LH = LogStart(LOG_METH2,"Expanded nodes: ");

    #endif
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    bool searching = true;
    while (!(Q.Empty()) && searching)
    {
        CT.Ping(Q.Handle(),1); 
        TNode u = Q.Delete();

        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"%ld ",u);
            LogAppend(LH,CT.logBuffer);
        }

        #endif
        
        while (I.Active(u))
        {
            TArc a = I.Read(u);
            TNode v = EndNode(a);

            if (d[v]==InfFloat && Eligible(a))
            {               
                d[v] = d[u]+1;
                CT.Ping(Q.Handle(),1); 
                Q.Insert(v);
                P[v] = a;

                if (v==t) searching = false;
            }
        }
    }

    #if defined(_LOGGING_)

    if (CT.logMeth>1) LogEnd(LH);

    #endif

    Close(H);

    CloseFold();

    SetBounds(TimerSPTree,d[t],d[t]);
    CT.Ping(Handle(),m); 

    CloseFold(ModSPTree,NO_INDENT);
    
    if (t==NoNode) return 0;
    else return (d[t]!=InfFloat);
}


bool abstractMixedGraph::ShortestPath(TNode s,TNode t)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("ShortestPath",s);

    if (t>=n && t!=NoNode) NoSuchNode("ShortestPath",t);

    #endif
    
    OpenFold(ModSPTree,NO_INDENT);
    
    LogEntry(LOG_METH,"Computing distance labels and shortest path tree...");

    bool ret = false;

    switch (CT.methSearch)
    {
        case 0:
        {
            ret = FIFOLabelCorrecting(s,t);
            break;
        }
        case 1:
        {
            ret = Dijkstra(s,t);
            break;
        }
        case 2:
        {
            ret = BellmanFord(s,t);
            break;
        }
        case 3:
        {
            #if defined(_FAILSAVE_)

            if (!CLength() || MaxLength()<0)
            {
                CloseFold(ModSPTree,NO_INDENT);
                Error(ERR_REJECTED,"ShortestPath","Non-trivial length labels");
            }

            #endif

            ret = CLength()*BFS(s,t);
            break;
        }
        case 4:
        {
            TNode x = DAGSearch(DAG_SPTREE,s,t);
            
            if (x!=NoNode)
            {
                #if defined(_FAILSAVE_)

                CloseFold(ModSPTree,NO_INDENT);
                Error(ERR_REJECTED,"ShortestPath","Not a DAG");

                #endif
            }

            ret = (t==NoNode || d[t]<InfFloat);
            
            break;
        }
        default:
        {
            CloseFold(ModSPTree,NO_INDENT);
            UnknownOption("ShortestPath",CT.methSearch);
        }
    }
    
    CloseFold(ModSPTree,NO_INDENT);

    return ret;
}


bool abstractMixedGraph::Dijkstra(TNode s,TNode t)
    throw(ERRange,ERRejected)
{                       
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("Dijkstra",s);

    if (t>=n && t!=NoNode) NoSuchNode("Dijkstra",t);

    #endif
    
    OpenFold(ModDikjstra,SHOW_TITLE);
    
    InitLabels();
    InitPredecessors();

    goblinQueue<TNode,TFloat> *Q = NULL;

    if (nHeap!=NULL)
    {
        Q = nHeap;
        Q -> Init();
    }
    else Q = NewNodeHeap();
    
    Q -> Insert(s,0);
    d[s] = 0;
    
    #if defined(_LOGGING_)
    
    THandle LH = LogStart(LOG_METH2,"Expanded nodes:");
    
    #endif
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    while (!(Q->Empty()))
    {
        TNode u = Q->Delete();

        if (u==t) break;
        
        #if defined(_LOGGING_)

        if (CT.logMeth>1 && I.Active(u))
        {
            sprintf(CT.logBuffer," %ld[%g]",u,d[u]);
            LogAppend(LH,CT.logBuffer);
        }

        #endif

        while (I.Active(u))
        {
            TArc a = I.Read(u);
            TFloat l = RedLength(a);
            TNode v = EndNode(a);
            TFloat dd = d[u]+l;

            if (d[v]>dd && Eligible(a))
            {               
                #if defined(_FAILSAVE_)

                if (l<0)
                {
                    CloseFold(ModDikjstra);
                    Error(ERR_REJECTED,"Dijkstra","Negative arc length");
                }

                #endif

                if (d[v]==InfFloat) Q->Insert(v,dd);
                else Q->ChangeKey(v,dd);

                d[v] = dd;
                P[v] = a;
            }
        }
    }

    #if defined(_LOGGING_)

    if (CT.logMeth>1) LogEnd(LH);

    #endif

    Close(H);

    if (nHeap==NULL) delete Q;

    if (t!=NoNode) SetBounds(TimerSPTree,d[t],d[t]);

    CT.Ping(Handle(),m); 

    CloseFold(ModDikjstra);

    if (t==NoNode) return 0;
    else return (d[t]!=InfFloat);
}


TNode abstractMixedGraph::VoronoiRegions()
    throw(ERRejected)
{                       
    LogEntry(LOG_METH,"Computing Voronoi regions...");
    OpenFold();
    
    InitLabels();
    InitPredecessors();
    InitPartition();

    goblinQueue<TNode,TFloat> *Q = NULL;

    if (nHeap!=NULL)
    {
        Q = nHeap;
        Q -> Init();
    }
    else Q = NewNodeHeap();
    
    TNode nTerminals = 0;

    for (TNode i=0;i<n;i++)
    {
        Bud(i);

        if (Demand(i)!=0)
        {
            nTerminals ++;
            Q -> Insert(i,0);
            d[i] = 0;
        }
    }
    
    #if defined(_LOGGING_)
    
    THandle LH = LogStart(LOG_METH2,"Expanded nodes:");
    
    #endif
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    while (!(Q->Empty()))
    {
        TNode u = Q->Delete();
        
        #if defined(_LOGGING_)

        if (CT.logMeth>=2 && I.Active(u))
        {
            sprintf(CT.logBuffer," %ld[%g]",u,d[u]);
            LogAppend(LH,CT.logBuffer);
        }

        #endif

        while (I.Active(u))
        {
            TArc a = I.Read(u);
            TFloat l = RedLength(a);
            TNode v = EndNode(a);
            TFloat dd = d[u]+l;

            if (d[v]>dd && Eligible(a))
            {               
                #if defined(_FAILSAVE_)

                if (l<0) Error(ERR_REJECTED,"ComputeRegions","Negative arc length");

                #endif

                if (d[v]==InfFloat) Q->Insert(v,dd);
                else Q->ChangeKey(v,dd);

                d[v] = dd;
                P[v] = a;
            }
        }
    }

    #if defined(_LOGGING_)

    if (CT.logMeth>1) LogEnd(LH);

    #endif

    Close(H);

    if (nHeap==NULL) delete Q;

    CloseFold();
    CT.Ping(Handle(),m); 

    for (TNode i=0;i<n;i++)
    {
        if (P[i]==NoArc) continue;

        TNode u = Find(StartNode(P[i]));
        TNode v = Find(i);

        if (u!=v) Merge(u,v);
    }

    return nTerminals;
}


bool abstractMixedGraph::BellmanFord(TNode s,TNode t)
    throw(ERRange,ERCheck)
{                       
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("BellmanFord",s);

    if (t>=n && t!=NoNode) NoSuchNode("BellmanFord",t);

    #endif
    
    OpenFold(ModBellmanFord,SHOW_TITLE);

    InitLabels();
    InitPredecessors();

    d[s] = 0;
    TNode i = 1;
    bool Updates = true;

    while (Updates && i<2*n)
    {
        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"Iteration %ld:",i);
            LogEntry(LOG_METH2,CT.logBuffer);
        }

        #endif
        
        OpenFold();
        Updates = 0;

        for (TArc a=0;a<2*m;a++)
        {
            TNode u = StartNode(a);
            TFloat thisLabel = d[u] + Length(a);
            TNode v = EndNode(a);

            if (d[v]>thisLabel && d[u]!=InfFloat && P[u]!=(a^1) && Eligible(a))
            {               
                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,"Label d[%ld] updated to %g",
                        v,thisLabel);
                    LogEntry(LOG_METH2,CT.logBuffer);
                }

                #endif
                
                d[v] = thisLabel;
                P[v] = a;
                Updates = 1;
            }
        }

        i++;
        CloseFold();
        CT.Ping(Handle(),m);  
    }
    
    if (i==2*n)
    {
        CloseFold(ModBellmanFord);
        Error(ERR_CHECK,"BellmanFord","Negative length cycles");
    }
    
    if (t==NoNode)
    {
        CloseFold(ModBellmanFord);
        return false;
    }
    
    if (d[t]!=InfFloat) SetBounds(TimerSPTree,d[t],d[t]);
    else SetBounds(TimerSPTree,-InfFloat,-InfFloat);
    
    CloseFold(ModBellmanFord);

    return (d[t]!=InfFloat);
}


bool abstractMixedGraph::FIFOLabelCorrecting(TNode s,TNode t)
    throw(ERRange,ERCheck)
{                       
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("FIFOLabelCorrecting",s);

    if (t>=n && t!=NoNode) NoSuchNode("FIFOLabelCorrecting",t);

    #endif
    
    OpenFold(ModFIFOLabelCorrect,NO_INDENT);

    TNode v = NegativeCycle(s);
    
    if (v!=NoNode)
    {
        CloseFold(ModFIFOLabelCorrect,NO_INDENT);
        Error(ERR_CHECK,"FIFOLabelCorrecting","Negative length cycles");
    }

    if (t==NoNode)
    {
        CloseFold(ModFIFOLabelCorrect,NO_INDENT);
        return false;
    }
    
    OpenFold(ModFIFOLabelCorrect);

    if (d[t]!=InfFloat) SetBounds(TimerSPTree,d[t],d[t]);
    else SetBounds(TimerSPTree,-InfFloat,-InfFloat);

    CloseFold(ModFIFOLabelCorrect);

    CloseFold(ModFIFOLabelCorrect,NO_INDENT);

    return (d[t]!=InfFloat);
}


TNode abstractMixedGraph::NegativeCycle(TNode s) throw(ERRange)
{                       
    #if defined(_FAILSAVE_)

    if (s>=n && s!=NoNode) NoSuchNode("NegativeCycle",s);

    #endif
    
    OpenFold(ModFIFOLabelCorrect,SHOW_TITLE);
    
    InitLabels();
    InitPredecessors();
    
    bool* active = new bool[n];
    
    if (s!=NoNode)
    {
        d[s] = 0;
        for (TNode v=0;v<n;v++) active[v] = false;
        active[s] = true;
    }
    else
    {
        for (TNode v=0;v<n;v++)
        {
            active[v] = true;
            d[v] = 0;
        }
    }

    TNode itCount = 1;
    TNode root = NoNode;
    
    bool Updates = true;
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    while (Updates && itCount<=2*n)
    {
        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"Iteration %ld:",itCount);
            LogEntry(LOG_METH2,CT.logBuffer);
        }

        #endif
        
        OpenFold();
        Updates = 0;
        for (TNode v=0;v<n;v++)
        {
            if (active[v])
            {
                active[v] = false;

                while (I.Active(v))
                {
                    TArc a = I.Read(v);
                    TNode w = EndNode(a);
                    TFloat thisLabel = d[v] + Length(a) + mu;

                    if (d[w]>thisLabel && d[v]!=InfFloat && P[v]!=(a^1) && Eligible(a))
                    {               
                        #if defined(_LOGGING_)

                        if (CT.logMeth>1)
                        {
                            sprintf(CT.logBuffer,"Label d[%ld] updated to %g",
                                w,thisLabel);
                            LogEntry(LOG_METH2,CT.logBuffer);
                        }

                        #endif

                        d[w] = thisLabel;
                        P[w] = a;

                        if (itCount<2*n) Updates = 1;
                        else root = v;

                        active[w] = true;
                    }
                }

                I.Reset(v);
            }
        }

        itCount++;

        CloseFold();
        CT.Ping(Handle(),m);
    }

    Close(H);
    delete[] active;
    
    if (root!=NoNode)
    {
        for (TNode i=0;i<n;i++) root = StartNode(P[root]);
        
        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"...Node on a negative cycle: %ld",root);
            LogEntry(LOG_RES,CT.logBuffer);
        }

        #endif
    }

    CloseFold(ModFIFOLabelCorrect);

    return root;
}


TNode abstractDiGraph::MinimumMeanCycle() throw()
{
    OpenFold(ModKarpMeanCycle,SHOW_TITLE);
    
    InitPotentials();
    TFloat *dist = new TFloat[(n+1)*n];
    TArc* pred = new TArc[(n+1)*n];
    
    for (TNode v=0;v<n;v++) dist[0*n+v] = 0;
    
    for (TNode k=1;k<=n;k++)
    {
        for (TNode v=0;v<n;v++) dist[k*n+v] = InfFloat;
	
        for (TArc a=0;a<2*m;a++)
        {
            TNode u = StartNode(a);
            TNode v = EndNode(a);
            TFloat l = Length(a);
	    
    	    if (Eligible(a) && dist[(k-1)*n+u]!=InfFloat && (dist[k*n+v]>dist[(k-1)*n+u]+l))
	        {
            	dist[k*n+v] = dist[(k-1)*n+u]+l;
                pred[k*n+v] = a;
	        }
        }
    }
    
    TNode minNode = NoNode;
    mu = InfFloat;
    for (TNode v=0;v<n;v++)
    {
        TFloat thisMaxMean = -InfFloat;
	
        if ((dist[n*n+v])<InfFloat)
        {
            for (TNode k=0;k<n;k++)
            {
                if ((dist[k*n+v])<InfFloat)
                {
                    TFloat thisMean = (dist[n*n+v]-dist[k*n+v])/(n-k);
                    if (thisMean>thisMaxMean) thisMaxMean = thisMean;
                }
	        }
        }
	
        if (thisMaxMean<mu && thisMaxMean>-InfFloat)
        {
            mu = thisMaxMean;
            minNode = v;
        }
    }
    
    if (mu<InfFloat)
    {
        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"...Minimum ratio: %g",mu);
            LogEntry(LOG_RES,CT.logBuffer);
            LogEntry(LOG_METH2,"Extracting minimum mean cycle...");
        }

        #endif
	
        InitPredecessors();
    
        TNode k = n;

        while (P[minNode]==NoArc)
        {
            P[minNode] = pred[k*n+minNode];
            minNode = StartNode(P[minNode]);
            k--;
        }
    }
    else
    {
        minNode = NoNode;
        LogEntry(LOG_RES,"...Graph is acyclic");
    }
    
    delete[] dist;
    delete[] pred;

    CloseFold(ModKarpMeanCycle);

    return minNode;
}


TNode abstractDiGraph::TopSort() throw()
{
    LogEntry(LOG_METH,"Computing topological ordering...");
    
    OpenFold(ModSPTree,NO_INDENT);
    
    TNode ret = DAGSearch(DAG_TOPSORT);
    
    CloseFold(ModSPTree,NO_INDENT);
    
    return ret;
}


TNode abstractDiGraph::CriticalPath() throw()
{
    LogEntry(LOG_METH,"Computing critical path...");
    
    OpenFold(ModSPTree,NO_INDENT);
    
    TNode ret = DAGSearch(DAG_CRITICAL);
    
    CloseFold(ModSPTree,NO_INDENT);
    
    return ret;
}


TNode abstractDiGraph::DAGSearch(TDAGSearch opt,TNode s,TNode t)
    throw(ERRange)
{
    if (opt==DAG_TOPSORT) OpenFold(ModDAGSearch);
    else OpenFold(ModDAGSearch,SHOW_TITLE);
    
    staticQueue<TNode,TFloat> Q(n,CT);
    TArc* idg = new TArc[n];

    for (TNode v=0;v<n;v++) idg[v] = 0;

    for (TArc a=0;a<2*m;a++)
        if (Eligible(a)) 
        {
            TNode w = EndNode(a);
            idg[w]++;
        }

    #if defined(_LOGGING_)

    THandle LH = NoHandle;

    #endif

    switch (opt)
    {
        case DAG_TOPSORT:
        {
            InitColours(NoNode);
            break;
        }
        case DAG_CRITICAL:
        {
            InitPredecessors();
            InitLabels(-InfFloat);

            #if defined(_LOGGING_)

            LH = LogStart(LOG_METH2,"Expanded nodes:");

            #endif

            break;
        }
        case DAG_SPTREE:
        {
            InitColours(NoNode);
            InitLabels(InfFloat);
            InitPredecessors();
            d[s] = 0;

            #if defined(_LOGGING_)

            LH = LogStart(LOG_METH2,"Expanded nodes:");

            #endif

            break;
        }
    }

    for (TNode v=0;v<n;v++)
    {
        if (idg[v]==0)
        {
            Q.Insert(v);
            if (opt==DAG_CRITICAL) d[v] = 0;
        }
    }
    
    TNode nr = 0;
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    while (!Q.Empty())
    {
        TNode v = Q.Delete();
        
        if (opt!=DAG_CRITICAL) colour[v] = nr;
        
        nr++;

        #if defined(_LOGGING_)

        if (opt!=DAG_TOPSORT && CT.logMeth>1 && d[v]<InfFloat)
        {
            sprintf(CT.logBuffer," %ld[%g]",v,d[v]);
            LogAppend(LH,CT.logBuffer);
        }

        #endif
        
        while (I.Active(v))
        {
            TArc a = I.Read(v);

            if (!Eligible(a)) continue;

            TNode w = EndNode(a);
            idg[w]--;
            
            if ((opt==DAG_SPTREE && d[v]<InfFloat && d[w]>d[v]+Length(a)) ||
                (opt==DAG_CRITICAL && d[w]<d[v]+Length(a))
               )
            {
                d[w] = d[v]+Length(a);
                P[w] = a;
            }

            if (idg[w]==0) Q.Insert(w);
        }
    }

    #if defined(_LOGGING_)

    if (CT.logMeth>1) LogEnd(LH);

    #endif

    Close(H);
    delete[] idg;

    CT.Ping(Handle(),m);
    
    CloseFold(ModDAGSearch);

    if (CT.logRes && nr<n) LogEntry(LOG_RES,"...Graph contains cycles");

    switch (opt)
    {
        case DAG_TOPSORT:
        case DAG_SPTREE:
        {
            if (nr<n)
            {
                for (TNode v=0;v<n;v++)
                    if (colour[v]==NoNode) return v;
            }
        
            if (opt==DAG_SPTREE)
            {
                if (t!=NoNode) SetBounds(TimerSPTree,d[t],d[t]);
            }
            else
            {
                if (CT.logRes) LogEntry(LOG_RES,"...Graph is acyclic");
            }

            return NoNode;
        }
        case DAG_CRITICAL:
        {
            if (nr<n) return NoNode;
            
            TNode maxNode = NoNode;
            for (TNode v=0;v<n;v++)
            {
                if ((maxNode==NoNode && d[v]>-InfFloat) ||
                    (maxNode!=NoNode && d[v]>d[maxNode])
                   )
                {
                    maxNode = v;
                }
            }
            
            if (CT.logRes)
            {
                sprintf(CT.logBuffer,"...Minimum project time: %g",d[maxNode]);
                LogEntry(LOG_RES,CT.logBuffer);
            }

            return maxNode;
        }
    }

    return NoNode;
}
