
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, May 2001.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "bbstsp.h"
#include "spsgra.h"
#include "dnsgra.h"
#include "binheap.h"
#include "dynq.h"
    

BBSTSP::BBSTSP(abstractGraph& GC,TNode r) throw() :
    goblinBBNode<TArc,TFloat>(GC.M(),GC.Context()), G(GC)
{
    if (CT.methCandidates>=0 && G.IsDense()) SetCandidateGraph();
    else X = new graph(G,OPT_CLONE);

    H = X->Investigate();
    selected = 0;
    root = r;
    depth = TArc(X->N()*log(double(X->N())*0.1));

    for (TNode v=0;v<G.N();v++) X->SetPotential(v,G.Pi(v));

    for (TArc a=0;a<n;a++)
        if (X->StartNode(2*a)==X->EndNode(2*a)) Lower(a);

    for (TNode v=0;v<G.N();v++) CheckNode(v);
    
    LogEntry(LOG_MEM,"(symmetric TSP)");
}


BBSTSP::BBSTSP(BBSTSP& Node) throw() :
    goblinBBNode<TArc,TFloat>(Node.G.M(),Node.Context(),Node.scheme), G(Node.G)
{
    X = new graph(*Node.X,OPT_CLONE);
    H = X->Investigate();
    unfixed = Node.Unfixed();
    selected = Node.selected;
    root = Node.root;
    depth = TArc(X->N()*log(double(X->N())*0.1));
    
    for (TNode v=0;v<G.N();v++) X->SetPotential(v,G.Pi(v));

    for (TArc a=0;a<X->M();a++) X->SetSub(2*a,Node.X -> Sub(2*a));

    objective = Node.Objective();
    solved = true;
    
    LogEntry(LOG_MEM,"(symmetric TSP)");
}


BBSTSP::~BBSTSP() throw()
{
    X -> Close(H);
    delete X;
    
    LogEntry(LOG_MEM,"(symmetric TSP)");
}


void BBSTSP::SetCandidateGraph() throw()
{
    LogEntry(LOG_METH,"Constructing candidate graph...");

    CT.SuppressLogging();
    graph* Y = new graph(G,OPT_CLONE);
    Y -> SetCUCap(1);
    Y -> InitSubgraph();
    CT.RestoreLogging();

    for (TNode v=0;v<G.N();v++)
        if (G.Pred(v)!=NoArc)
        {
            TNode u = G.StartNode(G.Pred(v));
            TNode w = G.EndNode(G.Pred(v));
            Y -> AddArc(Y->Adjacency(u,w),1);
        }
        
    for (int i=0;i<20;i++)
    {
        CT.SuppressLogging();
        TFloat ret = Y->TSPHeuristicsRandom();
        CT.RestoreLogging();
        if (ret<InfFloat)
        {    
            for (TNode v=0;v<G.N();v++)
                if (Y->Sub(Y->Pred(v))==0) Y->AddArc(Y->Pred(v),1);

            #if defined(_LOGGING_)

            if (CT.logMeth>=2)
            {
                sprintf(CT.logBuffer,"Tour of length %g added",ret);
                LogEntry(LOG_METH2,CT.logBuffer);
            }
            
            #endif
        }
    }

    binaryHeap<TArc,TFloat> Q(2*Y->M(),CT);
    H = Y->Investigate();

    for (TNode v=0;v<G.N();v++)
    {
        while (Y->Active(H,v))
        {
            TArc a = Y->Read(H,v);
            Q.Insert(a,Y->Length(a));
        }
        
        int i = 0;
        while (!Q.Empty())
        {
            TArc a = Q.Delete();       
            if (Y->Sub(a)==0 && (i<CT.methCandidates || G.Sub(a)>0))
            {
                Y->AddArc(a,1);
                i++;
            }
        }
    }

    Y->Close(H);
    
    X = new graph(*Y,OPT_SUB);
    X -> SetCUCap(1);
    n = unfixed = X->M();
        
    if (CT.traceLevel==2) X -> Display();
    
    CT.SuppressLogging();
    delete Y;
    CT.RestoreLogging();

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Candidate subgraph has %ld arcs",X->M());
        LogEntry(LOG_RES,CT.logBuffer);
    }
}


unsigned long BBSTSP::Size() throw()
{
    return
          sizeof(BBSTSP)
        + goblinDataObject::Allocated()
        + goblinBBNode<TArc,TFloat>::Allocated()
        + Allocated();
}


unsigned long BBSTSP::Allocated() throw()
{
    return 0;
}


TArc BBSTSP::SelectVariable() throw()
{
    TArc ret0 = NoArc;
    TArc ret1 = NoArc;
    for (TNode v=0;v<X->N() && ret1==NoArc;v++)
    {
        char fixedIncidences = 0;
        X -> Reset(H,v);
        TArc a0 = NoArc;
        while (X->Active(H,v) && fixedIncidences<2)
        {
            TArc a = X->Read(H,v);
            if (X->Sub(a)==1 && X->LCap(a)==0 &&
                (a0==NoArc || X->Length(a)<X->Length(a0))) a0 = a;
            if (X->LCap(a)==1) fixedIncidences++;
        }
        
        if (fixedIncidences==0 && a0!=NoArc) ret0 = a0;
        if (fixedIncidences==1 && a0!=NoArc) ret1 = a0;
    }
    
    if (ret1!=NoArc) return ret1>>1;
    if (ret0!=NoArc) return ret0>>1;

    #if defined(_FAILSAVE_)

    InternalError("SelectVariable","No branching variable found");

    #endif

    throw ERInternal();
}


char BBSTSP::Direction1(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=n) NoSuchArc("Direction1",a);

    #endif
    
    return 1;
}


char BBSTSP::Direction2(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=n) NoSuchArc("Direction2",a);

    #endif
    
    return 1;
}


goblinBBNode<TNode,TFloat> *BBSTSP::Clone() throw()
{
    return new BBSTSP(*this);
}


void BBSTSP::Raise(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=n) NoSuchArc("Raise",a);

    #endif
    
    if (X->Sub(2*a)==0)
    {
        X -> AddArc(2*a,1);
        solved = 0;
    }

    X -> SetLCap(2*a,1);
    unfixed--;
    selected++;
    
    CheckNode(X->StartNode(2*a));
    CheckNode(X->EndNode(2*a));
    
    if (unfixed==0) solved = 0;
}


void BBSTSP::Lower(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=n) NoSuchArc("Lower",a);

    #endif
    
    if (X->Sub(2*a)==1)
    {
        X -> OmitArc(2*a,1);
        solved = 0;
    }

    X -> SetUCap(2*a,0);
    unfixed--;
    
    CheckNode(X->StartNode(2*a));
    CheckNode(X->EndNode(2*a));
    
    if (unfixed==0) solved = 0;
}


void BBSTSP::CheckNode(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=X->N()) NoSuchNode("CheckNode",v);

    #endif
    
    char fixedIncidences = 0;
    X -> Reset(H,v);
    while (X->Active(H,v) && fixedIncidences<2)
    {
        TArc a2 = X->Read(H,v);
        if (X->LCap(a2)==1) fixedIncidences++;
    }

    if (fixedIncidences==2)
    {
        X -> Reset(H,v);
        while (X->Active(H,v))
        {
            TArc a2 = X->Read(H,v);
            if (X->LCap(a2)==0 && X->UCap(a2)==1) Lower(a2>>1);
        }
        return;
    }

    TArc a1 = NoArc;
    TArc a2 = NoArc;
    char incidences = 0;
    X -> Reset(H,v);
    while (incidences<3 && X->Active(H,v))
    {
        TArc a = X->Read(H,v);
        if (X->UCap(a)==1)
        {
            incidences++;
            if (X->LCap(a)==0)
            {
                if (a1==NoArc) a1 = a;
                else if (a2==NoArc) a2 = a;
            }
        }
    }
    
    if (incidences<=2)
    {
        if (a1!=NoArc) Raise(a1>>1);
        if (a2!=NoArc && X->UCap(a2)==1 && X->LCap(a2)==0) Raise(a2>>1);
    }
}


TFloat BBSTSP::SolveRelaxation() throw()
{
    for (TNode v=0;v<X->N();v++)
    {
        char fixedIncidences = 0;
        X -> Reset(H,v);
        while (X->Active(H,v) && fixedIncidences<3)
        {
            TArc a2 = X->Read(H,v);
            if (X->LCap(a2)==1) fixedIncidences++;
        }
        
        if (fixedIncidences>2) return InfFloat;
    }

    CT.SuppressLogging();
    
    if (X->CutNodes())
    {
        CT.RestoreLogging();
        return InfFloat;
    }
    
    TFloat objective = InfFloat;
    try
    {
        if (scheme==NULL || scheme->Phase()==2)
        {
            X -> InitSubgraph();
            X -> ReleasePredecessors();
            objective = X->TSPSubOpt1Tree(root,InfFloat,CT.methRelaxTSP2|4);
            X -> MinTree(root,abstractMixedGraph::MTREE_1TREE);
        }
        else
        {
            objective = X->MinTree(root,abstractMixedGraph::MTREE_1TREE);
            if (scheme!=NULL && objective<scheme->bestLower && scheme->Phase()>0)
            {
                X -> InitSubgraph();
                X -> ReleasePredecessors();
                objective = X->TSPSubOpt1Tree(root,InfFloat,CT.methRelaxTSP2|4);
                X -> MinTree(root,abstractMixedGraph::MTREE_1TREE);
            }
        }
    }
    catch (ERRejected)
    {
        CT.RestoreLogging();
        return InfFloat;
    }
    
    CT.RestoreLogging();

    if (unfixed==0 && !Feasible()) return InfFloat;
        
    return objective;
}


bool BBSTSP::Feasible() throw()
{
    try
    {
        CT.SuppressLogging();
        X -> ExtractCycles();
        CT.RestoreLogging();
    }
    catch (ERCheck)
    {
        X -> ReleaseDegrees();
        CT.RestoreLogging();
        return 0;
    }
    
    return 1;
}


TFloat BBSTSP::LocalSearch() throw()
{
    G.InitPredecessors();

    for (TNode v=0;v<G.N();v++)
    {
        TArc a = X->Pred(v);
        TNode u = X->StartNode(a);
        TArc a2 = G.Adjacency(u,v);
        G.SetPred(v,a2);
    }
    
    CT.SuppressLogging();
    objective = G.TSPLocalSearch();
    CT.RestoreLogging();
    SetUpperBound(TimerTsp,objective);

    if (CT.traceLevel==3) G.Display();
    
    return objective;
}


void BBSTSP::SaveSolution() throw()
{
}


TFloat abstractGraph::TSPBranchAndBound(TNode r,TFloat upperBound) throw()
{
    LogEntry(LOG_METH,"TSP branch and bound...");
    OpenFold(ModTSP);
        
    BBSTSP *rootNode = new BBSTSP(*this,r);

    if (P==NULL) upperBound = InfFloat;

    char mode = 1;
    if (CT.methCandidates<0) mode = 2;

    goblinBBScheme<TNode,TFloat>* scheme =
        new goblinBBScheme<TNode,TFloat>(rootNode,upperBound,mode);
    TFloat ret = scheme->bestBound;

    CloseFold(ModTSP);
    
    if (ret!=InfFloat)
    {
        SetUpperBound(TimerTsp,ret);

        if (CT.logRes)
        {
            sprintf(CT.logBuffer,
                "...Optimal tour has Length %g",ret);
            LogEntry(LOG_RES,CT.logBuffer);
        }

        delete scheme;
        return ret;
    }
    
    LogEntry(LOG_RES,"...Problem is infeasible");

    delete scheme;
    return InfFloat;
}


TFloat abstractGraph::TSPSubOpt1Tree(TNode r,TFloat bestUpper,TOption level) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (r>=n && r!=NoNode) NoSuchNode("TSPSubOpt1Tree",r);

    #endif
    
    OpenFold(ModSubgradOptTSP,SHOW_TITLE);

    TArc *bestTour = P;
    P = NULL;
    
    bool resolve = false;
    if (pi==NULL) InitPotentials();
    else
    {
        TFloat sum = 0;

        for (TNode v=0;v<n;v++) sum += pi[v];
        
        for (TNode v=0;v<n;v++) pi[v] -= sum/n;
        
        LogEntry(LOG_METH,"Using existing potentials..");
        resolve = true;
    }

    TFloat *bestPi = pi;
    pi = NULL;
    InitPotentials();
    TFloat *oldPi = pi;
    pi = NULL;
    InitPotentials();

    for (TNode v=0;v<n;v++) pi[v] = bestPi[v];
    
    TFloat *oldDeg = new TFloat[n];
    
    TFloat bestLower = -InfFloat;
    TFloat lastBound = -InfFloat;
    TFloat gap = InfFloat;
    TNode dev = n;
    TNode bestDev = n;
    unsigned int step = 0;

    sprintf(CT.logBuffer,"Computing minimum %ld-tree...",r);
    LogEntry(LOG_METH,CT.logBuffer);
    CT.SuppressLogging();
    TFloat thisBound = MinTree(r,abstractMixedGraph::MTREE_1TREE);
    CT.RestoreLogging();

    if ((level&3)==0)
    {
        CloseFold(ModSubgradOptTSP);
        ReleasePredecessors();
        P = bestTour;
        ReleasePotentials();
        pi = oldPi;
        ReleasePotentials();
        pi = bestPi;

        if (CT.logRes && bestUpper!=InfFloat && thisBound!=-InfFloat)
        {
            sprintf(CT.logBuffer,"Found gap is [%g,%g] or %g percent",
                ceil(thisBound),bestUpper,(bestUpper/ceil(thisBound)-1)*100);
            LogEntry(LOG_RES,CT.logBuffer);
        }

        return ceil(thisBound);
    }

    // Compute initial step length

    TFloat t0 = thisBound/n;
    if (t0<0) t0 *= -1;
    if (resolve && t0>200 && (level&3)==1) t0 = 200;
    TFloat t = t0;

    unsigned int count = 0;
    while (bestUpper>=ceil(bestLower-CT.epsilon)+1
           && t>0.02 && dev>0 && step<5000
           && CT.SolverRunning())
    {
        step++;
        
        #if defined(_LOGGING_)

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"Step number %d",step);
            LogEntry(LOG_METH2,CT.logBuffer);
            sprintf(CT.logBuffer,"Step length: %g",t);
            LogEntry(LOG_METH2,CT.logBuffer);
        }

        #endif
        
        if (step>1)
        {
            sprintf(CT.logBuffer,"Computing minimum %ld-tree...",r);
            LogEntry(LOG_METH,CT.logBuffer);
            CT.SuppressLogging();
            InitSubgraph();
            thisBound = MinTree(r,abstractMixedGraph::MTREE_1TREE);
            CT.RestoreLogging();

            if (CT.logRes>=2)
            {
                sprintf(CT.logBuffer,"...Tree length is %g",thisBound);
                LogEntry(LOG_RES2,CT.logBuffer);
            }
        }
        
        TFloat *swapPi = oldPi;
        dev = 0;
        for (TNode v=0;v<n;v++)
        {
            if (Deg(v)!=2) dev++;

            if (step==1) swapPi[v] = pi[v]+t*(Deg(v)-2);
            else swapPi[v] = pi[v]+t*(0.7*Deg(v)+0.3*oldDeg[v]-2);
        }
        
        #if defined(_LOGGING_)

        if (CT.logMeth>=2)
        {
            sprintf(CT.logBuffer,"Nodes with incorrect degrees: %ld",dev);
            LogEntry(LOG_METH2,CT.logBuffer);
        }

        #endif

        bool newGap = false;
        if (thisBound>bestLower+CT.epsilon)
        {
            bestLower = thisBound;
            
            // Do not export the lower bound if the procedure was called
            // for a branch&bound subproblem
            if (!(level&4)) SetLowerBound(TimerTsp,ceil(bestLower));

            newGap = true;
            for (TNode v=0;v<n;v++) bestPi[v] = pi[v];
        }
        
        if (CLCap() && CUCap() && (step==1 || dev<bestDev || dev<5+n/100))
        {
            bestDev = dev;
            TFloat thisUpper = InfFloat;
            try
            {
                sprintf(CT.logBuffer,"Transforming to Hamiltonian cycle...");
                LogEntry(LOG_METH,CT.logBuffer);
                thisUpper = TSPHeuristicsTree(r);
            }
            catch (ERRejected) {};

            if (thisUpper<bestUpper)
            {
                #if defined(_LOGGING_)

                LogEntry(LOG_METH,"Saving tour...");

                #endif

                TArc *swapTour = bestTour;
                bestTour = P;
                P = swapTour;

                bestUpper = thisUpper;
                SetUpperBound(TimerTsp,bestUpper);
                newGap = true;
            }
        }
        
        if (newGap)
        {
            gap = bestUpper/ceil(bestLower-CT.epsilon)-1;

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"Gap decreases to %g percent",gap*100);
                LogEntry(LOG_METH2,CT.logBuffer);
            }

            #endif
        }
        
        oldPi = pi;
        pi = swapPi;
        
        TFloat *swapDeg = oldDeg;
        oldDeg = sDeg;
        sDeg = swapDeg;
        
        if (!newGap && ((level&3)==1 || thisBound<=lastBound))
        {
            if ((level&3)==1)
            {
                t *= 0.9;
                count = 0;
            }
            else t *= 0.98;
        }
        else
        {
            count++;
            if (count>3 && newGap) t *= 1.15;
        }
        
        lastBound = thisBound;
    }
    
    ReleasePotentials();
    pi = oldPi;
    ReleasePotentials();
    pi = bestPi;
    
    delete[] sDeg;
    sDeg = oldDeg;
    
    ReleasePredecessors();
    if (bestTour!=NULL) P = bestTour;
    if (bestUpper==InfFloat) ReleasePredecessors();

    if (CT.logMeth==1)
    {
        sprintf(CT.logBuffer,"Total number of iterations: %d",step);
        LogEntry(LOG_METH,CT.logBuffer);
    }

    CloseFold(ModSubgradOptTSP);

    if (CT.logRes && bestUpper!=InfFloat && bestLower!=-InfFloat)
    {
        sprintf(CT.logBuffer,"...Final gap is [%g,%g] or %g percent",
            ceil(bestLower),bestUpper,gap*100);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    return bestLower;
}


TFloat denseGraph::TSPHeuristics(TNode r) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (r>=n && r!=NoNode) NoSuchNode("TSPHeuristics",r);

    #endif
    
    OpenFold(ModTSP,NO_INDENT);
    
    LogEntry(LOG_METH,"Computing heuristic tour...");
    
    TFloat ret = InfFloat;

    try
    {
        switch (CT.methTSP)
        {
            case 0:
            {
                ret = TSPHeuristicsRandom();
                break;
            }
            case 1:
            {
                if (r==NoNode)
                    ret = TSPHeuristicsInsert();
                else
                    ret = TSPHeuristicsInsert(r);
                break;
            }
            case 2:
            {
                if (r==NoNode)
                    ret = TSPHeuristicsTree();
                else
                    ret = TSPHeuristicsTree(r);
                break;
            }
            case 3:
            {
                ret = TSPHeuristicsChristofides(r);
                break;
            }
            default:
            {
                UnknownOption("TSPHeuristics",CT.methTSP);
            }
        }
    }
    catch (ERRejected) {};

    CloseFold(ModTSP,NO_INDENT);

    if (ret==InfFloat) throw ERRejected();
    else return ret;
}


TFloat abstractGraph::TSPHeuristics(TNode r) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (!CLCap() || MaxLCap()!=0)
        Error(ERR_REJECTED,"TSPHeuristics","Non-trivial lower bounds");
    
    #endif

    OpenFold(ModTSP,NO_INDENT);

    LogEntry(LOG_METH,"Transforming to dense graph...");
    
    denseGraph G(n,0,CT);

    for (TArc i=0;i<G.M();i++)
    {
        TArc a = Adjacency(G.StartNode(2*i),G.EndNode(2*i));
        if (CLength())
        {
            if (a!=NoArc) G.SetLength(2*i,1);
            else G.SetLength(2*i,2);
        } 
        else
        {
            if (a!=NoArc) G.SetLength(2*i,Length(a));
            else G.SetLength(2*i,(MaxLength()+1)*n);
        }
    }
    
    if (Dim()>0)
    {
        TNode v = NoNode;
        for (v=0;v<n;v++)
        {
            G.SetC(v,0,C(v,0));
            G.SetC(v,1,C(v,1));
        }
    }

    int swapSolve = CT.methSolve;
    int swapSubgraph = CT.subgraph;
    CT.methSolve = 1;
    CT.subgraph = 0;
    
    G.TSP(r);
    
    InitPredecessors();
    TFloat length = 0;

    if (r==NoNode) r = 0;
    TNode v = r;

    while (true)
    {
        TNode u = G.StartNode(G.Pred(v));
        TArc a = Adjacency(u,v);
        
        if (a!=NoArc)
        {
            P[v] = a;
            length += Length(a);
        }
        else
        {
            CT.methSolve = swapSolve;
            CT.subgraph = swapSubgraph;
            ReleasePredecessors();
            throw ERRejected();
        }

        v = u;

        if (v==r) break;
    }
    
    CT.methSolve = swapSolve;
    CT.subgraph = swapSubgraph;
    
    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Tour has Length %g",length);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    CloseFold(ModTSP,NO_INDENT);
        
    return length;
}


TFloat abstractGraph::TSPHeuristicsChristofides(TNode r) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (r>=n && r!=NoNode) NoSuchNode("TSPHeuristicsChristofides",r);

    #endif
    
    OpenFold(ModChristofides,SHOW_TITLE);
    
    if (r==NoNode) r = TNode(CT.Rand(n));
    InitSubgraph();
    MinTree(r);
    
    TNode *map = new TNode[n];
    TNode no = 0;
    
    for (TNode v=0;v<n;v++)
    {
    	if (((TArc)Deg(v))%2) map[v] = no++;
        else map[v] = NoNode;
    }
    
    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"Spanning tree has %ld odd vertices",no);
        LogEntry(LOG_RES,CT.logBuffer);
    }
    
    LogEntry(LOG_METH,"Constructing matching problem...");
    OpenFold();

    denseGraph *G = new denseGraph(no,0,CT);
    TNode *revmap = new TNode[no];
    
    for (TNode u=0;u<n;u++)
    {
        if (map[u]!=NoNode)
        {
	        revmap[map[u]] = u;
            
            if (Dim()>=2)
            {
                G -> SetC(map[u],0,C(u,0));
                G -> SetC(map[u],1,C(u,1));
            }
            
            for (TNode v=0;v<u;v++)
            {
                if (map[v]!=NoNode)
                {
                    if (CT.methGeometry==0)
                    {
                        TArc a = Adjacency(u,v);
                        TArc a0 = G->Adjacency(map[u],map[v]);
                        if (u==v) G->SetLength(a0,InfFloat);
                        else G->SetLength(a0,Length(a));
                    }
                }
            }
        }
    }

    CloseFold();
    CT.Ping(G->Handle(),m);
    
    G -> MinCMatching(1);
    CT.Ping(G->Handle(),n*n*n);
    
    LogEntry(LOG_METH,"Constructing Eulerian cycle...");
    OpenFold();

    dynamicQueue<TArc,TFloat> Q(2*m,CT);

    TNode root = r;
    TNode u = root;

    THandle H = Investigate();
    THandle Ho = G->Investigate();

    while (Q.Cardinality()<n-1+no/2)
    {
        if (Active(H,u))
        {
            // Add tree arc
            
            TArc a = Read(H,u);
            if (Sub(a)>0)
            {
                OmitArc(a,1);
                Q.Insert(a);
                u = EndNode(a);

                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,
                        "Adding arc %ld with end node %ld",a,u);
                    LogEntry(LOG_METH2,CT.logBuffer);
                }

                #endif
            }
        }
        else
        {
            TNode uo = map[u];
            if (uo!=NoNode && G->Active(Ho,uo))
            {
                // Arc matching arc
                
                TNode ao = G->Read(Ho,uo);
                if (G->Sub(ao)>0)
                {
                    G -> OmitArc(ao,1);
                    TNode v = revmap[EndNode(ao)];
                    TArc a = Adjacency(u,v);
                    Q.Insert(a);

                    #if defined(_LOGGING_)

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,
                            "Adding arc %ld with end node %ld",a,v);
                        LogEntry(LOG_METH2,CT.logBuffer);
                    }

                    #endif

                    u = v;
                }
            }
            else
            {
                while (!Active(H,u) && (map[u]==NoNode || !G->Active(Ho,map[u])))
                {
                    TArc a = Q.Delete();
                    Q.Insert(a);
                    u = EndNode(a);

                    #if defined(_LOGGING_)

                    if (CT.logMeth>1)
                    {
                        sprintf(CT.logBuffer,
                            "Shifting arc %ld with end node %ld",a,u);
                        LogEntry(LOG_METH2,CT.logBuffer);
                    }

                    #endif
                }
                
                root = u;
            }
        }
    }

    G -> Close(Ho);
    Close(H);

    CloseFold();
    LogEntry(LOG_METH,"Extracting tour...");
    OpenFold();

    InitPredecessors();
    TNode last = root;
    TFloat sum = 0;

    while (!Q.Empty())
    {
        TArc a = Q.Delete();
        TNode v = EndNode(a);
        if (v!=root && P[v]==NoArc)
        {
            a = Adjacency(last,v);
            P[v] = a;
            last = v;
            sum += Length(a);
        }
    }
    TArc a = Adjacency(last,root);
    P[root] = a;
    sum += Length(a);

    CloseFold();

    SetUpperBound(TimerTsp,sum);

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Tour has length %g",sum);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    CT.Ping(Handle(),n*n*n);
    
    delete G;
    delete[] map;
    delete[] revmap;
    
    if (CT.methLocal) sum = TSPLocalSearch();
    
    CloseFold(ModChristofides);

    return sum;
}


TFloat abstractGraph::TSPLocalSearch() throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (P==NULL) Error(ERR_REJECTED,"TSPLocalSearch","Missing tour");

    #endif
    
    OpenFold(ModTSP,NO_INDENT);

    LogEntry(LOG_METH,"Searching for local optimality...");
    
    while (CT.SolverRunning() && TSP2Exchange(-MaxLength())) {};
    
    bool unchanged = false;
    while (CT.SolverRunning() && !unchanged)
    {
        unchanged = !TSP2Exchange(0);
        if (unchanged) unchanged = !TSPNodeExchange(0);
    }
    
    TArc a = P[0];
    TFloat l = Length(a);
    TNode x = StartNode(a);
    TNode count = 1;

    while (x!=0)
    {
        a = P[x];
        l += Length(a);
        x = StartNode(a);
        count++;
    }

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Tour has length %g",l);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    SetUpperBound(TimerTsp,l);

    CloseFold(ModTSP,NO_INDENT);
    
    if (count<n)
    {
        InternalError("TSPLocalSearch","Tour is incomplete");
    }

    return l;
}


bool abstractGraph::TSP2Exchange(TFloat limit) throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (P==NULL) Error(ERR_REJECTED,"TSP2Exchange","Missing tour");

    #endif
    
    OpenFold(Mod2OptTSP);

    TNode r = CT.Rand(n);
    TNode v1 = r;
    TArc a1 = P[r];
    TNode u1 = StartNode(a1);

    bool unchanged = true;
    while (unchanged && u1!=r)
    {
        TNode v2 = StartNode(P[u1]);
        TArc a2 = P[v2];
        TNode u2 = StartNode(a2);

        while (unchanged && u2!=r && u2!=v1)
        {
            TArc a1prime = Adjacency(u1,u2);
            TArc a2prime = Adjacency(v1,v2);

            TFloat diff = InfFloat;
            if (a1prime!=NoArc && a2prime!=NoArc)
                diff = Length(a1prime)+Length(a2prime)
                     - Length(a1)-Length(a2);

            if (diff<limit)
            {
                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,
                        "Local improvement (%g units, 2-exchange)",-diff);
                    LogEntry(LOG_METH2,CT.logBuffer);
                    sprintf(CT.logBuffer,
                        "New tour: ... %ld %ld ... %ld %ld ...",
                        u1,u2,v1,v2);
                    LogEntry(LOG_METH2,CT.logBuffer);
                }

                #endif

                TNode x = u2;
                TArc a = P[x];

                // Flip arc directions

                while (x!=v1)
                {
                    TNode y = StartNode(a);
                    TArc an = P[y];
                    P[y] = (a^1);
                    x = y;
                    a = an;
                }

                P[v2] = a2prime;
                P[u2] = a1prime;

                unchanged = false;

                CT.Ping(Handle(),n*n);
            }

            v2 = u2;
            a2 = P[u2];
            u2 = StartNode(a2);
        }

        v1 = u1;
        a1 = P[u1];
        u1 = StartNode(a1);
    }

    CloseFold(Mod2OptTSP);

    return !unchanged;
}
