
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, May 2001.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "bbstable.h"
#include "spsgra.h"
#include "absbig.h"
    

BBStableSet::BBStableSet(abstractMixedGraph &GC) throw() :
    goblinBBNode<TNode,TFloat>(GC.N(),GC.Context()), G(GC)
{
    int savedMaxBBNodes = CT.maxBBNodes;
    CT.maxBBNodes = 1;
    int savedMaxBBIterations = CT.maxBBIterations;
    CT.maxBBIterations = 1;
    SetUpperBound(TimerStable,G.CliqueCover(G.N()));
    CT.maxBBNodes = savedMaxBBNodes;
    CT.maxBBIterations = savedMaxBBIterations;

    for (TNode v=0;v<n;v++) G.SetDist(v,TFloat(G.Colour(v)));

    if (CT.traceLevel==3) G.Display();

    colour = new char[n];
    currentVar = 0;
    selected = 0;
    H = G.Investigate();
    for (TNode v=0;v<n;v++) colour[v] = 1;
    
    LogEntry(LOG_MEM,"(stable sets)");
}


BBStableSet::BBStableSet(BBStableSet &Node) throw() :
    goblinBBNode<TNode,TFloat>(Node.G.N(),Node.Context(),Node.scheme), G(Node.G)
{
    colour = new char[n];
    currentVar = Node.currentVar;
    selected = Node.selected;
    for (TNode v=0;v<n;v++)
    {
        colour[v] = Node.colour[v];
        if (colour[v]!=1) unfixed--;
    }
    
    H = G.Investigate();
    
    LogEntry(LOG_MEM,"(stable sets)");
}


BBStableSet::~BBStableSet() throw()
{
    delete[] colour;
    G.Close(H);
    
    LogEntry(LOG_MEM,"(stable sets)");
}


unsigned long BBStableSet::Size() throw()
{
    return
          sizeof(BBStableSet)
        + goblinDataObject::Allocated()
        + goblinBBNode<TNode,TFloat>::Allocated()
        + Allocated();
}


unsigned long BBStableSet::Allocated() throw()
{
    return n*sizeof(char);
}


TNode BBStableSet::SelectVariable() throw()
{
    TNode* reducedDeg = new TNode[n];
    for (TNode i=0;i<n;i++) reducedDeg[i] = 0;

    for (TArc a=0;a<2*G.M();a++)
    {
        TNode u = G.StartNode(a);
        TNode v = G.EndNode(a);
        if (colour[u]==1 && colour[v]==1) reducedDeg[u]++;
    }
    
    TNode minNode = NoNode;
    TNode minDeg = 0;
    for (TNode i=0;i<n;i++)
    {
        if (colour[i]==1 && (minNode==NoNode || reducedDeg[i]<minDeg))
        {
            minNode = i;
            minDeg = reducedDeg[i];
        }
    }

    delete[] reducedDeg;

    if (minNode!=NoNode) return minNode;
    
    #if defined(_FAILSAVE_)

    InternalError("SelectVariable","Solution is fixed");

    #endif

    throw ERInternal();
}


char BBStableSet::Direction1(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Direction1",v);

    #endif
    
    return 1;
}


char BBStableSet::Direction2(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Direction2",v);

    #endif
    
    return 1;
}


goblinBBNode<TNode,TFloat> *BBStableSet::Clone() throw()
{
    return new BBStableSet(*this);
}


void BBStableSet::Raise(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Raise",v);

    #endif
    
    colour[v] = 2;
    unfixed--;
    selected++;
    solved = false;
    
    G.Reset(H,v);
    while (G.Active(H,v))
    {
        TArc a = G.Read(H,v);
        TNode w = G.EndNode(a);
        if (colour[w]==1)
        {
            colour[w] = 0;
            unfixed--;
        }
        
        #if defined(_FAILSAVE_)

        if (colour[w]==2)
        {
            sprintf(CT.logBuffer,"Conflicting nodes: %ld, %ld",v,w);
            InternalError1("Raise");
        }

        #endif
    }
}


void BBStableSet::Lower(TNode v) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=n) NoSuchNode("Lower",v);

    #endif
    
    colour[v] = 0;
    unfixed--;
    solved = false;
}


TFloat BBStableSet::SolveRelaxation() throw()
{
    TNode maxColour = 0;

    for (TNode v=0;v<n;v++)
        if (TNode(G.Dist(v))>maxColour) maxColour = TNode(G.Dist(v));
    
    TFloat objective = selected;
    for (TNode c=0;c<=maxColour;c++)
    {
        bool blocked = true;
        for (TNode v=0;v<n;v++)
        {
            if (colour[v]==1 && TNode(G.Dist(v))==c) blocked = false;
        }
        
        if (!blocked) objective += 1;
    }

    return -objective;
}


void BBStableSet::SaveSolution() throw()
{
    for (TNode v=0;v<n;v++) G.SetColour(v,colour[v]/2);

    if (CT.traceLevel==3) G.Display();
}


TNode abstractMixedGraph::StableSet() throw()
{
    LogEntry(LOG_METH,"Computing maximum independent nodes set...");
    OpenFold(ModStable);
        
    BBStableSet* rootNode = new BBStableSet(*this);

    SetLowerBound(TimerStable,0);
    goblinBBScheme<TNode,TFloat> scheme(rootNode,0);
    
    TNode ret = TNode(-scheme.bestBound);

    if (CT.logRes)
    {
        if (ret!=0)
        {
            sprintf(CT.logBuffer,"...Stable set has cardinality %ld",
                TNode(-scheme.bestBound));
            LogEntry(LOG_RES,CT.logBuffer);
        }
        else
        {
            LogEntry(LOG_RES,"...No stable set constructed");
            SetColour(0,1);
            for (TNode v=1;v<n;v++) SetColour(v,0);
            ret = 1;
        }
    }

    SetLowerBound(TimerStable,ret);
    CloseFold(ModStable);
    
    return TNode(ret);
}


TNode abstractMixedGraph::Clique() throw()
{
    LogEntry(LOG_METH,"Computing maximum clique...");
    OpenFold(ModStable);

    complementaryGraph* G = new complementaryGraph(*this,(TOption)0);

    if (colour!=NULL)
        for (TNode v=0;v<n;v++) G->SetColour(v,Colour(v));
    
    BBStableSet* rootNode = new BBStableSet(*G);

    SetLowerBound(TimerStable,0);

    goblinBBScheme<TNode,TFloat>* scheme
        = new goblinBBScheme<TNode,TFloat>(rootNode,0);
    
    TNode ret = TNode(-scheme->bestBound);

    if (CT.logRes)
    {
        if (ret!=0)
        {
            sprintf(CT.logBuffer,"...Clique has cardinality %ld",ret);
            LogEntry(LOG_RES,CT.logBuffer);
        }
        else
        {
            LogEntry(LOG_RES,"...No clique constructed");
            G->SetColour(0,1);
            for (TNode v=1;v<n;v++) G->SetColour(v,0);
            ret = 1;
            if (G->M()>0)
            {
                G->SetColour(StartNode(0),1);
                G->SetColour(EndNode(0),1);
                ret = 2;
            }
        }
    }

    SetLowerBound(TimerStable,ret);

    delete scheme;
    
    for (TNode v=0;v<n;v++) SetColour(v,G->Colour(v));
    
    delete G;
    
    CloseFold(ModStable);
    
    return ret;
}


TNode abstractMixedGraph::VertexCover() throw()
{
    LogEntry(LOG_METH,"Computing minimum vertex cover...");
    OpenFold(ModStable);
        
    TNode ret = StableSet();
    
    for (TNode v=0;v<n;v++) SetColour(v,1-Colour(v));

    CloseFold(ModStable);
    
    return n-ret;
}


TNode abstractBiGraph::StableSet() throw()
{
    LogEntry(LOG_METH,"Computing maximum stable set...");
    OpenFold(ModStable);

    for (TNode v=0;v<n1;v++)
        if (n1<n2) SetColour(v,0);
        else SetColour(v,1);
    for (TNode v=n1;v<n;v++)
        if (n2<=n1) SetColour(v,0);
        else SetColour(v,1);

    CloseFold(ModStable);
    
    if (n1<n2) return n2;
    else return n1;
}


TNode abstractBiGraph::VertexCover() throw()
{
    LogEntry(LOG_METH,"Computing minimum vertex cover...");
    OpenFold(ModStable);

    for (TNode v=0;v<n1;v++)
        if (n1>=n2) SetColour(v,0);
        else SetColour(v,1);
    for (TNode v=n1;v<n;v++)
        if (n2>n1) SetColour(v,0);
        else SetColour(v,1);

    CloseFold(ModStable);
    
    if (n1>=n2) return n2;
    else return n1;
}


TNode abstractBiGraph::NodeColouring(TNode dummy) throw()
{
    LogEntry(LOG_METH,"Computing minimum node colouring...");
    OpenFold(ModStable);

    for (TNode v=0;v<n;v++)
        if (v<n1 || m==0) SetColour(v,0);
        else SetColour(v,1);

    CloseFold(ModStable);

    if (m==0) return 1;
    else return 2;
}
