
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, July 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsgra.h"
#include "spsdig.h"


TFloat abstractMixedGraph::SteinerTree(TNode r) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (r>=n && r!=NoNode) NoSuchNode("TSP",r);

    #endif
    
    OpenFold(ModSteiner,NO_INDENT);

    if (r!=NoNode)
    {
        sprintf(CT.logBuffer,"Computing minimum %ld-Steiner tree...",r);
        LogEntry(LOG_METH,CT.logBuffer);
    }
    else LogEntry(LOG_METH,"Computing minimum Steiner tree...");
    
    TArc* savedTree = NULL;
    TFloat savedLength = 0;
    if (P!=NULL)
    {
        bool feasible = true;
    
        TNode rank = 0;
        for (TNode v=0;v<n;v++)
        {
            TArc a = P[v];
            if (a!=NoArc)
            {
                savedLength += Length(a);
                rank++;
                if (v==r) feasible = false;
            }
        }
        
        if (rank<n-1) feasible = false;
        
        // Must add a check for connectivity
        
        if (feasible)
        {
            savedLength -= SteinerTrimLeaves();
            savedTree = P;
            P = NULL;
            if (CT.logMeth)
            {
                sprintf(CT.logBuffer,"Initial tree has length %g",savedLength);
                LogEntry(LOG_METH,CT.logBuffer);
            }
            SetUpperBound(TimerSteiner,savedLength);
        }
    }
    
    LogEntry(LOG_METH,"Applying Heuristics...");
    TFloat bestUpper = InfFloat;
    try
    {
        bestUpper = SteinerHeuristics(r);
    }
    catch (ERRejected) {};
    
    if (savedTree!=NULL)
    {
        if (savedLength<bestUpper)
        {
            if (P!=NULL) ReleasePredecessors();
            P = savedTree;
            bestUpper = savedLength;
        }
        else delete[] savedTree;
    }
    
    if (CT.methSolve==2)
    {
        LogEntry(LOG_METH,"Exhaustive search...");
        bestUpper = SteinerEnumerate(r);
    }
        
    CloseFold(ModSteiner,NO_INDENT);

    if (bestUpper==InfFloat) throw ERRejected();
    
    return bestUpper;
}


TFloat abstractMixedGraph::SteinerHeuristics(TNode r) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (r!=NoNode && (r>=n || Demand(r)==0))
    {
        sprintf(CT.logBuffer,"Inappropriate root node: %ld",r);
        Error(ERR_RANGE,"SteinerEnumerate",CT.logBuffer);
    }

    #endif

    OpenFold(ModSteiner);
    
    for (TNode v=0;r==NoNode && v<n;v++)
        if (Demand(v)!=0) r = v;
    
    TFloat ret = MinTree(r);
    
    ret -= SteinerTrimLeaves();

    SetUpperBound(TimerSteiner,ret);

    CloseFold(ModSteiner);

    sprintf(CT.logBuffer,"...Steiner tree has length %g",ret);
    LogEntry(LOG_RES,CT.logBuffer);
    
    return ret;
}


TFloat abstractMixedGraph::SteinerTrimLeaves() throw()
{
    LogEntry(LOG_METH,"Trimming the leaves...");
    OpenFold(ModSteiner);
    
    TNode* odg = new TNode[n];
    
    for (TNode v=0;v<n;v++) odg[v] = 0;
    
    for (TNode v=0;v<n;v++)
        if (P[v]!=NoArc) odg[StartNode(P[v])]++;
    
    TFloat diff = 0;
    for (TNode v=0;v<n;v++)
    {
        TNode w = v;
        while (P[w]!=NoArc && Demand(w)==0 && odg[w]==0)
        {
            diff += Length(P[w]);
            TNode x = StartNode(P[w]);
            odg[x]--;
            P[w] = NoArc;

            sprintf(CT.logBuffer,"Node %ld deleted",w);
            LogEntry(LOG_METH,CT.logBuffer);

            w = x;
        }
    }

    delete[] odg;

    CT.Ping(Handle(),m);
    
    CloseFold(ModSteiner);

    if (CT.logRes && diff!=0)
    {
        sprintf(CT.logBuffer,"...Tree length decreases by %g",diff);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    return diff;
}


TFloat abstractMixedGraph::SteinerEnumerate(TNode r) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (r!=NoNode && (r>=n || Demand(r)==0))
    {
        sprintf(CT.logBuffer,"Inappropriate root node: %ld",r);
        Error(ERR_RANGE,"SteinerEnumerate",CT.logBuffer);
    }

    #endif
    
    OpenFold(ModSteiner);

    LogEntry(LOG_METH,"(Steiner node enumeration)");

    sprintf(CT.logBuffer,"Root node is %ld",r);
    LogEntry(LOG_METH,CT.logBuffer);

    TNode k = 0;
    for (TNode v=0;v<n;v++)
        if (Demand(v)==0) k++;
    
    sprintf(CT.logBuffer,"%ld Steiner nodes detected",k);
    LogEntry(LOG_METH,CT.logBuffer);
    CT.SuppressLogging();

    TFloat ret = InfFloat;

    if (k==0) ret = MinTree(r);

    if (k==n)
    {
        InitPredecessors();
        ret = 0;
    }
    
    InitColours(0);
    bool searching = (k!=0 && k!=n);
    unsigned long itCount = 0;
    while (CT.SolverRunning() && searching)
    {
        itCount++;
    
        completeOrientation G(*this);

        for (TArc a=0;a<G.M();a++)
        {
            TArc a0 = G.OriginalArc(2*a);
            if (colour[StartNode(a0)]==1) G.SetLength(2*a,InfFloat);
            if (colour[EndNode(a0)]==1) G.SetLength(2*a,0);
        }

        TFloat thisRet = G.MinTree(r);
        if (thisRet<ret)
        {
            TNode uncovered = 0;
            for (TNode v=0;v<n;v++)
                if (G.Demand(v)!=0 && G.Pred(v)==NoArc) uncovered++;

            if (uncovered==1)
            {
                ret = thisRet;
                InitPredecessors();
                for (TNode v=0;v<n;v++)
                    if (colour[v]==0 && v!=r && G.Pred(v)!=NoArc)
                        P[v] = G.OriginalArc(G.Pred(v));
                CT.RestoreLogging();
                SetUpperBound(TimerSteiner,ret);
                CT.SuppressLogging();
            }
        }
        
        TNode v = n;
        searching = false;
        while (v>0 && !searching)
        {
            v--;
            if (Demand(v)==0)
            {
                if (colour[v]==1) colour[v] = 0;
                else
                {
                    colour[v] = 1;
                    searching = true;
                }
            }
        }
        
        if (LowerBound(TimerSteiner)>=ret) searching = false;
    }
    
    CT.RestoreLogging();
    sprintf(CT.logBuffer,"%ld subproblems solved",itCount);
    LogEntry(LOG_METH,CT.logBuffer);
    
    CloseFold(ModSteiner);

    sprintf(CT.logBuffer,"...Best tree has length %g",ret);
    LogEntry(LOG_RES,CT.logBuffer);
    
    return ret;
}


TFloat abstractGraph::SteinerHeuristics(TNode r) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (r!=NoNode && (r>=n || Demand(r)==0))
    {
        sprintf(CT.logBuffer,"Inappropriate root node: %ld",r);
        Error(ERR_RANGE,"SteinerHeuristics",CT.logBuffer);
    }
    
    for (TNode v=0;r==NoNode && v<n;v++)
        if (Demand(v)!=0) r = v;

    if (r==NoNode)
        Error(ERR_REJECTED,"SteinerHeuristics","No terminal node found");

    #endif
    
    OpenFold(ModMehlhorn);

    TFloat lower = 0;
    {
        voronoiDiagram G(*this);

        LogEntry(LOG_METH,"Computing minimum spanning tree...");
        OpenFold();
        lower = ceil(G.Prim2()/(2*(1-1/TFloat(G.N()))));
        CloseFold();
        SetLowerBound(TimerSteiner,lower);
        G.UpdateSubgraph();
    }
    
    CT.Ping(Handle(),m); 

    #if defined(_LOGGING_)

    LogEntry(LOG_METH2,"Extracting tree from subgraph...");

    #endif

    InitPredecessors();
    TFloat length = 0;
    
    THandle H = Investigate();
    goblinIterator &I = Iterator(H);
    TNode w = r;
    while (I.Active(w) || w!=r)
    {
        if (I.Active(w))
        {
            TArc a = I.Read(w);
            TNode u = EndNode(a);
            if (Sub(a)>0 && a!=(P[w]^1))
            {
                if (P[u]==NoArc)
                {
                    P[u] = a;
                    length += Length(a);
                    if (u!=r) w = u;
                }
            }
        }
        else w = StartNode(P[w]);
    };
    Close(H);

    SetUpperBound(TimerSteiner,length);
    CT.Ping(Handle(),m);
    CloseFold(ModMehlhorn);

    if (CT.logRes)
    {
        sprintf(CT.logBuffer,"...Steiner tree has length %g",length);
        LogEntry(LOG_RES,CT.logBuffer);
    }

    return length;
}


TFloat abstractGraph::SteinerEnumerate(TNode r) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (r!=NoNode && (r>=n || Demand(r)==0))
    {
        sprintf(CT.logBuffer,"Inappropriate root node: %ld",r);
        Error(ERR_RANGE,"SteinerEnumerate",CT.logBuffer);
    }
    
    for (TNode v=0;r==NoNode && v<n;v++)
        if (Demand(v)!=0) r = v;

    if (r==NoNode)
        Error(ERR_REJECTED,"SteinerEnumerate","No terminal node found");

    #endif
    
    OpenFold(ModSteiner);

    LogEntry(LOG_METH,"(Steiner node enumeration)");

    sprintf(CT.logBuffer,"Root node is %ld",r);
    LogEntry(LOG_METH,CT.logBuffer);

    TNode k = 0;
    for (TNode v=0;v<n;v++)
        if (Demand(v)==0) k++;
    
    sprintf(CT.logBuffer,"%ld Steiner nodes detected",k);
    LogEntry(LOG_METH,CT.logBuffer);
    CT.SuppressLogging();

    TFloat ret = InfFloat;

    if (k==0) ret = Prim2(r);

    if (k==n)
    {
        InitPredecessors();
        ret = 0;
    }
    
    InitColours(0);
    bool searching = (k!=0 && k!=n);
    unsigned long itCount = 0;
    while (CT.SolverRunning() && searching)
    {
        itCount++;

        graph G(*this,OPT_CLONE);

        for (TArc a=0;a<2*m;a++)
            if (colour[StartNode(a)]==1) G.SetLength(a,InfFloat);

        if (G.Prim2(r)<InfFloat)
        {
            TFloat thisRet = 0;
            for (TNode v=0;v<n;v++)
                if (colour[v]==0 && v!=r && G.Pred(v)!=NoArc)
                    thisRet += Length(G.Pred(v));
            
            if (thisRet<ret)
            {
                TNode uncovered = 0;
                for (TNode v=0;v<n;v++)
                    if (G.Demand(v)!=0 && G.Pred(v)==NoArc) uncovered++;

                if (uncovered==1)
                {
                    ret = thisRet;
                    InitPredecessors();
                    for (TNode v=0;v<n;v++)
                        if (colour[v]==0 && v!=r && G.Pred(v)!=NoArc)
                            P[v] = G.Pred(v);
                    CT.RestoreLogging();
                    SetUpperBound(TimerSteiner,ret);
                    CT.SuppressLogging();
                }
            }
        }
        
        TNode v = n;
        searching = false;
        while (v>0 && !searching)
        {
            v--;
            if (Demand(v)==0)
            {
                if (colour[v]==1) colour[v] = 0;
                else
                {
                    colour[v] = 1;
                    searching = true;
                }
            }
        }
        
        if (LowerBound(TimerSteiner)>=ret) searching = false;
    }
    
    CT.RestoreLogging();
    sprintf(CT.logBuffer,"%ld subproblems solved",itCount);
    LogEntry(LOG_METH,CT.logBuffer);
    
    CloseFold(ModSteiner);

    sprintf(CT.logBuffer,"...Best tree has length %g",ret);
    LogEntry(LOG_RES,CT.logBuffer);
    
    return ret;
}
