
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, July 2003.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsgra.h"
#include "dnsgra.h"


void abstractGraph::ComputeTJoin() throw()
{
    #if defined(_FAILSAVE_)

    if (MaxLCap()>0)
        Error(ERR_REJECTED,"ComputeTJoin","Non-trivial lower bounds");

    #endif
    
    OpenFold(ModTJoin,NO_INDENT);
    
    if (CT.traceLevel==2) Display();

    LogEntry(LOG_METH,"Computing minimum cost T-join...");
    OpenFold();
    
    TNode* map = new TNode[n];

    TNode no = 0;
    
    for (TNode v=0;v<n;v++)
    {
    	if (Demand(v)>0) map[v] = no++;
        else map[v] = NoNode;
    }
    
    if (CT.logMeth)
    {
        sprintf(CT.logBuffer,"Problem has %ld odd vertices",no);
        LogEntry(LOG_METH,CT.logBuffer);
    }
    
    if (no==0)
    {
        delete[] map;
        CloseFold();
        CloseFold(ModTJoin,NO_INDENT);
        return;
    }
    
    if (no%1)
    {
        delete[] map;
        CloseFold();
        CloseFold(ModTJoin,NO_INDENT);
        Error(ERR_REJECTED,"ComputeTJoin","Number of odd nodes must be even");
    }
    
    LogEntry(LOG_METH,"Transforming into matching problem...");
    OpenFold();
    
    denseGraph* G = new denseGraph(no,0,CT);
    denseGraphStructure &Y = G->Struct();

    TNode* revmap = new TNode[no];

    for (TNode u=0;u<n;u++)
    {
        if (map[u]!=NoNode)
        {
            if (Dim()>=2)
            {
                Y.SetC(map[u],0,C(u,0));
                Y.SetC(map[u],1,C(u,1));
            }
            
            revmap[map[u]] = u;

            LogEntry(LOG_METH,"Computing shortest path tree...");

            Dijkstra(u);

            for (TNode v=0;v<=u;v++)
            {
            	if (map[v]!=NoNode)
                {
                    TArc a = G->Adjacency(map[u],map[v]);
                    Y.SetLength(a,d[v]);
                }
            }
        }
    }

    CloseFold();
    
    CT.Ping(G->Handle(),n*m);
    G -> MinCMatching((TCap)1);
    CT.Ping(G->Handle(),n*n*n);
    
    LogEntry(LOG_METH,"Transforming matching into disjoint paths...");
    OpenFold();

    for (TArc a=0;a<G->M();a++)
    {
        if (G->Sub(2*a)>0)
        {
            TNode u = revmap[G->StartNode(2*a)];
            TNode v = revmap[G->EndNode(2*a)];

            LogEntry(LOG_METH,"Computing shortest path tree...");

            Dijkstra(u);

            #if defined(_LOGGING_)

            THandle LH = NoHandle;

            if (CT.logMeth>1)
            {
                LogEntry(LOG_METH2,"Adding the path (in reverse order):");
                sprintf(CT.logBuffer,"  (%ld",v);
                LH = LogStart(LOG_METH2,CT.logBuffer);
            }

            #endif

            while (v!=u)
            {
                TArc a2 = P[v];

                if (Sub(a2)==UCap(a2)) OmitArc(a2,UCap(a2));
                else AddArc(a2,UCap(a2));

                v = StartNode(a2);

                #if defined(_LOGGING_)

                if (CT.logMeth>1)
                {
                    sprintf(CT.logBuffer,",%ld",v);
                    LogAppend(LH,CT.logBuffer);
                }

                #endif
            }

            #if defined(_LOGGING_)

            if (CT.logMeth>1) LogEnd(LH,")");

            #endif
        }
    }

    CloseFold();
    delete G;

    CT.Ping(Handle(),n*n*n);
    
    delete[] map;
    delete[] revmap;
    
    CloseFold();

    CloseFold(ModTJoin,NO_INDENT);
};


void abstractGraph::MinCTJoin() throw()
{
    OpenFold(ModTJoin,NO_INDENT);
    
    LogEntry(LOG_METH,"Eliminating negative length labels...");
    
    graph G(*this,OPT_CLONE);
    
    for (TArc a=0;a<2*m;a++)
    {
        if (Length(a)<0)
        {
            TNode v = StartNode(a);

            if (G.Demand(v)>0) G.SetDemand(v,0);
            else G.SetDemand(v,1);

            if (a&1) G.SetLength(a,-Length(a));
        }
    }
    
    G.ComputeTJoin();
    
    LogEntry(LOG_METH,"Flipping negative length arcs...");
    
    for (TArc a=0;a<m;a++)
    {
        TArc a2 = 2*a;

        if ((G.Sub(a2)>0 && Length(a2)>=0) || (G.Sub(a2)==0 && Length(a2)<0))
            AddArc(a2,UCap(a2)-Sub(a2));
        else OmitArc(a2,Sub(a2));
    }

    CloseFold(ModTJoin,NO_INDENT);
};


void abstractGraph::MinCTJoin(TNode s,TNode t) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("MinCTJoin",s);
    
    if (t>=n) NoSuchNode("MinCTJoin",t);   

    #endif
    
    OpenFold(ModTJoin,NO_INDENT);
    
    LogEntry(LOG_METH,"Eliminating negative length labels...");
    
    graph G(*this,OPT_CLONE);
    
    for (TNode v=0;v<n;v++)
    {
    	if (v==s || v==t) G.SetDemand(v,1);
        else G.SetDemand(v,0);
    }
    
    for (TArc a=0;a<2*m;a++)
    {
        if (Length(a)<0)
        {
            TNode v = StartNode(a);

            if (G.Demand(v)>0) G.SetDemand(v,0);
            else G.SetDemand(v,1);

            if (a&1) G.SetLength(a,-Length(a));
        }
    }
    
    G.ComputeTJoin();
    
    LogEntry(LOG_METH,"Flipping negative length arcs...");
    
    for (TArc a=0;a<m;a++)
    {
        TArc a2 = 2*a;

        if ((G.Sub(a2)>0 && Length(a2)>=0) || (G.Sub(a2)==0 && Length(a2)<0))
            AddArc(a2,UCap(a2)-Sub(a2));
        else OmitArc(a2,Sub(a2));
    }

    CloseFold(ModTJoin,NO_INDENT);
};
