
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Markus Schwank, November 2001.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsdig.h"


TCap abstractMixedGraph::TreePacking(TNode r) throw(ERRejected,ERRange)
{
    Error(ERR_REJECTED,"TreePacking","Method applies to directed graphs only");
    throw ERRejected();
}


TCap abstractDiGraph::TreePacking(TNode r) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (r>=n && r!=NoNode) NoSuchNode("TreePacking",r);

    #endif
    
    OpenFold(ModTreePack,NO_INDENT);
    
    if (CT.logMeth)
    {
        LogEntry(LOG_METH,"");
        LogEntry(LOG_METH,"------------------------------");
        LogEntry(LOG_METH,"Arborescence Packing Algorithm");
        LogEntry(LOG_METH,"by Markus Schwank, November 01");
        LogEntry(LOG_METH,"------------------------------");
        LogEntry(LOG_METH,"");
    }

    InitColours();
    long unsigned step = 0;

    abstractDiGraph *G = TreePKGInit();

    LogEntry(LOG_METH,"Computing number of arborescences...");
    OpenFold();

    TCap totalmulti = G->StrongEdgeConnectivity(r);

    sprintf(CT.logBuffer,"...total multiplicity: %g",totalmulti);
    LogEntry(LOG_METH,CT.logBuffer);

    CloseFold();

    for (TNode v=0;v<n;v++) colour[v] = G->Colour(v);

    TFloat returnmulti = totalmulti;

    LogEntry(LOG_METH,"Packing with arborescences...");
    OpenFold();

    SetBounds(TimerTreePack,0,totalmulti);

    while (totalmulti>0)
    {
        step++;

   	    sprintf(CT.logBuffer,
            "Computing the %ldth spanning %ld-arborescence...",step,r);
        LogEntry(LOG_METH,CT.logBuffer);
        OpenFold();

        TreePKGStripTree(G,(&totalmulti),r);
        if (CT.traceLevel==3) Display();

        CloseFold();
   	    sprintf(CT.logBuffer,"...remaining multiplicity: %g",totalmulti);
        LogEntry(LOG_METH,CT.logBuffer);

        SetLowerBound(TimerTreePack,UpperBound(TimerTreePack)-totalmulti);
    }

    delete G;
    ReleasePredecessors();

    CloseFold();

    CloseFold(ModTreePack,NO_INDENT);
    
    return returnmulti;
}


abstractDiGraph* abstractDiGraph::TreePKGInit() throw()
{
    return new completeOrientation(*this);
}


TCap abstractDiGraph::TreePKGStripTree(abstractDiGraph* G,TCap *totalMulti,
    TNode r) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (r>=n && r!=NoNode) NoSuchNode("TreePKGStripTree",r);

    #endif
    
    CT.SuppressLogging();

    TFloat rank = 0;
    TCap totalmulti = TCap(*totalMulti);

    TNode* p = new TNode[n];
    TNode* nr = new TNode[n];

    for (TNode v=0;v<n;v++)
    {
        nr[v] = 0;
        p[v]  = 0;
    }

    nr[r] = 1;
    TNode x = r;

    bool* used = new bool[2*m];

    for (TArc a=0;a<2*m;a++) used[a] = false;

    THandle H = Investigate();
    goblinIterator &I = Iterator(H);

    while (rank<n-1)
    {
        while (I.Active(x) && rank<n-1)
        {
            TArc a = I.Read(x);

            if (a%2==0 && !used[a] && G->UCap(a)>0)
            {
                used[a] = true;
                TNode w = EndNode(a);

                if (nr[w]==0)
                {
                    G -> SetUCap(a,G->UCap(a)-1);

                    if (G->StrongEdgeConnectivity(r)>=totalmulti-1)
                    {
                        rank++;
                        p[w] = x;
                        nr[w] = 1;
                        x = w;
                    }
                    else G->SetUCap(a,G->UCap(a)+1);
                }
            }
        }

        x = p[x];

        I.Reset(x);
    }

    Close(H);

    delete[] used;
    delete[] nr;

    CT.RestoreLogging();

    LogEntry(LOG_METH,"Computing the minimum tree arc capacity...");

    TCap minCap = InfCap;

    InitPredecessors();

    for (TNode v=0;v<n;v++)
    {
        if (v!=r)
        {
            P[v] = Adjacency(p[v],v);

            TCap thisCap = G->UCap(P[v])+1;
            G->SetUCap(P[v],thisCap);

            if (thisCap<minCap) minCap = thisCap;
        }
    }

    delete[] p;

    sprintf(CT.logBuffer,
        "...Minimum arc capacity: %g",minCap);
    LogEntry(LOG_RES,CT.logBuffer);

    LogEntry(LOG_METH,"Computing tree capacity...");
    CT.SuppressLogging();

   	TFloat cap = minCap;

    for (TNode v=0;v<n;v++)
        if (v!=r) G->SetUCap(P[v],G->UCap(P[v])-minCap);

    OpenFold();

    while (G->StrongEdgeConnectivity(r)!=totalmulti-cap && cap>0)
    {
        cap--;
   	    for (TNode v=0;v<n;v++)
            if (v!=r)
            {
                TArc a = P[v];
                G->SetUCap(a,G->UCap(a)+1);
            }
   	}
    
   	CloseFold();

    CT.RestoreLogging();

    sprintf(CT.logBuffer,
        "...Tree has capacity: %g",cap);
    LogEntry(LOG_RES,CT.logBuffer);

   	(*totalMulti) -= cap;
    
    return cap;
}
