
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, January 2000.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsbal.h"


balancedFNW::balancedFNW(TNode nn,TOption options,
    goblinController& thisContext) throw() :
    goblinDataObject(thisContext),
    abstractMixedGraph(2*nn,(TArc)0),
    abstractBalancedFNW(nn),
    X(*this,options)
{                                                   
    X.SetCOrientation(1);
    X.NewSubgraph();
    
    LogEntry(LOG_MEM,"...Balanced flow network instanciated");
}

     
balancedFNW::balancedFNW(const char* impFileName,goblinController& thisContext)
    throw(ERFile,ERParse) :
    goblinDataObject(thisContext),
    abstractMixedGraph((TNode)0,(TArc)0),
    X(*this,(TOption)0)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    LogEntry(LOG_IO,"Loading balanced flow network...");
    if (!CT.logIO && CT.logMem)
        LogEntry(LOG_MEM,"Loading balanced flow network...");
    
    goblinImport F(impFileName,CT);
    F.Scan("balanced_fnw");
    F.Scan("definition");

    F.Scan("nodes");
    TNode *nodes = F.GetTNodeTuple(3);
    n = nodes[0];
    ni = nodes[2];
    n1 = n>>1;
    delete[] nodes;    
    
    F.Scan("arcs");
    TArc *arcs = F.GetTArcTuple(1);
    m = arcs[0];
    delete[] arcs;    

    CheckLimits();
    X.ReDimension(n,ni,m);
        
    X.ReadIncidences(F);
    X.ReadUCap(F);
    X.ReadLCap(F);
    X.ReadDemand(F);
    X.ReadOrientation(F);
    
    F.Scan(); // definition
    
    X.ReadObjectives(F);
    X.ReadGeometry(F);
    X.ReadLayout(F);
    ReadSolutions(F);
    F.ReadConfiguration();

    if (CT.sourceNode<n) SetSourceNode(CT.sourceNode);
    else SetSourceNode(NoNode);

    if (CT.targetNode<n) SetTargetNode(CT.targetNode);
    else SetTargetNode(NoNode);

    if (CT.rootNode<n) SetRootNode(CT.rootNode);
    else SetRootNode(NoNode);
    
    F.Scan(); // balanced_fnw

    X.SetCOrientation(1);
             
    int l = strlen(impFileName)-4;
    char* tmpLabel = new char[l+1];
    memcpy(tmpLabel,impFileName,l);
    tmpLabel[l] = 0;
    SetLabel(tmpLabel);
    delete[] tmpLabel;

    CT.SetMaster(Handle());

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
}


unsigned long balancedFNW::Size() throw()
{   
    return
          sizeof(balancedFNW)
        + goblinDataObject::Allocated()
        + abstractMixedGraph::Allocated()
        + abstractDiGraph::Allocated()
        + abstractBalancedFNW::Allocated();
}


unsigned long balancedFNW::Allocated() throw()
{   
    return 0;
}


TArc balancedFNW::InsertArc(TNode u,TNode v) throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    X.InsertArc(v^1,u^1,X.MaxUCap(),0,X.MaxLength());
    m++;
    return X.InsertArc(u,v,X.MaxUCap(),0,X.MaxLength());
}


TArc balancedFNW::InsertArc(TNode u,TNode v,TCap uu,TFloat cc)
    throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    X.InsertArc(v^1,u^1,uu,0,cc);
    m++;
    return X.InsertArc(u,v,uu,0,cc);
}



TArc balancedFNW::InsertArc(TNode u,TNode v,TCap uu,TCap ll,TFloat cc)
    throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    X.InsertArc(v^1,u^1,uu,ll,cc);
    m++;
    return X.InsertArc(u,v,uu,ll,cc);
}


void balancedFNW::DeleteNode(TNode v) throw(ERRange)
{
    ReleaseAll();

    if (v&1)
    {
        X.DeleteNode(v);
        X.DeleteNode(v^1);
    }
    else
    {
        X.DeleteNode(v^1);
        X.DeleteNode(v);
    }

    n = X.N()-X.NI();
    ni = X.NI();
    m = X.M();
}


TFloat balancedFNW::Length(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("Length",a);

    #endif

    if (X.CLength())
    {
        TFloat tmpLength = X.MaxLength();

        if (a&1) return -tmpLength;
        else return tmpLength;
    }
    
    TFloat tmpLength = X.Length(a);
    
    if (a&1) return -tmpLength;
    else return tmpLength;
}


void balancedFNW::Symmetrize() throw()
{ 
    LogEntry(LOG_MAN,"Symmetrising flow...");

    TArc m0 = m>>1;
    for (TArc a=0;a<m0;a++)
    {
        TFloat Lambda = (Flow(4*a+2)-Flow(4*a))/2;
        if (Lambda>0)
        {
            Push(4*a,Lambda);
            Push(4*a+3,Lambda);
        }
        else
        {
            Push(4*a+1,-Lambda);
            Push(4*a+2,-Lambda);
        }
    }
}


void balancedFNW::Relax() throw()
{ 
    LogEntry(LOG_MAN,"Relaxing symmetry...");
}


balancedFNW::~balancedFNW() throw()
{
    LogEntry(LOG_MEM,"...Balanced flow network disallocated");
    
    if (CT.traceLevel==2) Display();
}


splitGraph::splitGraph(abstractDiGraph &G,TNode s,TNode t) throw() :
    goblinDataObject(G.Context()),
    abstractMixedGraph(2*G.N()+2,(TArc)0),
    balancedFNW(G.N()+1)
{                             
    #if defined(_FAILSAVE_)

    if (s>=n) NoSuchNode("SplitGraph",s);
    
    if (t>=n) NoSuchNode("SplitGraph",t);

    #endif

    LogEntry(LOG_MAN,"Generating split graph...");
    
    X.ReSize(2*G.N()+2,2*G.M()+4);
    
    if (G.Dim()>=2)
    {
        for (TNode v=0;v<G.N();v++)
        {
            X.SetC(2*v,0,G.C(v,0)+10);
            X.SetC(2*v,1,G.C(v,1));
            X.SetC(2*v+1,0,G.CMax(0)-G.C(v,0)+10);
            X.SetC(2*v+1,1,2*G.CMax(1)-G.C(v,1)+10);
        }
    
        X.SetC(2*G.N(),0,G.CMax(0)+20);
        X.SetC(2*G.N(),1,G.CMax(1)+5);
        X.SetC(2*G.N()+1,0,0);
        X.SetC(2*G.N()+1,1,G.CMax(1)+5);
    }

    TFloat scap = 0;
    TFloat tcap = 0;

    for (TArc a=0;a<G.M();a++)
    {
        TNode u = G.StartNode(2*a);
        TNode v = G.EndNode(2*a);
        InsertArc(2*u,2*v,G.UCap(2*a),G.LCap(2*a),G.Length(2*a));
        if (u==s) scap += G.UCap(2*a);
        if (v==t) tcap += G.UCap(2*a);
    }

    InsertArc(n-1,2*s,scap,0,0);
    InsertArc(2*t,n-2,tcap,0,0);
}

