
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsbig.h"


biGraph::biGraph(TNode nn1,TNode nn2,TOption options,
    goblinController& thisContext) throw() :
    goblinDataObject(thisContext), abstractMixedGraph(nn1+nn2,(TArc)0),
    abstractBiGraph(nn1,nn2), X(*this)
{
    SetCDemand(1);
    
    LogEntry(LOG_MEM,"...Sparse bigraph instanciated");
}


biGraph::biGraph(const char* impFileName,goblinController& thisContext)
    throw(ERFile,ERParse) :
    goblinDataObject(thisContext), abstractMixedGraph((TNode)0,(TArc)0),
    abstractBiGraph((TNode)0,(TNode)0), X(*this)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    LogEntry(LOG_IO,"Loading bigraph...");
    if (!CT.logIO && CT.logMem) LogEntry(LOG_MEM,"Loading bigraph...");
    
    goblinImport F(impFileName,CT);
    
    F.Scan("bigraph");
    F.Scan("definition");

    F.Scan("nodes");
    TNode *nodes = F.GetTNodeTuple(3);
    n = nodes[0];
    n1 = nodes[1];
    ni = nodes[2];
    n2 = n-n1;
    delete[] nodes;    
    
    F.Scan("arcs");
    TArc *arcs = F.GetTArcTuple(1);
    m = arcs[0];
    delete[] arcs;    

    CheckLimits();
    X.ReDimension(n,ni,m);
        
    X.ReadIncidences(F);
    X.ReadUCap(F);
    X.ReadLCap(F);
    X.ReadDemand(F);
    X.ReadOrientation(F);
    
    F.Scan(); // definition
    
    X.ReadObjectives(F);
    X.ReadGeometry(F);
    X.ReadLayout(F);
    ReadSolutions(F);
    F.ReadConfiguration();

    if (CT.sourceNode<n) SetSourceNode(CT.sourceNode);
    else SetSourceNode(NoNode);

    if (CT.targetNode<n) SetTargetNode(CT.targetNode);
    else SetTargetNode(NoNode);

    if (CT.rootNode<n) SetRootNode(CT.rootNode);
    else SetRootNode(NoNode);
    
    F.Scan(); // bigraph
    
    int l = strlen(impFileName)-4;
    char* tmpLabel = new char[l+1];
    memcpy(tmpLabel,impFileName,l);
    tmpLabel[l] = 0;
    SetLabel(tmpLabel);
    delete[] tmpLabel;
    CT.SetMaster(Handle());

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
}


biGraph::biGraph(abstractBiGraph &G,TOption options) throw() :
    goblinDataObject(G.Context()), abstractMixedGraph(G.N(),(TArc)0),
    abstractBiGraph(G.N1(),(TNode)0), X(*this)
{
    X.ReSize(G.N(),G.M());
    
    if (options & OPT_CLONE)
    {
        for (TNode u=0;u<n;u++)
        {
            SetDemand(u,G.Demand(u));

            for (char i=0;i<G.Dim();i++) SetC(u,i,G.C(u,i));
        }

        for (TArc a=0;a<G.M();a++)
        {
            TNode u = G.StartNode(2*a);
            TNode v = G.EndNode(2*a);
            TCap tmpCap = InfCap;

            if (options & OPT_SUB) tmpCap = TCap(G.Sub(2*a));
            else tmpCap = G.UCap(2*a);

            InsertArc(u,v,tmpCap,G.LCap(2*a),G.Length(2*a));
        }
        
        for (TNode u=0;u<n;u++)
        {
            TArc a = G.First(u);
            do
            {
                TArc a2 = G.Right(a,u);
                SetRight(a,a2);
                a = a2;
            }
            while (a!=G.First(u));
            
            SetFirst(u,a);
        }
        
        LogEntry(LOG_MEM,"...Bigraph clone generated");
    }
    else
    {
        LogEntry(LOG_MAN,"Computing underlying bigraph...");
        
        TNode *adjacent = new TNode[n];
        n1 = n2 = 0;

        for (TNode w=0;w<n;w++)
        {
            adjacent[w] = NoNode;

            if (G.Outer(w)) n1++;
            else n2++;
        }
    
        THandle H = G.Investigate();
        goblinIterator &I = G.Iterator(H);
        for (TNode u=0;u<n;u++)
        {
            for (char i=0;i<G.Dim();i++) SetC(u,i,G.C(u,i));

            if (G.Outer(u))
            {
                while (I.Active(u))
                {
                    TArc a = I.Read(u);
                    TNode v = G.EndNode(a);
                    TCap tmpCap = InfCap;

                    if (options & OPT_SUB) tmpCap = TCap(G.Sub(a));
                    else tmpCap = G.UCap(a);

                    if (tmpCap>0 && u<v && G.Outer(v) &&
                        (adjacent[v]!=u || (options & OPT_PARALLELS)))
                    {
                        InsertArc(u,v,tmpCap,G.LCap(a),G.Length(a));
                        adjacent[v] = u;
                    }
                }
            }
        }
        G.Close(H);

        delete adjacent;

        X.ReSize(N(),M());
    }
    
    if (CT.traceLevel==2) Display();
}


unsigned long biGraph::Size() throw()
{   
    return
          sizeof(biGraph)
        + goblinDataObject::Allocated()
        + abstractMixedGraph::Allocated()
        + abstractGraph::Allocated()
        + abstractBiGraph::Allocated();
}


unsigned long biGraph::Allocated() throw()
{   
    return 0;
}


TNode biGraph::SwapNode(TNode u) throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=n) NoSuchNode("SwapNode",u);

    // if (First(u)!=NoArc)
    //    Error(ERR_REJECTED,"SwapNode","Node must be isolated");

    #endif
    
    ReleaseAll();

    TNode v = n1;
    
    if (u<n1)
    {
        n1--;
        v = n1;
    }
    else
    {
        n1++;
    }
    
    if (u!=v) X.SwapNodes(u,v);
   
    return v;
}


TArc biGraph::InsertArc(TNode u,TNode v) throw(ERRange,ERRejected)
{
    ReleaseAdjacencies(); 
    ReleaseEmbedding();
    m++; 
    return X.InsertArc(u,v);
}


TArc biGraph::InsertArc(TNode u,TNode v,TCap uu,TFloat cc) throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    return X.InsertArc(u,v,uu,0,cc);
}


TArc biGraph::InsertArc(TNode u,TNode v,TCap uu,TCap ll,TFloat cc) throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    return X.InsertArc(u,v,uu,ll,cc);
}


void biGraph::DeleteArc(TArc a) throw(ERRange)
{
    ReleaseAll();
    X.DeleteArc(a);
    ni = X.NI();
    m = X.M();
}


void biGraph::DeleteNode(TNode v) throw(ERRange)
{
    ReleaseAll();

    if (v<n1) SwapNode(v);

    X.DeleteNode(v);

    n = X.N()-X.NI();
    ni = X.NI();
    m = X.M();
}


biGraph::~biGraph() throw()
{
    LogEntry(LOG_MEM,"...Sparse bigraph disallocated");
    
    if (CT.traceLevel==2) Display();
}


inducedBigraph::inducedBigraph(abstractMixedGraph& G,TNode c1,TNode c2,
    TOption options) throw(ERRejected) :
    goblinDataObject(G.Context()),
    abstractMixedGraph(TNode(1),TArc(0)),
    abstractBiGraph(TNode(1),TNode(0)),
    biGraph(TNode(1),TNode(0),TOption(0),G.Context())
{
    X.ReSize(G.N(),G.M());

    TNode* mapNodes = new TNode[G.N()];

    for (TNode v=0;v<G.N();v++) mapNodes[v] = NoNode;

    bool first = true;    

    for (TNode u=0;u<G.N();u++)
    {
        if (G.Colour(u)==c1)
        {
            if (!first) InsertNode();
            else first = false;
            
            mapNodes[u] = n-1;

            SetDemand(n-1,G.Demand(u));
            for (char i=0;i<G.Dim();i++) SetC(n-1,i,G.C(u,i));
        }
    }

    n1 = n;
    
    for (TNode u=0;u<G.N();u++)
    {
        if (G.Colour(u)==c2)
        {
            if (!first) InsertNode();
            else first = false;
            
            mapNodes[u] = n-1;

            SetDemand(n-1,G.Demand(u));
            for (char i=0;i<G.Dim();i++) SetC(n-1,i,G.C(u,i));
        }
    }

    goblinHashTable<TArc,TArc>* Adj = NULL;

    if (!(options & OPT_PARALLELS))
        Adj = new goblinHashTable<TArc,TArc>(n*n,G.M(),NoArc,CT);

    for (TArc a=0;a<2*G.M();a++)
    {
        TNode u = G.StartNode(a);
        TNode v = G.EndNode(a);
        TNode u2 = mapNodes[u];
        TNode v2 = mapNodes[v];

        if (G.Colour(u)!=c1 || G.Colour(v)!=c2) continue;
    
        TCap tmpCap = InfCap;

        if (options & OPT_SUB)
            tmpCap = TCap(G.Sub(a));
        else tmpCap = G.UCap(a);

        if (tmpCap<=0) continue;

        TFloat l = G.Length(a);

        if (options & OPT_PARALLELS)
        {
            InsertArc(u2,v2,tmpCap,G.LCap(a),l);
            continue;
        }

        TArc j1 = n*u2+v2;
        TArc a1 = Adj -> Key(j1);

        if (a1!=NoArc)
        {
            if (Length(2*a1)>l)
            {
                SetLength(2*a1,l);
                SetUCap(2*a1,tmpCap);
                SetLCap(2*a1,G.LCap(a));
            }

            continue;
        }

        TArc j2 = n*v2+u2;
        TArc a2 = Adj -> Key(j2);

        if (G.Orientation(a)==0 && a2!=NoArc)
        {
            if (Length(2*a2)>l)
            {
                SetLength(2*a2,l);
                SetUCap(2*a2,tmpCap);
                SetLCap(2*a2,G.LCap(a));
            }

            continue;
        }

        a1 = InsertArc(u2,v2,tmpCap,G.LCap(a),l);
        Adj -> ChangeKey(j1,a1);
    }

    delete[] mapNodes;

    if (Adj!=NULL) delete Adj;

    X.ReSize(N(),M());

    LogEntry(LOG_MEM,"...Induced bigraph instanciated");
}
