
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsdig.h"


diGraph::diGraph(TNode nn,TOption options,goblinController& thisContext,
    bool mmode) throw() :
    goblinDataObject(thisContext), abstractMixedGraph(nn,(TArc)0), X(*this)
{
    X.SetCDemand(0);
    X.SetCOrientation(1);
    mode = mmode;
    
    LogEntry(LOG_MEM,"...Sparse digraph instanciated");
}


diGraph::diGraph(const char* impFileName,goblinController& thisContext)
    throw(ERFile,ERParse) :
    goblinDataObject(thisContext), abstractMixedGraph((TNode)0,(TArc)0), X(*this)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    LogEntry(LOG_IO,"Loading digraph...");
    if (!CT.logIO && CT.logMem) LogEntry(LOG_MEM,"Loading digraph...");
    
    goblinImport F(impFileName,CT);

    F.Scan("digraph");
    F.Scan("definition");

    F.Scan("nodes");
    TNode *nodes = F.GetTNodeTuple(3);
    n = nodes[0];
    ni = nodes[2];
    delete[] nodes;    
    
    F.Scan("arcs");
    TArc *arcs = F.GetTArcTuple(1);
    m = arcs[0];
    delete[] arcs;
        
    CheckLimits();
    X.ReDimension(n,ni,m);
    
    X.ReadIncidences(F);
    X.ReadUCap(F);
    X.ReadLCap(F);
    X.ReadDemand(F);
    X.ReadOrientation(F);
    
    F.Scan(); // definition
    
    X.ReadObjectives(F);
    X.ReadGeometry(F);
    X.ReadLayout(F);
    ReadSolutions(F);
    F.ReadConfiguration();

    if (CT.sourceNode<n) SetSourceNode(CT.sourceNode);
    else SetSourceNode(NoNode);

    if (CT.targetNode<n) SetTargetNode(CT.targetNode);
    else SetTargetNode(NoNode);

    if (CT.rootNode<n) SetRootNode(CT.rootNode);
    else SetRootNode(NoNode);
    
    F.Scan(); // digraph

    X.SetCOrientation(1);
    
    int l = strlen(impFileName)-4;
    char* tmpLabel = new char[l+1];
    memcpy(tmpLabel,impFileName,l);
    tmpLabel[l] = 0;
    SetLabel(tmpLabel);
    delete[] tmpLabel;
    CT.SetMaster(Handle());

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
}


diGraph::diGraph(abstractMixedGraph &G,TOption options) throw() :
    goblinDataObject(G.Context()), abstractMixedGraph(G.N(),(TArc)0), X(*this)
{
    X.SetCOrientation(1);
    
    X.ReSize(G.N(),2*G.M());
    
    if (options & OPT_CLONE)
    {
        for (TNode u=0;u<n;u++)
        {
            SetDemand(u,G.Demand(u));

            for (char i=0;i<G.Dim();i++) SetC(u,i,G.C(u,i));
        }

        for (TArc a=0;a<G.M();a++)
        {
            TNode u = G.StartNode(2*a);
            TNode v = G.EndNode(2*a);
            TCap tmpCap = InfCap;

            if (options & OPT_SUB)
                tmpCap = TCap(G.Sub(2*a));
            else tmpCap = G.UCap(2*a);

            InsertArc(u,v,tmpCap,G.LCap(2*a),G.Length(2*a));
        }
        
        for (TNode u=0;u<n;u++)
        {
            TArc a = G.First(u);
            do
            {
                TArc a2 = G.Right(a,u);
                SetRight(a,a2);
                a = a2;
            }
            while (a!=G.First(u));
            
            SetFirst(u,a);
        }
        
        LogEntry(LOG_MEM,"...Digraph clone generated");
    }
    else
    {
        LogEntry(LOG_MAN,"Computing orientation...");
        
        TNode* adjacent = new TNode[n];
        for (TNode w=0;w<n;w++) adjacent[w] = NoNode;
        
        THandle H = G.Investigate();
        for (TNode u=0;u<n;u++)
        {
            SetDemand(u,G.Demand(u));

            for (char i=0;i<G.Dim();i++) SetC(u,i,G.C(u,i));
        
            while (G.Active(H,u))
            {
                TArc a = G.Read(H,u);
                TCap tmpCap = InfCap;

                if (options & OPT_SUB) tmpCap = TCap(G.Sub(a));
                else tmpCap = G.UCap(a);

                TNode v = G.EndNode(a);

                if (tmpCap>0 && !G.Blocking(a) && ((options
                    & OPT_PARALLELS) || adjacent[v]!=u))
                {
                    if (G.IsBipartite())
                    {
                        if (u>v) continue; 
                        InsertArc(u,v,tmpCap,G.LCap(a),G.Length(a&(a^1)));
                    }
                    else InsertArc(u,v,tmpCap,G.LCap(a),G.Length(a&(a^1)));

                    adjacent[v] = u;
                }
            }
        }
        G.Close(H);

        delete[] adjacent;

        X.ReSize(N(),M());
    }
    
    if (CT.traceLevel==2) Display();
}


TArc diGraph::InsertArc(TNode u,TNode v) throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    return X.InsertArc(u,v);
}


TArc diGraph::InsertArc(TNode u,TNode v,TCap uu,TFloat cc)
    throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    return X.InsertArc(u,v,uu,0,cc);
}


TArc diGraph::InsertArc(TNode u,TNode v,TCap uu,TCap ll,TFloat cc)
    throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    return X.InsertArc(u,v,uu,ll,cc);
}


void diGraph::DeleteArc(TArc a) throw(ERRange)
{
    ReleaseAll();
    X.DeleteArc(a);
    ni = X.NI();
    m = X.M();
}

    
diGraph::~diGraph() throw()
{
    LogEntry(LOG_MEM,"...Sparse digraph disallocated");
    
    if (CT.traceLevel==2 && mode==0) Display();
}


unsigned long diGraph::Size() throw()
{   
    return 
          sizeof(diGraph)
        + goblinDataObject::Allocated()
        + abstractMixedGraph::Allocated()
        + abstractDiGraph::Allocated();
}


unsigned long diGraph::Allocated() throw()
{   
    return 0;
}


completeOrientation::completeOrientation(abstractMixedGraph &G,TOption options)
    throw() :
    goblinDataObject(G.Context()),
    abstractMixedGraph(G.N(),(TArc)0),
    diGraph(G.N(),0,G.Context(),1)
{                             
    LogEntry(LOG_MAN,"Orienting graph arcs...");
    
    for (TArc a=0;a<2*G.M();a++)
        if (!G.Blocking(a) || (options & OPT_REVERSE))
            InsertArc(G.StartNode(a),G.EndNode(a),G.UCap(a),0,G.Length(a&(a^1)));
            
    for (TNode v=0;v<n;v++)
    {
        SetDemand(v,G.Demand(v));

        for (char i=0;i<G.Dim();i++) SetC(v,i,G.C(v,i));
    }
    
    type = 2;

    if (G.IsDirected()) type = 0;
    if (G.IsUndirected()) type = 1;
    
    if (type==2)
    {
        origin = new TArc[m];
        TArc i = 0;

        for (TArc a=0;a<2*G.M();a++)
            if (!G.Blocking(a) || (options & OPT_REVERSE)) origin[i++] = a;

        LogEntry(LOG_MEM,"...Arc mapping allocated");
    }
    else origin = NULL;
}


completeOrientation::~completeOrientation() throw()
{
    if (origin!=NULL) delete[] origin;
    
    LogEntry(LOG_MEM,"...Complete orientation disallocated");
    
    if (CT.traceLevel==2) Display();
}


inducedOrientation::inducedOrientation(abstractMixedGraph &G,TOption options)
    throw() :
    goblinDataObject(G.Context()),
    abstractMixedGraph(G.N(),(TArc)0),
    diGraph(G.N(),0,G.Context(),1)
{                             
    LogEntry(LOG_MAN,"Orienting graph arcs by colours...");
    
    for (TArc a=0;a<2*G.M();a++)
        if (!G.Blocking(a) &&
            (G.Blocking(a^1) || G.Colour(G.StartNode(a))<G.Colour(G.EndNode(a)))
           )
        {
            InsertArc(G.StartNode(a),G.EndNode(a),G.UCap(a),0,G.Length(a&(a^1)));
        }
            
    for (TNode v=0;v<n;v++)
    {
        SetDemand(v,G.Demand(v));

        for (char i=0;i<G.Dim();i++) SetC(v,i,G.C(v,i));
    }
}


unsigned long completeOrientation::Size() throw()
{   
    return 
          sizeof(completeOrientation)
        + goblinDataObject::Allocated()
        + abstractMixedGraph::Allocated()
        + abstractDiGraph::Allocated()
        + diGraph::Allocated()
        + completeOrientation::Allocated();
}


unsigned long completeOrientation::Allocated() throw()
{   
    if (origin!=NULL) return sizeof(TArc)*m;
    else return 0;
}


TArc completeOrientation::OriginalArc(TArc a) throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*m) NoSuchArc("OriginalArc",a);

    #endif
    
    if (type==1) return (a>>1);
    if (origin!=NULL) return origin[a>>1];
    return a;
}


nodeSplitting::nodeSplitting(abstractMixedGraph &G,TOption options)
    throw() :
    goblinDataObject(G.Context()),
    abstractMixedGraph(2*G.N(),(TArc)0),
    diGraph(2*G.N(),0,G.Context(),1)
{                             
    LogEntry(LOG_MAN,"Splitting graph nodes...");
    
    for (TArc a=0;a<2*G.M();a++)
        if (!G.Blocking(a) || options)
            InsertArc(2*G.StartNode(a)+1,2*G.EndNode(a),G.UCap(a),0,0);

    for (TNode v=0;v<G.N();v++) InsertArc(2*v,2*v+1,G.Demand(v),0,0);
    
    if (G.Dim()>=2)
    {
        for (TNode v=0;v<G.N();v++)
        {
            SetC(2*v,0,(G.C(v,0)));
            SetC(2*v,1,(G.C(v,1)));
            SetC(2*v+1,0,(G.C(v,0))+5);
            SetC(2*v+1,1,(G.C(v,1))+3);
        }
    }
    
    if (CT.traceLevel==2) Display();
}
