
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, December 1998.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsgra.h"


graph::graph(TNode nn,TOption options,goblinController& thisContext,bool mmode)
    throw() : goblinDataObject(thisContext), abstractMixedGraph(nn,(TArc)0),
    X(*this)
{
    SetCDemand(1);
    mode = mmode;
    
    LogEntry(LOG_MEM,"...Sparse graph instanciated");
}
    

graph::graph(const char* impFileName,goblinController& thisContext)
    throw(ERFile,ERParse) :
    goblinDataObject(thisContext), abstractMixedGraph((TNode)0,(TArc)0),
    X(*this)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    LogEntry(LOG_IO,"Loading graph...");

    if (!CT.logIO && CT.logMem) LogEntry(LOG_MEM,"Loading graph...");

    goblinImport F(impFileName,CT);
    
    F.Scan("graph");
    F.Scan("definition");

    F.Scan("nodes");
    TNode *nodes = F.GetTNodeTuple(3);
    n = nodes[0];
    ni = nodes[2];
    delete[] nodes;    
    
    F.Scan("arcs");
    TArc *arcs = F.GetTArcTuple(1);
    m = arcs[0];
    delete[] arcs;
    
    CheckLimits();    
    X.ReDimension(n,ni,m);
    
    X.ReadIncidences(F);
    X.ReadUCap(F);
    X.ReadLCap(F);
    X.ReadDemand(F);
    X.ReadOrientation(F);
    
    F.Scan(); // definition
    
    X.ReadObjectives(F);
    X.ReadGeometry(F);
    X.ReadLayout(F);
    ReadSolutions(F);
    F.ReadConfiguration();

    if (CT.sourceNode<n) SetSourceNode(CT.sourceNode);
    else SetSourceNode(NoNode);

    if (CT.targetNode<n) SetTargetNode(CT.targetNode);
    else SetTargetNode(NoNode);

    if (CT.rootNode<n) SetRootNode(CT.rootNode);
    else SetRootNode(NoNode);
    
    F.Scan(); // graph
    
    int l = strlen(impFileName)-4;
    char* tmpLabel = new char[l+1];
    memcpy(tmpLabel,impFileName,l);
    tmpLabel[l] = 0;
    SetLabel(tmpLabel);
    delete[] tmpLabel;
    CT.SetMaster(Handle());

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
}


graph::graph(abstractMixedGraph& G,TOption options) throw() :
    goblinDataObject(G.Context()), abstractMixedGraph(G.N(),TArc(0)), X(*this)
{
    X.ReSize(G.N(),G.M());
    
    if (options & OPT_CLONE)
    {
        for (TNode u=0;u<n;u++)
        {
            SetDemand(u,G.Demand(u));

            for (char i=0;i<G.Dim();i++) SetC(u,i,G.C(u,i));
        }

        for (TArc a=0;a<G.M();a++)
        {
            TNode u = G.StartNode(2*a);
            TNode v = G.EndNode(2*a);
            TCap tmpCap = InfCap;

            if (options & OPT_SUB)
                tmpCap = (TCap)G.Sub(2*a);
            else tmpCap = G.UCap(2*a);

            InsertArc(u,v,tmpCap,G.LCap(2*a),G.Length(2*a));
        }
        
        for (TNode u=0;u<n;u++)
        {
            TArc a = G.First(u);
            do
            {
                TArc a2 = G.Right(a,u);
                SetRight(a,a2);
                a = a2;
            }
            while (a!=G.First(u));
            
            SetFirst(u,a);
        }

        LogEntry(LOG_MEM,"...Graph clone generated");
    }
    else
    {
        LogEntry(LOG_MAN,"Computing underlying graph...");
        
        TNode* adjacent = new TNode[n];
        for (TNode w=0;w<n;w++) adjacent[w] = NoNode;
    
        THandle H = G.Investigate();
        goblinIterator &I = G.Iterator(H);
        for (TNode u=0;u<n;u++)
        {
            for (char i=0;i<G.Dim();i++) SetC(u,i,G.C(u,i));
        
            while (I.Active(u))
            {
                TArc a = I.Read(u);
                TNode v = G.EndNode(a);
                TCap tmpCap = InfCap;

                if (options & OPT_SUB)
                    tmpCap = (TCap)G.Sub(a);
                else tmpCap = G.UCap(a);

                if (tmpCap>0 && u<v && (adjacent[v]!=u ||
                    (options & OPT_PARALLELS)))
                {
                    InsertArc(u,v,tmpCap,G.LCap(a),G.Length(a));
                    adjacent[v] = u;
                }
            }
        }
        G.Close(H);

        delete[] adjacent;

        X.ReSize(N(),M());
    }
}
    

unsigned long graph::Size() throw()
{   
    return
          sizeof(graph)
        + goblinDataObject::Allocated()
        + abstractMixedGraph::Allocated()
        + abstractGraph::Allocated();
}


unsigned long graph::Allocated() throw()
{   
    return 0;
}


TArc graph::InsertArc(TNode u,TNode v) throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    return X.InsertArc(u,v);
}


TArc graph::InsertArc(TNode u,TNode v,TCap uu,TFloat cc)
    throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    return X.InsertArc(u,v,uu,0,cc);
}


TArc graph::InsertArc(TNode u,TNode v,TCap uu,TCap ll,TFloat cc)
    throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    return X.InsertArc(u,v,uu,ll,cc);
}


void graph::DeleteArc(TArc a) throw(ERRange)
{
    ReleaseAll();
    X.DeleteArc(a);
    ni = X.NI();
    m = X.M();
}


graph::~graph() throw()
{
    LogEntry(LOG_MEM,"...Sparse graph disallocated");
    
    if (CT.traceLevel==2 && !mode) Display();
}
    

complementaryGraph::complementaryGraph(abstractMixedGraph &G,TOption options)
    throw(ERRejected) :
    goblinDataObject(G.Context()),
    abstractMixedGraph(G.N(),(TArc)0),
    graph(G.N(),0,G.Context())
{                             
    #if defined(_FAILSAVE_)

    if (G.N()*(G.N()-1)/2>=CT.MaxArc())
        Error(ERR_REJECTED,"complementaryGraph","Number of arcs is out of range");

    #endif

    LogEntry(LOG_MAN,"Generating complementary graph...");

    if (G.Dim()>=2)
    {
        for (TNode v=0;v<n;v++)
        {
            SetC(v,0,(G.C(v,0)));
            SetC(v,1,(G.C(v,1)));
        }
    }
    
    for (TNode u=0;u<n;u++)
    {
        for (TNode v=u+1;v<n;v++)
        {
            if (G.Adjacency(u,v)==NoArc && G.Adjacency(v,u)==NoArc)
            {
                if (CT.Rand(2)) InsertArc(v,u);
                else InsertArc(u,v);
            }
        }
    }
    
    if (CT.traceLevel==2) Display();
}


lineGraph::lineGraph(abstractMixedGraph &G,TOption options) throw(ERRejected) :
    goblinDataObject(G.Context()),
    abstractMixedGraph(G.M(),(TArc)0),
    graph(G.M(),0,G.Context())
{                             
    #if defined(_FAILSAVE_)

    if (G.M()>=CT.MaxNode())
        Error(ERR_REJECTED,"lineGraph","Number of arcs is out of range");

    #endif

    LogEntry(LOG_MAN,"Generating line graph...");

    if (G.Dim()>=2)
    {
        for (TNode v=0;v<G.M();v++)
        {
            SetC(v,0,(G.C(G.StartNode((TArc)2*v),0)+G.C(G.EndNode((TArc)2*v),0))/2);
            SetC(v,1,(G.C(G.StartNode((TArc)2*v),1)+G.C(G.EndNode((TArc)2*v),1))/2);
        }
    }
    
    for (TArc a1=0;a1<G.M()-1;a1++)
    {
        for (TArc a2=a1+1;a2<G.M();a2++)
        {
            if (G.StartNode(2*a1)==G.StartNode(2*a2) ||
                G.StartNode(2*a1)==G.EndNode(2*a2)   ||
                G.EndNode(2*a1)==G.StartNode(2*a2)   ||
                G.EndNode(2*a1)==G.EndNode(2*a2))
            {
                if (CT.Rand(2)) InsertArc((TNode)a1,(TNode)a2);
                else InsertArc((TNode)a2,(TNode)a1);
            }
        }
    }
    
    if (CT.traceLevel==2) Display();
}


dualGraph::dualGraph(abstractMixedGraph &G,TOption options) throw(ERRejected) :
    goblinDataObject(G.Context()),
    abstractMixedGraph(G.M()-G.N()+2,0),
    graph(G.M()-G.N()+2,0,G.Context())
{                             
    #if defined(_FAILSAVE_)

    if (G.M()-G.N()+2>=CT.MaxNode())
        Error(ERR_REJECTED,"dualGraph","Number of regions is out of range");

    #endif

    LogEntry(LOG_MAN,"Generating dual graph...");

    TNode* region = new TNode[2*G.M()];
    for (TArc a=0;a<2*G.M();a++) region[a] = NoNode;
    
    OpenFold();
    
    TNode regionIndex = 0;
    for (TArc a=0;a<2*G.M();a++)
    {
        if (region[a]!=NoNode) continue;

        TArc regionLength = 1;
        TArc a2 = G.Right(a^1,G.EndNode(a));
        region[a2] = regionIndex;
        TFloat sumX = G.C(G.StartNode(a2),0);
        TFloat sumY = G.C(G.StartNode(a2),1);

        #if defined(_LOGGING_)

        THandle LH = NoHandle;

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"Region %ld = (%ld",regionIndex,a2);
            LH = LogStart(LOG_METH2,CT.logBuffer);
        }

        #endif

        while (a2!=a)
        {
            regionLength++;
            a2 = G.Right(a2^1,G.EndNode(a2));
            region[a2] = regionIndex;
            sumX += G.C(G.StartNode(a2),0);
            sumY += G.C(G.StartNode(a2),1);

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,",%ld",a2);
                LogAppend(LH,CT.logBuffer);
            }

            #endif
        }

        #if defined(_LOGGING_)

        if (CT.logMeth>1) LogEnd(LH,")");

        #endif
        
        if (G.Dim()>=2)
        {
            SetC(regionIndex,0,sumX/regionLength);
            SetC(regionIndex,1,sumY/regionLength);
        }
        
        regionIndex++;
    }
    
    #if defined(_FAILSAVE_)

    if (G.M()-G.N()+2!=regionIndex)
    {
        CloseFold();
        Error(ERR_REJECTED,"dualGraph","Not a planar embedding");
    }

    #endif

    TArc* predArc = new TArc[2*G.M()];
    
    for (TArc a=0;a<G.M();a++)
    {
        TNode u = region[2*a];
        TNode v = region[2*a+1];

        InsertArc(v,u);
        
        predArc[2*a]   = G.Right(2*a,G.StartNode(2*a));
        predArc[2*a+1] = G.Right(2*a+1,G.StartNode(2*a+1));
    }

    X.Planarize(predArc);
    
    CloseFold();

    delete[] region;
    delete[] predArc;
        
    if (G.Dim()>=2) AutoArcAlignment();
    
    if (CT.traceLevel==2) Display();
}


tiling::tiling(abstractMixedGraph &G,TOption options,TNode x,TNode y)
    throw() : goblinDataObject(G.Context()),
    abstractMixedGraph((TNode)(x+1)*(y+1)+x*y*(G.N()-4),(TArc)0),
    graph((TNode)(x+1)*(y+1)+x*y*(G.N()-4),options,G.Context())
{                             
    ReSize(n,2*x*y+x+y+x*y*G.M());
    
    TNode offset = (x+1)*(y+1);
    for (TNode j=0;j<y;j++)
    {
        for (TNode i=0;i<x;i++)
        {
            for (TNode v0=0;v0<G.N();v0++)
            {
                if ((j==0 || !(v0==1 || v0==0)) &&
                    (i==0 || !(v0==2 || v0==0))    )
                {
                    TNode v = Index(i,j,x,offset,v0);

                    if (G.Dim()>=2)
                    {
                        SetC(v,0,i*(G.CMax(0))+(G.C(v0,0)));
                        SetC(v,1,j*(G.CMax(1))+(G.C(v0,1)));
                    }
                }
            }
            
            for (TArc a=0;a<G.M();a++)
            {
            	TNode u0 = G.StartNode(2*a);
            	TNode v0 = G.EndNode(2*a);
		
                TNode u = Index(i,j,x,offset,u0);
                TNode v = Index(i,j,x,offset,v0);
		
                if ((j==0 || !((u0==0 && v0==1) || (v0==0 && u0==1))) &&
                    (i==0 || !((u0==0 && v0==2) || (v0==0 && u0==2)))    )
                {
                    if (CT.Rand(2)) InsertArc(u,v);
                    else InsertArc(v,u);
                }
            }
            offset += G.N()-4;
        }
    }
    
    if (CT.traceLevel==2) Display();
}


TNode tiling::Index (TNode i,TNode j,TNode x,TNode offset,TNode w) throw()
{
    switch (w)
    {
        case 0: return j*(x+1)+i;
        case 1: return j*(x+1)+i+1;
        case 2: return (j+1)*(x+1)+i;
        case 3: return (j+1)*(x+1)+i+1;
        default: return offset+w-4;
    }
}


voronoiDiagram::voronoiDiagram(abstractMixedGraph& GG) throw() :
    goblinDataObject(GG.Context()),
    abstractMixedGraph(GG.VoronoiRegions(),(TArc)0),
    graph((TNode)0,TOption(0),GG.Context(),true),
    G(GG)
{
    LogEntry(LOG_MAN,"Contracting partial trees...");

    ReSize(G.N(),G.M());
    
    TNode* mapNodes = new TNode[G.N()];

    for (TNode v=0;v<G.N();v++) mapNodes[v] = NoNode;

    TNode i = 0;

    for (TNode v=0;v<G.N();v++)
    {
        if (mapNodes[G.Find(v)]==NoNode) mapNodes[G.Find(v)] = i++;
        mapNodes[v] = mapNodes[G.Find(v)];
        
        if (G.Demand(v)!=0)
        {
            for (char i=0;i<G.Dim();i++) SetC(mapNodes[v],i,G.C(v,i));
        }
    }
    
    revMap = new TArc[G.M()];
    goblinHashTable<TArc,TArc> Adj(i*i,G.M(),NoArc,CT);

    for (TArc a=0;a<G.M();a++)
    {
        TNode u = G.StartNode(2*a);
        TNode v = G.EndNode(2*a);
        TNode u2 = mapNodes[G.Find(u)];
        TNode v2 = mapNodes[G.Find(v)];
        
        if (u2!=v2)
        {
            TFloat l = G.Dist(u)+G.Dist(v)+G.Length(2*a);

            TArc j = i*u2+v2;

            if (u2>v2) j = i*v2+u2;
            
            TArc a2 = Adj.Key(j);

            if (a2==NoArc)
            {
                a2 = InsertArc(u2,v2,TCap(1),l);
                Adj.ChangeKey(j,a2);
                revMap[a2] = a;
            }
            else
            {
                if (Length(2*a2)>l)
                {
                    SetLength(2*a2,l);
                    revMap[a2] = a;
                }
            }
        }
    }

    delete[] mapNodes;

    ReSize(N(),M());
    
    if (CT.traceLevel==2) Display();
}


TFloat voronoiDiagram::UpdateSubgraph() throw()
{
    LogEntry(LOG_METH,"Mapping tree to original graph...");

    TFloat ret = 0;
    
    G.InitSubgraph();
    
    for (TNode v=0;v<n;v++)
    {
        if (P[v]==NoArc) continue;
        
        TArc a = revMap[P[v]>>1];

        if (G.Sub(2*a)==0) G.AddArc(2*a,1);
        
        TNode u = G.StartNode(2*a);

        while (G.Demand(u)==0)
        {
            TArc a = G.Pred(u);
            if (G.Sub(a)==0) G.AddArc(a,1);
            u = G.StartNode(a);
        }
        
        u = G.EndNode(2*a);

        while (G.Demand(u)==0)
        {
            TArc a = G.Pred(u);
            if (G.Sub(a)==0) G.AddArc(a,1);
            u = G.StartNode(a);
        }
    }
    
    return ret;
}


voronoiDiagram::~voronoiDiagram() throw()
{
    if (CT.traceLevel==2) Display();
    
    delete[] revMap;

    LogEntry(LOG_MAN,"...Voronoi diagram deleted");
}


triangularGraph::triangularGraph(TNode nn,TOption options,
    goblinController &thisContext) throw() :
    goblinDataObject(thisContext),
    abstractMixedGraph((TNode)0,(TArc)0),
    graph((TNode)0,options,thisContext)
{
    LogEntry(LOG_MAN,"Generating triangular graph...");
    
    TNode** node = new TNode*[nn];
    
    for (TNode i=0;i<nn;i++)
    {
        node[i] = new TNode[nn];
        for (TNode j=i+1;j<nn;j++)
        {
            node[i][j] = InsertNode();
        
            for (TNode k=0;k<i;k++)   InsertArc(node[i][j],node[k][j]);
            for (TNode k=i+1;k<j;k++) InsertArc(node[i][j],node[i][k]);
            for (TNode k=0;k<i;k++)   InsertArc(node[i][j],node[k][i]);
        }
    }
    
    delete[] node;
}
