
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, September 2001.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsmix.h"


// ------------------------------------------------------


mixedGraph::mixedGraph(TNode nn,TOption options,goblinController& thisContext)
    throw() : goblinDataObject(thisContext), abstractMixedGraph(nn,(TArc)0),
    X(*this)
{
    X.SetCDemand(0);
    
    LogEntry(LOG_MEM,"...Mixed graph instanciated");
}


mixedGraph::mixedGraph(const char* impFileName,goblinController &thisContext)
    throw(ERFile,ERParse) : goblinDataObject(thisContext),
    abstractMixedGraph(TNode(0),TArc(0)), X(*this)
{
    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Enable();
    
    #endif

    LogEntry(LOG_IO,"Loading mixed graph...");
    if (!CT.logIO && CT.logMem) LogEntry(LOG_MEM,"Loading mixed graph...");
    
    goblinImport F(impFileName,CT);

    F.Scan("mixed");
    F.Scan("definition");

    F.Scan("nodes");
    TNode *nodes = F.GetTNodeTuple(3);
    n = nodes[0];
    ni = nodes[2];
    delete[] nodes;    
    
    F.Scan("arcs");
    TArc *arcs = F.GetTArcTuple(1);
    m = arcs[0];
    delete[] arcs;
        
    CheckLimits();
    X.ReDimension(n,ni,m);
    
    X.ReadIncidences(F);
    X.ReadUCap(F);
    X.ReadLCap(F);
    X.ReadDemand(F);
    X.ReadOrientation(F);
    
    F.Scan(); // definition
    
    X.ReadObjectives(F);
    X.ReadGeometry(F);
    X.ReadLayout(F);
    ReadSolutions(F);
    F.ReadConfiguration();

    if (CT.sourceNode<n) SetSourceNode(CT.sourceNode);
    else SetSourceNode(NoNode);

    if (CT.targetNode<n) SetTargetNode(CT.targetNode);
    else SetTargetNode(NoNode);

    if (CT.rootNode<n) SetRootNode(CT.rootNode);
    else SetRootNode(NoNode);
    
    F.Scan(); // mixed

    int l = strlen(impFileName)-4;
    char* tmpLabel = new char[l+1];
    memcpy(tmpLabel,impFileName,l);
    tmpLabel[l] = 0;
    SetLabel(tmpLabel);
    delete[] tmpLabel;
    CT.SetMaster(Handle());
    
    if (CT.traceLevel==2) Display();

    #if defined(_TIMERS_)

    CT.globalTimer[TimerIO] -> Disable();
    
    #endif
}


unsigned long mixedGraph::Size() throw()
{   
    return 
          sizeof(mixedGraph)
        + goblinDataObject::Allocated()
        + abstractMixedGraph::Allocated();
}


unsigned long mixedGraph::Allocated() throw()
{   
    return 0;
}


TArc mixedGraph::InsertArc(TNode u,TNode v) throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    return X.InsertArc(u,v);
}


TArc mixedGraph::InsertArc(TNode u,TNode v,TCap uu,TFloat cc)
    throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    return X.InsertArc(u,v,uu,0,cc);
}


TArc mixedGraph::InsertArc(TNode u,TNode v,TCap uu,TCap ll,TFloat cc)
    throw(ERRange,ERRejected)
{
    ReleaseAdjacencies();
    ReleaseEmbedding();
    m++;
    return X.InsertArc(u,v,uu,ll,cc);
}


void mixedGraph::DeleteArc(TArc a) throw(ERRange)
{
    ReleaseAll();
    X.DeleteArc(a);
    ni = X.NI();
    m = X.M();
}


mixedGraph::~mixedGraph() throw()
{
    LogEntry(LOG_MEM,"...Mixed graph disallocated");
    
    if (CT.traceLevel==2) Display();
}


inducedSubgraph::inducedSubgraph(abstractMixedGraph& G,TNode c,TOption options)
    throw(ERRejected) :
    goblinDataObject(G.Context()),
    abstractMixedGraph(TNode(1),TArc(0)),
    mixedGraph(TNode(1),TOption(0),G.Context())
{
    X.ReSize(G.N(),G.M());

    TNode* mapNodes = new TNode[G.N()];
    for (TNode v=0;v<G.N();v++) mapNodes[v] = NoNode;

    bool first = true;    
    for (TNode u=0;u<G.N();u++)
    {
        if (G.Colour(u)==c)
        {
            if (!first) InsertNode();
            else first = false;
            
            mapNodes[u] = n-1;

            SetDemand(n-1,G.Demand(u));
            for (char i=0;i<G.Dim();i++) SetC(n-1,i,G.C(u,i));
        };
    };

    goblinHashTable<TArc,TArc>* Adj = NULL;

    if (!(options & OPT_PARALLELS))
        Adj = new goblinHashTable<TArc,TArc>(2*n*n,G.M(),NoArc,CT);

    for (TArc a=0;a<G.M();a++)
    {
        TNode u = G.StartNode(2*a);
        TNode v = G.EndNode(2*a);
        TNode u2 = mapNodes[u];
        TNode v2 = mapNodes[v];

        if (u2==NoNode || v2==NoNode) continue;
    
        TCap tmpCap = InfCap;

        if (options & OPT_SUB)
            tmpCap = TCap(G.Sub(2*a));
        else tmpCap = G.UCap(2*a);

        if (tmpCap<=0) continue;

        TFloat l = G.Length(2*a);

        if (options & OPT_PARALLELS)
        {
            TArc a1 = InsertArc(u2,v2,tmpCap,G.LCap(2*a),l);
            SetOrientation(2*a1,G.Orientation(2*a));
            continue;
        }

        TArc j1 = 2*(n*u2+v2)+G.Orientation(2*a);
        TArc a1 = Adj -> Key(j1);

        if (a1!=NoArc)
        {
            if (Length(2*a1)>l)
            {
                SetLength(2*a1,l);
                SetUCap(2*a1,tmpCap);
                SetLCap(2*a1,G.LCap(2*a));
            }

            continue;
        }

        TArc j2 = 2*(n*v2+u2)+G.Orientation(2*a);
        TArc a2 = Adj -> Key(j2);

        if (G.Orientation(2*a)==0 && a2!=NoArc)
        {
            if (Length(2*a2)>l)
            {
                SetLength(2*a2,l);
                SetUCap(2*a2,tmpCap);
                SetLCap(2*a2,G.LCap(2*a));
            }

            continue;
        }

        a1 = InsertArc(u2,v2,tmpCap,G.LCap(2*a),l);
        SetOrientation(2*a1,G.Orientation(2*a));
        Adj -> ChangeKey(j1,a1);
    }

    delete[] mapNodes;
    if (Adj!=NULL) delete Adj;

    X.ReSize(N(),M());

    LogEntry(LOG_MEM,"...Induced subgraph instanciated");
}


colourContraction::colourContraction(abstractMixedGraph& G,TOption options)
    throw() :
    goblinDataObject(G.Context()),
    abstractMixedGraph(TNode(1),TArc(0)),
    mixedGraph(TNode(1),TOption(0),G.Context())
{
    LogEntry(LOG_MAN,"Contracting colours...");

    ReSize(G.N(),G.M());
    
    TNode* mapColour = new TNode[G.N()];
    for (TNode v=0;v<G.N();v++) mapColour[v] = NoNode;

    TNode j = 0;
    for (TNode v=0;v<G.N();v++)
    {
        if (mapColour[G.Colour(v)]==NoNode)
        {
            if (j>0) InsertNode();

            mapColour[G.Colour(v)] = j++;

            for (char i=0;i<G.Dim();i++)
                SetC(mapColour[G.Colour(v)],i,G.C(v,i));
        }
    }
    
    for (TNode v=0;v<G.N();v++)
        SetColour(mapColour[G.Colour(v)],G.Colour(v));

    goblinHashTable<TArc,TArc>* Adj = NULL;

    if (!(options & OPT_PARALLELS))
        Adj = new goblinHashTable<TArc,TArc>(2*n*n,G.M(),NoArc,CT);

    for (TArc a=0;a<G.M();a++)
    {
        TNode u = G.StartNode(2*a);
        TNode v = G.EndNode(2*a);
        TNode u2 = mapColour[G.Colour(u)];
        TNode v2 = mapColour[G.Colour(v)];

        if (u2==v2) continue;

        TCap tmpCap = InfCap;

        if (options & OPT_SUB)
            tmpCap = TCap(G.Sub(2*a));
        else tmpCap = G.UCap(2*a);

        if (tmpCap<=0) continue;

        TFloat l = G.Length(2*a);

        if (options & OPT_PARALLELS)
        {
            TArc a1 = InsertArc(u2,v2,tmpCap,G.LCap(2*a),l);
            SetOrientation(2*a1,G.Orientation(2*a));
            continue;
        }

        TArc j1 = 2*(n*u2+v2)+G.Orientation(2*a);
        TArc a1 = Adj -> Key(j1);
        
        if (a1!=NoArc)
        {
            if (Length(2*a1)>l)
            {
                SetLength(2*a1,l);
                SetUCap(2*a1,tmpCap);
                SetLCap(2*a1,G.LCap(2*a));
            }

            continue;
        }
        
        TArc j2 = 2*(n*v2+u2)+G.Orientation(2*a);
        TArc a2 = Adj -> Key(j2);

        if (G.Orientation(2*a)==0 && a2!=NoArc)
        {
            if (Length(2*a2)>l)
            {
                SetLength(2*a2,l);
                SetUCap(2*a2,tmpCap);
                SetLCap(2*a2,G.LCap(2*a));
            }

            continue;
        }
        
        a1 = InsertArc(u2,v2,tmpCap,G.LCap(2*a),l);
        SetOrientation(2*a1,G.Orientation(2*a));
        Adj -> ChangeKey(j1,a1);
    }

    delete[] mapColour;

    if (Adj!=NULL) delete Adj;

    ReSize(N(),M());
    
    if (CT.traceLevel==2) Display();
}


explicitSurfaceGraph::explicitSurfaceGraph(abstractMixedGraph& G,
    shrinkingFamily<TNode>& S,TFloat* modLength,TArc* inArc) throw() :
    goblinDataObject(G.Context()),abstractMixedGraph(2*G.N(),(TArc)0),
    mixedGraph(2*G.N(),0,G.Context())
{
    InitPredecessors();

    for (TArc a=0;a<G.M();a++)
    {
        TNode u = G.StartNode(2*a);
        TNode v = G.EndNode(2*a);
        TCap thisCap = 0;
        if (S.Set(u)!=S.Set(v)) thisCap = G.UCap(2*a);
        TArc a2 = NoArc;
        if (modLength!=NULL)
            a2 = InsertArc(S.Set(u),S.Set(v),thisCap,0,modLength[2*a]);
        else
            a2 = InsertArc(S.Set(u),S.Set(v),thisCap,0,0);
        SetOrientation(2*a2,G.Orientation(2*a));
    }
    
    if (G.Dim()>=2)
    {
        TNode v = 0;
        for (;v<n;v++)
        {
            SetC(v,0,-InfFloat);
            SetC(v,1,-InfFloat);
            P[v] = inArc[v];
        }

        for (v=0;v<G.N();v++)
        {
            TNode w = S.Set(v);
            if (First(w)!=NoArc)
            {
                SetC(w,0,(G.C(v,0)));
                SetC(w,1,(G.C(v,1)));
            }
        }
    }
    
    AutoArcAlignment();
}
