
/*
**  This file forms part of the GOBLIN C++ Class Library.
**
**  Written by Christian Fremuth-Paeger, September 1999.
**  
**  Copying, compiling, distribution and modification
**  of this source code is permitted only in accordance
**  with the GOBLIN general licence information.
*/             


#include "spsstruc.h"


sparseGraphStructure::sparseGraphStructure(abstractMixedGraph &GC,
        TOption options) throw() :
    goblinDataObject(GC.Context()), genericGraphStructure(GC)
{
    ni = G.NI();

    SN = new TNode[2*mMax];

    right = new TArc[2*mMax];
    first = new TArc[nMax];
    left = NULL;

    align = NULL;
    interpolate = NULL;
    
    for (TNode v=0;v<nMax;v++) {first[v] = NoArc;};

    sub = NULL;

    LogEntry(LOG_MEM,"...Sparse graph structure instanciated");
}
    

void sparseGraphStructure::ReadIncidences(goblinImport &F)
    throw(ERParse)
{
    SN = new TNode[2*mMax];

    right = new TArc[2*mMax];
    first = new TArc[nMax];
    left = NULL;

    char* token = F.Scan("incidences");

    for (TArc a=0;a<2*mMax;a++) SN[a] = NoNode;

    for (TNode v=0;v<nAct-ni;v++)
    {
        token = F.Scan();   /* node v */
        TArc* AL = F.GetTArcTuple(0);
        TArc maxIndex = F.Length();
        
        for (TNode i=0;i<maxIndex;i++)
        {
            TArc a = AL[i];
            SN[a] = v;
            if (i<maxIndex-1) right[a] = AL[i+1];
        }

        if (maxIndex>0)
        {
            right[AL[maxIndex-1]] = AL[0];
            first[v] = AL[0];
        }
        else first[v] = NoArc;
        
        delete[] AL;
    }

    for (TNode v=nAct-ni;v<nAct;v++) first[v] = NoArc;

    token = F.Scan();

    LogEntry(LOG_MEM,"...Incidence lists allocated");
}


void sparseGraphStructure::ReadLayout(goblinImport& F)
    throw(ERParse)
{
    F.Scan("layout");
    
    F.Scan("style");
    char* sty = F.GetCharTuple(1);

    if (sty[0]<=2) CT.arcStyle = sty[0];
    else
    {
        sprintf(CT.logBuffer,"Unknown arc style option: %d",sty[0]);
        Error(ERR_PARSE,"ReadLayout",CT.logBuffer);
    }

    delete[] sty;
    
    F.Scan("align");
    align = F.GetTNodeTuple(mAct);

    if (F.Constant())
    {
        delete[] align;
        align = NULL;
    }
    
    F.Scan("interpolate");
    interpolate = F.GetTNodeTuple(nAct);

    if (F.Constant())
    {
        delete[] interpolate;
        interpolate = NULL;
    }
    
    F.Scan(); // layout
}


void sparseGraphStructure::ReDimension(TNode nn,TNode nni,TArc mm)
    throw(ERRejected)
{   
    #if defined(_FAILSAVE_)

    if (nMax>0 || mMax>0)
        Error(ERR_REJECTED,"ReDimension","Graph structure must be initial");

    #endif
    
    nMax = nAct = nn+nni;
    mMax = mAct = mm;
    ni = nni;
}


void sparseGraphStructure::ReSize(TNode nn,TArc mm)
    throw(ERRejected)
{   
    #if defined(_FAILSAVE_)

    if (nn<nAct || mm<mAct || nn>CT.MaxNode() || 2*mm>CT.MaxArc())
        Error(ERR_REJECTED,"ReSize","Dimensions are out of range");

    #endif
    
    if (length!=NULL)
    {
        length = (TFloat *)GoblinRealloc(length,mm*sizeof(TFloat));
        for (TArc a=mMax;a<mm;a++) length[a] = clength;
    }

    if (ucap!=NULL)
    {
        ucap = (TCap *)GoblinRealloc(ucap,mm*sizeof(TCap));
        for (TArc a=mMax;a<mm;a++) ucap[a] = cucap;
    }

    if (lcap!=NULL)
    {
        lcap = (TCap *)GoblinRealloc(lcap,mm*sizeof(TCap));
        for (TArc a=mMax;a<mm;a++) lcap[a] = clcap;
    }

    if (demand!=NULL)
    {
        demand = (TCap *)GoblinRealloc(demand,nn*sizeof(TCap));
        for (TNode v=nMax;v<nn;v++) demand[v] = cDemand;
    }

    if (x!=NULL)
    {
        x = (TFloat *)GoblinRealloc(x,nn*sizeof(TFloat));
        for (TNode v=nMax;v<nn;v++) x[v] = 0;
    }

    if (y!=NULL)
    {
        y = (TFloat *)GoblinRealloc(y,nn*sizeof(TFloat));
        for (TNode v=nMax;v<nn;v++) y[v] = 0;
    }

    if (orientation!=NULL)
    {
        orientation = (char*)GoblinRealloc(orientation,mm*sizeof(char));
        for (TArc a=mMax;a<mm;a++) orientation[a] = cOrientation;
    }

    if (sub!=NULL)
    {
        sub = (TFloat *)GoblinRealloc(sub,mm*sizeof(TFloat));
        for (TArc a=mMax;a<mm;a++) sub[a] = clcap;
    }

    SN = (TNode *)GoblinRealloc(SN,2*mm*sizeof(TNode));
       
    first = (TArc *)GoblinRealloc(first,nn*sizeof(TArc));

    for (TNode v=nMax;v<nn;v++) first[v] = NoArc;
       
    right = (TArc *)GoblinRealloc(right,2*mm*sizeof(TArc));

    if (left!=NULL)
    {
        left = (TArc *)GoblinRealloc(left,2*mm*sizeof(TArc));
        for (TArc a=mMax;a<mm;a++) sub[a] = clcap;
    }
    
    if (align!=NULL)
    {
        align = (TNode *)GoblinRealloc(align,mm*sizeof(TNode));
        for (TArc a=mMax;a<mm;a++) align[a] = NoNode;
    }

    if (interpolate!=NULL)
    {
        interpolate = (TNode *)GoblinRealloc(demand,nn*sizeof(TNode));
        for (TNode v=nMax;v<nn;v++) interpolate[v] = NoNode;
    }
    
    nMax = nn;
    mMax = mm;
}


unsigned long sparseGraphStructure::Size()
    throw()
{   
    return
          sizeof(sparseGraphStructure)
        + goblinDataObject::Allocated()
        + genericGraphStructure::Allocated()
        + sparseGraphStructure::Allocated();
}


unsigned long sparseGraphStructure::Allocated()
    throw()
{   
    unsigned long tmpSize
        = 2*mMax*sizeof(TNode)             // SN[]
        + nMax*sizeof(TArc)                // first[]
        + 2*mMax*sizeof(TArc);             // right[]
        
    if (left!=NULL)         tmpSize += 2*mMax*sizeof(TArc);
    if (sub!=NULL)          tmpSize += mMax*sizeof(TFloat);
    if (align!=NULL)        tmpSize += mMax*sizeof(TNode);
    if (interpolate!=NULL)  tmpSize += nMax*sizeof(TNode);

    return tmpSize;
}
    

TArc sparseGraphStructure::InsertArc(TNode u,TNode v,TCap uu,TCap ll,TFloat cc)
    throw(ERRange,ERRejected)
{
    G.ReleaseIterators();
    
    #if defined(_FAILSAVE_)

    if (G.IsReferenced())
        Error(ERR_REJECTED,"InsertArc","Object is referenced");

    if (u>=nAct) NoSuchNode("InsertArc",u);

    if (v>=nAct) NoSuchNode("InsertArc",v);

    if (2*mMax>CT.MaxArc()-2)
        Error(ERR_REJECTED,"InsertArc","Number of arcs is out of range");

    #endif
    
    G.ReleaseAdjacencies();
    mAct++;
    
    if (mMax+1==mAct)
    {
        SN = (TNode*)GoblinRealloc(SN,(2*(mMax+1))*sizeof(TNode));
        right = (TArc*)GoblinRealloc(right,(2*(mMax+1))*sizeof(TArc));
       
        if (left!=NULL)
            left = (TArc*)GoblinRealloc(left,2*(mMax+1)*sizeof(TArc));
       
        if (ucap!=NULL)
        {
            ucap = (TCap*)GoblinRealloc(ucap,(mMax+1)*sizeof(TCap));
            ucap[mMax] = cucap;
        }

        if (length!=NULL)
        {
            length = (TFloat*)GoblinRealloc(length,(mMax+1)*sizeof(TFloat));
            length[mMax] = clength;       
        }
    
        if (lcap!=NULL)
        {
            lcap = (TCap*)GoblinRealloc(lcap,(mMax+1)*sizeof(TCap));
            lcap[mMax] = 0;
        }

        if (orientation!=NULL)
        {
            orientation = (char*)GoblinRealloc(orientation,(mMax+1)*sizeof(char));
            orientation[mMax] = cOrientation;
        }

        if (sub!=NULL)
        {
            sub = (TFloat*)GoblinRealloc(sub,(mMax+1)*sizeof(TFloat));
            sub[mMax] = 0;
        }
        
        if (align!=NULL)
        {
            align = (TNode*)GoblinRealloc(align,(mMax+1)*sizeof(TNode));
            align[mMax] = NoNode;
        }
        
        Error(MSG_WARN,"InsertArc","Non-Buffered arc insertion");
        mMax++;
    }

    TArc a = mAct-1;
    TArc af = 2*a;
    TArc ar = 2*a+1;
    
    SN[af] = u;
    SN[ar] = v;
 
    if (first[u]==NoArc)
    {
        first[u] = af;
        right[af] = af;
        if (left!=NULL) left[af] = af;
    }
    else
    {
        if (left==NULL)
        {
            right[af] = right[first[u]];
            right[first[u]] = af;
            first[u] = af;
        }
        else
        {
            TArc last = left[first[u]];
            left[first[u]] = af;
            left[af] = last;
            right[last] = af;
            right[af] = first[u];
        }
    }
    
    if (first[v]==NoArc)
    {
        first[v] = ar;
        right[ar] = ar;
        if (left!=NULL) left[ar] = ar;
    }
    else
    {
        if (left==NULL)
        {
            right[ar] = right[first[v]];
            right[first[v]] = ar;
            first[v] = ar;
        }
        else
        {
            TArc last = left[first[v]];
            left[first[v]] = ar;
            left[ar] = last;
            right[last] = ar;
            right[ar] = first[v];
        }
    }

    if (sub!=NULL)
    {
        sub[a] = ll;
        if (ll>0) G.ReleaseDegrees();
    }
    
    SetLength(2*a,cc);
    SetUCap(2*a,uu);
    SetLCap(2*a,ll);

    return a;
}
    

TArc sparseGraphStructure::InsertArc(TNode v,TNode w)
    throw(ERRange,ERRejected)
{
    TCap c = clength;
    if (CT.randLength) c = TFloat(CT.SignedRand());
    
    TFloat u = cucap;
    if (CT.randUCap) u = TCap(1+CT.UnsignedRand());
    
    TFloat l = clcap;
    if (CT.randLCap) l = TCap(CT.Rand(1+int(u)));
    
    return InsertArc(v,w,u,l,c);
}


TNode sparseGraphStructure::InsertNode()
    throw(ERRejected)
{
    G.ReleaseIterators();
    
    #if defined(_FAILSAVE_)

    if (G.IsReferenced())
        Error(ERR_REJECTED,"InsertNode","Object is referenced");

    if (nMax>CT.MaxNode()-1)
        Error(ERR_REJECTED,"InsertNode","Number of nodes is out of range");

    #endif
    
    G.ReleaseAdjacencies();
    nAct++;
    
    if (nMax+1==nAct)
    {
        first = (TArc *)GoblinRealloc(first,(nMax+1)*sizeof(TArc));
        first[nMax] = NoArc;
       
        if (x!=NULL)
        {
            x = (TFloat *)GoblinRealloc(x,(nMax+1)*sizeof(TFloat));
            x[nMax] = 0;
        }
       
        if (y!=NULL)
        {
            y = (TFloat *)GoblinRealloc(y,(nMax+1)*sizeof(TFloat));
            y[nMax] = 0;
        }

        if (demand!=NULL)
        {
            demand = (TCap *)GoblinRealloc(demand,(nMax+1)*sizeof(TCap));
            demand[nMax] = cDemand;       
        }

        if (interpolate!=NULL)
        {
            interpolate = (TNode *)GoblinRealloc(interpolate,(nMax+1)*sizeof(TNode));
            interpolate[nMax] = NoNode;
        }

        Error(MSG_WARN,"InsertNode","Non-Buffered node insertion");
        nMax++;
    }
    
    if (ni>0) SwapNodes(nAct-1,nAct-ni-1);

    return nAct-ni-1;
}


TNode sparseGraphStructure::InsertArtificalNode()
    throw(ERRejected)
{
    #if defined(_FAILSAVE_)

    if (nMax>CT.MaxNode()-1)
        Error(ERR_REJECTED,"InsertNode","Number of nodes is out of range");

    #endif
    
    nAct++;
    ni++;
    
    if (nMax+1==nAct)
    {
        first = (TArc *)GoblinRealloc(first,(nMax+1)*sizeof(TArc));
        first[nMax] = NoArc;
       
        if (x!=NULL)
        {
            x = (TFloat *)GoblinRealloc(x,(nMax+1)*sizeof(TFloat));
            x[nMax] = 0;
        }
       
        if (y!=NULL)
        {
            y = (TFloat *)GoblinRealloc(y,(nMax+1)*sizeof(TFloat));
            y[nMax] = 0;
        }

        if (demand!=NULL)
        {
            demand = (TCap *)GoblinRealloc(demand,(nMax+1)*sizeof(TCap));
            demand[nMax] = cDemand;       
        }

        if (interpolate!=NULL)
        {
            interpolate = (TNode *)GoblinRealloc(interpolate,(nMax+1)*sizeof(TNode));
            interpolate[nMax] = NoNode;
        }

        Error(MSG_WARN,"InsertArtificialNode","Non-Buffered node insertion");
        nMax++;
    }

    return nAct-1;
}


TNode sparseGraphStructure::InsertAlignmentPoint(TArc a)
    throw(ERRejected,ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("InsertAlignmentPoint",a);

    if (align!=NULL && align[a>>1]!=NoNode)
        Error(ERR_REJECTED,
            "InsertAlignmentPoint","Redefinition of alignment point");

    #endif
    
    TNode u = InsertArtificalNode();
    
    if (align==NULL)
    {
        align = new TNode[mMax];
        for (TArc i=0;i<mMax;i++) align[i] = NoNode;
        LogEntry(LOG_MEM,"...Alignment points allocated");
    }

    align[a>>1] = u;
    if (interpolate!=NULL) interpolate[u] = NoNode;
    
    return u;
}


TNode sparseGraphStructure::InsertBendNode(TNode v) throw(ERRejected,ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=nAct) NoSuchNode("InsertBendNodes",v);

    #endif
    
    TNode u = InsertArtificalNode();
    
    if (interpolate==NULL)
    {
        interpolate = new TNode[nMax];
        for (TNode i=0;i<nMax;i++) interpolate[i] = NoNode;
        LogEntry(LOG_MEM,"...Interpolation points allocated");
    }
    
    interpolate[u] = interpolate[v];
    interpolate[v] = u;
    
    return u;
}


void sparseGraphStructure::ExplicitParallels() throw()
{
    TArc mOld = mAct;
    for (TArc a=0;a<mOld;a++)
    {
        TNode u = StartNode(2*a);
        TNode v = EndNode(2*a);
        TCap multiplicity = ceil(UCap(2*a));
        TCap thisLower = LCap(2*a);
        TFloat thisSub = Sub(2*a);
        TFloat thisLength = Length(2*a);

        SetUCap(2*a,UCap(2*a)-multiplicity+1);
        
        for (TCap i=1;i<multiplicity;i++)
        {
            if (thisLower==0)
            {
                InsertArc(u,v,1,0,thisLength);
            }
            else if (thisLower<1)
            {
                InsertArc(u,v,1,thisLower,thisLength);
                thisLower = 0;
            }
            else
            {
                InsertArc(u,v,1,1,thisLength);
                thisLower -= 1;
            }
            
            if (!sub) continue;

            if (thisSub==0)
            {
                sub[mAct-1] = 0;
            }
            else if (thisSub<1)
            {
                sub[mAct-1] = thisSub;
                thisSub = 0;
            }
            else
            {
                sub[mAct-1] = 1;
                thisSub -= 1;
            }
        }

        sub[a] = thisSub;
        SetLCap(2*a,thisLower);
    }
}


void sparseGraphStructure::SwapArcs(TArc a1,TArc a2)
    throw(ERRange)
{   
    #if defined(_FAILSAVE_)

    if (a1>=2*mAct) NoSuchArc("SwapArcs",a1);

    if (a2>=2*mAct) NoSuchArc("SwapArcs",a2);

    #endif
    
    if (a1==a2) return;
    
    TNode u1 = SN[a1];
    TNode u2 = SN[a2];
    TNode v1 = SN[a1^1];
    TNode v2 = SN[a2^1];
    
    SN[a1] = u2;
    SN[a2] = u1;
    SN[a1^1] = v2;
    SN[a2^1] = v1;
    
    if (u1!=NoNode)
    {
        if (first[u1]==a1) first[u1] = a2;
        if (first[v1]==(a1^1)) first[v1] = (a2^1);
    }

    if (u2!=NoNode)
    {
        if (first[u2]==a2 && u1!=u2) first[u2] = a1;
        if (first[v2]==(a2^1) && v1!=v2) first[v2] = (a1^1);
    }
    
    TArc a1l = Left(a1);
    TArc a2l = Left(a2);
    TArc a1r = Right(a1);
    TArc a2r = Right(a2);
    
    if (a1l!=a2 && a2l!=a1)
    {
        if (a2!=a2r)
        {
            left[a1] = a2l;
            right[a1] = a2r;

            if (a1r!=NoArc)
            {
                left[a1r] = a2;
                right[a1l] = a2;
            }
        }
        else
        {
            left[a1] = a1;
            right[a1] = a1;
        }
    
        if (a1!=a1r)
        {
            left[a2] = a1l;
            right[a2] = a1r;

            if (a2r!=NoArc)
            {
                left[a2r] = a1;
                right[a2l] = a1;
            }
        }
        else
        {
            left[a2] = a2;
            right[a2] = a2;
        }
    }

    a1 = (a1^1);
    a2 = (a2^1);
    a1l = Left(a1);
    a2l = Left(a2);
    a1r = Right(a1);
    a2r = Right(a2);
    
    if (a1l!=a2 && a2l!=a1 && a2!=(a1^1))
    {
        if (a2!=a2r)
        {
            left[a1] = a2l;
            right[a1] = a2r;

            if (a1r!=NoArc)
            {
                left[a1r] = a2;
                right[a1l] = a2;
            }
        }
        else
        {
            left[a1] = a1;
            right[a1] = a1;
        }
    
        if (a1!=a1r)
        {
            left[a2] = a1l;
            right[a2] = a1r;

            if (a2r!=NoArc)
            {
                left[a2r] = a1;
                right[a2l] = a1;
            }
        }
        else
        {
            left[a2] = a2;
            right[a2] = a2;
        }
    }
    
    if (ucap!=NULL)
    {
        TCap swap = ucap[a1>>1];
        ucap[a1>>1] = ucap[a2>>1];
        ucap[a2>>1] = swap;
    }
    
    if (lcap!=NULL)
    {
        TCap swap = lcap[a1>>1];
        lcap[a1>>1] = lcap[a2>>1];
        lcap[a2>>1] = swap;
    }
    
    if (length!=NULL)
    {
        TFloat swap = length[a1>>1];
        length[a1>>1] = length[a2>>1];
        length[a2>>1] = swap;
    }
    
    if (orientation!=NULL)
    {
        char swap = orientation[a1>>1];
        orientation[a1>>1] = orientation[a2>>1];
        orientation[a2>>1] = swap;
    }
    
    if (sub!=NULL)
    {
        TFloat swap = sub[a1>>1];
        sub[a1>>1] = sub[a2>>1];
        sub[a2>>1] = swap;
    }
    
    if (align!=NULL)
    {
        TNode swap = align[a1>>1];
        align[a1>>1] = align[a2>>1];
        align[a2>>1] = swap;
    }
}


void sparseGraphStructure::SwapNodes(TNode u,TNode v)
    throw(ERRange)
{   
    #if defined(_FAILSAVE_)

    if (u>=nAct) NoSuchNode("SwapNodes",u);

    if (v>=nAct) NoSuchNode("SwapNodes",v);

    #endif
    
    if (u==v) return;
    
    TArc  a =  first[u];

    if (a!=NoArc)
    {
        SN[a] = v;

        while (right[a]!=first[u])
        {
            a = right[a];
            SN[a] = v;
        }
    }

    a = first[v];

    if (a!=NoArc)
    {
        SN[a] = u;

        while (right[a]!=first[v])
        {
            a = right[a];
            SN[a] = u;
        }
    }
    
    TArc swap = first[u];
    first[u] = first[v];
    first[v] = swap;

    if (demand!=NULL)
    {
        TCap swap = demand[u];
        demand[u] = demand[v];
        demand[v] = swap;
    }
    
    if (x!=NULL)
    {
        TFloat swap = x[u];
        x[u] = x[v];
        x[v] = swap;
    }
    
    if (y!=NULL)
    {
        TFloat swap = y[u];
        y[u] = y[v];
        y[v] = swap;
    }
    
    if (align!=NULL)
    {
        for (TArc a=0;a<mAct;a++)
        {
            if (align[a]==u) align[a] = v;
            else if (align[a]==v) align[a] = u;
        }

        if (interpolate!=NULL)
        {
            TNode swap = interpolate[u];
            interpolate[u] = interpolate[v];
            interpolate[v] = swap;

            for (TNode w=nAct-ni;w<nAct;w++)
            {
                if (interpolate[w]==u) interpolate[w] = v;
                else if (interpolate[w]==v) interpolate[w] = u;
            }
        }
    }
}


void sparseGraphStructure::FlipArc(TArc a)
    throw(ERRange)
{
    SwapArcs(a,a^1);
    
    if (interpolate!=NULL)
    {
        TNode v = align[a>>1];
        
        if (v!=NoNode && interpolate[v]!=NoNode)
        {
            // Revert the order of bend nodes

            TNode* revThread = new TNode[nAct];
            TNode w = interpolate[v];
            revThread[w] = NoNode;
            
            while (interpolate[w]!=NoNode)
            {
                revThread[interpolate[w]] = w;
                w = interpolate[w];
            }

            interpolate[v] = w;
            
            while (w!=NoNode)
            {
                interpolate[w] = revThread[w];
                w = revThread[w];
            }

            delete[] revThread;
        }
    }
}


void sparseGraphStructure::CancelArc(TArc ad)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (ad>=2*mAct) NoSuchArc("CancelArc",ad);
    
    if (SN[ad]==NoNode) CancelledArc("CancelArc",ad);

    #endif
            
    if (sub!=NULL) G.AdjustDegrees(2*(ad>>1)+1,sub[ad]);
    
    TArc al = Left(ad);
    TArc ar = right[ad];

    if (ad!=ar)
    {
        right[al] = ar;
        left[ar] = al;

        if (first[SN[ad]]==ad) first[SN[ad]] = al;
    }
    else first[SN[ad]] = NoArc;
    
    SN[ad] = NoNode;
    right[ad] = NoArc;
    left[ad] = NoArc;
    
    ad = ad^1;
    
    al = left[ad];
    ar = right[ad];

    if (ad!=ar)
    {
        right[al] = ar;
        left[ar] = al;

        if (first[SN[ad]]==ad) first[SN[ad]] = al;
    }
    else first[SN[ad]] = NoArc;
    
    SN[ad] = NoNode;
    right[ad] = NoArc;
    left[ad] = NoArc;
}


void sparseGraphStructure::CancelNode(TNode v)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=nAct) NoSuchNode("CancelNode",v);

    #endif

    while (first[v]!=NoArc) CancelArc(first[v]);

    SetC(v,0,-InfFloat);
}


void sparseGraphStructure::ReleaseBendNodes(TArc a)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("ReleaseBendNodes",a);

    #endif
    
    if (align!=NULL)
    {
        TNode w = Align(a);

        while (w!=NoNode)
        {
            TNode x = Interpolate(w);
            DeleteNode(w);
            if (x!=nAct) w = x;
        }

        align[a>>1] = NoNode;
    }
}


void sparseGraphStructure::DeleteArc(TArc a)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("DeleteArc",a);

    #endif
    
    TArc ad = 2*mAct-2;

    if (SN[a]!=NoNode) CancelArc(a);
    
    ReleaseBendNodes(a);
    SwapArcs(a,ad);

    mAct--;
}


void sparseGraphStructure::DeleteNode(TNode v)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=nAct) NoSuchNode("DeleteNode",v);

    #endif
    
    if (v<nAct-ni)
    {
        CancelNode(v);
        SwapNodes(v,nAct-ni-1);

        if (ni>0) SwapNodes(nAct-1,nAct-ni-1);

        nAct--;
        DeleteArcs();
    }
    else
    {
        TArc i = 0;

        for (;i<mAct && align[i]!=v;i++) {};
        
        if (i!=mAct)
        {
            if (!interpolate) align[i] = NoNode;
            else  align[i] = interpolate[v];
        }
        else
        {
            TNode j = nAct-ni;
            
            for (;j<nAct && Interpolate(j)!=v;j++) {};

            interpolate[j] = interpolate[v];
        }

        SwapNodes(v,nAct-1);
        nAct--;
        ni--;
    }
}


void sparseGraphStructure::DeleteArcs()
    throw()
{
    TArc a = 0;

    while (a<2*mAct)
    {
        if (SN[a]==NoNode) DeleteArc(a);
        else a+=2;
    }
}


void sparseGraphStructure::DeleteNodes()
    throw()
{
    TNode v = 0;

    while (v<nAct-ni)
    {
        if (first[v]==NoArc) DeleteNode(v);
        else v++;
    }
}


void sparseGraphStructure::ContractArc(TArc ac)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (ac>=2*mAct) NoSuchArc("ContractArc",ac);
    
    if (SN[ac]==SN[ac^1])
    {
        sprintf(CT.logBuffer,"Arc is a loop: %ld",ac);
        Error(ERR_REJECTED,"ContractArc",CT.logBuffer);
    }

    #endif

    TNode u = SN[ac];
    TNode v = SN[ac^1];
    TArc a = (ac^1);

    while (right[a]!=(ac^1))
    {
        SN[a] = u;
        a = right[a];
    }

    SN[a] = u;

    right[a] = right[ac];

    if (left!=NULL) left[right[ac]] = a;

    right[ac] = (ac^1);

    if (left!=NULL) left[ac^1] = ac;

    first[v] = NoArc;

    CancelArc(ac);

    SetC(v,0,-InfFloat);
}


void sparseGraphStructure::IdentifyNodes(TNode u,TNode v)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (u>=nAct) NoSuchNode("InsertArc",u);

    if (v>=nAct) NoSuchNode("InsertArc",v);

    #endif
    
    TArc a1 = first[u];
    TArc a2 = first[v];
    TArc a = a2;

    while (right[a]!=a2)
    {
        SN[a] = u;
        a = right[a];
    }

    SN[a] = u;

    right[a] = right[a1];

    if (left!=NULL) left[right[a1]] = a;

    right[a1] = a2;

    if (left!=NULL) left[a2] = a1;

    first[v] = NoArc;

    SetC(v,0,-InfFloat);
}


TNode sparseGraphStructure::StartNode(TArc a)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("StartNode",a);

    #endif
    
    return SN[a];
}


TNode sparseGraphStructure::EndNode(TArc a)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("EndNode",a);

    #endif
    
    return SN[a^1];
}


TArc sparseGraphStructure::First(TNode v)
    throw(ERRange)
{   
    #if defined(_FAILSAVE_)

    if (v>=nAct) NoSuchNode("First",v);

    #endif
    
    return first[v];
}


bool sparseGraphStructure::HiddenNode(TNode v)
    throw(ERRange)
{   
    #if defined(_FAILSAVE_)

    if (v>=nAct) NoSuchNode("HiddenNode",v);

    #endif
    
    return C(v,0)==-InfFloat;
}


bool sparseGraphStructure::HiddenArc(TArc a)
    throw(ERRange)
{   
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("HiddenArc",a);

    #endif
    
    return (   SN[a]==NoNode
            || HiddenNode(SN[a])
            || HiddenNode(SN[a^1])
           );
}


void sparseGraphStructure::SetFirst(TNode v,TArc a)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (v>=nAct) NoSuchNode("SetFirst",v);

    if (SN[a]!=v) Error(ERR_REJECTED,"SetFirst","Mismatching start node");

    #endif
    
    first[v] = a;
}


TArc sparseGraphStructure::Right(TArc a)
    throw(ERRange)
{   
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("Right",a);

    #endif
    
    return right[a];
}


void sparseGraphStructure::SetRight(TArc a1,TArc a2)
    throw(ERRange,ERRejected)
{
    if (a1==a2) return;
    
    #if defined(_FAILSAVE_)

    if (a1>=2*mAct) NoSuchArc("SetRight",a1);
    if (a2>=2*mAct) NoSuchArc("SetRight",a2);

    if (SN[a1]!=SN[a2]) Error(ERR_REJECTED,"SetRight","Mismatching start nodes");

    #endif
    
    TArc a2l = Left(a2);
    TArc a1r = Right(a1);
    TArc a2r = Right(a2);

    if (a1r==a2) return;
    
    right[a1]  = a2;
    left[a2]   = a1;
    right[a2]  = a1r;
    left[a1r]  = a2;
    right[a2l] = a2r;
    left[a2r]  = a2l;
}


TArc sparseGraphStructure::Left(TArc ac)
    throw(ERRange)
{   
    #if defined(_FAILSAVE_)

    if (ac>=2*mAct) NoSuchArc("Left",ac);

    #endif
    
    if (left==NULL)
    {
        left = new TArc[2*mMax];
        LogEntry(LOG_MEM,"Generating reverse incidence lists...");
        
        for (TArc a=0;a<2*mAct;a++) left[a] = NoArc;
        
        for (TArc a=0;a<2*mAct;a++)
        {
            if (right[a]!=NoArc)
            {
                if (left[right[a]]==NoArc) left[right[a]] = a;
                else
                {
                    InternalError("Left","Inconsistent incidence lists");
                }
            }
        }
    }

    return left[ac];
}


void sparseGraphStructure::Planarize(TArc* pred)
    throw()
{
    for (TArc a=0;a<2*mAct;a++)
    {
        TArc a2 = (pred[a]^1);

        // Skip the arcs which have been added to obtain 2-connectivity
        while (a2>=2*mAct) a2 = (pred[a2]^1);

        right[a] = a2;
        if (left) left[a2] = a;
    }
}


TArc sparseGraphStructure::GrowExteriorFace()
    throw(ERRejected)
{
    LogEntry(LOG_METH,"Selecting exterior face...");

    TNode* region = new TNode[2*G.M()];
    for (TArc a=0;a<2*G.M();a++) region[a] = NoNode;
    
    OpenFold();
    
    TNode regionIndex = 0;
    TArc retArc = NoArc;
    TNode maxLength = 0;

    for (TArc a=0;a<2*G.M();a++)
    {
        if (region[a]!=NoNode) continue;

        TArc regionLength = 1;
        TArc a2 = G.Right(a^1,SN[a^1]);
        region[a2] = regionIndex;

        #if defined(_LOGGING_)

        THandle LH = NoHandle;

        if (CT.logMeth>1)
        {
            sprintf(CT.logBuffer,"Region %ld = (%ld[%ld]%ld",
                regionIndex,SN[a2],a2,SN[a2^1]);
            LH = LogStart(LOG_METH2,CT.logBuffer);
        }

        #endif

        while (a2!=a)
        {
            regionLength++;
            a2 = G.Right(a2^1,SN[a2^1]);
            region[a2] = regionIndex;

            #if defined(_LOGGING_)

            if (CT.logMeth>1)
            {
                sprintf(CT.logBuffer,"[%ld]%ld",a2,SN[a2^1]);
                LogAppend(LH,CT.logBuffer);
            }

            #endif
        }

        #if defined(_LOGGING_)

        if (CT.logMeth>1) LogEnd(LH,")");

        #endif
        
        if (regionLength>maxLength)
        {
            retArc = a;
            maxLength = regionLength;
        }
        
        regionIndex++;
    }
    
    CloseFold();
    
    #if defined(_FAILSAVE_)

    if (G.M()-G.N()+2!=regionIndex)
        Error(ERR_REJECTED,"RefineUnboundedRegion","Not a planar embedding");

    #endif
    
    sprintf(CT.logBuffer,"...Exterior region has length %ld",maxLength);
    LogEntry(LOG_RES,CT.logBuffer);
    LogEntry(LOG_METH,"Eliminating exterior cut arcs...");
    OpenFold();
    
    TArc a = retArc;
    maxLength = 0;
    TArc mBridges = 0;
    
    while (true)
    {
        TArc bridge = Right(a^1);
        TNode r = EndNode(Right(bridge));
        
        if (Right(bridge)==bridge) r = EndNode(Left(bridge^1));
        
        bool contractible = true;
        
        if (region[bridge]==region[bridge^1])
        {
            if (bridge&1) mBridges++;
            
            sprintf(CT.logBuffer,"Arc %ld is a bridge",bridge);
            LogEntry(LOG_METH2,CT.logBuffer);
        }
        else contractible = !G.MarkComponent(bridge,r);
        
        if (!contractible)
        {
            // The end nodes of bridge define a cut set.
            // The coloured vertices form a respective connected component.
            // Swap bridge to the interior. By that, the number of nodes in
            // the unbounded region strictly increases.

            TArc a2 = bridge;
            TArc a3 = bridge^1;

            do {a2 = Left(a2);}
            while (G.Colour(SN[a2^1])==2);

            do {a3 = Left(a3);}
            while (G.Colour(SN[a3^1])==1);

            if (a2!=bridge && a2!=Left(bridge) &&
                a3!=(bridge^1) && a3!=Left(bridge^1))
            {
                if (bridge==retArc) retArc = Right(retArc^1);
            
                SetRight(a2,bridge);
                SetRight(a3,bridge^1);
                LogEntry(LOG_METH,"...Cut arc swapped to the interior");
            }
            else if (a2!=bridge && a2!=Left(bridge) &&
                     a3!=(bridge^1) && a3==Left(bridge^1))
            {
                SetRight(a2,bridge);
                LogEntry(LOG_METH,"...Block swapped to the exterior");
            }
            else
            {
                a = bridge;
                maxLength++;

                if (a2==bridge || a2==Left(bridge))
                    sprintf(CT.logBuffer,"...%ld is a cut node",SN[bridge^1]);
                else sprintf(CT.logBuffer,"...%ld is a cut node",SN[bridge]);

                LogEntry(LOG_METH,CT.logBuffer);

                if (a==retArc) break;
            }
        }
        else
        {
            a = bridge;
            maxLength++;

            if (a==retArc) break;
        }
    }
    
    CloseFold();

    if (mBridges<G.M() && maxLength<G.N()+mBridges)
    {
        sprintf(CT.logBuffer,"...Exterior region has length %ld",maxLength);
        LogEntry(LOG_RES,CT.logBuffer);
    }
    else LogEntry(LOG_RES,"...Graph is outerplanar");

    return retArc;
}


TNode sparseGraphStructure::Align(TArc a)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("Align",a);

    #endif

    if (align==NULL) return NoNode;
    else return align[a>>1];
}


TNode sparseGraphStructure::Interpolate(TNode v)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (v>=nAct) NoSuchNode("Interpolate",v);

    #endif

    if (interpolate==NULL) return NoNode;
    else return interpolate[v];
}


void sparseGraphStructure::ReleaseIncidences()
    throw()
{   
    if (left!=NULL)
    {
        delete[] left;
        left = NULL;
        LogEntry(LOG_MEM,"...Reverse incidences disallocated");
    }
}


void sparseGraphStructure::NewSubgraph()
    throw(ERRejected)
{
    if (sub==NULL)
    {
        sub = new TFloat[mMax];
        LogEntry(LOG_MEM,"...Subgraph allocated");
    
        for (TArc a=0;a<mAct;a++)
        {
            TCap ll = LCap(2*a);
            sub[a] = ll;

            if (ll>0) G.AdjustDegrees(2*a,ll);
        }
    }
    else Error(ERR_REJECTED,"NewSubgraph","A subgraph is already present");
}


void sparseGraphStructure::ReleaseSubgraph()
    throw()
{
    if (sub!=NULL)
    {
        delete[] sub;
        LogEntry(LOG_MEM,"...Subgraph disallocated");
        sub = NULL;
    }
}


void sparseGraphStructure::AddArc(TArc a,TFloat Lambda)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("AddArc",a);

    if (Lambda<0 || Lambda>UCap(a)-Sub(a))
        AmountOutOfRange("AddArc",Lambda);

    #endif
 
    if (sub==NULL) NewSubgraph();

    sub[a>>1] = sub[a>>1]+Lambda;
    G.AdjustDegrees(a,Lambda);
}


void sparseGraphStructure::OmitArc(TArc a,TFloat Lambda)
    throw(ERRange,ERRejected)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("OmitArc",a);

    if (Lambda<0 || Lambda>Sub(a)-LCap(a))
        AmountOutOfRange("OmitArc",Lambda);

    #endif

    if (sub==NULL) NewSubgraph();

    sub[a>>1] = sub[a>>1]-Lambda;
    G.AdjustDegrees(a,-Lambda);
}


TFloat sparseGraphStructure::Sub(TArc a)
    throw(ERRange)
{
    #if defined(_FAILSAVE_)

    if (a>=2*mAct) NoSuchArc("Sub",a);

    #endif

    if (sub==NULL) return LCap(a);
    else return sub[a>>1];
}


sparseGraphStructure::~sparseGraphStructure()
    throw()
{
    ReleaseIncidences();
    ReleaseSubgraph();
    
    delete[] SN;
    delete[] right;    
    delete[] first;    
    delete[] align;
    delete[] interpolate;

    LogEntry(LOG_MEM,"...Sparse graph structure disallocated");
}
